/*
 * QEMU model of the ARM SMMU-500
 *
 * Copyright (c) 2014 Xilinx Inc.
 *
 * Partially autogenerated by xregqemu.py 2014-08-25.
 * Written by Edgar E. Iglesias.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "qemu/osdep.h"
#include "hw/sysbus.h"
#include "hw/register-dep.h"
#include "qemu/bitops.h"
#include "qemu/log.h"
#include "qapi/error.h"
#include "sysemu/dma.h"

#include "hw/fdt_generic_util.h"

#ifndef XILINX_SMMU500_ERR_DEBUG
#define XILINX_SMMU500_ERR_DEBUG 0
#endif

#define TYPE_XILINX_SMMU500 "arm.mmu-500"
#define TYPE_XILINX_SMMU500_IOMMU_MEMORY_REGION "arm.mmu-500-iommu-memory-region"

#define XILINX_SMMU500(obj) \
     OBJECT_CHECK(SMMU, (obj), TYPE_XILINX_SMMU500)

#define DEBUG_DEV_SMMU 0
#define DEBUG_DEV_SMMU_PTW 0

#define D(...) do {             \
    if (DEBUG_DEV_SMMU) {       \
        qemu_log(__VA_ARGS__);  \
    }                           \
} while (0);

#define D_PTW(...) do {         \
    if (DEBUG_DEV_SMMU_PTW) {   \
        qemu_log(__VA_ARGS__);  \
    }                           \
} while (0);

DEP_REG32(SMMU_SCR0, 0x0)
    DEP_FIELD(SMMU_SCR0, NSCFG, 2, 28)
    DEP_FIELD(SMMU_SCR0, WACFG, 2, 26)
    DEP_FIELD(SMMU_SCR0, RACFG, 2, 24)
    DEP_FIELD(SMMU_SCR0, SHCFG, 2, 22)
    DEP_FIELD(SMMU_SCR0, SMCFCFG, 1, 21)
    DEP_FIELD(SMMU_SCR0, MTCFG, 1, 20)
    DEP_FIELD(SMMU_SCR0, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_SCR0, BSU, 2, 14)
    DEP_FIELD(SMMU_SCR0, FB, 1, 13)
    DEP_FIELD(SMMU_SCR0, PTM, 1, 12)
    DEP_FIELD(SMMU_SCR0, USFCFG, 1, 10)
    DEP_FIELD(SMMU_SCR0, GSE, 1, 9)
    DEP_FIELD(SMMU_SCR0, STALLD, 1, 8)
    DEP_FIELD(SMMU_SCR0, TRANSIENTCFG, 2, 6)
    DEP_FIELD(SMMU_SCR0, GCFGFIE, 1, 5)
    DEP_FIELD(SMMU_SCR0, GCFGFRE, 1, 4)
    DEP_FIELD(SMMU_SCR0, GFIE, 1, 2)
    DEP_FIELD(SMMU_SCR0, GFRE, 1, 1)
    DEP_FIELD(SMMU_SCR0, CLIENTPD, 1, 0)
DEP_REG32(SMMU_SCR1, 0x4)
    DEP_FIELD(SMMU_SCR1, NSCAFRO, 1, 28)
    DEP_FIELD(SMMU_SCR1, SPMEN, 1, 27)
    DEP_FIELD(SMMU_SCR1, SIF, 1, 26)
    DEP_FIELD(SMMU_SCR1, GEFRO, 1, 25)
    DEP_FIELD(SMMU_SCR1, GASRAE, 1, 24)
    DEP_FIELD(SMMU_SCR1, NSNUMIRPTO, 8, 16)
    DEP_FIELD(SMMU_SCR1, NSNUMSMRGO, 6, 8)
    DEP_FIELD(SMMU_SCR1, NSNUMCBO, 5, 0)
DEP_REG32(SMMU_SACR, 0x10)
    DEP_FIELD(SMMU_SACR, NORMALIZE, 1, 27)
    DEP_FIELD(SMMU_SACR, CACHE_LOCK, 1, 26)
    DEP_FIELD(SMMU_SACR, PAGESIZE, 1, 16)
    DEP_FIELD(SMMU_SACR, S2CRB_TLBEN, 1, 10)
    DEP_FIELD(SMMU_SACR, MMUDISB_TLBEN, 1, 9)
    DEP_FIELD(SMMU_SACR, SMTNMB_TLBEN, 1, 8)
    DEP_FIELD(SMMU_SACR, S1WC2EN, 1, 2)
DEP_REG32(SMMU_SIDR0, 0x20)
    DEP_FIELD(SMMU_SIDR0, SES, 1, 31)
    DEP_FIELD(SMMU_SIDR0, S1TS, 1, 30)
    DEP_FIELD(SMMU_SIDR0, S2TS, 1, 29)
    DEP_FIELD(SMMU_SIDR0, NTS, 1, 28)
    DEP_FIELD(SMMU_SIDR0, SMS, 1, 27)
    DEP_FIELD(SMMU_SIDR0, ATOSNS, 1, 26)
    DEP_FIELD(SMMU_SIDR0, PTFS, 2, 24)
    DEP_FIELD(SMMU_SIDR0, NUMIRPT, 8, 16)
    DEP_FIELD(SMMU_SIDR0, CTTW, 1, 14)
    DEP_FIELD(SMMU_SIDR0, BTM, 1, 13)
    DEP_FIELD(SMMU_SIDR0, NUMSIDB, 4, 9)
    DEP_FIELD(SMMU_SIDR0, NUMSMRG, 8, 0)
DEP_REG32(SMMU_SIDR1, 0x24)
    DEP_FIELD(SMMU_SIDR1, PAGESIZE, 1, 31)
    DEP_FIELD(SMMU_SIDR1, NUMPAGENDXB, 3, 28)
    DEP_FIELD(SMMU_SIDR1, NUMS2CB, 8, 16)
    DEP_FIELD(SMMU_SIDR1, SMCD, 1, 15)
    DEP_FIELD(SMMU_SIDR1, SSDTP, 1, 12)
    DEP_FIELD(SMMU_SIDR1, NUMSSDNDXB, 4, 8)
    DEP_FIELD(SMMU_SIDR1, NUMCB, 8, 0)
DEP_REG32(SMMU_SIDR2, 0x28)
    DEP_FIELD(SMMU_SIDR2, PTFSV8_64KB, 1, 14)
    DEP_FIELD(SMMU_SIDR2, PTFSV8_16KB, 1, 13)
    DEP_FIELD(SMMU_SIDR2, TFSV8_4KB, 1, 12)
    DEP_FIELD(SMMU_SIDR2, UBS, 4, 8)
    DEP_FIELD(SMMU_SIDR2, OAS, 4, 4)
    DEP_FIELD(SMMU_SIDR2, IAS, 4, 0)
DEP_REG32(SMMU_SIDR7, 0x3c)
    DEP_FIELD(SMMU_SIDR7, MAJOR, 4, 4)
    DEP_FIELD(SMMU_SIDR7, MINOR, 4, 0)
DEP_REG32(SMMU_SGFAR_LOW, 0x40)
DEP_REG32(SMMU_SGFAR_HIGH, 0x44)
    DEP_FIELD(SMMU_SGFAR_HIGH, FADDR, 17, 0)
DEP_REG32(SMMU_SGFSR, 0x48)
    DEP_FIELD(SMMU_SGFSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_SGFSR, UUT, 1, 8)
    DEP_FIELD(SMMU_SGFSR, PF, 1, 7)
    DEP_FIELD(SMMU_SGFSR, EF, 1, 6)
    DEP_FIELD(SMMU_SGFSR, CAF, 1, 5)
    DEP_FIELD(SMMU_SGFSR, UCIF, 1, 4)
    DEP_FIELD(SMMU_SGFSR, UCBF, 1, 3)
    DEP_FIELD(SMMU_SGFSR, SMCF, 1, 2)
    DEP_FIELD(SMMU_SGFSR, USF, 1, 1)
    DEP_FIELD(SMMU_SGFSR, ICF, 1, 0)
DEP_REG32(SMMU_SGFSRRESTORE, 0x4c)
    DEP_FIELD(SMMU_SGFSRRESTORE, MULTI, 1, 31)
    DEP_FIELD(SMMU_SGFSRRESTORE, UUT, 1, 8)
    DEP_FIELD(SMMU_SGFSRRESTORE, PF, 1, 7)
    DEP_FIELD(SMMU_SGFSRRESTORE, EF, 1, 6)
    DEP_FIELD(SMMU_SGFSRRESTORE, CAF, 1, 5)
    DEP_FIELD(SMMU_SGFSRRESTORE, UCIF, 1, 4)
    DEP_FIELD(SMMU_SGFSRRESTORE, UCBF, 1, 3)
    DEP_FIELD(SMMU_SGFSRRESTORE, SMCF, 1, 2)
    DEP_FIELD(SMMU_SGFSRRESTORE, USF, 1, 1)
    DEP_FIELD(SMMU_SGFSRRESTORE, ICF, 1, 0)
DEP_REG32(SMMU_SGFSYNR0, 0x50)
    DEP_FIELD(SMMU_SGFSYNR0, ATS, 1, 6)
    DEP_FIELD(SMMU_SGFSYNR0, NSATTR, 1, 5)
    DEP_FIELD(SMMU_SGFSYNR0, NSSTATE, 1, 4)
    DEP_FIELD(SMMU_SGFSYNR0, IND, 1, 3)
    DEP_FIELD(SMMU_SGFSYNR0, PNU, 1, 2)
    DEP_FIELD(SMMU_SGFSYNR0, WNR, 1, 1)
DEP_REG32(SMMU_SGFSYNR1, 0x54)
    DEP_FIELD(SMMU_SGFSYNR1, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_SGFSYNR1, STREAMID, 15, 0)
DEP_REG32(SMMU_STLBIALL, 0x60)
DEP_REG32(SMMU_TLBIVMID, 0x64)
    DEP_FIELD(SMMU_TLBIVMID, VMID, 8, 0)
DEP_REG32(SMMU_TLBIALLNSNH, 0x68)
DEP_REG32(SMMU_STLBGSYNC, 0x70)
DEP_REG32(SMMU_STLBGSTATUS, 0x74)
    DEP_FIELD(SMMU_STLBGSTATUS, GSACTIVE, 1, 0)
DEP_REG32(SMMU_DBGRPTRTBU, 0x80)
    DEP_FIELD(SMMU_DBGRPTRTBU, TBU_ID, 3, 24)
    DEP_FIELD(SMMU_DBGRPTRTBU, TLB_POINTER, 12, 4)
    DEP_FIELD(SMMU_DBGRPTRTBU, TLB_ENTRY_POINTER, 4, 0)
DEP_REG32(SMMU_DBGRDATATBU, 0x84)
DEP_REG32(SMMU_DBGRPTRTCU, 0x88)
    DEP_FIELD(SMMU_DBGRPTRTCU, DATASRC, 2, 26)
    DEP_FIELD(SMMU_DBGRPTRTCU, WAY_RAM, 2, 24)
    DEP_FIELD(SMMU_DBGRPTRTCU, TLB_POINTER, 9, 4)
    DEP_FIELD(SMMU_DBGRPTRTCU, TLB_ENTRY_POINTER, 4, 0)
DEP_REG32(SMMU_DBGRDATATCU, 0x8c)
DEP_REG32(SMMU_STLBIVALM_LOW, 0xa0)
DEP_REG32(SMMU_STLBIVALM_HIGH, 0xa4)
    DEP_FIELD(SMMU_STLBIVALM_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_STLBIVAM_LOW, 0xa8)
DEP_REG32(SMMU_STLBIVAM_HIGH, 0xac)
    DEP_FIELD(SMMU_STLBIVAM_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_STLBIALLM, 0xbc)
DEP_REG32(SMMU_NSCR0, 0x400)
    DEP_FIELD(SMMU_NSCR0, WACFG, 2, 26)
    DEP_FIELD(SMMU_NSCR0, RACFG, 2, 24)
    DEP_FIELD(SMMU_NSCR0, SHCFG, 2, 22)
    DEP_FIELD(SMMU_NSCR0, SMCFCFG, 1, 21)
    DEP_FIELD(SMMU_NSCR0, MTCFG, 1, 20)
    DEP_FIELD(SMMU_NSCR0, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_NSCR0, BSU, 2, 14)
    DEP_FIELD(SMMU_NSCR0, FB, 1, 13)
    DEP_FIELD(SMMU_NSCR0, PTM, 1, 12)
    DEP_FIELD(SMMU_NSCR0, VMIDPNE, 1, 11)
    DEP_FIELD(SMMU_NSCR0, USFCFG, 1, 10)
    DEP_FIELD(SMMU_NSCR0, GSE, 1, 9)
    DEP_FIELD(SMMU_NSCR0, STALLD, 1, 8)
    DEP_FIELD(SMMU_NSCR0, TRANSIENTCFG, 2, 6)
    DEP_FIELD(SMMU_NSCR0, GCFGFIE, 1, 5)
    DEP_FIELD(SMMU_NSCR0, GCFGFRE, 1, 4)
    DEP_FIELD(SMMU_NSCR0, GFIE, 1, 2)
    DEP_FIELD(SMMU_NSCR0, GFRE, 1, 1)
    DEP_FIELD(SMMU_NSCR0, CLIENTPD, 1, 0)
DEP_REG32(SMMU_NSACR, 0x410)
    DEP_FIELD(SMMU_NSACR, CACHE_LOCK, 1, 26)
    DEP_FIELD(SMMU_NSACR, DP4K_TBUDISB, 1, 25)
    DEP_FIELD(SMMU_NSACR, DP4K_TCUDISB, 1, 24)
    DEP_FIELD(SMMU_NSACR, S2CRB_TLBEN, 1, 10)
    DEP_FIELD(SMMU_NSACR, MMUDISB_TLBEN, 1, 9)
    DEP_FIELD(SMMU_NSACR, SMTNMB_TLBEN, 1, 8)
    DEP_FIELD(SMMU_NSACR, IPA2PA_CEN, 1, 4)
    DEP_FIELD(SMMU_NSACR, S2WC2EN, 1, 3)
    DEP_FIELD(SMMU_NSACR, S1WC2EN, 1, 2)
DEP_REG32(SMMU_NSGFAR_LOW, 0x440)
DEP_REG32(SMMU_NSGFAR_HIGH, 0x444)
    DEP_FIELD(SMMU_NSGFAR_HIGH, FADDR, 17, 0)
DEP_REG32(SMMU_NSGFSR, 0x448)
    DEP_FIELD(SMMU_NSGFSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_NSGFSR, UUT, 1, 8)
    DEP_FIELD(SMMU_NSGFSR, EF, 1, 6)
    DEP_FIELD(SMMU_NSGFSR, CAF, 1, 5)
    DEP_FIELD(SMMU_NSGFSR, UCIF, 1, 4)
    DEP_FIELD(SMMU_NSGFSR, UCBF, 1, 3)
    DEP_FIELD(SMMU_NSGFSR, SMCF, 1, 2)
    DEP_FIELD(SMMU_NSGFSR, USF, 1, 1)
    DEP_FIELD(SMMU_NSGFSR, ICF, 1, 0)
DEP_REG32(SMMU_NSGFSRRESTORE, 0x44c)
    DEP_FIELD(SMMU_NSGFSRRESTORE, MULTI, 1, 31)
    DEP_FIELD(SMMU_NSGFSRRESTORE, UUT, 1, 8)
    DEP_FIELD(SMMU_NSGFSRRESTORE, EF, 1, 6)
    DEP_FIELD(SMMU_NSGFSRRESTORE, CAF, 1, 5)
    DEP_FIELD(SMMU_NSGFSRRESTORE, UCIF, 1, 4)
    DEP_FIELD(SMMU_NSGFSRRESTORE, UCBF, 1, 3)
    DEP_FIELD(SMMU_NSGFSRRESTORE, SMCF, 1, 2)
    DEP_FIELD(SMMU_NSGFSRRESTORE, USF, 1, 1)
    DEP_FIELD(SMMU_NSGFSRRESTORE, ICF, 1, 0)
DEP_REG32(SMMU_NSGFSYNR0, 0x450)
    DEP_FIELD(SMMU_NSGFSYNR0, ATS, 1, 6)
    DEP_FIELD(SMMU_NSGFSYNR0, IND, 1, 3)
    DEP_FIELD(SMMU_NSGFSYNR0, PNU, 1, 2)
    DEP_FIELD(SMMU_NSGFSYNR0, WNR, 1, 1)
    DEP_FIELD(SMMU_NSGFSYNR0, NESTED, 1, 0)
DEP_REG32(SMMU_NSGFSYNDR1, 0x454)
    DEP_FIELD(SMMU_NSGFSYNDR1, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_NSGFSYNDR1, STREAMID, 15, 0)
DEP_REG32(SMMU_NSTLBGSYNC, 0x470)
DEP_REG32(SMMU_NSTLBGSTATUS, 0x474)
    DEP_FIELD(SMMU_NSTLBGSTATUS, GSACTIVE, 1, 0)
DEP_REG32(SMMU_SMR0, 0x800)
    DEP_FIELD(SMMU_SMR0, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR0, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR0, ID, 15, 0)
DEP_REG32(SMMU_SMR1, 0x804)
    DEP_FIELD(SMMU_SMR1, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR1, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR1, ID, 15, 0)
DEP_REG32(SMMU_SMR2, 0x808)
    DEP_FIELD(SMMU_SMR2, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR2, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR2, ID, 15, 0)
DEP_REG32(SMMU_SMR3, 0x80c)
    DEP_FIELD(SMMU_SMR3, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR3, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR3, ID, 15, 0)
DEP_REG32(SMMU_SMR4, 0x810)
    DEP_FIELD(SMMU_SMR4, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR4, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR4, ID, 15, 0)
DEP_REG32(SMMU_SMR5, 0x814)
    DEP_FIELD(SMMU_SMR5, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR5, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR5, ID, 15, 0)
DEP_REG32(SMMU_SMR6, 0x818)
    DEP_FIELD(SMMU_SMR6, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR6, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR6, ID, 15, 0)
DEP_REG32(SMMU_SMR7, 0x81c)
    DEP_FIELD(SMMU_SMR7, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR7, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR7, ID, 15, 0)
DEP_REG32(SMMU_SMR8, 0x820)
    DEP_FIELD(SMMU_SMR8, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR8, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR8, ID, 15, 0)
DEP_REG32(SMMU_SMR9, 0x824)
    DEP_FIELD(SMMU_SMR9, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR9, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR9, ID, 15, 0)
DEP_REG32(SMMU_SMR10, 0x828)
    DEP_FIELD(SMMU_SMR10, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR10, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR10, ID, 15, 0)
DEP_REG32(SMMU_SMR11, 0x82c)
    DEP_FIELD(SMMU_SMR11, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR11, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR11, ID, 15, 0)
DEP_REG32(SMMU_SMR12, 0x830)
    DEP_FIELD(SMMU_SMR12, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR12, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR12, ID, 15, 0)
DEP_REG32(SMMU_SMR13, 0x834)
    DEP_FIELD(SMMU_SMR13, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR13, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR13, ID, 15, 0)
DEP_REG32(SMMU_SMR14, 0x838)
    DEP_FIELD(SMMU_SMR14, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR14, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR14, ID, 15, 0)
DEP_REG32(SMMU_SMR15, 0x83c)
    DEP_FIELD(SMMU_SMR15, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR15, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR15, ID, 15, 0)
DEP_REG32(SMMU_SMR16, 0x840)
    DEP_FIELD(SMMU_SMR16, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR16, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR16, ID, 15, 0)
DEP_REG32(SMMU_SMR17, 0x844)
    DEP_FIELD(SMMU_SMR17, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR17, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR17, ID, 15, 0)
DEP_REG32(SMMU_SMR18, 0x848)
    DEP_FIELD(SMMU_SMR18, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR18, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR18, ID, 15, 0)
DEP_REG32(SMMU_SMR19, 0x84c)
    DEP_FIELD(SMMU_SMR19, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR19, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR19, ID, 15, 0)
DEP_REG32(SMMU_SMR20, 0x850)
    DEP_FIELD(SMMU_SMR20, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR20, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR20, ID, 15, 0)
DEP_REG32(SMMU_SMR21, 0x854)
    DEP_FIELD(SMMU_SMR21, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR21, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR21, ID, 15, 0)
DEP_REG32(SMMU_SMR22, 0x858)
    DEP_FIELD(SMMU_SMR22, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR22, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR22, ID, 15, 0)
DEP_REG32(SMMU_SMR23, 0x85c)
    DEP_FIELD(SMMU_SMR23, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR23, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR23, ID, 15, 0)
DEP_REG32(SMMU_SMR24, 0x860)
    DEP_FIELD(SMMU_SMR24, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR24, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR24, ID, 15, 0)
DEP_REG32(SMMU_SMR25, 0x864)
    DEP_FIELD(SMMU_SMR25, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR25, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR25, ID, 15, 0)
DEP_REG32(SMMU_SMR26, 0x868)
    DEP_FIELD(SMMU_SMR26, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR26, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR26, ID, 15, 0)
DEP_REG32(SMMU_SMR27, 0x86c)
    DEP_FIELD(SMMU_SMR27, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR27, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR27, ID, 15, 0)
DEP_REG32(SMMU_SMR28, 0x870)
    DEP_FIELD(SMMU_SMR28, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR28, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR28, ID, 15, 0)
DEP_REG32(SMMU_SMR29, 0x874)
    DEP_FIELD(SMMU_SMR29, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR29, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR29, ID, 15, 0)
DEP_REG32(SMMU_SMR30, 0x878)
    DEP_FIELD(SMMU_SMR30, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR30, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR30, ID, 15, 0)
DEP_REG32(SMMU_SMR31, 0x87c)
    DEP_FIELD(SMMU_SMR31, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR31, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR31, ID, 15, 0)
DEP_REG32(SMMU_SMR32, 0x880)
    DEP_FIELD(SMMU_SMR32, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR32, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR32, ID, 15, 0)
DEP_REG32(SMMU_SMR33, 0x884)
    DEP_FIELD(SMMU_SMR33, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR33, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR33, ID, 15, 0)
DEP_REG32(SMMU_SMR34, 0x888)
    DEP_FIELD(SMMU_SMR34, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR34, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR34, ID, 15, 0)
DEP_REG32(SMMU_SMR35, 0x88c)
    DEP_FIELD(SMMU_SMR35, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR35, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR35, ID, 15, 0)
DEP_REG32(SMMU_SMR36, 0x890)
    DEP_FIELD(SMMU_SMR36, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR36, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR36, ID, 15, 0)
DEP_REG32(SMMU_SMR37, 0x894)
    DEP_FIELD(SMMU_SMR37, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR37, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR37, ID, 15, 0)
DEP_REG32(SMMU_SMR38, 0x898)
    DEP_FIELD(SMMU_SMR38, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR38, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR38, ID, 15, 0)
DEP_REG32(SMMU_SMR39, 0x89c)
    DEP_FIELD(SMMU_SMR39, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR39, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR39, ID, 15, 0)
DEP_REG32(SMMU_SMR40, 0x8a0)
    DEP_FIELD(SMMU_SMR40, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR40, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR40, ID, 15, 0)
DEP_REG32(SMMU_SMR41, 0x8a4)
    DEP_FIELD(SMMU_SMR41, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR41, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR41, ID, 15, 0)
DEP_REG32(SMMU_SMR42, 0x8a8)
    DEP_FIELD(SMMU_SMR42, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR42, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR42, ID, 15, 0)
DEP_REG32(SMMU_SMR43, 0x8ac)
    DEP_FIELD(SMMU_SMR43, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR43, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR43, ID, 15, 0)
DEP_REG32(SMMU_SMR44, 0x8b0)
    DEP_FIELD(SMMU_SMR44, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR44, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR44, ID, 15, 0)
DEP_REG32(SMMU_SMR45, 0x8b4)
    DEP_FIELD(SMMU_SMR45, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR45, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR45, ID, 15, 0)
DEP_REG32(SMMU_SMR46, 0x8b8)
    DEP_FIELD(SMMU_SMR46, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR46, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR46, ID, 15, 0)
DEP_REG32(SMMU_SMR47, 0x8bc)
    DEP_FIELD(SMMU_SMR47, VALID, 1, 31)
    DEP_FIELD(SMMU_SMR47, MASK, 15, 16)
    DEP_FIELD(SMMU_SMR47, ID, 15, 0)
DEP_REG32(SMMU_S2CR0, 0xc00)
    DEP_FIELD(SMMU_S2CR0, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR0, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR0, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR0, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR0, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR0, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR0, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR0, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR0, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR0, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR0, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR0, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR1, 0xc04)
    DEP_FIELD(SMMU_S2CR1, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR1, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR1, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR1, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR1, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR1, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR1, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR1, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR1, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR1, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR1, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR1, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR2, 0xc08)
    DEP_FIELD(SMMU_S2CR2, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR2, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR2, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR2, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR2, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR2, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR2, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR2, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR2, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR2, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR2, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR2, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR3, 0xc0c)
    DEP_FIELD(SMMU_S2CR3, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR3, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR3, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR3, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR3, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR3, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR3, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR3, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR3, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR3, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR3, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR3, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR4, 0xc10)
    DEP_FIELD(SMMU_S2CR4, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR4, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR4, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR4, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR4, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR4, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR4, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR4, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR4, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR4, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR4, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR4, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR5, 0xc14)
    DEP_FIELD(SMMU_S2CR5, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR5, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR5, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR5, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR5, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR5, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR5, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR5, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR5, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR5, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR5, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR5, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR6, 0xc18)
    DEP_FIELD(SMMU_S2CR6, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR6, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR6, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR6, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR6, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR6, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR6, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR6, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR6, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR6, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR6, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR6, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR7, 0xc1c)
    DEP_FIELD(SMMU_S2CR7, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR7, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR7, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR7, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR7, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR7, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR7, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR7, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR7, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR7, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR7, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR7, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR8, 0xc20)
    DEP_FIELD(SMMU_S2CR8, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR8, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR8, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR8, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR8, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR8, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR8, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR8, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR8, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR8, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR8, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR8, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR9, 0xc24)
    DEP_FIELD(SMMU_S2CR9, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR9, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR9, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR9, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR9, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR9, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR9, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR9, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR9, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR9, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR9, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR9, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR10, 0xc28)
    DEP_FIELD(SMMU_S2CR10, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR10, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR10, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR10, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR10, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR10, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR10, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR10, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR10, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR10, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR10, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR10, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR11, 0xc2c)
    DEP_FIELD(SMMU_S2CR11, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR11, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR11, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR11, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR11, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR11, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR11, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR11, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR11, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR11, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR11, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR11, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR12, 0xc30)
    DEP_FIELD(SMMU_S2CR12, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR12, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR12, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR12, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR12, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR12, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR12, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR12, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR12, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR12, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR12, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR12, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR13, 0xc34)
    DEP_FIELD(SMMU_S2CR13, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR13, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR13, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR13, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR13, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR13, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR13, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR13, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR13, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR13, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR13, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR13, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR14, 0xc38)
    DEP_FIELD(SMMU_S2CR14, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR14, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR14, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR14, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR14, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR14, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR14, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR14, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR14, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR14, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR14, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR14, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR15, 0xc3c)
    DEP_FIELD(SMMU_S2CR15, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR15, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR15, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR15, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR15, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR15, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR15, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR15, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR15, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR15, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR15, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR15, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR16, 0xc40)
    DEP_FIELD(SMMU_S2CR16, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR16, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR16, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR16, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR16, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR16, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR16, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR16, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR16, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR16, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR16, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR16, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR17, 0xc44)
    DEP_FIELD(SMMU_S2CR17, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR17, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR17, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR17, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR17, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR17, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR17, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR17, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR17, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR17, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR17, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR17, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR18, 0xc48)
    DEP_FIELD(SMMU_S2CR18, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR18, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR18, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR18, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR18, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR18, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR18, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR18, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR18, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR18, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR18, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR18, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR19, 0xc4c)
    DEP_FIELD(SMMU_S2CR19, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR19, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR19, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR19, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR19, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR19, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR19, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR19, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR19, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR19, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR19, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR19, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR20, 0xc50)
    DEP_FIELD(SMMU_S2CR20, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR20, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR20, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR20, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR20, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR20, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR20, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR20, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR20, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR20, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR20, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR20, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR21, 0xc54)
    DEP_FIELD(SMMU_S2CR21, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR21, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR21, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR21, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR21, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR21, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR21, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR21, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR21, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR21, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR21, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR21, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR22, 0xc58)
    DEP_FIELD(SMMU_S2CR22, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR22, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR22, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR22, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR22, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR22, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR22, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR22, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR22, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR22, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR22, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR22, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR23, 0xc5c)
    DEP_FIELD(SMMU_S2CR23, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR23, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR23, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR23, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR23, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR23, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR23, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR23, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR23, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR23, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR23, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR23, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR24, 0xc60)
    DEP_FIELD(SMMU_S2CR24, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR24, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR24, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR24, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR24, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR24, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR24, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR24, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR24, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR24, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR24, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR24, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR25, 0xc64)
    DEP_FIELD(SMMU_S2CR25, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR25, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR25, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR25, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR25, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR25, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR25, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR25, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR25, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR25, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR25, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR25, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR26, 0xc68)
    DEP_FIELD(SMMU_S2CR26, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR26, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR26, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR26, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR26, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR26, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR26, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR26, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR26, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR26, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR26, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR26, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR27, 0xc6c)
    DEP_FIELD(SMMU_S2CR27, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR27, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR27, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR27, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR27, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR27, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR27, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR27, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR27, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR27, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR27, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR27, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR28, 0xc70)
    DEP_FIELD(SMMU_S2CR28, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR28, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR28, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR28, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR28, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR28, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR28, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR28, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR28, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR28, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR28, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR28, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR29, 0xc74)
    DEP_FIELD(SMMU_S2CR29, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR29, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR29, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR29, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR29, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR29, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR29, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR29, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR29, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR29, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR29, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR29, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR30, 0xc78)
    DEP_FIELD(SMMU_S2CR30, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR30, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR30, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR30, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR30, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR30, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR30, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR30, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR30, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR30, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR30, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR30, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR31, 0xc7c)
    DEP_FIELD(SMMU_S2CR31, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR31, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR31, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR31, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR31, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR31, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR31, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR31, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR31, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR31, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR31, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR31, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR32, 0xc80)
    DEP_FIELD(SMMU_S2CR32, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR32, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR32, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR32, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR32, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR32, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR32, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR32, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR32, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR32, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR32, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR32, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR33, 0xc84)
    DEP_FIELD(SMMU_S2CR33, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR33, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR33, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR33, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR33, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR33, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR33, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR33, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR33, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR33, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR33, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR33, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR34, 0xc88)
    DEP_FIELD(SMMU_S2CR34, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR34, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR34, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR34, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR34, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR34, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR34, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR34, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR34, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR34, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR34, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR34, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR35, 0xc8c)
    DEP_FIELD(SMMU_S2CR35, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR35, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR35, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR35, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR35, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR35, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR35, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR35, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR35, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR35, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR35, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR35, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR36, 0xc90)
    DEP_FIELD(SMMU_S2CR36, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR36, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR36, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR36, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR36, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR36, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR36, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR36, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR36, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR36, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR36, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR36, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR37, 0xc94)
    DEP_FIELD(SMMU_S2CR37, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR37, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR37, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR37, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR37, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR37, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR37, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR37, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR37, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR37, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR37, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR37, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR38, 0xc98)
    DEP_FIELD(SMMU_S2CR38, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR38, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR38, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR38, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR38, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR38, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR38, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR38, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR38, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR38, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR38, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR38, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR39, 0xc9c)
    DEP_FIELD(SMMU_S2CR39, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR39, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR39, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR39, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR39, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR39, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR39, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR39, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR39, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR39, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR39, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR39, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR40, 0xca0)
    DEP_FIELD(SMMU_S2CR40, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR40, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR40, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR40, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR40, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR40, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR40, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR40, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR40, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR40, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR40, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR40, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR41, 0xca4)
    DEP_FIELD(SMMU_S2CR41, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR41, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR41, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR41, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR41, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR41, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR41, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR41, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR41, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR41, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR41, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR41, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR42, 0xca8)
    DEP_FIELD(SMMU_S2CR42, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR42, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR42, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR42, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR42, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR42, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR42, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR42, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR42, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR42, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR42, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR42, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR43, 0xcac)
    DEP_FIELD(SMMU_S2CR43, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR43, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR43, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR43, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR43, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR43, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR43, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR43, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR43, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR43, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR43, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR43, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR44, 0xcb0)
    DEP_FIELD(SMMU_S2CR44, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR44, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR44, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR44, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR44, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR44, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR44, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR44, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR44, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR44, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR44, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR44, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR45, 0xcb4)
    DEP_FIELD(SMMU_S2CR45, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR45, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR45, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR45, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR45, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR45, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR45, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR45, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR45, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR45, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR45, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR45, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR46, 0xcb8)
    DEP_FIELD(SMMU_S2CR46, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR46, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR46, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR46, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR46, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR46, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR46, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR46, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR46, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR46, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR46, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR46, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_S2CR47, 0xcbc)
    DEP_FIELD(SMMU_S2CR47, TRANSIENTCFG, 2, 28)
    DEP_FIELD(SMMU_S2CR47, INSTCFG_1, 1, 27)
    DEP_FIELD(SMMU_S2CR47, INSTCFG_0_FB, 1, 26)
    DEP_FIELD(SMMU_S2CR47, PRIVCFG_BSU, 2, 24)
    DEP_FIELD(SMMU_S2CR47, WACFG, 2, 22)
    DEP_FIELD(SMMU_S2CR47, RACFG, 2, 20)
    DEP_FIELD(SMMU_S2CR47, NSCFG, 2, 18)
    DEP_FIELD(SMMU_S2CR47, TYPE, 2, 16)
    DEP_FIELD(SMMU_S2CR47, MEM_ATTR, 4, 12)
    DEP_FIELD(SMMU_S2CR47, MTCFG, 1, 11)
    DEP_FIELD(SMMU_S2CR47, SHCFG, 2, 8)
    DEP_FIELD(SMMU_S2CR47, CBNDX_VMID, 8, 0)
DEP_REG32(SMMU_PIDR4, 0xfd0)
    DEP_FIELD(SMMU_PIDR4, FOURKB_COUNT, 4, 4)
    DEP_FIELD(SMMU_PIDR4, JEP106_CONTINUATION_CODE, 4, 0)
DEP_REG32(SMMU_PIDR5, 0xfd4)
DEP_REG32(SMMU_PIDR6, 0xfd8)
DEP_REG32(SMMU_PIDR7, 0xfdc)
DEP_REG32(SMMU_PIDR0, 0xfe0)
    DEP_FIELD(SMMU_PIDR0, PARTNUMBER0, 8, 0)
DEP_REG32(SMMU_PIDR1, 0xfe4)
    DEP_FIELD(SMMU_PIDR1, JEP106_IDENTITY_CODE, 4, 4)
    DEP_FIELD(SMMU_PIDR1, PARTNUMBER1, 4, 0)
DEP_REG32(SMMU_PIDR2, 0xfe8)
    DEP_FIELD(SMMU_PIDR2, ARCHITECTURE_REVISION, 4, 4)
    DEP_FIELD(SMMU_PIDR2, JEDEC, 1, 3)
    DEP_FIELD(SMMU_PIDR2, JEP106_IDENTITY_CODE, 3, 0)
DEP_REG32(SMMU_PIDR3, 0xfec)
    DEP_FIELD(SMMU_PIDR3, REVAND, 4, 4)
    DEP_FIELD(SMMU_PIDR3, CUSTOMER_MODIFIED, 4, 0)
DEP_REG32(SMMU_CIDR0, 0xff0)
    DEP_FIELD(SMMU_CIDR0, PREAMBLE, 8, 0)
DEP_REG32(SMMU_CIDR1, 0xff4)
    DEP_FIELD(SMMU_CIDR1, PREAMBLE, 8, 0)
DEP_REG32(SMMU_CIDR2, 0xff8)
    DEP_FIELD(SMMU_CIDR2, PREAMBLE, 8, 0)
DEP_REG32(SMMU_CIDR3, 0xffc)
    DEP_FIELD(SMMU_CIDR3, PREAMBLE, 8, 0)
DEP_REG32(SMMU_CBAR0, 0x1000)
    DEP_FIELD(SMMU_CBAR0, IRPTNDX, 8, 24)
    DEP_FIELD(SMMU_CBAR0, WACFG, 2, 22)
    DEP_FIELD(SMMU_CBAR0, RACFG, 2, 20)
    DEP_FIELD(SMMU_CBAR0, BSU, 2, 18)
    DEP_FIELD(SMMU_CBAR0, TYPE, 2, 16)
    DEP_FIELD(SMMU_CBAR0, MEMATTR_CBNDX_7_4, 4, 12)
    DEP_FIELD(SMMU_CBAR0, FB_CBNDX_3, 1, 11)
    DEP_FIELD(SMMU_CBAR0, HYPC_CBNDX_2, 1, 10)
    DEP_FIELD(SMMU_CBAR0, BPSHCFG_CBNDX_1_0, 2, 8)
    DEP_FIELD(SMMU_CBAR0, VMID, 8, 0)
DEP_REG32(SMMU_CBAR1, 0x1004)
    DEP_FIELD(SMMU_CBAR1, IRPTNDX, 8, 24)
    DEP_FIELD(SMMU_CBAR1, WACFG, 2, 22)
    DEP_FIELD(SMMU_CBAR1, RACFG, 2, 20)
    DEP_FIELD(SMMU_CBAR1, BSU, 2, 18)
    DEP_FIELD(SMMU_CBAR1, TYPE, 2, 16)
    DEP_FIELD(SMMU_CBAR1, MEMATTR_CBNDX_7_4, 4, 12)
    DEP_FIELD(SMMU_CBAR1, FB_CBNDX_3, 1, 11)
    DEP_FIELD(SMMU_CBAR1, HYPC_CBNDX_2, 1, 10)
    DEP_FIELD(SMMU_CBAR1, BPSHCFG_CBNDX_1_0, 2, 8)
    DEP_FIELD(SMMU_CBAR1, VMID, 8, 0)
DEP_REG32(SMMU_CBAR2, 0x1008)
    DEP_FIELD(SMMU_CBAR2, IRPTNDX, 8, 24)
    DEP_FIELD(SMMU_CBAR2, WACFG, 2, 22)
    DEP_FIELD(SMMU_CBAR2, RACFG, 2, 20)
    DEP_FIELD(SMMU_CBAR2, BSU, 2, 18)
    DEP_FIELD(SMMU_CBAR2, TYPE, 2, 16)
    DEP_FIELD(SMMU_CBAR2, MEMATTR_CBNDX_7_4, 4, 12)
    DEP_FIELD(SMMU_CBAR2, FB_CBNDX_3, 1, 11)
    DEP_FIELD(SMMU_CBAR2, HYPC_CBNDX_2, 1, 10)
    DEP_FIELD(SMMU_CBAR2, BPSHCFG_CBNDX_1_0, 2, 8)
    DEP_FIELD(SMMU_CBAR2, VMID, 8, 0)
DEP_REG32(SMMU_CBAR3, 0x100c)
    DEP_FIELD(SMMU_CBAR3, IRPTNDX, 8, 24)
    DEP_FIELD(SMMU_CBAR3, WACFG, 2, 22)
    DEP_FIELD(SMMU_CBAR3, RACFG, 2, 20)
    DEP_FIELD(SMMU_CBAR3, BSU, 2, 18)
    DEP_FIELD(SMMU_CBAR3, TYPE, 2, 16)
    DEP_FIELD(SMMU_CBAR3, MEMATTR_CBNDX_7_4, 4, 12)
    DEP_FIELD(SMMU_CBAR3, FB_CBNDX_3, 1, 11)
    DEP_FIELD(SMMU_CBAR3, HYPC_CBNDX_2, 1, 10)
    DEP_FIELD(SMMU_CBAR3, BPSHCFG_CBNDX_1_0, 2, 8)
    DEP_FIELD(SMMU_CBAR3, VMID, 8, 0)
DEP_REG32(SMMU_CBAR4, 0x1010)
    DEP_FIELD(SMMU_CBAR4, IRPTNDX, 8, 24)
    DEP_FIELD(SMMU_CBAR4, WACFG, 2, 22)
    DEP_FIELD(SMMU_CBAR4, RACFG, 2, 20)
    DEP_FIELD(SMMU_CBAR4, BSU, 2, 18)
    DEP_FIELD(SMMU_CBAR4, TYPE, 2, 16)
    DEP_FIELD(SMMU_CBAR4, MEMATTR_CBNDX_7_4, 4, 12)
    DEP_FIELD(SMMU_CBAR4, FB_CBNDX_3, 1, 11)
    DEP_FIELD(SMMU_CBAR4, HYPC_CBNDX_2, 1, 10)
    DEP_FIELD(SMMU_CBAR4, BPSHCFG_CBNDX_1_0, 2, 8)
    DEP_FIELD(SMMU_CBAR4, VMID, 8, 0)
DEP_REG32(SMMU_CBAR5, 0x1014)
    DEP_FIELD(SMMU_CBAR5, IRPTNDX, 8, 24)
    DEP_FIELD(SMMU_CBAR5, WACFG, 2, 22)
    DEP_FIELD(SMMU_CBAR5, RACFG, 2, 20)
    DEP_FIELD(SMMU_CBAR5, BSU, 2, 18)
    DEP_FIELD(SMMU_CBAR5, TYPE, 2, 16)
    DEP_FIELD(SMMU_CBAR5, MEMATTR_CBNDX_7_4, 4, 12)
    DEP_FIELD(SMMU_CBAR5, FB_CBNDX_3, 1, 11)
    DEP_FIELD(SMMU_CBAR5, HYPC_CBNDX_2, 1, 10)
    DEP_FIELD(SMMU_CBAR5, BPSHCFG_CBNDX_1_0, 2, 8)
    DEP_FIELD(SMMU_CBAR5, VMID, 8, 0)
DEP_REG32(SMMU_CBAR6, 0x1018)
    DEP_FIELD(SMMU_CBAR6, IRPTNDX, 8, 24)
    DEP_FIELD(SMMU_CBAR6, WACFG, 2, 22)
    DEP_FIELD(SMMU_CBAR6, RACFG, 2, 20)
    DEP_FIELD(SMMU_CBAR6, BSU, 2, 18)
    DEP_FIELD(SMMU_CBAR6, TYPE, 2, 16)
    DEP_FIELD(SMMU_CBAR6, MEMATTR_CBNDX_7_4, 4, 12)
    DEP_FIELD(SMMU_CBAR6, FB_CBNDX_3, 1, 11)
    DEP_FIELD(SMMU_CBAR6, HYPC_CBNDX_2, 1, 10)
    DEP_FIELD(SMMU_CBAR6, BPSHCFG_CBNDX_1_0, 2, 8)
    DEP_FIELD(SMMU_CBAR6, VMID, 8, 0)
DEP_REG32(SMMU_CBAR7, 0x101c)
    DEP_FIELD(SMMU_CBAR7, IRPTNDX, 8, 24)
    DEP_FIELD(SMMU_CBAR7, WACFG, 2, 22)
    DEP_FIELD(SMMU_CBAR7, RACFG, 2, 20)
    DEP_FIELD(SMMU_CBAR7, BSU, 2, 18)
    DEP_FIELD(SMMU_CBAR7, TYPE, 2, 16)
    DEP_FIELD(SMMU_CBAR7, MEMATTR_CBNDX_7_4, 4, 12)
    DEP_FIELD(SMMU_CBAR7, FB_CBNDX_3, 1, 11)
    DEP_FIELD(SMMU_CBAR7, HYPC_CBNDX_2, 1, 10)
    DEP_FIELD(SMMU_CBAR7, BPSHCFG_CBNDX_1_0, 2, 8)
    DEP_FIELD(SMMU_CBAR7, VMID, 8, 0)
DEP_REG32(SMMU_CBAR8, 0x1020)
    DEP_FIELD(SMMU_CBAR8, IRPTNDX, 8, 24)
    DEP_FIELD(SMMU_CBAR8, WACFG, 2, 22)
    DEP_FIELD(SMMU_CBAR8, RACFG, 2, 20)
    DEP_FIELD(SMMU_CBAR8, BSU, 2, 18)
    DEP_FIELD(SMMU_CBAR8, TYPE, 2, 16)
    DEP_FIELD(SMMU_CBAR8, MEMATTR_CBNDX_7_4, 4, 12)
    DEP_FIELD(SMMU_CBAR8, FB_CBNDX_3, 1, 11)
    DEP_FIELD(SMMU_CBAR8, HYPC_CBNDX_2, 1, 10)
    DEP_FIELD(SMMU_CBAR8, BPSHCFG_CBNDX_1_0, 2, 8)
    DEP_FIELD(SMMU_CBAR8, VMID, 8, 0)
DEP_REG32(SMMU_CBAR9, 0x1024)
    DEP_FIELD(SMMU_CBAR9, IRPTNDX, 8, 24)
    DEP_FIELD(SMMU_CBAR9, WACFG, 2, 22)
    DEP_FIELD(SMMU_CBAR9, RACFG, 2, 20)
    DEP_FIELD(SMMU_CBAR9, BSU, 2, 18)
    DEP_FIELD(SMMU_CBAR9, TYPE, 2, 16)
    DEP_FIELD(SMMU_CBAR9, MEMATTR_CBNDX_7_4, 4, 12)
    DEP_FIELD(SMMU_CBAR9, FB_CBNDX_3, 1, 11)
    DEP_FIELD(SMMU_CBAR9, HYPC_CBNDX_2, 1, 10)
    DEP_FIELD(SMMU_CBAR9, BPSHCFG_CBNDX_1_0, 2, 8)
    DEP_FIELD(SMMU_CBAR9, VMID, 8, 0)
DEP_REG32(SMMU_CBAR10, 0x1028)
    DEP_FIELD(SMMU_CBAR10, IRPTNDX, 8, 24)
    DEP_FIELD(SMMU_CBAR10, WACFG, 2, 22)
    DEP_FIELD(SMMU_CBAR10, RACFG, 2, 20)
    DEP_FIELD(SMMU_CBAR10, BSU, 2, 18)
    DEP_FIELD(SMMU_CBAR10, TYPE, 2, 16)
    DEP_FIELD(SMMU_CBAR10, MEMATTR_CBNDX_7_4, 4, 12)
    DEP_FIELD(SMMU_CBAR10, FB_CBNDX_3, 1, 11)
    DEP_FIELD(SMMU_CBAR10, HYPC_CBNDX_2, 1, 10)
    DEP_FIELD(SMMU_CBAR10, BPSHCFG_CBNDX_1_0, 2, 8)
    DEP_FIELD(SMMU_CBAR10, VMID, 8, 0)
DEP_REG32(SMMU_CBAR11, 0x102c)
    DEP_FIELD(SMMU_CBAR11, IRPTNDX, 8, 24)
    DEP_FIELD(SMMU_CBAR11, WACFG, 2, 22)
    DEP_FIELD(SMMU_CBAR11, RACFG, 2, 20)
    DEP_FIELD(SMMU_CBAR11, BSU, 2, 18)
    DEP_FIELD(SMMU_CBAR11, TYPE, 2, 16)
    DEP_FIELD(SMMU_CBAR11, MEMATTR_CBNDX_7_4, 4, 12)
    DEP_FIELD(SMMU_CBAR11, FB_CBNDX_3, 1, 11)
    DEP_FIELD(SMMU_CBAR11, HYPC_CBNDX_2, 1, 10)
    DEP_FIELD(SMMU_CBAR11, BPSHCFG_CBNDX_1_0, 2, 8)
    DEP_FIELD(SMMU_CBAR11, VMID, 8, 0)
DEP_REG32(SMMU_CBAR12, 0x1030)
    DEP_FIELD(SMMU_CBAR12, IRPTNDX, 8, 24)
    DEP_FIELD(SMMU_CBAR12, WACFG, 2, 22)
    DEP_FIELD(SMMU_CBAR12, RACFG, 2, 20)
    DEP_FIELD(SMMU_CBAR12, BSU, 2, 18)
    DEP_FIELD(SMMU_CBAR12, TYPE, 2, 16)
    DEP_FIELD(SMMU_CBAR12, MEMATTR_CBNDX_7_4, 4, 12)
    DEP_FIELD(SMMU_CBAR12, FB_CBNDX_3, 1, 11)
    DEP_FIELD(SMMU_CBAR12, HYPC_CBNDX_2, 1, 10)
    DEP_FIELD(SMMU_CBAR12, BPSHCFG_CBNDX_1_0, 2, 8)
    DEP_FIELD(SMMU_CBAR12, VMID, 8, 0)
DEP_REG32(SMMU_CBAR13, 0x1034)
    DEP_FIELD(SMMU_CBAR13, IRPTNDX, 8, 24)
    DEP_FIELD(SMMU_CBAR13, WACFG, 2, 22)
    DEP_FIELD(SMMU_CBAR13, RACFG, 2, 20)
    DEP_FIELD(SMMU_CBAR13, BSU, 2, 18)
    DEP_FIELD(SMMU_CBAR13, TYPE, 2, 16)
    DEP_FIELD(SMMU_CBAR13, MEMATTR_CBNDX_7_4, 4, 12)
    DEP_FIELD(SMMU_CBAR13, FB_CBNDX_3, 1, 11)
    DEP_FIELD(SMMU_CBAR13, HYPC_CBNDX_2, 1, 10)
    DEP_FIELD(SMMU_CBAR13, BPSHCFG_CBNDX_1_0, 2, 8)
    DEP_FIELD(SMMU_CBAR13, VMID, 8, 0)
DEP_REG32(SMMU_CBAR14, 0x1038)
    DEP_FIELD(SMMU_CBAR14, IRPTNDX, 8, 24)
    DEP_FIELD(SMMU_CBAR14, WACFG, 2, 22)
    DEP_FIELD(SMMU_CBAR14, RACFG, 2, 20)
    DEP_FIELD(SMMU_CBAR14, BSU, 2, 18)
    DEP_FIELD(SMMU_CBAR14, TYPE, 2, 16)
    DEP_FIELD(SMMU_CBAR14, MEMATTR_CBNDX_7_4, 4, 12)
    DEP_FIELD(SMMU_CBAR14, FB_CBNDX_3, 1, 11)
    DEP_FIELD(SMMU_CBAR14, HYPC_CBNDX_2, 1, 10)
    DEP_FIELD(SMMU_CBAR14, BPSHCFG_CBNDX_1_0, 2, 8)
    DEP_FIELD(SMMU_CBAR14, VMID, 8, 0)
DEP_REG32(SMMU_CBAR15, 0x103c)
    DEP_FIELD(SMMU_CBAR15, IRPTNDX, 8, 24)
    DEP_FIELD(SMMU_CBAR15, WACFG, 2, 22)
    DEP_FIELD(SMMU_CBAR15, RACFG, 2, 20)
    DEP_FIELD(SMMU_CBAR15, BSU, 2, 18)
    DEP_FIELD(SMMU_CBAR15, TYPE, 2, 16)
    DEP_FIELD(SMMU_CBAR15, MEMATTR_CBNDX_7_4, 4, 12)
    DEP_FIELD(SMMU_CBAR15, FB_CBNDX_3, 1, 11)
    DEP_FIELD(SMMU_CBAR15, HYPC_CBNDX_2, 1, 10)
    DEP_FIELD(SMMU_CBAR15, BPSHCFG_CBNDX_1_0, 2, 8)
    DEP_FIELD(SMMU_CBAR15, VMID, 8, 0)
DEP_REG32(SMMU_CBFRSYNRA0, 0x1400)
    DEP_FIELD(SMMU_CBFRSYNRA0, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_CBFRSYNRA0, STREAMID, 15, 0)
DEP_REG32(SMMU_CBFRSYNRA1, 0x1404)
    DEP_FIELD(SMMU_CBFRSYNRA1, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_CBFRSYNRA1, STREAMID, 15, 0)
DEP_REG32(SMMU_CBFRSYNRA2, 0x1408)
    DEP_FIELD(SMMU_CBFRSYNRA2, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_CBFRSYNRA2, STREAMID, 15, 0)
DEP_REG32(SMMU_CBFRSYNRA3, 0x140c)
    DEP_FIELD(SMMU_CBFRSYNRA3, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_CBFRSYNRA3, STREAMID, 15, 0)
DEP_REG32(SMMU_CBFRSYNRA4, 0x1410)
    DEP_FIELD(SMMU_CBFRSYNRA4, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_CBFRSYNRA4, STREAMID, 15, 0)
DEP_REG32(SMMU_CBFRSYNRA5, 0x1414)
    DEP_FIELD(SMMU_CBFRSYNRA5, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_CBFRSYNRA5, STREAMID, 15, 0)
DEP_REG32(SMMU_CBFRSYNRA6, 0x1418)
    DEP_FIELD(SMMU_CBFRSYNRA6, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_CBFRSYNRA6, STREAMID, 15, 0)
DEP_REG32(SMMU_CBFRSYNRA7, 0x141c)
    DEP_FIELD(SMMU_CBFRSYNRA7, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_CBFRSYNRA7, STREAMID, 15, 0)
DEP_REG32(SMMU_CBFRSYNRA8, 0x1420)
    DEP_FIELD(SMMU_CBFRSYNRA8, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_CBFRSYNRA8, STREAMID, 15, 0)
DEP_REG32(SMMU_CBFRSYNRA9, 0x1424)
    DEP_FIELD(SMMU_CBFRSYNRA9, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_CBFRSYNRA9, STREAMID, 15, 0)
DEP_REG32(SMMU_CBFRSYNRA10, 0x1428)
    DEP_FIELD(SMMU_CBFRSYNRA10, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_CBFRSYNRA10, STREAMID, 15, 0)
DEP_REG32(SMMU_CBFRSYNRA11, 0x142c)
    DEP_FIELD(SMMU_CBFRSYNRA11, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_CBFRSYNRA11, STREAMID, 15, 0)
DEP_REG32(SMMU_CBFRSYNRA12, 0x1430)
    DEP_FIELD(SMMU_CBFRSYNRA12, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_CBFRSYNRA12, STREAMID, 15, 0)
DEP_REG32(SMMU_CBFRSYNRA13, 0x1434)
    DEP_FIELD(SMMU_CBFRSYNRA13, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_CBFRSYNRA13, STREAMID, 15, 0)
DEP_REG32(SMMU_CBFRSYNRA14, 0x1438)
    DEP_FIELD(SMMU_CBFRSYNRA14, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_CBFRSYNRA14, STREAMID, 15, 0)
DEP_REG32(SMMU_CBFRSYNRA15, 0x143c)
    DEP_FIELD(SMMU_CBFRSYNRA15, SSD_INDEX, 15, 16)
    DEP_FIELD(SMMU_CBFRSYNRA15, STREAMID, 15, 0)
DEP_REG32(SMMU_CBA2R0, 0x1800)
    DEP_FIELD(SMMU_CBA2R0, MONC, 1, 1)
    DEP_FIELD(SMMU_CBA2R0, VA64, 1, 0)
DEP_REG32(SMMU_CBA2R1, 0x1804)
    DEP_FIELD(SMMU_CBA2R1, MONC, 1, 1)
    DEP_FIELD(SMMU_CBA2R1, VA64, 1, 0)
DEP_REG32(SMMU_CBA2R2, 0x1808)
    DEP_FIELD(SMMU_CBA2R2, MONC, 1, 1)
    DEP_FIELD(SMMU_CBA2R2, VA64, 1, 0)
DEP_REG32(SMMU_CBA2R3, 0x180c)
    DEP_FIELD(SMMU_CBA2R3, MONC, 1, 1)
    DEP_FIELD(SMMU_CBA2R3, VA64, 1, 0)
DEP_REG32(SMMU_CBA2R4, 0x1810)
    DEP_FIELD(SMMU_CBA2R4, MONC, 1, 1)
    DEP_FIELD(SMMU_CBA2R4, VA64, 1, 0)
DEP_REG32(SMMU_CBA2R5, 0x1814)
    DEP_FIELD(SMMU_CBA2R5, MONC, 1, 1)
    DEP_FIELD(SMMU_CBA2R5, VA64, 1, 0)
DEP_REG32(SMMU_CBA2R6, 0x1818)
    DEP_FIELD(SMMU_CBA2R6, MONC, 1, 1)
    DEP_FIELD(SMMU_CBA2R6, VA64, 1, 0)
DEP_REG32(SMMU_CBA2R7, 0x181c)
    DEP_FIELD(SMMU_CBA2R7, MONC, 1, 1)
    DEP_FIELD(SMMU_CBA2R7, VA64, 1, 0)
DEP_REG32(SMMU_CBA2R8, 0x1820)
    DEP_FIELD(SMMU_CBA2R8, MONC, 1, 1)
    DEP_FIELD(SMMU_CBA2R8, VA64, 1, 0)
DEP_REG32(SMMU_CBA2R9, 0x1824)
    DEP_FIELD(SMMU_CBA2R9, MONC, 1, 1)
    DEP_FIELD(SMMU_CBA2R9, VA64, 1, 0)
DEP_REG32(SMMU_CBA2R10, 0x1828)
    DEP_FIELD(SMMU_CBA2R10, MONC, 1, 1)
    DEP_FIELD(SMMU_CBA2R10, VA64, 1, 0)
DEP_REG32(SMMU_CBA2R11, 0x182c)
    DEP_FIELD(SMMU_CBA2R11, MONC, 1, 1)
    DEP_FIELD(SMMU_CBA2R11, VA64, 1, 0)
DEP_REG32(SMMU_CBA2R12, 0x1830)
    DEP_FIELD(SMMU_CBA2R12, MONC, 1, 1)
    DEP_FIELD(SMMU_CBA2R12, VA64, 1, 0)
DEP_REG32(SMMU_CBA2R13, 0x1834)
    DEP_FIELD(SMMU_CBA2R13, MONC, 1, 1)
    DEP_FIELD(SMMU_CBA2R13, VA64, 1, 0)
DEP_REG32(SMMU_CBA2R14, 0x1838)
    DEP_FIELD(SMMU_CBA2R14, MONC, 1, 1)
    DEP_FIELD(SMMU_CBA2R14, VA64, 1, 0)
DEP_REG32(SMMU_CBA2R15, 0x183c)
    DEP_FIELD(SMMU_CBA2R15, MONC, 1, 1)
    DEP_FIELD(SMMU_CBA2R15, VA64, 1, 0)
DEP_REG32(SMMU_ITCTRL, 0x2000)
    DEP_FIELD(SMMU_ITCTRL, TBU_INDEX, 3, 4)
    DEP_FIELD(SMMU_ITCTRL, MODULE, 1, 3)
    DEP_FIELD(SMMU_ITCTRL, RAM_DATA, 1, 2)
    DEP_FIELD(SMMU_ITCTRL, RAM_MODE, 1, 1)
    DEP_FIELD(SMMU_ITCTRL, INTGMODE, 1, 0)
DEP_REG32(SMMU_ITIP, 0x2004)
    DEP_FIELD(SMMU_ITIP, SPINDEN, 1, 0)
DEP_REG32(SMMU_ITOP_GLBL, 0x2008)
    DEP_FIELD(SMMU_ITOP_GLBL, TCU_RAM_DATA, 4, 16)
    DEP_FIELD(SMMU_ITOP_GLBL, GLBLSF1, 1, 9)
    DEP_FIELD(SMMU_ITOP_GLBL, GLBLNSF1, 1, 1)
DEP_REG32(SMMU_ITOP_PERF_INDEX, 0x200c)
    DEP_FIELD(SMMU_ITOP_PERF_INDEX, WAY_IPA2PA_PF, 2, 30)
    DEP_FIELD(SMMU_ITOP_PERF_INDEX, IPA2PA_PF_INDEX, 7, 16)
    DEP_FIELD(SMMU_ITOP_PERF_INDEX, WAY_MTLB_WC, 2, 14)
    DEP_FIELD(SMMU_ITOP_PERF_INDEX, MTLB_WC_INDEX, 12, 0)
DEP_REG32(SMMU_ITOP_CXT0TO31_RAM0, 0x2010)
DEP_REG32(SMMU_TBUQOS0, 0x2100)
    DEP_FIELD(SMMU_TBUQOS0, QOSTBU5, 4, 20)
    DEP_FIELD(SMMU_TBUQOS0, QOSTBU4, 4, 16)
    DEP_FIELD(SMMU_TBUQOS0, QOSTBU3, 4, 12)
    DEP_FIELD(SMMU_TBUQOS0, QOSTBU2, 4, 8)
    DEP_FIELD(SMMU_TBUQOS0, QOSTBU1, 4, 4)
    DEP_FIELD(SMMU_TBUQOS0, QOSTBU0, 4, 0)
DEP_REG32(SMMU_PER, 0x2200)
    DEP_FIELD(SMMU_PER, PER_TCU, 8, 8)
    DEP_FIELD(SMMU_PER, PER_TBU, 8, 0)
DEP_REG32(SMMU_TBU_PWR_STATUS, 0x2204)
DEP_REG32(PMEVCNTR0, 0x3000)
DEP_REG32(PMEVCNTR1, 0x3004)
DEP_REG32(PMEVCNTR2, 0x3008)
DEP_REG32(PMEVCNTR3, 0x300c)
DEP_REG32(PMEVCNTR4, 0x3010)
DEP_REG32(PMEVCNTR5, 0x3014)
DEP_REG32(PMEVCNTR6, 0x3018)
DEP_REG32(PMEVCNTR7, 0x301c)
DEP_REG32(PMEVCNTR8, 0x3020)
DEP_REG32(PMEVCNTR9, 0x3024)
DEP_REG32(PMEVCNTR10, 0x3028)
DEP_REG32(PMEVCNTR11, 0x302c)
DEP_REG32(PMEVCNTR12, 0x3030)
DEP_REG32(PMEVCNTR13, 0x3034)
DEP_REG32(PMEVCNTR14, 0x3038)
DEP_REG32(PMEVCNTR15, 0x303c)
DEP_REG32(PMEVCNTR16, 0x3040)
DEP_REG32(PMEVCNTR17, 0x3044)
DEP_REG32(PMEVCNTR18, 0x3048)
DEP_REG32(PMEVCNTR19, 0x304c)
DEP_REG32(PMEVCNTR20, 0x3050)
DEP_REG32(PMEVCNTR21, 0x3054)
DEP_REG32(PMEVCNTR22, 0x3058)
DEP_REG32(PMEVCNTR23, 0x305c)
DEP_REG32(PMEVTYPER0, 0x3400)
    DEP_FIELD(PMEVTYPER0, P, 1, 31)
    DEP_FIELD(PMEVTYPER0, U, 1, 30)
    DEP_FIELD(PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(PMEVTYPER1, 0x3404)
    DEP_FIELD(PMEVTYPER1, P, 1, 31)
    DEP_FIELD(PMEVTYPER1, U, 1, 30)
    DEP_FIELD(PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(PMEVTYPER2, 0x3408)
    DEP_FIELD(PMEVTYPER2, P, 1, 31)
    DEP_FIELD(PMEVTYPER2, U, 1, 30)
    DEP_FIELD(PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(PMEVTYPER3, 0x340c)
    DEP_FIELD(PMEVTYPER3, P, 1, 31)
    DEP_FIELD(PMEVTYPER3, U, 1, 30)
    DEP_FIELD(PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(PMEVTYPER4, 0x3410)
    DEP_FIELD(PMEVTYPER4, P, 1, 31)
    DEP_FIELD(PMEVTYPER4, U, 1, 30)
    DEP_FIELD(PMEVTYPER4, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER4, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER4, EVENT, 5, 0)
DEP_REG32(PMEVTYPER5, 0x3414)
    DEP_FIELD(PMEVTYPER5, P, 1, 31)
    DEP_FIELD(PMEVTYPER5, U, 1, 30)
    DEP_FIELD(PMEVTYPER5, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER5, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER5, EVENT, 5, 0)
DEP_REG32(PMEVTYPER6, 0x3418)
    DEP_FIELD(PMEVTYPER6, P, 1, 31)
    DEP_FIELD(PMEVTYPER6, U, 1, 30)
    DEP_FIELD(PMEVTYPER6, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER6, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER6, EVENT, 5, 0)
DEP_REG32(PMEVTYPER7, 0x341c)
    DEP_FIELD(PMEVTYPER7, P, 1, 31)
    DEP_FIELD(PMEVTYPER7, U, 1, 30)
    DEP_FIELD(PMEVTYPER7, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER7, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER7, EVENT, 5, 0)
DEP_REG32(PMEVTYPER8, 0x3420)
    DEP_FIELD(PMEVTYPER8, P, 1, 31)
    DEP_FIELD(PMEVTYPER8, U, 1, 30)
    DEP_FIELD(PMEVTYPER8, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER8, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER8, EVENT, 5, 0)
DEP_REG32(PMEVTYPER9, 0x3424)
    DEP_FIELD(PMEVTYPER9, P, 1, 31)
    DEP_FIELD(PMEVTYPER9, U, 1, 30)
    DEP_FIELD(PMEVTYPER9, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER9, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER9, EVENT, 5, 0)
DEP_REG32(PMEVTYPER10, 0x3428)
    DEP_FIELD(PMEVTYPER10, P, 1, 31)
    DEP_FIELD(PMEVTYPER10, U, 1, 30)
    DEP_FIELD(PMEVTYPER10, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER10, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER10, EVENT, 5, 0)
DEP_REG32(PMEVTYPER11, 0x342c)
    DEP_FIELD(PMEVTYPER11, P, 1, 31)
    DEP_FIELD(PMEVTYPER11, U, 1, 30)
    DEP_FIELD(PMEVTYPER11, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER11, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER11, EVENT, 5, 0)
DEP_REG32(PMEVTYPER12, 0x3430)
    DEP_FIELD(PMEVTYPER12, P, 1, 31)
    DEP_FIELD(PMEVTYPER12, U, 1, 30)
    DEP_FIELD(PMEVTYPER12, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER12, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER12, EVENT, 5, 0)
DEP_REG32(PMEVTYPER13, 0x3434)
    DEP_FIELD(PMEVTYPER13, P, 1, 31)
    DEP_FIELD(PMEVTYPER13, U, 1, 30)
    DEP_FIELD(PMEVTYPER13, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER13, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER13, EVENT, 5, 0)
DEP_REG32(PMEVTYPER14, 0x3438)
    DEP_FIELD(PMEVTYPER14, P, 1, 31)
    DEP_FIELD(PMEVTYPER14, U, 1, 30)
    DEP_FIELD(PMEVTYPER14, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER14, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER14, EVENT, 5, 0)
DEP_REG32(PMEVTYPER15, 0x343c)
    DEP_FIELD(PMEVTYPER15, P, 1, 31)
    DEP_FIELD(PMEVTYPER15, U, 1, 30)
    DEP_FIELD(PMEVTYPER15, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER15, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER15, EVENT, 5, 0)
DEP_REG32(PMEVTYPER16, 0x3440)
    DEP_FIELD(PMEVTYPER16, P, 1, 31)
    DEP_FIELD(PMEVTYPER16, U, 1, 30)
    DEP_FIELD(PMEVTYPER16, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER16, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER16, EVENT, 5, 0)
DEP_REG32(PMEVTYPER17, 0x3444)
    DEP_FIELD(PMEVTYPER17, P, 1, 31)
    DEP_FIELD(PMEVTYPER17, U, 1, 30)
    DEP_FIELD(PMEVTYPER17, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER17, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER17, EVENT, 5, 0)
DEP_REG32(PMEVTYPER18, 0x3448)
    DEP_FIELD(PMEVTYPER18, P, 1, 31)
    DEP_FIELD(PMEVTYPER18, U, 1, 30)
    DEP_FIELD(PMEVTYPER18, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER18, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER18, EVENT, 5, 0)
DEP_REG32(PMEVTYPER19, 0x344c)
    DEP_FIELD(PMEVTYPER19, P, 1, 31)
    DEP_FIELD(PMEVTYPER19, U, 1, 30)
    DEP_FIELD(PMEVTYPER19, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER19, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER19, EVENT, 5, 0)
DEP_REG32(PMEVTYPER20, 0x3450)
    DEP_FIELD(PMEVTYPER20, P, 1, 31)
    DEP_FIELD(PMEVTYPER20, U, 1, 30)
    DEP_FIELD(PMEVTYPER20, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER20, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER20, EVENT, 5, 0)
DEP_REG32(PMEVTYPER21, 0x3454)
    DEP_FIELD(PMEVTYPER21, P, 1, 31)
    DEP_FIELD(PMEVTYPER21, U, 1, 30)
    DEP_FIELD(PMEVTYPER21, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER21, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER21, EVENT, 5, 0)
DEP_REG32(PMEVTYPER22, 0x3458)
    DEP_FIELD(PMEVTYPER22, P, 1, 31)
    DEP_FIELD(PMEVTYPER22, U, 1, 30)
    DEP_FIELD(PMEVTYPER22, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER22, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER22, EVENT, 5, 0)
DEP_REG32(PMEVTYPER23, 0x345c)
    DEP_FIELD(PMEVTYPER23, P, 1, 31)
    DEP_FIELD(PMEVTYPER23, U, 1, 30)
    DEP_FIELD(PMEVTYPER23, NSP, 1, 29)
    DEP_FIELD(PMEVTYPER23, NSU, 1, 28)
    DEP_FIELD(PMEVTYPER23, EVENT, 5, 0)
DEP_REG32(PMCGCR0, 0x3800)
    DEP_FIELD(PMCGCR0, CGNC, 4, 24)
    DEP_FIELD(PMCGCR0, SIDG, 7, 16)
    DEP_FIELD(PMCGCR0, X, 1, 12)
    DEP_FIELD(PMCGCR0, E, 1, 11)
    DEP_FIELD(PMCGCR0, CBAEN, 1, 10)
    DEP_FIELD(PMCGCR0, TCEFCFG, 2, 8)
    DEP_FIELD(PMCGCR0, NDX, 4, 0)
DEP_REG32(PMCGCR1, 0x3804)
    DEP_FIELD(PMCGCR1, CGNC, 4, 24)
    DEP_FIELD(PMCGCR1, SIDG, 7, 16)
    DEP_FIELD(PMCGCR1, X, 1, 12)
    DEP_FIELD(PMCGCR1, E, 1, 11)
    DEP_FIELD(PMCGCR1, CBAEN, 1, 10)
    DEP_FIELD(PMCGCR1, TCEFCFG, 2, 8)
    DEP_FIELD(PMCGCR1, NDX, 4, 0)
DEP_REG32(PMCGCR2, 0x3808)
    DEP_FIELD(PMCGCR2, CGNC, 4, 24)
    DEP_FIELD(PMCGCR2, SIDG, 7, 16)
    DEP_FIELD(PMCGCR2, X, 1, 12)
    DEP_FIELD(PMCGCR2, E, 1, 11)
    DEP_FIELD(PMCGCR2, CBAEN, 1, 10)
    DEP_FIELD(PMCGCR2, TCEFCFG, 2, 8)
    DEP_FIELD(PMCGCR2, NDX, 4, 0)
DEP_REG32(PMCGCR3, 0x380c)
    DEP_FIELD(PMCGCR3, CGNC, 4, 24)
    DEP_FIELD(PMCGCR3, SIDG, 7, 16)
    DEP_FIELD(PMCGCR3, X, 1, 12)
    DEP_FIELD(PMCGCR3, E, 1, 11)
    DEP_FIELD(PMCGCR3, CBAEN, 1, 10)
    DEP_FIELD(PMCGCR3, TCEFCFG, 2, 8)
    DEP_FIELD(PMCGCR3, NDX, 4, 0)
DEP_REG32(PMCGCR4, 0x3810)
    DEP_FIELD(PMCGCR4, CGNC, 4, 24)
    DEP_FIELD(PMCGCR4, SIDG, 7, 16)
    DEP_FIELD(PMCGCR4, X, 1, 12)
    DEP_FIELD(PMCGCR4, E, 1, 11)
    DEP_FIELD(PMCGCR4, CBAEN, 1, 10)
    DEP_FIELD(PMCGCR4, TCEFCFG, 2, 8)
    DEP_FIELD(PMCGCR4, NDX, 4, 0)
DEP_REG32(PMCGCR5, 0x3814)
    DEP_FIELD(PMCGCR5, CGNC, 4, 24)
    DEP_FIELD(PMCGCR5, SIDG, 7, 16)
    DEP_FIELD(PMCGCR5, X, 1, 12)
    DEP_FIELD(PMCGCR5, E, 1, 11)
    DEP_FIELD(PMCGCR5, CBAEN, 1, 10)
    DEP_FIELD(PMCGCR5, TCEFCFG, 2, 8)
    DEP_FIELD(PMCGCR5, NDX, 4, 0)
DEP_REG32(PMCGSMR0, 0x3a00)
    DEP_FIELD(PMCGSMR0, MASK, 10, 16)
    DEP_FIELD(PMCGSMR0, ID, 10, 0)
DEP_REG32(PMCGSMR1, 0x3a04)
    DEP_FIELD(PMCGSMR1, MASK, 10, 16)
    DEP_FIELD(PMCGSMR1, ID, 10, 0)
DEP_REG32(PMCGSMR2, 0x3a08)
    DEP_FIELD(PMCGSMR2, MASK, 10, 16)
    DEP_FIELD(PMCGSMR2, ID, 10, 0)
DEP_REG32(PMCGSMR3, 0x3a0c)
    DEP_FIELD(PMCGSMR3, MASK, 10, 16)
    DEP_FIELD(PMCGSMR3, ID, 10, 0)
DEP_REG32(PMCGSMR4, 0x3a10)
    DEP_FIELD(PMCGSMR4, MASK, 10, 16)
    DEP_FIELD(PMCGSMR4, ID, 10, 0)
DEP_REG32(PMCGSMR5, 0x3a14)
    DEP_FIELD(PMCGSMR5, MASK, 10, 16)
    DEP_FIELD(PMCGSMR5, ID, 10, 0)
DEP_REG32(PMCNTENSET, 0x3c00)
    DEP_FIELD(PMCNTENSET, P23, 1, 23)
    DEP_FIELD(PMCNTENSET, P22, 1, 22)
    DEP_FIELD(PMCNTENSET, P21, 1, 21)
    DEP_FIELD(PMCNTENSET, P20, 1, 20)
    DEP_FIELD(PMCNTENSET, P19, 1, 19)
    DEP_FIELD(PMCNTENSET, P18, 1, 18)
    DEP_FIELD(PMCNTENSET, P17, 1, 17)
    DEP_FIELD(PMCNTENSET, P16, 1, 16)
    DEP_FIELD(PMCNTENSET, P15, 1, 15)
    DEP_FIELD(PMCNTENSET, P14, 1, 14)
    DEP_FIELD(PMCNTENSET, P13, 1, 13)
    DEP_FIELD(PMCNTENSET, P12, 1, 12)
    DEP_FIELD(PMCNTENSET, P11, 1, 11)
    DEP_FIELD(PMCNTENSET, P10, 1, 10)
    DEP_FIELD(PMCNTENSET, P9, 1, 9)
    DEP_FIELD(PMCNTENSET, P8, 1, 8)
    DEP_FIELD(PMCNTENSET, P7, 1, 7)
    DEP_FIELD(PMCNTENSET, P6, 1, 6)
    DEP_FIELD(PMCNTENSET, P5, 1, 5)
    DEP_FIELD(PMCNTENSET, P4, 1, 4)
    DEP_FIELD(PMCNTENSET, P3, 1, 3)
    DEP_FIELD(PMCNTENSET, P2, 1, 2)
    DEP_FIELD(PMCNTENSET, P1, 1, 1)
    DEP_FIELD(PMCNTENSET, P0, 1, 0)
DEP_REG32(PMCNTENCLR, 0x3c20)
    DEP_FIELD(PMCNTENCLR, P23, 1, 23)
    DEP_FIELD(PMCNTENCLR, P22, 1, 22)
    DEP_FIELD(PMCNTENCLR, P21, 1, 21)
    DEP_FIELD(PMCNTENCLR, P20, 1, 20)
    DEP_FIELD(PMCNTENCLR, P19, 1, 19)
    DEP_FIELD(PMCNTENCLR, P18, 1, 18)
    DEP_FIELD(PMCNTENCLR, P17, 1, 17)
    DEP_FIELD(PMCNTENCLR, P16, 1, 16)
    DEP_FIELD(PMCNTENCLR, P15, 1, 15)
    DEP_FIELD(PMCNTENCLR, P14, 1, 14)
    DEP_FIELD(PMCNTENCLR, P13, 1, 13)
    DEP_FIELD(PMCNTENCLR, P12, 1, 12)
    DEP_FIELD(PMCNTENCLR, P11, 1, 11)
    DEP_FIELD(PMCNTENCLR, P10, 1, 10)
    DEP_FIELD(PMCNTENCLR, P9, 1, 9)
    DEP_FIELD(PMCNTENCLR, P8, 1, 8)
    DEP_FIELD(PMCNTENCLR, P7, 1, 7)
    DEP_FIELD(PMCNTENCLR, P6, 1, 6)
    DEP_FIELD(PMCNTENCLR, P5, 1, 5)
    DEP_FIELD(PMCNTENCLR, P4, 1, 4)
    DEP_FIELD(PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(PMCNTENCLR, P0, 1, 0)
DEP_REG32(PMINTENSET, 0x3c40)
    DEP_FIELD(PMINTENSET, P23, 1, 23)
    DEP_FIELD(PMINTENSET, P22, 1, 22)
    DEP_FIELD(PMINTENSET, P21, 1, 21)
    DEP_FIELD(PMINTENSET, P20, 1, 20)
    DEP_FIELD(PMINTENSET, P19, 1, 19)
    DEP_FIELD(PMINTENSET, P18, 1, 18)
    DEP_FIELD(PMINTENSET, P17, 1, 17)
    DEP_FIELD(PMINTENSET, P16, 1, 16)
    DEP_FIELD(PMINTENSET, P15, 1, 15)
    DEP_FIELD(PMINTENSET, P14, 1, 14)
    DEP_FIELD(PMINTENSET, P13, 1, 13)
    DEP_FIELD(PMINTENSET, P12, 1, 12)
    DEP_FIELD(PMINTENSET, P11, 1, 11)
    DEP_FIELD(PMINTENSET, P10, 1, 10)
    DEP_FIELD(PMINTENSET, P9, 1, 9)
    DEP_FIELD(PMINTENSET, P8, 1, 8)
    DEP_FIELD(PMINTENSET, P7, 1, 7)
    DEP_FIELD(PMINTENSET, P6, 1, 6)
    DEP_FIELD(PMINTENSET, P5, 1, 5)
    DEP_FIELD(PMINTENSET, P4, 1, 4)
    DEP_FIELD(PMINTENSET, P3, 1, 3)
    DEP_FIELD(PMINTENSET, P2, 1, 2)
    DEP_FIELD(PMINTENSET, P1, 1, 1)
    DEP_FIELD(PMINTENSET, P0, 1, 0)
DEP_REG32(PMINTENCLR, 0x3c60)
    DEP_FIELD(PMINTENCLR, P23, 1, 23)
    DEP_FIELD(PMINTENCLR, P22, 1, 22)
    DEP_FIELD(PMINTENCLR, P21, 1, 21)
    DEP_FIELD(PMINTENCLR, P20, 1, 20)
    DEP_FIELD(PMINTENCLR, P19, 1, 19)
    DEP_FIELD(PMINTENCLR, P18, 1, 18)
    DEP_FIELD(PMINTENCLR, P17, 1, 17)
    DEP_FIELD(PMINTENCLR, P16, 1, 16)
    DEP_FIELD(PMINTENCLR, P15, 1, 15)
    DEP_FIELD(PMINTENCLR, P14, 1, 14)
    DEP_FIELD(PMINTENCLR, P13, 1, 13)
    DEP_FIELD(PMINTENCLR, P12, 1, 12)
    DEP_FIELD(PMINTENCLR, P11, 1, 11)
    DEP_FIELD(PMINTENCLR, P10, 1, 10)
    DEP_FIELD(PMINTENCLR, P9, 1, 9)
    DEP_FIELD(PMINTENCLR, P8, 1, 8)
    DEP_FIELD(PMINTENCLR, P7, 1, 7)
    DEP_FIELD(PMINTENCLR, P6, 1, 6)
    DEP_FIELD(PMINTENCLR, P5, 1, 5)
    DEP_FIELD(PMINTENCLR, P4, 1, 4)
    DEP_FIELD(PMINTENCLR, P3, 1, 3)
    DEP_FIELD(PMINTENCLR, P2, 1, 2)
    DEP_FIELD(PMINTENCLR, P1, 1, 1)
    DEP_FIELD(PMINTENCLR, P0, 1, 0)
DEP_REG32(PMOVSCLR, 0x3c80)
    DEP_FIELD(PMOVSCLR, P23, 1, 23)
    DEP_FIELD(PMOVSCLR, P22, 1, 22)
    DEP_FIELD(PMOVSCLR, P21, 1, 21)
    DEP_FIELD(PMOVSCLR, P20, 1, 20)
    DEP_FIELD(PMOVSCLR, P19, 1, 19)
    DEP_FIELD(PMOVSCLR, P18, 1, 18)
    DEP_FIELD(PMOVSCLR, P17, 1, 17)
    DEP_FIELD(PMOVSCLR, P16, 1, 16)
    DEP_FIELD(PMOVSCLR, P15, 1, 15)
    DEP_FIELD(PMOVSCLR, P14, 1, 14)
    DEP_FIELD(PMOVSCLR, P13, 1, 13)
    DEP_FIELD(PMOVSCLR, P12, 1, 12)
    DEP_FIELD(PMOVSCLR, P11, 1, 11)
    DEP_FIELD(PMOVSCLR, P10, 1, 10)
    DEP_FIELD(PMOVSCLR, P9, 1, 9)
    DEP_FIELD(PMOVSCLR, P8, 1, 8)
    DEP_FIELD(PMOVSCLR, P7, 1, 7)
    DEP_FIELD(PMOVSCLR, P6, 1, 6)
    DEP_FIELD(PMOVSCLR, P5, 1, 5)
    DEP_FIELD(PMOVSCLR, P4, 1, 4)
    DEP_FIELD(PMOVSCLR, P3, 1, 3)
    DEP_FIELD(PMOVSCLR, P2, 1, 2)
    DEP_FIELD(PMOVSCLR, P1, 1, 1)
    DEP_FIELD(PMOVSCLR, P0, 1, 0)
DEP_REG32(PMOVSSET, 0x3cc0)
    DEP_FIELD(PMOVSSET, P23, 1, 23)
    DEP_FIELD(PMOVSSET, P22, 1, 22)
    DEP_FIELD(PMOVSSET, P21, 1, 21)
    DEP_FIELD(PMOVSSET, P20, 1, 20)
    DEP_FIELD(PMOVSSET, P19, 1, 19)
    DEP_FIELD(PMOVSSET, P18, 1, 18)
    DEP_FIELD(PMOVSSET, P17, 1, 17)
    DEP_FIELD(PMOVSSET, P16, 1, 16)
    DEP_FIELD(PMOVSSET, P15, 1, 15)
    DEP_FIELD(PMOVSSET, P14, 1, 14)
    DEP_FIELD(PMOVSSET, P13, 1, 13)
    DEP_FIELD(PMOVSSET, P12, 1, 12)
    DEP_FIELD(PMOVSSET, P11, 1, 11)
    DEP_FIELD(PMOVSSET, P10, 1, 10)
    DEP_FIELD(PMOVSSET, P9, 1, 9)
    DEP_FIELD(PMOVSSET, P8, 1, 8)
    DEP_FIELD(PMOVSSET, P7, 1, 7)
    DEP_FIELD(PMOVSSET, P6, 1, 6)
    DEP_FIELD(PMOVSSET, P5, 1, 5)
    DEP_FIELD(PMOVSSET, P4, 1, 4)
    DEP_FIELD(PMOVSSET, P3, 1, 3)
    DEP_FIELD(PMOVSSET, P2, 1, 2)
    DEP_FIELD(PMOVSSET, P1, 1, 1)
    DEP_FIELD(PMOVSSET, P0, 1, 0)
DEP_REG32(PMCFGR, 0x3e00)
    DEP_FIELD(PMCFGR, NCG, 8, 24)
    DEP_FIELD(PMCFGR, UEN, 1, 19)
    DEP_FIELD(PMCFGR, EX, 1, 16)
    DEP_FIELD(PMCFGR, CCD, 1, 15)
    DEP_FIELD(PMCFGR, CC, 1, 14)
    DEP_FIELD(PMCFGR, SIZE, 6, 8)
    DEP_FIELD(PMCFGR, N, 8, 0)
DEP_REG32(PMCR, 0x3e04)
    DEP_FIELD(PMCR, IMP, 8, 24)
    DEP_FIELD(PMCR, X, 1, 4)
    DEP_FIELD(PMCR, P, 1, 1)
    DEP_FIELD(PMCR, E, 1, 0)
DEP_REG32(PMCEID0, 0x3e20)
    DEP_FIELD(PMCEID0, EVENT0X12, 1, 17)
    DEP_FIELD(PMCEID0, EVENT0X11, 1, 16)
    DEP_FIELD(PMCEID0, EVENT0X10, 1, 15)
    DEP_FIELD(PMCEID0, EVENT0X0A, 1, 9)
    DEP_FIELD(PMCEID0, EVENT0X09, 1, 8)
    DEP_FIELD(PMCEID0, EVENT0X08, 1, 7)
    DEP_FIELD(PMCEID0, EVENT0X01, 1, 1)
    DEP_FIELD(PMCEID0, EVENT0X00, 1, 0)
DEP_REG32(PMAUTHSTATUS, 0x3fb8)
    DEP_FIELD(PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(PMAUTHSTATUS, NSE, 1, 0)
DEP_REG32(PMDEVTYPE, 0x3fcc)
    DEP_FIELD(PMDEVTYPE, T, 4, 4)
    DEP_FIELD(PMDEVTYPE, C, 4, 0)
DEP_REG32(SMMU_CB0_SCTLR, 0x10000)
    DEP_FIELD(SMMU_CB0_SCTLR, NSCFG, 2, 28)
    DEP_FIELD(SMMU_CB0_SCTLR, WACFG, 2, 26)
    DEP_FIELD(SMMU_CB0_SCTLR, RACFG, 2, 24)
    DEP_FIELD(SMMU_CB0_SCTLR, SHCFG, 2, 22)
    DEP_FIELD(SMMU_CB0_SCTLR, FB, 1, 21)
    DEP_FIELD(SMMU_CB0_SCTLR, MTCFG, 1, 20)
    DEP_FIELD(SMMU_CB0_SCTLR, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_CB0_SCTLR, TRANSIENTCFG, 2, 14)
    DEP_FIELD(SMMU_CB0_SCTLR, PTW, 1, 13)
    DEP_FIELD(SMMU_CB0_SCTLR, ASIDPNE, 1, 12)
    DEP_FIELD(SMMU_CB0_SCTLR, UWXN, 1, 10)
    DEP_FIELD(SMMU_CB0_SCTLR, WXN, 1, 9)
    DEP_FIELD(SMMU_CB0_SCTLR, HUPCF, 1, 8)
    DEP_FIELD(SMMU_CB0_SCTLR, CFCFG, 1, 7)
    DEP_FIELD(SMMU_CB0_SCTLR, CFIE, 1, 6)
    DEP_FIELD(SMMU_CB0_SCTLR, CFRE, 1, 5)
    DEP_FIELD(SMMU_CB0_SCTLR, E, 1, 4)
    DEP_FIELD(SMMU_CB0_SCTLR, AFFD, 1, 3)
    DEP_FIELD(SMMU_CB0_SCTLR, AFE, 1, 2)
    DEP_FIELD(SMMU_CB0_SCTLR, TRE, 1, 1)
    DEP_FIELD(SMMU_CB0_SCTLR, M, 1, 0)
DEP_REG32(SMMU_CB0_ACTLR, 0x10004)
    DEP_FIELD(SMMU_CB0_ACTLR, CPRE, 1, 1)
    DEP_FIELD(SMMU_CB0_ACTLR, CMTLB, 1, 0)
DEP_REG32(SMMU_CB0_RESUME, 0x10008)
    DEP_FIELD(SMMU_CB0_RESUME, TNR, 1, 0)
DEP_REG32(SMMU_CB0_TCR2, 0x10010)
    DEP_FIELD(SMMU_CB0_TCR2, NSCFG1, 1, 30)
    DEP_FIELD(SMMU_CB0_TCR2, SEP, 3, 15)
    DEP_FIELD(SMMU_CB0_TCR2, NSCFG0, 1, 14)
    DEP_FIELD(SMMU_CB0_TCR2, TBI1, 1, 6)
    DEP_FIELD(SMMU_CB0_TCR2, TBI0, 1, 5)
    DEP_FIELD(SMMU_CB0_TCR2, AS, 1, 4)
    DEP_FIELD(SMMU_CB0_TCR2, PASIZE, 3, 0)
DEP_REG32(SMMU_CB0_TTBR0_LOW, 0x10020)
    DEP_FIELD(SMMU_CB0_TTBR0_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB0_TTBR0_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB0_TTBR0_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB0_TTBR0_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB0_TTBR0_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB0_TTBR0_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB0_TTBR0_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB0_TTBR0_HIGH, 0x10024)
    DEP_FIELD(SMMU_CB0_TTBR0_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB0_TTBR0_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB0_TTBR1_LOW, 0x10028)
    DEP_FIELD(SMMU_CB0_TTBR1_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB0_TTBR1_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB0_TTBR1_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB0_TTBR1_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB0_TTBR1_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB0_TTBR1_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB0_TTBR1_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB0_TTBR1_HIGH, 0x1002c)
    DEP_FIELD(SMMU_CB0_TTBR1_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB0_TTBR1_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB0_TCR_LPAE, 0x10030)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, EAE, 1, 31)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, NSCFG1_TG1, 1, 30)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, SH1, 2, 28)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, ORGN1, 2, 26)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, IRGN1, 2, 24)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, EPD1, 1, 23)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, A1, 1, 22)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, T1SZ_5_3, 3, 19)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, T1SZ_2_0_PASIZE, 3, 16)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, NSCFG0_TG0, 1, 14)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, SH0, 2, 12)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, ORGN0, 2, 10)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, IRGN0, 2, 8)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, SL0_1_EPD0, 1, 7)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, SL0_0, 1, 6)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, PD1_T0SZ_5, 1, 5)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, S_PD0_T0SZ_4, 1, 4)
    DEP_FIELD(SMMU_CB0_TCR_LPAE, T0SZ_3_0, 4, 0)
DEP_REG32(SMMU_CB0_CONTEXTIDR, 0x10034)
    DEP_FIELD(SMMU_CB0_CONTEXTIDR, PROCID, 24, 8)
    DEP_FIELD(SMMU_CB0_CONTEXTIDR, ASID, 8, 0)
DEP_REG32(SMMU_CB0_PRRR_MAIR0, 0x10038)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, NOS7, 1, 31)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, NOS6, 1, 30)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, NOS5, 1, 29)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, NOS4, 1, 28)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, NOS3, 1, 27)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, NOS2, 1, 26)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, NOS1, 1, 25)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, NOS0, 1, 24)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, NS1, 1, 19)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, NS0, 1, 18)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, DS1, 1, 17)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, DS0, 1, 16)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, TR7, 2, 14)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, TR6, 2, 12)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, TR5, 2, 10)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, TR4, 2, 8)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, TR3, 2, 6)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, TR2, 2, 4)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, TR1, 2, 2)
    DEP_FIELD(SMMU_CB0_PRRR_MAIR0, TR0, 2, 0)
DEP_REG32(SMMU_CB0_NMRR_MAIR1, 0x1003c)
    DEP_FIELD(SMMU_CB0_NMRR_MAIR1, OR7, 2, 30)
    DEP_FIELD(SMMU_CB0_NMRR_MAIR1, OR6, 2, 28)
    DEP_FIELD(SMMU_CB0_NMRR_MAIR1, OR5, 2, 26)
    DEP_FIELD(SMMU_CB0_NMRR_MAIR1, OR4, 2, 24)
    DEP_FIELD(SMMU_CB0_NMRR_MAIR1, OR3, 2, 22)
    DEP_FIELD(SMMU_CB0_NMRR_MAIR1, OR2, 2, 20)
    DEP_FIELD(SMMU_CB0_NMRR_MAIR1, OR1, 2, 18)
    DEP_FIELD(SMMU_CB0_NMRR_MAIR1, OR0, 2, 16)
    DEP_FIELD(SMMU_CB0_NMRR_MAIR1, IR7, 2, 14)
    DEP_FIELD(SMMU_CB0_NMRR_MAIR1, IR6, 2, 12)
    DEP_FIELD(SMMU_CB0_NMRR_MAIR1, IR5, 2, 10)
    DEP_FIELD(SMMU_CB0_NMRR_MAIR1, IR4, 2, 8)
    DEP_FIELD(SMMU_CB0_NMRR_MAIR1, IR3, 2, 6)
    DEP_FIELD(SMMU_CB0_NMRR_MAIR1, IR2, 2, 4)
    DEP_FIELD(SMMU_CB0_NMRR_MAIR1, IR1, 2, 2)
    DEP_FIELD(SMMU_CB0_NMRR_MAIR1, IR0, 2, 0)
DEP_REG32(SMMU_CB0_FSR, 0x10058)
    DEP_FIELD(SMMU_CB0_FSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_CB0_FSR, SS, 1, 30)
    DEP_FIELD(SMMU_CB0_FSR, FORMAT, 2, 9)
    DEP_FIELD(SMMU_CB0_FSR, UUT, 1, 8)
    DEP_FIELD(SMMU_CB0_FSR, ASF, 1, 7)
    DEP_FIELD(SMMU_CB0_FSR, TLBLKF, 1, 6)
    DEP_FIELD(SMMU_CB0_FSR, TLBMCF, 1, 5)
    DEP_FIELD(SMMU_CB0_FSR, EF, 1, 4)
    DEP_FIELD(SMMU_CB0_FSR, PF, 1, 3)
    DEP_FIELD(SMMU_CB0_FSR, AFF, 1, 2)
    DEP_FIELD(SMMU_CB0_FSR, TF, 1, 1)
DEP_REG32(SMMU_CB0_FSRRESTORE, 0x1005c)
DEP_REG32(SMMU_CB0_FAR_LOW, 0x10060)
DEP_REG32(SMMU_CB0_FAR_HIGH, 0x10064)
    DEP_FIELD(SMMU_CB0_FAR_HIGH, BITS, 17, 0)
DEP_REG32(SMMU_CB0_FSYNR0, 0x10068)
    DEP_FIELD(SMMU_CB0_FSYNR0, S1CBNDX, 4, 16)
    DEP_FIELD(SMMU_CB0_FSYNR0, AFR, 1, 11)
    DEP_FIELD(SMMU_CB0_FSYNR0, PTWF, 1, 10)
    DEP_FIELD(SMMU_CB0_FSYNR0, ATOF, 1, 9)
    DEP_FIELD(SMMU_CB0_FSYNR0, NSATTR, 1, 8)
    DEP_FIELD(SMMU_CB0_FSYNR0, IND, 1, 6)
    DEP_FIELD(SMMU_CB0_FSYNR0, PNU, 1, 5)
    DEP_FIELD(SMMU_CB0_FSYNR0, WNR, 1, 4)
    DEP_FIELD(SMMU_CB0_FSYNR0, PLVL, 2, 0)
DEP_REG32(SMMU_CB0_IPAFAR_LOW, 0x10070)
    DEP_FIELD(SMMU_CB0_IPAFAR_LOW, IPAFAR_L, 20, 12)
    DEP_FIELD(SMMU_CB0_IPAFAR_LOW, FAR_RO, 12, 0)
DEP_REG32(SMMU_CB0_IPAFAR_HIGH, 0x10074)
    DEP_FIELD(SMMU_CB0_IPAFAR_HIGH, BITS, 16, 0)
DEP_REG32(SMMU_CB0_TLBIVA_LOW, 0x10600)
DEP_REG32(SMMU_CB0_TLBIVA_HIGH, 0x10604)
    DEP_FIELD(SMMU_CB0_TLBIVA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB0_TLBIVA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB0_TLBIVAA_LOW, 0x10608)
DEP_REG32(SMMU_CB0_TLBIVAA_HIGH, 0x1060c)
    DEP_FIELD(SMMU_CB0_TLBIVAA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB0_TLBIVAA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB0_TLBIASID, 0x10610)
    DEP_FIELD(SMMU_CB0_TLBIASID, ASID, 16, 0)
DEP_REG32(SMMU_CB0_TLBIALL, 0x10618)
DEP_REG32(SMMU_CB0_TLBIVAL_LOW, 0x10620)
DEP_REG32(SMMU_CB0_TLBIVAL_HIGH, 0x10624)
    DEP_FIELD(SMMU_CB0_TLBIVAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB0_TLBIVAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB0_TLBIVAAL_LOW, 0x10628)
DEP_REG32(SMMU_CB0_TLBIVAAL_HIGH, 0x1062c)
    DEP_FIELD(SMMU_CB0_TLBIVAAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB0_TLBIVAAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB0_TLBIIPAS2_LOW, 0x10630)
DEP_REG32(SMMU_CB0_TLBIIPAS2_HIGH, 0x10634)
    DEP_FIELD(SMMU_CB0_TLBIIPAS2_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB0_TLBIIPAS2L_LOW, 0x10638)
DEP_REG32(SMMU_CB0_TLBIIPAS2L_HIGH, 0x1063c)
    DEP_FIELD(SMMU_CB0_TLBIIPAS2L_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB0_TLBSYNC, 0x107f0)
DEP_REG32(SMMU_CB0_TLBSTATUS, 0x107f4)
    DEP_FIELD(SMMU_CB0_TLBSTATUS, SACTIVE, 1, 0)
DEP_REG32(SMMU_CB0_PMEVCNTR0, 0x10e00)
DEP_REG32(SMMU_CB0_PMEVCNTR1, 0x10e04)
DEP_REG32(SMMU_CB0_PMEVCNTR2, 0x10e08)
DEP_REG32(SMMU_CB0_PMEVCNTR3, 0x10e0c)
DEP_REG32(SMMU_CB0_PMEVTYPER0, 0x10e80)
    DEP_FIELD(SMMU_CB0_PMEVTYPER0, P, 1, 31)
    DEP_FIELD(SMMU_CB0_PMEVTYPER0, U, 1, 30)
    DEP_FIELD(SMMU_CB0_PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(SMMU_CB0_PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(SMMU_CB0_PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(SMMU_CB0_PMEVTYPER1, 0x10e84)
    DEP_FIELD(SMMU_CB0_PMEVTYPER1, P, 1, 31)
    DEP_FIELD(SMMU_CB0_PMEVTYPER1, U, 1, 30)
    DEP_FIELD(SMMU_CB0_PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(SMMU_CB0_PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(SMMU_CB0_PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(SMMU_CB0_PMEVTYPER2, 0x10e88)
    DEP_FIELD(SMMU_CB0_PMEVTYPER2, P, 1, 31)
    DEP_FIELD(SMMU_CB0_PMEVTYPER2, U, 1, 30)
    DEP_FIELD(SMMU_CB0_PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(SMMU_CB0_PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(SMMU_CB0_PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(SMMU_CB0_PMEVTYPER3, 0x10e8c)
    DEP_FIELD(SMMU_CB0_PMEVTYPER3, P, 1, 31)
    DEP_FIELD(SMMU_CB0_PMEVTYPER3, U, 1, 30)
    DEP_FIELD(SMMU_CB0_PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(SMMU_CB0_PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(SMMU_CB0_PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(SMMU_CB0_PMCFGR, 0x10f00)
    DEP_FIELD(SMMU_CB0_PMCFGR, NCG, 8, 24)
    DEP_FIELD(SMMU_CB0_PMCFGR, UEN, 1, 19)
    DEP_FIELD(SMMU_CB0_PMCFGR, EX, 1, 16)
    DEP_FIELD(SMMU_CB0_PMCFGR, CCD, 1, 15)
    DEP_FIELD(SMMU_CB0_PMCFGR, CC, 1, 14)
    DEP_FIELD(SMMU_CB0_PMCFGR, SIZE, 6, 8)
    DEP_FIELD(SMMU_CB0_PMCFGR, N, 8, 0)
DEP_REG32(SMMU_CB0_PMCR, 0x10f04)
    DEP_FIELD(SMMU_CB0_PMCR, IMP, 8, 24)
    DEP_FIELD(SMMU_CB0_PMCR, X, 1, 4)
    DEP_FIELD(SMMU_CB0_PMCR, P, 1, 1)
    DEP_FIELD(SMMU_CB0_PMCR, E, 1, 0)
DEP_REG32(SMMU_CB0_PMCEID, 0x10f20)
    DEP_FIELD(SMMU_CB0_PMCEID, EVENT0X12, 1, 17)
    DEP_FIELD(SMMU_CB0_PMCEID, EVENT0X11, 1, 16)
    DEP_FIELD(SMMU_CB0_PMCEID, EVENT0X10, 1, 15)
    DEP_FIELD(SMMU_CB0_PMCEID, EVENT0X0A, 1, 9)
    DEP_FIELD(SMMU_CB0_PMCEID, EVENT0X09, 1, 8)
    DEP_FIELD(SMMU_CB0_PMCEID, EVENT0X08, 1, 7)
    DEP_FIELD(SMMU_CB0_PMCEID, EVENT0X01, 1, 1)
    DEP_FIELD(SMMU_CB0_PMCEID, EVENT0X00, 1, 0)
DEP_REG32(SMMU_CB0_PMCNTENSE, 0x10f40)
    DEP_FIELD(SMMU_CB0_PMCNTENSE, P3, 1, 3)
    DEP_FIELD(SMMU_CB0_PMCNTENSE, P2, 1, 2)
    DEP_FIELD(SMMU_CB0_PMCNTENSE, P1, 1, 1)
    DEP_FIELD(SMMU_CB0_PMCNTENSE, P0, 1, 0)
DEP_REG32(SMMU_CB0_PMCNTENCLR, 0x10f44)
    DEP_FIELD(SMMU_CB0_PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB0_PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB0_PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB0_PMCNTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB0_PMCNTENSET, 0x10f48)
    DEP_FIELD(SMMU_CB0_PMCNTENSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB0_PMCNTENSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB0_PMCNTENSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB0_PMCNTENSET, P0, 1, 0)
DEP_REG32(SMMU_CB0_PMINTENCLR, 0x10f4c)
    DEP_FIELD(SMMU_CB0_PMINTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB0_PMINTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB0_PMINTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB0_PMINTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB0_PMOVSCLR, 0x10f50)
    DEP_FIELD(SMMU_CB0_PMOVSCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB0_PMOVSCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB0_PMOVSCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB0_PMOVSCLR, P0, 1, 0)
DEP_REG32(SMMU_CB0_PMOVSSET, 0x10f58)
    DEP_FIELD(SMMU_CB0_PMOVSSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB0_PMOVSSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB0_PMOVSSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB0_PMOVSSET, P0, 1, 0)
DEP_REG32(SMMU_CB0_PMAUTHSTATUS, 0x10fb8)
    DEP_FIELD(SMMU_CB0_PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(SMMU_CB0_PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(SMMU_CB0_PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(SMMU_CB0_PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(SMMU_CB0_PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(SMMU_CB0_PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(SMMU_CB0_PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(SMMU_CB0_PMAUTHSTATUS, NSE, 1, 0)
DEP_REG32(SMMU_CB1_SCTLR, 0x11000)
    DEP_FIELD(SMMU_CB1_SCTLR, NSCFG, 2, 28)
    DEP_FIELD(SMMU_CB1_SCTLR, WACFG, 2, 26)
    DEP_FIELD(SMMU_CB1_SCTLR, RACFG, 2, 24)
    DEP_FIELD(SMMU_CB1_SCTLR, SHCFG, 2, 22)
    DEP_FIELD(SMMU_CB1_SCTLR, FB, 1, 21)
    DEP_FIELD(SMMU_CB1_SCTLR, MTCFG, 1, 20)
    DEP_FIELD(SMMU_CB1_SCTLR, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_CB1_SCTLR, TRANSIENTCFG, 2, 14)
    DEP_FIELD(SMMU_CB1_SCTLR, PTW, 1, 13)
    DEP_FIELD(SMMU_CB1_SCTLR, ASIDPNE, 1, 12)
    DEP_FIELD(SMMU_CB1_SCTLR, UWXN, 1, 10)
    DEP_FIELD(SMMU_CB1_SCTLR, WXN, 1, 9)
    DEP_FIELD(SMMU_CB1_SCTLR, HUPCF, 1, 8)
    DEP_FIELD(SMMU_CB1_SCTLR, CFCFG, 1, 7)
    DEP_FIELD(SMMU_CB1_SCTLR, CFIE, 1, 6)
    DEP_FIELD(SMMU_CB1_SCTLR, CFRE, 1, 5)
    DEP_FIELD(SMMU_CB1_SCTLR, E, 1, 4)
    DEP_FIELD(SMMU_CB1_SCTLR, AFFD, 1, 3)
    DEP_FIELD(SMMU_CB1_SCTLR, AFE, 1, 2)
    DEP_FIELD(SMMU_CB1_SCTLR, TRE, 1, 1)
    DEP_FIELD(SMMU_CB1_SCTLR, M, 1, 0)
DEP_REG32(SMMU_CB1_ACTLR, 0x11004)
    DEP_FIELD(SMMU_CB1_ACTLR, CPRE, 1, 1)
    DEP_FIELD(SMMU_CB1_ACTLR, CMTLB, 1, 0)
DEP_REG32(SMMU_CB1_RESUME, 0x11008)
    DEP_FIELD(SMMU_CB1_RESUME, TNR, 1, 0)
DEP_REG32(SMMU_CB1_TCR2, 0x11010)
    DEP_FIELD(SMMU_CB1_TCR2, NSCFG1, 1, 30)
    DEP_FIELD(SMMU_CB1_TCR2, SEP, 3, 15)
    DEP_FIELD(SMMU_CB1_TCR2, NSCFG0, 1, 14)
    DEP_FIELD(SMMU_CB1_TCR2, TBI1, 1, 6)
    DEP_FIELD(SMMU_CB1_TCR2, TBI0, 1, 5)
    DEP_FIELD(SMMU_CB1_TCR2, AS, 1, 4)
    DEP_FIELD(SMMU_CB1_TCR2, PASIZE, 3, 0)
DEP_REG32(SMMU_CB1_TTBR0_LOW, 0x11020)
    DEP_FIELD(SMMU_CB1_TTBR0_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB1_TTBR0_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB1_TTBR0_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB1_TTBR0_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB1_TTBR0_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB1_TTBR0_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB1_TTBR0_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB1_TTBR0_HIGH, 0x11024)
    DEP_FIELD(SMMU_CB1_TTBR0_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB1_TTBR0_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB1_TTBR1_LOW, 0x11028)
    DEP_FIELD(SMMU_CB1_TTBR1_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB1_TTBR1_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB1_TTBR1_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB1_TTBR1_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB1_TTBR1_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB1_TTBR1_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB1_TTBR1_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB1_TTBR1_HIGH, 0x1102c)
    DEP_FIELD(SMMU_CB1_TTBR1_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB1_TTBR1_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB1_TCR_LPAE, 0x11030)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, EAE, 1, 31)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, NSCFG1_TG1, 1, 30)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, SH1, 2, 28)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, ORGN1, 2, 26)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, IRGN1, 2, 24)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, EPD1, 1, 23)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, A1, 1, 22)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, T1SZ_5_3, 3, 19)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, T1SZ_2_0_PASIZE, 3, 16)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, NSCFG0_TG0, 1, 14)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, SH0, 2, 12)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, ORGN0, 2, 10)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, IRGN0, 2, 8)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, SL0_1_EPD0, 1, 7)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, SL0_0, 1, 6)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, PD1_T0SZ_5, 1, 5)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, S_PD0_T0SZ_4, 1, 4)
    DEP_FIELD(SMMU_CB1_TCR_LPAE, T0SZ_3_0, 4, 0)
DEP_REG32(SMMU_CB1_CONTEXTIDR, 0x11034)
    DEP_FIELD(SMMU_CB1_CONTEXTIDR, PROCID, 24, 8)
    DEP_FIELD(SMMU_CB1_CONTEXTIDR, ASID, 8, 0)
DEP_REG32(SMMU_CB1_PRRR_MAIR0, 0x11038)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, NOS7, 1, 31)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, NOS6, 1, 30)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, NOS5, 1, 29)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, NOS4, 1, 28)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, NOS3, 1, 27)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, NOS2, 1, 26)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, NOS1, 1, 25)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, NOS0, 1, 24)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, NS1, 1, 19)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, NS0, 1, 18)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, DS1, 1, 17)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, DS0, 1, 16)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, TR7, 2, 14)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, TR6, 2, 12)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, TR5, 2, 10)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, TR4, 2, 8)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, TR3, 2, 6)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, TR2, 2, 4)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, TR1, 2, 2)
    DEP_FIELD(SMMU_CB1_PRRR_MAIR0, TR0, 2, 0)
DEP_REG32(SMMU_CB1_NMRR_MAIR1, 0x1103c)
    DEP_FIELD(SMMU_CB1_NMRR_MAIR1, OR7, 2, 30)
    DEP_FIELD(SMMU_CB1_NMRR_MAIR1, OR6, 2, 28)
    DEP_FIELD(SMMU_CB1_NMRR_MAIR1, OR5, 2, 26)
    DEP_FIELD(SMMU_CB1_NMRR_MAIR1, OR4, 2, 24)
    DEP_FIELD(SMMU_CB1_NMRR_MAIR1, OR3, 2, 22)
    DEP_FIELD(SMMU_CB1_NMRR_MAIR1, OR2, 2, 20)
    DEP_FIELD(SMMU_CB1_NMRR_MAIR1, OR1, 2, 18)
    DEP_FIELD(SMMU_CB1_NMRR_MAIR1, OR0, 2, 16)
    DEP_FIELD(SMMU_CB1_NMRR_MAIR1, IR7, 2, 14)
    DEP_FIELD(SMMU_CB1_NMRR_MAIR1, IR6, 2, 12)
    DEP_FIELD(SMMU_CB1_NMRR_MAIR1, IR5, 2, 10)
    DEP_FIELD(SMMU_CB1_NMRR_MAIR1, IR4, 2, 8)
    DEP_FIELD(SMMU_CB1_NMRR_MAIR1, IR3, 2, 6)
    DEP_FIELD(SMMU_CB1_NMRR_MAIR1, IR2, 2, 4)
    DEP_FIELD(SMMU_CB1_NMRR_MAIR1, IR1, 2, 2)
    DEP_FIELD(SMMU_CB1_NMRR_MAIR1, IR0, 2, 0)
DEP_REG32(SMMU_CB1_FSR, 0x11058)
    DEP_FIELD(SMMU_CB1_FSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_CB1_FSR, SS, 1, 30)
    DEP_FIELD(SMMU_CB1_FSR, FORMAT, 2, 9)
    DEP_FIELD(SMMU_CB1_FSR, UUT, 1, 8)
    DEP_FIELD(SMMU_CB1_FSR, ASF, 1, 7)
    DEP_FIELD(SMMU_CB1_FSR, TLBLKF, 1, 6)
    DEP_FIELD(SMMU_CB1_FSR, TLBMCF, 1, 5)
    DEP_FIELD(SMMU_CB1_FSR, EF, 1, 4)
    DEP_FIELD(SMMU_CB1_FSR, PF, 1, 3)
    DEP_FIELD(SMMU_CB1_FSR, AFF, 1, 2)
    DEP_FIELD(SMMU_CB1_FSR, TF, 1, 1)
DEP_REG32(SMMU_CB1_FSRRESTORE, 0x1105c)
DEP_REG32(SMMU_CB1_FAR_LOW, 0x11060)
DEP_REG32(SMMU_CB1_FAR_HIGH, 0x11064)
    DEP_FIELD(SMMU_CB1_FAR_HIGH, BITS, 17, 0)
DEP_REG32(SMMU_CB1_FSYNR0, 0x11068)
    DEP_FIELD(SMMU_CB1_FSYNR0, S1CBNDX, 4, 16)
    DEP_FIELD(SMMU_CB1_FSYNR0, AFR, 1, 11)
    DEP_FIELD(SMMU_CB1_FSYNR0, PTWF, 1, 10)
    DEP_FIELD(SMMU_CB1_FSYNR0, ATOF, 1, 9)
    DEP_FIELD(SMMU_CB1_FSYNR0, NSATTR, 1, 8)
    DEP_FIELD(SMMU_CB1_FSYNR0, IND, 1, 6)
    DEP_FIELD(SMMU_CB1_FSYNR0, PNU, 1, 5)
    DEP_FIELD(SMMU_CB1_FSYNR0, WNR, 1, 4)
    DEP_FIELD(SMMU_CB1_FSYNR0, PLVL, 2, 0)
DEP_REG32(SMMU_CB1_IPAFAR_LOW, 0x11070)
    DEP_FIELD(SMMU_CB1_IPAFAR_LOW, IPAFAR_L, 20, 12)
    DEP_FIELD(SMMU_CB1_IPAFAR_LOW, FAR_RO, 12, 0)
DEP_REG32(SMMU_CB1_IPAFAR_HIGH, 0x11074)
    DEP_FIELD(SMMU_CB1_IPAFAR_HIGH, BITS, 16, 0)
DEP_REG32(SMMU_CB1_TLBIVA_LOW, 0x11600)
DEP_REG32(SMMU_CB1_TLBIVA_HIGH, 0x11604)
    DEP_FIELD(SMMU_CB1_TLBIVA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB1_TLBIVA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB1_TLBIVAA_LOW, 0x11608)
DEP_REG32(SMMU_CB1_TLBIVAA_HIGH, 0x1160c)
    DEP_FIELD(SMMU_CB1_TLBIVAA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB1_TLBIVAA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB1_TLBIASID, 0x11610)
    DEP_FIELD(SMMU_CB1_TLBIASID, ASID, 16, 0)
DEP_REG32(SMMU_CB1_TLBIALL, 0x11618)
DEP_REG32(SMMU_CB1_TLBIVAL_LOW, 0x11620)
DEP_REG32(SMMU_CB1_TLBIVAL_HIGH, 0x11624)
    DEP_FIELD(SMMU_CB1_TLBIVAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB1_TLBIVAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB1_TLBIVAAL_LOW, 0x11628)
DEP_REG32(SMMU_CB1_TLBIVAAL_HIGH, 0x1162c)
    DEP_FIELD(SMMU_CB1_TLBIVAAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB1_TLBIVAAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB1_TLBIIPAS2_LOW, 0x11630)
DEP_REG32(SMMU_CB1_TLBIIPAS2_HIGH, 0x11634)
    DEP_FIELD(SMMU_CB1_TLBIIPAS2_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB1_TLBIIPAS2L_LOW, 0x11638)
DEP_REG32(SMMU_CB1_TLBIIPAS2L_HIGH, 0x1163c)
    DEP_FIELD(SMMU_CB1_TLBIIPAS2L_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB1_TLBSYNC, 0x117f0)
DEP_REG32(SMMU_CB1_TLBSTATUS, 0x117f4)
    DEP_FIELD(SMMU_CB1_TLBSTATUS, SACTIVE, 1, 0)
DEP_REG32(SMMU_CB1_PMEVCNTR0, 0x11e00)
DEP_REG32(SMMU_CB1_PMEVCNTR1, 0x11e04)
DEP_REG32(SMMU_CB1_PMEVCNTR2, 0x11e08)
DEP_REG32(SMMU_CB1_PMEVCNTR3, 0x11e0c)
DEP_REG32(SMMU_CB1_PMEVTYPER0, 0x11e80)
    DEP_FIELD(SMMU_CB1_PMEVTYPER0, P, 1, 31)
    DEP_FIELD(SMMU_CB1_PMEVTYPER0, U, 1, 30)
    DEP_FIELD(SMMU_CB1_PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(SMMU_CB1_PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(SMMU_CB1_PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(SMMU_CB1_PMEVTYPER1, 0x11e84)
    DEP_FIELD(SMMU_CB1_PMEVTYPER1, P, 1, 31)
    DEP_FIELD(SMMU_CB1_PMEVTYPER1, U, 1, 30)
    DEP_FIELD(SMMU_CB1_PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(SMMU_CB1_PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(SMMU_CB1_PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(SMMU_CB1_PMEVTYPER2, 0x11e88)
    DEP_FIELD(SMMU_CB1_PMEVTYPER2, P, 1, 31)
    DEP_FIELD(SMMU_CB1_PMEVTYPER2, U, 1, 30)
    DEP_FIELD(SMMU_CB1_PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(SMMU_CB1_PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(SMMU_CB1_PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(SMMU_CB1_PMEVTYPER3, 0x11e8c)
    DEP_FIELD(SMMU_CB1_PMEVTYPER3, P, 1, 31)
    DEP_FIELD(SMMU_CB1_PMEVTYPER3, U, 1, 30)
    DEP_FIELD(SMMU_CB1_PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(SMMU_CB1_PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(SMMU_CB1_PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(SMMU_CB1_PMCFGR, 0x11f00)
    DEP_FIELD(SMMU_CB1_PMCFGR, NCG, 8, 24)
    DEP_FIELD(SMMU_CB1_PMCFGR, UEN, 1, 19)
    DEP_FIELD(SMMU_CB1_PMCFGR, EX, 1, 16)
    DEP_FIELD(SMMU_CB1_PMCFGR, CCD, 1, 15)
    DEP_FIELD(SMMU_CB1_PMCFGR, CC, 1, 14)
    DEP_FIELD(SMMU_CB1_PMCFGR, SIZE, 6, 8)
    DEP_FIELD(SMMU_CB1_PMCFGR, N, 8, 0)
DEP_REG32(SMMU_CB1_PMCR, 0x11f04)
    DEP_FIELD(SMMU_CB1_PMCR, IMP, 8, 24)
    DEP_FIELD(SMMU_CB1_PMCR, X, 1, 4)
    DEP_FIELD(SMMU_CB1_PMCR, P, 1, 1)
    DEP_FIELD(SMMU_CB1_PMCR, E, 1, 0)
DEP_REG32(SMMU_CB1_PMCEID, 0x11f20)
    DEP_FIELD(SMMU_CB1_PMCEID, EVENT0X12, 1, 17)
    DEP_FIELD(SMMU_CB1_PMCEID, EVENT0X11, 1, 16)
    DEP_FIELD(SMMU_CB1_PMCEID, EVENT0X10, 1, 15)
    DEP_FIELD(SMMU_CB1_PMCEID, EVENT0X0A, 1, 9)
    DEP_FIELD(SMMU_CB1_PMCEID, EVENT0X09, 1, 8)
    DEP_FIELD(SMMU_CB1_PMCEID, EVENT0X08, 1, 7)
    DEP_FIELD(SMMU_CB1_PMCEID, EVENT0X01, 1, 1)
    DEP_FIELD(SMMU_CB1_PMCEID, EVENT0X00, 1, 0)
DEP_REG32(SMMU_CB1_PMCNTENSE, 0x11f40)
    DEP_FIELD(SMMU_CB1_PMCNTENSE, P3, 1, 3)
    DEP_FIELD(SMMU_CB1_PMCNTENSE, P2, 1, 2)
    DEP_FIELD(SMMU_CB1_PMCNTENSE, P1, 1, 1)
    DEP_FIELD(SMMU_CB1_PMCNTENSE, P0, 1, 0)
DEP_REG32(SMMU_CB1_PMCNTENCLR, 0x11f44)
    DEP_FIELD(SMMU_CB1_PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB1_PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB1_PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB1_PMCNTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB1_PMCNTENSET, 0x11f48)
    DEP_FIELD(SMMU_CB1_PMCNTENSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB1_PMCNTENSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB1_PMCNTENSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB1_PMCNTENSET, P0, 1, 0)
DEP_REG32(SMMU_CB1_PMINTENCLR, 0x11f4c)
    DEP_FIELD(SMMU_CB1_PMINTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB1_PMINTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB1_PMINTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB1_PMINTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB1_PMOVSCLR, 0x11f50)
    DEP_FIELD(SMMU_CB1_PMOVSCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB1_PMOVSCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB1_PMOVSCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB1_PMOVSCLR, P0, 1, 0)
DEP_REG32(SMMU_CB1_PMOVSSET, 0x11f58)
    DEP_FIELD(SMMU_CB1_PMOVSSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB1_PMOVSSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB1_PMOVSSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB1_PMOVSSET, P0, 1, 0)
DEP_REG32(SMMU_CB1_PMAUTHSTATUS, 0x11fb8)
    DEP_FIELD(SMMU_CB1_PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(SMMU_CB1_PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(SMMU_CB1_PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(SMMU_CB1_PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(SMMU_CB1_PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(SMMU_CB1_PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(SMMU_CB1_PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(SMMU_CB1_PMAUTHSTATUS, NSE, 1, 0)
DEP_REG32(SMMU_CB2_SCTLR, 0x12000)
    DEP_FIELD(SMMU_CB2_SCTLR, NSCFG, 2, 28)
    DEP_FIELD(SMMU_CB2_SCTLR, WACFG, 2, 26)
    DEP_FIELD(SMMU_CB2_SCTLR, RACFG, 2, 24)
    DEP_FIELD(SMMU_CB2_SCTLR, SHCFG, 2, 22)
    DEP_FIELD(SMMU_CB2_SCTLR, FB, 1, 21)
    DEP_FIELD(SMMU_CB2_SCTLR, MTCFG, 1, 20)
    DEP_FIELD(SMMU_CB2_SCTLR, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_CB2_SCTLR, TRANSIENTCFG, 2, 14)
    DEP_FIELD(SMMU_CB2_SCTLR, PTW, 1, 13)
    DEP_FIELD(SMMU_CB2_SCTLR, ASIDPNE, 1, 12)
    DEP_FIELD(SMMU_CB2_SCTLR, UWXN, 1, 10)
    DEP_FIELD(SMMU_CB2_SCTLR, WXN, 1, 9)
    DEP_FIELD(SMMU_CB2_SCTLR, HUPCF, 1, 8)
    DEP_FIELD(SMMU_CB2_SCTLR, CFCFG, 1, 7)
    DEP_FIELD(SMMU_CB2_SCTLR, CFIE, 1, 6)
    DEP_FIELD(SMMU_CB2_SCTLR, CFRE, 1, 5)
    DEP_FIELD(SMMU_CB2_SCTLR, E, 1, 4)
    DEP_FIELD(SMMU_CB2_SCTLR, AFFD, 1, 3)
    DEP_FIELD(SMMU_CB2_SCTLR, AFE, 1, 2)
    DEP_FIELD(SMMU_CB2_SCTLR, TRE, 1, 1)
    DEP_FIELD(SMMU_CB2_SCTLR, M, 1, 0)
DEP_REG32(SMMU_CB2_ACTLR, 0x12004)
    DEP_FIELD(SMMU_CB2_ACTLR, CPRE, 1, 1)
    DEP_FIELD(SMMU_CB2_ACTLR, CMTLB, 1, 0)
DEP_REG32(SMMU_CB2_RESUME, 0x12008)
    DEP_FIELD(SMMU_CB2_RESUME, TNR, 1, 0)
DEP_REG32(SMMU_CB2_TCR2, 0x12010)
    DEP_FIELD(SMMU_CB2_TCR2, NSCFG1, 1, 30)
    DEP_FIELD(SMMU_CB2_TCR2, SEP, 3, 15)
    DEP_FIELD(SMMU_CB2_TCR2, NSCFG0, 1, 14)
    DEP_FIELD(SMMU_CB2_TCR2, TBI1, 1, 6)
    DEP_FIELD(SMMU_CB2_TCR2, TBI0, 1, 5)
    DEP_FIELD(SMMU_CB2_TCR2, AS, 1, 4)
    DEP_FIELD(SMMU_CB2_TCR2, PASIZE, 3, 0)
DEP_REG32(SMMU_CB2_TTBR0_LOW, 0x12020)
    DEP_FIELD(SMMU_CB2_TTBR0_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB2_TTBR0_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB2_TTBR0_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB2_TTBR0_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB2_TTBR0_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB2_TTBR0_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB2_TTBR0_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB2_TTBR0_HIGH, 0x12024)
    DEP_FIELD(SMMU_CB2_TTBR0_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB2_TTBR0_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB2_TTBR1_LOW, 0x12028)
    DEP_FIELD(SMMU_CB2_TTBR1_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB2_TTBR1_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB2_TTBR1_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB2_TTBR1_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB2_TTBR1_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB2_TTBR1_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB2_TTBR1_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB2_TTBR1_HIGH, 0x1202c)
    DEP_FIELD(SMMU_CB2_TTBR1_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB2_TTBR1_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB2_TCR_LPAE, 0x12030)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, EAE, 1, 31)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, NSCFG1_TG1, 1, 30)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, SH1, 2, 28)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, ORGN1, 2, 26)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, IRGN1, 2, 24)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, EPD1, 1, 23)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, A1, 1, 22)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, T1SZ_5_3, 3, 19)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, T1SZ_2_0_PASIZE, 3, 16)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, NSCFG0_TG0, 1, 14)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, SH0, 2, 12)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, ORGN0, 2, 10)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, IRGN0, 2, 8)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, SL0_1_EPD0, 1, 7)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, SL0_0, 1, 6)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, PD1_T0SZ_5, 1, 5)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, S_PD0_T0SZ_4, 1, 4)
    DEP_FIELD(SMMU_CB2_TCR_LPAE, T0SZ_3_0, 4, 0)
DEP_REG32(SMMU_CB2_CONTEXTIDR, 0x12034)
    DEP_FIELD(SMMU_CB2_CONTEXTIDR, PROCID, 24, 8)
    DEP_FIELD(SMMU_CB2_CONTEXTIDR, ASID, 8, 0)
DEP_REG32(SMMU_CB2_PRRR_MAIR0, 0x12038)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, NOS7, 1, 31)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, NOS6, 1, 30)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, NOS5, 1, 29)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, NOS4, 1, 28)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, NOS3, 1, 27)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, NOS2, 1, 26)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, NOS1, 1, 25)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, NOS0, 1, 24)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, NS1, 1, 19)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, NS0, 1, 18)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, DS1, 1, 17)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, DS0, 1, 16)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, TR7, 2, 14)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, TR6, 2, 12)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, TR5, 2, 10)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, TR4, 2, 8)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, TR3, 2, 6)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, TR2, 2, 4)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, TR1, 2, 2)
    DEP_FIELD(SMMU_CB2_PRRR_MAIR0, TR0, 2, 0)
DEP_REG32(SMMU_CB2_NMRR_MAIR1, 0x1203c)
    DEP_FIELD(SMMU_CB2_NMRR_MAIR1, OR7, 2, 30)
    DEP_FIELD(SMMU_CB2_NMRR_MAIR1, OR6, 2, 28)
    DEP_FIELD(SMMU_CB2_NMRR_MAIR1, OR5, 2, 26)
    DEP_FIELD(SMMU_CB2_NMRR_MAIR1, OR4, 2, 24)
    DEP_FIELD(SMMU_CB2_NMRR_MAIR1, OR3, 2, 22)
    DEP_FIELD(SMMU_CB2_NMRR_MAIR1, OR2, 2, 20)
    DEP_FIELD(SMMU_CB2_NMRR_MAIR1, OR1, 2, 18)
    DEP_FIELD(SMMU_CB2_NMRR_MAIR1, OR0, 2, 16)
    DEP_FIELD(SMMU_CB2_NMRR_MAIR1, IR7, 2, 14)
    DEP_FIELD(SMMU_CB2_NMRR_MAIR1, IR6, 2, 12)
    DEP_FIELD(SMMU_CB2_NMRR_MAIR1, IR5, 2, 10)
    DEP_FIELD(SMMU_CB2_NMRR_MAIR1, IR4, 2, 8)
    DEP_FIELD(SMMU_CB2_NMRR_MAIR1, IR3, 2, 6)
    DEP_FIELD(SMMU_CB2_NMRR_MAIR1, IR2, 2, 4)
    DEP_FIELD(SMMU_CB2_NMRR_MAIR1, IR1, 2, 2)
    DEP_FIELD(SMMU_CB2_NMRR_MAIR1, IR0, 2, 0)
DEP_REG32(SMMU_CB2_FSR, 0x12058)
    DEP_FIELD(SMMU_CB2_FSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_CB2_FSR, SS, 1, 30)
    DEP_FIELD(SMMU_CB2_FSR, FORMAT, 2, 9)
    DEP_FIELD(SMMU_CB2_FSR, UUT, 1, 8)
    DEP_FIELD(SMMU_CB2_FSR, ASF, 1, 7)
    DEP_FIELD(SMMU_CB2_FSR, TLBLKF, 1, 6)
    DEP_FIELD(SMMU_CB2_FSR, TLBMCF, 1, 5)
    DEP_FIELD(SMMU_CB2_FSR, EF, 1, 4)
    DEP_FIELD(SMMU_CB2_FSR, PF, 1, 3)
    DEP_FIELD(SMMU_CB2_FSR, AFF, 1, 2)
    DEP_FIELD(SMMU_CB2_FSR, TF, 1, 1)
DEP_REG32(SMMU_CB2_FSRRESTORE, 0x1205c)
DEP_REG32(SMMU_CB2_FAR_LOW, 0x12060)
DEP_REG32(SMMU_CB2_FAR_HIGH, 0x12064)
    DEP_FIELD(SMMU_CB2_FAR_HIGH, BITS, 17, 0)
DEP_REG32(SMMU_CB2_FSYNR0, 0x12068)
    DEP_FIELD(SMMU_CB2_FSYNR0, S1CBNDX, 4, 16)
    DEP_FIELD(SMMU_CB2_FSYNR0, AFR, 1, 11)
    DEP_FIELD(SMMU_CB2_FSYNR0, PTWF, 1, 10)
    DEP_FIELD(SMMU_CB2_FSYNR0, ATOF, 1, 9)
    DEP_FIELD(SMMU_CB2_FSYNR0, NSATTR, 1, 8)
    DEP_FIELD(SMMU_CB2_FSYNR0, IND, 1, 6)
    DEP_FIELD(SMMU_CB2_FSYNR0, PNU, 1, 5)
    DEP_FIELD(SMMU_CB2_FSYNR0, WNR, 1, 4)
    DEP_FIELD(SMMU_CB2_FSYNR0, PLVL, 2, 0)
DEP_REG32(SMMU_CB2_IPAFAR_LOW, 0x12070)
    DEP_FIELD(SMMU_CB2_IPAFAR_LOW, IPAFAR_L, 20, 12)
    DEP_FIELD(SMMU_CB2_IPAFAR_LOW, FAR_RO, 12, 0)
DEP_REG32(SMMU_CB2_IPAFAR_HIGH, 0x12074)
    DEP_FIELD(SMMU_CB2_IPAFAR_HIGH, BITS, 16, 0)
DEP_REG32(SMMU_CB2_TLBIVA_LOW, 0x12600)
DEP_REG32(SMMU_CB2_TLBIVA_HIGH, 0x12604)
    DEP_FIELD(SMMU_CB2_TLBIVA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB2_TLBIVA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB2_TLBIVAA_LOW, 0x12608)
DEP_REG32(SMMU_CB2_TLBIVAA_HIGH, 0x1260c)
    DEP_FIELD(SMMU_CB2_TLBIVAA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB2_TLBIVAA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB2_TLBIASID, 0x12610)
    DEP_FIELD(SMMU_CB2_TLBIASID, ASID, 16, 0)
DEP_REG32(SMMU_CB2_TLBIALL, 0x12618)
DEP_REG32(SMMU_CB2_TLBIVAL_LOW, 0x12620)
DEP_REG32(SMMU_CB2_TLBIVAL_HIGH, 0x12624)
    DEP_FIELD(SMMU_CB2_TLBIVAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB2_TLBIVAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB2_TLBIVAAL_LOW, 0x12628)
DEP_REG32(SMMU_CB2_TLBIVAAL_HIGH, 0x1262c)
    DEP_FIELD(SMMU_CB2_TLBIVAAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB2_TLBIVAAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB2_TLBIIPAS2_LOW, 0x12630)
DEP_REG32(SMMU_CB2_TLBIIPAS2_HIGH, 0x12634)
    DEP_FIELD(SMMU_CB2_TLBIIPAS2_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB2_TLBIIPAS2L_LOW, 0x12638)
DEP_REG32(SMMU_CB2_TLBIIPAS2L_HIGH, 0x1263c)
    DEP_FIELD(SMMU_CB2_TLBIIPAS2L_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB2_TLBSYNC, 0x127f0)
DEP_REG32(SMMU_CB2_TLBSTATUS, 0x127f4)
    DEP_FIELD(SMMU_CB2_TLBSTATUS, SACTIVE, 1, 0)
DEP_REG32(SMMU_CB2_PMEVCNTR0, 0x12e00)
DEP_REG32(SMMU_CB2_PMEVCNTR1, 0x12e04)
DEP_REG32(SMMU_CB2_PMEVCNTR2, 0x12e08)
DEP_REG32(SMMU_CB2_PMEVCNTR3, 0x12e0c)
DEP_REG32(SMMU_CB2_PMEVTYPER0, 0x12e80)
    DEP_FIELD(SMMU_CB2_PMEVTYPER0, P, 1, 31)
    DEP_FIELD(SMMU_CB2_PMEVTYPER0, U, 1, 30)
    DEP_FIELD(SMMU_CB2_PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(SMMU_CB2_PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(SMMU_CB2_PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(SMMU_CB2_PMEVTYPER1, 0x12e84)
    DEP_FIELD(SMMU_CB2_PMEVTYPER1, P, 1, 31)
    DEP_FIELD(SMMU_CB2_PMEVTYPER1, U, 1, 30)
    DEP_FIELD(SMMU_CB2_PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(SMMU_CB2_PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(SMMU_CB2_PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(SMMU_CB2_PMEVTYPER2, 0x12e88)
    DEP_FIELD(SMMU_CB2_PMEVTYPER2, P, 1, 31)
    DEP_FIELD(SMMU_CB2_PMEVTYPER2, U, 1, 30)
    DEP_FIELD(SMMU_CB2_PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(SMMU_CB2_PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(SMMU_CB2_PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(SMMU_CB2_PMEVTYPER3, 0x12e8c)
    DEP_FIELD(SMMU_CB2_PMEVTYPER3, P, 1, 31)
    DEP_FIELD(SMMU_CB2_PMEVTYPER3, U, 1, 30)
    DEP_FIELD(SMMU_CB2_PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(SMMU_CB2_PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(SMMU_CB2_PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(SMMU_CB2_PMCFGR, 0x12f00)
    DEP_FIELD(SMMU_CB2_PMCFGR, NCG, 8, 24)
    DEP_FIELD(SMMU_CB2_PMCFGR, UEN, 1, 19)
    DEP_FIELD(SMMU_CB2_PMCFGR, EX, 1, 16)
    DEP_FIELD(SMMU_CB2_PMCFGR, CCD, 1, 15)
    DEP_FIELD(SMMU_CB2_PMCFGR, CC, 1, 14)
    DEP_FIELD(SMMU_CB2_PMCFGR, SIZE, 6, 8)
    DEP_FIELD(SMMU_CB2_PMCFGR, N, 8, 0)
DEP_REG32(SMMU_CB2_PMCR, 0x12f04)
    DEP_FIELD(SMMU_CB2_PMCR, IMP, 8, 24)
    DEP_FIELD(SMMU_CB2_PMCR, X, 1, 4)
    DEP_FIELD(SMMU_CB2_PMCR, P, 1, 1)
    DEP_FIELD(SMMU_CB2_PMCR, E, 1, 0)
DEP_REG32(SMMU_CB2_PMCEID, 0x12f20)
    DEP_FIELD(SMMU_CB2_PMCEID, EVENT0X12, 1, 17)
    DEP_FIELD(SMMU_CB2_PMCEID, EVENT0X11, 1, 16)
    DEP_FIELD(SMMU_CB2_PMCEID, EVENT0X10, 1, 15)
    DEP_FIELD(SMMU_CB2_PMCEID, EVENT0X0A, 1, 9)
    DEP_FIELD(SMMU_CB2_PMCEID, EVENT0X09, 1, 8)
    DEP_FIELD(SMMU_CB2_PMCEID, EVENT0X08, 1, 7)
    DEP_FIELD(SMMU_CB2_PMCEID, EVENT0X01, 1, 1)
    DEP_FIELD(SMMU_CB2_PMCEID, EVENT0X00, 1, 0)
DEP_REG32(SMMU_CB2_PMCNTENSE, 0x12f40)
    DEP_FIELD(SMMU_CB2_PMCNTENSE, P3, 1, 3)
    DEP_FIELD(SMMU_CB2_PMCNTENSE, P2, 1, 2)
    DEP_FIELD(SMMU_CB2_PMCNTENSE, P1, 1, 1)
    DEP_FIELD(SMMU_CB2_PMCNTENSE, P0, 1, 0)
DEP_REG32(SMMU_CB2_PMCNTENCLR, 0x12f44)
    DEP_FIELD(SMMU_CB2_PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB2_PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB2_PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB2_PMCNTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB2_PMCNTENSET, 0x12f48)
    DEP_FIELD(SMMU_CB2_PMCNTENSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB2_PMCNTENSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB2_PMCNTENSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB2_PMCNTENSET, P0, 1, 0)
DEP_REG32(SMMU_CB2_PMINTENCLR, 0x12f4c)
    DEP_FIELD(SMMU_CB2_PMINTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB2_PMINTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB2_PMINTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB2_PMINTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB2_PMOVSCLR, 0x12f50)
    DEP_FIELD(SMMU_CB2_PMOVSCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB2_PMOVSCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB2_PMOVSCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB2_PMOVSCLR, P0, 1, 0)
DEP_REG32(SMMU_CB2_PMOVSSET, 0x12f58)
    DEP_FIELD(SMMU_CB2_PMOVSSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB2_PMOVSSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB2_PMOVSSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB2_PMOVSSET, P0, 1, 0)
DEP_REG32(SMMU_CB2_PMAUTHSTATUS, 0x12fb8)
    DEP_FIELD(SMMU_CB2_PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(SMMU_CB2_PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(SMMU_CB2_PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(SMMU_CB2_PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(SMMU_CB2_PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(SMMU_CB2_PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(SMMU_CB2_PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(SMMU_CB2_PMAUTHSTATUS, NSE, 1, 0)
DEP_REG32(SMMU_CB3_SCTLR, 0x13000)
    DEP_FIELD(SMMU_CB3_SCTLR, NSCFG, 2, 28)
    DEP_FIELD(SMMU_CB3_SCTLR, WACFG, 2, 26)
    DEP_FIELD(SMMU_CB3_SCTLR, RACFG, 2, 24)
    DEP_FIELD(SMMU_CB3_SCTLR, SHCFG, 2, 22)
    DEP_FIELD(SMMU_CB3_SCTLR, FB, 1, 21)
    DEP_FIELD(SMMU_CB3_SCTLR, MTCFG, 1, 20)
    DEP_FIELD(SMMU_CB3_SCTLR, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_CB3_SCTLR, TRANSIENTCFG, 2, 14)
    DEP_FIELD(SMMU_CB3_SCTLR, PTW, 1, 13)
    DEP_FIELD(SMMU_CB3_SCTLR, ASIDPNE, 1, 12)
    DEP_FIELD(SMMU_CB3_SCTLR, UWXN, 1, 10)
    DEP_FIELD(SMMU_CB3_SCTLR, WXN, 1, 9)
    DEP_FIELD(SMMU_CB3_SCTLR, HUPCF, 1, 8)
    DEP_FIELD(SMMU_CB3_SCTLR, CFCFG, 1, 7)
    DEP_FIELD(SMMU_CB3_SCTLR, CFIE, 1, 6)
    DEP_FIELD(SMMU_CB3_SCTLR, CFRE, 1, 5)
    DEP_FIELD(SMMU_CB3_SCTLR, E, 1, 4)
    DEP_FIELD(SMMU_CB3_SCTLR, AFFD, 1, 3)
    DEP_FIELD(SMMU_CB3_SCTLR, AFE, 1, 2)
    DEP_FIELD(SMMU_CB3_SCTLR, TRE, 1, 1)
    DEP_FIELD(SMMU_CB3_SCTLR, M, 1, 0)
DEP_REG32(SMMU_CB3_ACTLR, 0x13004)
    DEP_FIELD(SMMU_CB3_ACTLR, CPRE, 1, 1)
    DEP_FIELD(SMMU_CB3_ACTLR, CMTLB, 1, 0)
DEP_REG32(SMMU_CB3_RESUME, 0x13008)
    DEP_FIELD(SMMU_CB3_RESUME, TNR, 1, 0)
DEP_REG32(SMMU_CB3_TCR2, 0x13010)
    DEP_FIELD(SMMU_CB3_TCR2, NSCFG1, 1, 30)
    DEP_FIELD(SMMU_CB3_TCR2, SEP, 3, 15)
    DEP_FIELD(SMMU_CB3_TCR2, NSCFG0, 1, 14)
    DEP_FIELD(SMMU_CB3_TCR2, TBI1, 1, 6)
    DEP_FIELD(SMMU_CB3_TCR2, TBI0, 1, 5)
    DEP_FIELD(SMMU_CB3_TCR2, AS, 1, 4)
    DEP_FIELD(SMMU_CB3_TCR2, PASIZE, 3, 0)
DEP_REG32(SMMU_CB3_TTBR0_LOW, 0x13020)
    DEP_FIELD(SMMU_CB3_TTBR0_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB3_TTBR0_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB3_TTBR0_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB3_TTBR0_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB3_TTBR0_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB3_TTBR0_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB3_TTBR0_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB3_TTBR0_HIGH, 0x13024)
    DEP_FIELD(SMMU_CB3_TTBR0_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB3_TTBR0_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB3_TTBR1_LOW, 0x13028)
    DEP_FIELD(SMMU_CB3_TTBR1_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB3_TTBR1_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB3_TTBR1_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB3_TTBR1_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB3_TTBR1_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB3_TTBR1_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB3_TTBR1_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB3_TTBR1_HIGH, 0x1302c)
    DEP_FIELD(SMMU_CB3_TTBR1_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB3_TTBR1_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB3_TCR_LPAE, 0x13030)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, EAE, 1, 31)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, NSCFG1_TG1, 1, 30)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, SH1, 2, 28)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, ORGN1, 2, 26)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, IRGN1, 2, 24)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, EPD1, 1, 23)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, A1, 1, 22)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, T1SZ_5_3, 3, 19)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, T1SZ_2_0_PASIZE, 3, 16)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, NSCFG0_TG0, 1, 14)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, SH0, 2, 12)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, ORGN0, 2, 10)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, IRGN0, 2, 8)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, SL0_1_EPD0, 1, 7)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, SL0_0, 1, 6)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, PD1_T0SZ_5, 1, 5)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, S_PD0_T0SZ_4, 1, 4)
    DEP_FIELD(SMMU_CB3_TCR_LPAE, T0SZ_3_0, 4, 0)
DEP_REG32(SMMU_CB3_CONTEXTIDR, 0x13034)
    DEP_FIELD(SMMU_CB3_CONTEXTIDR, PROCID, 24, 8)
    DEP_FIELD(SMMU_CB3_CONTEXTIDR, ASID, 8, 0)
DEP_REG32(SMMU_CB3_PRRR_MAIR0, 0x13038)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, NOS7, 1, 31)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, NOS6, 1, 30)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, NOS5, 1, 29)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, NOS4, 1, 28)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, NOS3, 1, 27)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, NOS2, 1, 26)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, NOS1, 1, 25)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, NOS0, 1, 24)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, NS1, 1, 19)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, NS0, 1, 18)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, DS1, 1, 17)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, DS0, 1, 16)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, TR7, 2, 14)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, TR6, 2, 12)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, TR5, 2, 10)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, TR4, 2, 8)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, TR3, 2, 6)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, TR2, 2, 4)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, TR1, 2, 2)
    DEP_FIELD(SMMU_CB3_PRRR_MAIR0, TR0, 2, 0)
DEP_REG32(SMMU_CB3_NMRR_MAIR1, 0x1303c)
    DEP_FIELD(SMMU_CB3_NMRR_MAIR1, OR7, 2, 30)
    DEP_FIELD(SMMU_CB3_NMRR_MAIR1, OR6, 2, 28)
    DEP_FIELD(SMMU_CB3_NMRR_MAIR1, OR5, 2, 26)
    DEP_FIELD(SMMU_CB3_NMRR_MAIR1, OR4, 2, 24)
    DEP_FIELD(SMMU_CB3_NMRR_MAIR1, OR3, 2, 22)
    DEP_FIELD(SMMU_CB3_NMRR_MAIR1, OR2, 2, 20)
    DEP_FIELD(SMMU_CB3_NMRR_MAIR1, OR1, 2, 18)
    DEP_FIELD(SMMU_CB3_NMRR_MAIR1, OR0, 2, 16)
    DEP_FIELD(SMMU_CB3_NMRR_MAIR1, IR7, 2, 14)
    DEP_FIELD(SMMU_CB3_NMRR_MAIR1, IR6, 2, 12)
    DEP_FIELD(SMMU_CB3_NMRR_MAIR1, IR5, 2, 10)
    DEP_FIELD(SMMU_CB3_NMRR_MAIR1, IR4, 2, 8)
    DEP_FIELD(SMMU_CB3_NMRR_MAIR1, IR3, 2, 6)
    DEP_FIELD(SMMU_CB3_NMRR_MAIR1, IR2, 2, 4)
    DEP_FIELD(SMMU_CB3_NMRR_MAIR1, IR1, 2, 2)
    DEP_FIELD(SMMU_CB3_NMRR_MAIR1, IR0, 2, 0)
DEP_REG32(SMMU_CB3_FSR, 0x13058)
    DEP_FIELD(SMMU_CB3_FSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_CB3_FSR, SS, 1, 30)
    DEP_FIELD(SMMU_CB3_FSR, FORMAT, 2, 9)
    DEP_FIELD(SMMU_CB3_FSR, UUT, 1, 8)
    DEP_FIELD(SMMU_CB3_FSR, ASF, 1, 7)
    DEP_FIELD(SMMU_CB3_FSR, TLBLKF, 1, 6)
    DEP_FIELD(SMMU_CB3_FSR, TLBMCF, 1, 5)
    DEP_FIELD(SMMU_CB3_FSR, EF, 1, 4)
    DEP_FIELD(SMMU_CB3_FSR, PF, 1, 3)
    DEP_FIELD(SMMU_CB3_FSR, AFF, 1, 2)
    DEP_FIELD(SMMU_CB3_FSR, TF, 1, 1)
DEP_REG32(SMMU_CB3_FSRRESTORE, 0x1305c)
DEP_REG32(SMMU_CB3_FAR_LOW, 0x13060)
DEP_REG32(SMMU_CB3_FAR_HIGH, 0x13064)
    DEP_FIELD(SMMU_CB3_FAR_HIGH, BITS, 17, 0)
DEP_REG32(SMMU_CB3_FSYNR0, 0x13068)
    DEP_FIELD(SMMU_CB3_FSYNR0, S1CBNDX, 4, 16)
    DEP_FIELD(SMMU_CB3_FSYNR0, AFR, 1, 11)
    DEP_FIELD(SMMU_CB3_FSYNR0, PTWF, 1, 10)
    DEP_FIELD(SMMU_CB3_FSYNR0, ATOF, 1, 9)
    DEP_FIELD(SMMU_CB3_FSYNR0, NSATTR, 1, 8)
    DEP_FIELD(SMMU_CB3_FSYNR0, IND, 1, 6)
    DEP_FIELD(SMMU_CB3_FSYNR0, PNU, 1, 5)
    DEP_FIELD(SMMU_CB3_FSYNR0, WNR, 1, 4)
    DEP_FIELD(SMMU_CB3_FSYNR0, PLVL, 2, 0)
DEP_REG32(SMMU_CB3_IPAFAR_LOW, 0x13070)
    DEP_FIELD(SMMU_CB3_IPAFAR_LOW, IPAFAR_L, 20, 12)
    DEP_FIELD(SMMU_CB3_IPAFAR_LOW, FAR_RO, 12, 0)
DEP_REG32(SMMU_CB3_IPAFAR_HIGH, 0x13074)
    DEP_FIELD(SMMU_CB3_IPAFAR_HIGH, BITS, 16, 0)
DEP_REG32(SMMU_CB3_TLBIVA_LOW, 0x13600)
DEP_REG32(SMMU_CB3_TLBIVA_HIGH, 0x13604)
    DEP_FIELD(SMMU_CB3_TLBIVA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB3_TLBIVA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB3_TLBIVAA_LOW, 0x13608)
DEP_REG32(SMMU_CB3_TLBIVAA_HIGH, 0x1360c)
    DEP_FIELD(SMMU_CB3_TLBIVAA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB3_TLBIVAA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB3_TLBIASID, 0x13610)
    DEP_FIELD(SMMU_CB3_TLBIASID, ASID, 16, 0)
DEP_REG32(SMMU_CB3_TLBIALL, 0x13618)
DEP_REG32(SMMU_CB3_TLBIVAL_LOW, 0x13620)
DEP_REG32(SMMU_CB3_TLBIVAL_HIGH, 0x13624)
    DEP_FIELD(SMMU_CB3_TLBIVAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB3_TLBIVAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB3_TLBIVAAL_LOW, 0x13628)
DEP_REG32(SMMU_CB3_TLBIVAAL_HIGH, 0x1362c)
    DEP_FIELD(SMMU_CB3_TLBIVAAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB3_TLBIVAAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB3_TLBIIPAS2_LOW, 0x13630)
DEP_REG32(SMMU_CB3_TLBIIPAS2_HIGH, 0x13634)
    DEP_FIELD(SMMU_CB3_TLBIIPAS2_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB3_TLBIIPAS2L_LOW, 0x13638)
DEP_REG32(SMMU_CB3_TLBIIPAS2L_HIGH, 0x1363c)
    DEP_FIELD(SMMU_CB3_TLBIIPAS2L_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB3_TLBSYNC, 0x137f0)
DEP_REG32(SMMU_CB3_TLBSTATUS, 0x137f4)
    DEP_FIELD(SMMU_CB3_TLBSTATUS, SACTIVE, 1, 0)
DEP_REG32(SMMU_CB3_PMEVCNTR0, 0x13e00)
DEP_REG32(SMMU_CB3_PMEVCNTR1, 0x13e04)
DEP_REG32(SMMU_CB3_PMEVCNTR2, 0x13e08)
DEP_REG32(SMMU_CB3_PMEVCNTR3, 0x13e0c)
DEP_REG32(SMMU_CB3_PMEVTYPER0, 0x13e80)
    DEP_FIELD(SMMU_CB3_PMEVTYPER0, P, 1, 31)
    DEP_FIELD(SMMU_CB3_PMEVTYPER0, U, 1, 30)
    DEP_FIELD(SMMU_CB3_PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(SMMU_CB3_PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(SMMU_CB3_PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(SMMU_CB3_PMEVTYPER1, 0x13e84)
    DEP_FIELD(SMMU_CB3_PMEVTYPER1, P, 1, 31)
    DEP_FIELD(SMMU_CB3_PMEVTYPER1, U, 1, 30)
    DEP_FIELD(SMMU_CB3_PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(SMMU_CB3_PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(SMMU_CB3_PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(SMMU_CB3_PMEVTYPER2, 0x13e88)
    DEP_FIELD(SMMU_CB3_PMEVTYPER2, P, 1, 31)
    DEP_FIELD(SMMU_CB3_PMEVTYPER2, U, 1, 30)
    DEP_FIELD(SMMU_CB3_PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(SMMU_CB3_PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(SMMU_CB3_PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(SMMU_CB3_PMEVTYPER3, 0x13e8c)
    DEP_FIELD(SMMU_CB3_PMEVTYPER3, P, 1, 31)
    DEP_FIELD(SMMU_CB3_PMEVTYPER3, U, 1, 30)
    DEP_FIELD(SMMU_CB3_PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(SMMU_CB3_PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(SMMU_CB3_PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(SMMU_CB3_PMCFGR, 0x13f00)
    DEP_FIELD(SMMU_CB3_PMCFGR, NCG, 8, 24)
    DEP_FIELD(SMMU_CB3_PMCFGR, UEN, 1, 19)
    DEP_FIELD(SMMU_CB3_PMCFGR, EX, 1, 16)
    DEP_FIELD(SMMU_CB3_PMCFGR, CCD, 1, 15)
    DEP_FIELD(SMMU_CB3_PMCFGR, CC, 1, 14)
    DEP_FIELD(SMMU_CB3_PMCFGR, SIZE, 6, 8)
    DEP_FIELD(SMMU_CB3_PMCFGR, N, 8, 0)
DEP_REG32(SMMU_CB3_PMCR, 0x13f04)
    DEP_FIELD(SMMU_CB3_PMCR, IMP, 8, 24)
    DEP_FIELD(SMMU_CB3_PMCR, X, 1, 4)
    DEP_FIELD(SMMU_CB3_PMCR, P, 1, 1)
    DEP_FIELD(SMMU_CB3_PMCR, E, 1, 0)
DEP_REG32(SMMU_CB3_PMCEID, 0x13f20)
    DEP_FIELD(SMMU_CB3_PMCEID, EVENT0X12, 1, 17)
    DEP_FIELD(SMMU_CB3_PMCEID, EVENT0X11, 1, 16)
    DEP_FIELD(SMMU_CB3_PMCEID, EVENT0X10, 1, 15)
    DEP_FIELD(SMMU_CB3_PMCEID, EVENT0X0A, 1, 9)
    DEP_FIELD(SMMU_CB3_PMCEID, EVENT0X09, 1, 8)
    DEP_FIELD(SMMU_CB3_PMCEID, EVENT0X08, 1, 7)
    DEP_FIELD(SMMU_CB3_PMCEID, EVENT0X01, 1, 1)
    DEP_FIELD(SMMU_CB3_PMCEID, EVENT0X00, 1, 0)
DEP_REG32(SMMU_CB3_PMCNTENSE, 0x13f40)
    DEP_FIELD(SMMU_CB3_PMCNTENSE, P3, 1, 3)
    DEP_FIELD(SMMU_CB3_PMCNTENSE, P2, 1, 2)
    DEP_FIELD(SMMU_CB3_PMCNTENSE, P1, 1, 1)
    DEP_FIELD(SMMU_CB3_PMCNTENSE, P0, 1, 0)
DEP_REG32(SMMU_CB3_PMCNTENCLR, 0x13f44)
    DEP_FIELD(SMMU_CB3_PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB3_PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB3_PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB3_PMCNTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB3_PMCNTENSET, 0x13f48)
    DEP_FIELD(SMMU_CB3_PMCNTENSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB3_PMCNTENSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB3_PMCNTENSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB3_PMCNTENSET, P0, 1, 0)
DEP_REG32(SMMU_CB3_PMINTENCLR, 0x13f4c)
    DEP_FIELD(SMMU_CB3_PMINTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB3_PMINTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB3_PMINTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB3_PMINTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB3_PMOVSCLR, 0x13f50)
    DEP_FIELD(SMMU_CB3_PMOVSCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB3_PMOVSCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB3_PMOVSCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB3_PMOVSCLR, P0, 1, 0)
DEP_REG32(SMMU_CB3_PMOVSSET, 0x13f58)
    DEP_FIELD(SMMU_CB3_PMOVSSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB3_PMOVSSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB3_PMOVSSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB3_PMOVSSET, P0, 1, 0)
DEP_REG32(SMMU_CB3_PMAUTHSTATUS, 0x13fb8)
    DEP_FIELD(SMMU_CB3_PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(SMMU_CB3_PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(SMMU_CB3_PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(SMMU_CB3_PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(SMMU_CB3_PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(SMMU_CB3_PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(SMMU_CB3_PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(SMMU_CB3_PMAUTHSTATUS, NSE, 1, 0)
DEP_REG32(SMMU_CB4_SCTLR, 0x14000)
    DEP_FIELD(SMMU_CB4_SCTLR, NSCFG, 2, 28)
    DEP_FIELD(SMMU_CB4_SCTLR, WACFG, 2, 26)
    DEP_FIELD(SMMU_CB4_SCTLR, RACFG, 2, 24)
    DEP_FIELD(SMMU_CB4_SCTLR, SHCFG, 2, 22)
    DEP_FIELD(SMMU_CB4_SCTLR, FB, 1, 21)
    DEP_FIELD(SMMU_CB4_SCTLR, MTCFG, 1, 20)
    DEP_FIELD(SMMU_CB4_SCTLR, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_CB4_SCTLR, TRANSIENTCFG, 2, 14)
    DEP_FIELD(SMMU_CB4_SCTLR, PTW, 1, 13)
    DEP_FIELD(SMMU_CB4_SCTLR, ASIDPNE, 1, 12)
    DEP_FIELD(SMMU_CB4_SCTLR, UWXN, 1, 10)
    DEP_FIELD(SMMU_CB4_SCTLR, WXN, 1, 9)
    DEP_FIELD(SMMU_CB4_SCTLR, HUPCF, 1, 8)
    DEP_FIELD(SMMU_CB4_SCTLR, CFCFG, 1, 7)
    DEP_FIELD(SMMU_CB4_SCTLR, CFIE, 1, 6)
    DEP_FIELD(SMMU_CB4_SCTLR, CFRE, 1, 5)
    DEP_FIELD(SMMU_CB4_SCTLR, E, 1, 4)
    DEP_FIELD(SMMU_CB4_SCTLR, AFFD, 1, 3)
    DEP_FIELD(SMMU_CB4_SCTLR, AFE, 1, 2)
    DEP_FIELD(SMMU_CB4_SCTLR, TRE, 1, 1)
    DEP_FIELD(SMMU_CB4_SCTLR, M, 1, 0)
DEP_REG32(SMMU_CB4_ACTLR, 0x14004)
    DEP_FIELD(SMMU_CB4_ACTLR, CPRE, 1, 1)
    DEP_FIELD(SMMU_CB4_ACTLR, CMTLB, 1, 0)
DEP_REG32(SMMU_CB4_RESUME, 0x14008)
    DEP_FIELD(SMMU_CB4_RESUME, TNR, 1, 0)
DEP_REG32(SMMU_CB4_TCR2, 0x14010)
    DEP_FIELD(SMMU_CB4_TCR2, NSCFG1, 1, 30)
    DEP_FIELD(SMMU_CB4_TCR2, SEP, 3, 15)
    DEP_FIELD(SMMU_CB4_TCR2, NSCFG0, 1, 14)
    DEP_FIELD(SMMU_CB4_TCR2, TBI1, 1, 6)
    DEP_FIELD(SMMU_CB4_TCR2, TBI0, 1, 5)
    DEP_FIELD(SMMU_CB4_TCR2, AS, 1, 4)
    DEP_FIELD(SMMU_CB4_TCR2, PASIZE, 3, 0)
DEP_REG32(SMMU_CB4_TTBR0_LOW, 0x14020)
    DEP_FIELD(SMMU_CB4_TTBR0_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB4_TTBR0_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB4_TTBR0_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB4_TTBR0_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB4_TTBR0_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB4_TTBR0_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB4_TTBR0_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB4_TTBR0_HIGH, 0x14024)
    DEP_FIELD(SMMU_CB4_TTBR0_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB4_TTBR0_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB4_TTBR1_LOW, 0x14028)
    DEP_FIELD(SMMU_CB4_TTBR1_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB4_TTBR1_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB4_TTBR1_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB4_TTBR1_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB4_TTBR1_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB4_TTBR1_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB4_TTBR1_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB4_TTBR1_HIGH, 0x1402c)
    DEP_FIELD(SMMU_CB4_TTBR1_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB4_TTBR1_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB4_TCR_LPAE, 0x14030)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, EAE, 1, 31)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, NSCFG1_TG1, 1, 30)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, SH1, 2, 28)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, ORGN1, 2, 26)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, IRGN1, 2, 24)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, EPD1, 1, 23)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, A1, 1, 22)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, T1SZ_5_3, 3, 19)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, T1SZ_2_0_PASIZE, 3, 16)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, NSCFG0_TG0, 1, 14)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, SH0, 2, 12)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, ORGN0, 2, 10)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, IRGN0, 2, 8)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, SL0_1_EPD0, 1, 7)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, SL0_0, 1, 6)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, PD1_T0SZ_5, 1, 5)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, S_PD0_T0SZ_4, 1, 4)
    DEP_FIELD(SMMU_CB4_TCR_LPAE, T0SZ_3_0, 4, 0)
DEP_REG32(SMMU_CB4_CONTEXTIDR, 0x14034)
    DEP_FIELD(SMMU_CB4_CONTEXTIDR, PROCID, 24, 8)
    DEP_FIELD(SMMU_CB4_CONTEXTIDR, ASID, 8, 0)
DEP_REG32(SMMU_CB4_PRRR_MAIR0, 0x14038)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, NOS7, 1, 31)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, NOS6, 1, 30)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, NOS5, 1, 29)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, NOS4, 1, 28)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, NOS3, 1, 27)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, NOS2, 1, 26)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, NOS1, 1, 25)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, NOS0, 1, 24)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, NS1, 1, 19)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, NS0, 1, 18)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, DS1, 1, 17)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, DS0, 1, 16)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, TR7, 2, 14)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, TR6, 2, 12)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, TR5, 2, 10)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, TR4, 2, 8)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, TR3, 2, 6)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, TR2, 2, 4)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, TR1, 2, 2)
    DEP_FIELD(SMMU_CB4_PRRR_MAIR0, TR0, 2, 0)
DEP_REG32(SMMU_CB4_NMRR_MAIR1, 0x1403c)
    DEP_FIELD(SMMU_CB4_NMRR_MAIR1, OR7, 2, 30)
    DEP_FIELD(SMMU_CB4_NMRR_MAIR1, OR6, 2, 28)
    DEP_FIELD(SMMU_CB4_NMRR_MAIR1, OR5, 2, 26)
    DEP_FIELD(SMMU_CB4_NMRR_MAIR1, OR4, 2, 24)
    DEP_FIELD(SMMU_CB4_NMRR_MAIR1, OR3, 2, 22)
    DEP_FIELD(SMMU_CB4_NMRR_MAIR1, OR2, 2, 20)
    DEP_FIELD(SMMU_CB4_NMRR_MAIR1, OR1, 2, 18)
    DEP_FIELD(SMMU_CB4_NMRR_MAIR1, OR0, 2, 16)
    DEP_FIELD(SMMU_CB4_NMRR_MAIR1, IR7, 2, 14)
    DEP_FIELD(SMMU_CB4_NMRR_MAIR1, IR6, 2, 12)
    DEP_FIELD(SMMU_CB4_NMRR_MAIR1, IR5, 2, 10)
    DEP_FIELD(SMMU_CB4_NMRR_MAIR1, IR4, 2, 8)
    DEP_FIELD(SMMU_CB4_NMRR_MAIR1, IR3, 2, 6)
    DEP_FIELD(SMMU_CB4_NMRR_MAIR1, IR2, 2, 4)
    DEP_FIELD(SMMU_CB4_NMRR_MAIR1, IR1, 2, 2)
    DEP_FIELD(SMMU_CB4_NMRR_MAIR1, IR0, 2, 0)
DEP_REG32(SMMU_CB4_FSR, 0x14058)
    DEP_FIELD(SMMU_CB4_FSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_CB4_FSR, SS, 1, 30)
    DEP_FIELD(SMMU_CB4_FSR, FORMAT, 2, 9)
    DEP_FIELD(SMMU_CB4_FSR, UUT, 1, 8)
    DEP_FIELD(SMMU_CB4_FSR, ASF, 1, 7)
    DEP_FIELD(SMMU_CB4_FSR, TLBLKF, 1, 6)
    DEP_FIELD(SMMU_CB4_FSR, TLBMCF, 1, 5)
    DEP_FIELD(SMMU_CB4_FSR, EF, 1, 4)
    DEP_FIELD(SMMU_CB4_FSR, PF, 1, 3)
    DEP_FIELD(SMMU_CB4_FSR, AFF, 1, 2)
    DEP_FIELD(SMMU_CB4_FSR, TF, 1, 1)
DEP_REG32(SMMU_CB4_FSRRESTORE, 0x1405c)
DEP_REG32(SMMU_CB4_FAR_LOW, 0x14060)
DEP_REG32(SMMU_CB4_FAR_HIGH, 0x14064)
    DEP_FIELD(SMMU_CB4_FAR_HIGH, BITS, 17, 0)
DEP_REG32(SMMU_CB4_FSYNR0, 0x14068)
    DEP_FIELD(SMMU_CB4_FSYNR0, S1CBNDX, 4, 16)
    DEP_FIELD(SMMU_CB4_FSYNR0, AFR, 1, 11)
    DEP_FIELD(SMMU_CB4_FSYNR0, PTWF, 1, 10)
    DEP_FIELD(SMMU_CB4_FSYNR0, ATOF, 1, 9)
    DEP_FIELD(SMMU_CB4_FSYNR0, NSATTR, 1, 8)
    DEP_FIELD(SMMU_CB4_FSYNR0, IND, 1, 6)
    DEP_FIELD(SMMU_CB4_FSYNR0, PNU, 1, 5)
    DEP_FIELD(SMMU_CB4_FSYNR0, WNR, 1, 4)
    DEP_FIELD(SMMU_CB4_FSYNR0, PLVL, 2, 0)
DEP_REG32(SMMU_CB4_IPAFAR_LOW, 0x14070)
    DEP_FIELD(SMMU_CB4_IPAFAR_LOW, IPAFAR_L, 20, 12)
    DEP_FIELD(SMMU_CB4_IPAFAR_LOW, FAR_RO, 12, 0)
DEP_REG32(SMMU_CB4_IPAFAR_HIGH, 0x14074)
    DEP_FIELD(SMMU_CB4_IPAFAR_HIGH, BITS, 16, 0)
DEP_REG32(SMMU_CB4_TLBIVA_LOW, 0x14600)
DEP_REG32(SMMU_CB4_TLBIVA_HIGH, 0x14604)
    DEP_FIELD(SMMU_CB4_TLBIVA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB4_TLBIVA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB4_TLBIVAA_LOW, 0x14608)
DEP_REG32(SMMU_CB4_TLBIVAA_HIGH, 0x1460c)
    DEP_FIELD(SMMU_CB4_TLBIVAA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB4_TLBIVAA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB4_TLBIASID, 0x14610)
    DEP_FIELD(SMMU_CB4_TLBIASID, ASID, 16, 0)
DEP_REG32(SMMU_CB4_TLBIALL, 0x14618)
DEP_REG32(SMMU_CB4_TLBIVAL_LOW, 0x14620)
DEP_REG32(SMMU_CB4_TLBIVAL_HIGH, 0x14624)
    DEP_FIELD(SMMU_CB4_TLBIVAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB4_TLBIVAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB4_TLBIVAAL_LOW, 0x14628)
DEP_REG32(SMMU_CB4_TLBIVAAL_HIGH, 0x1462c)
    DEP_FIELD(SMMU_CB4_TLBIVAAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB4_TLBIVAAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB4_TLBIIPAS2_LOW, 0x14630)
DEP_REG32(SMMU_CB4_TLBIIPAS2_HIGH, 0x14634)
    DEP_FIELD(SMMU_CB4_TLBIIPAS2_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB4_TLBIIPAS2L_LOW, 0x14638)
DEP_REG32(SMMU_CB4_TLBIIPAS2L_HIGH, 0x1463c)
    DEP_FIELD(SMMU_CB4_TLBIIPAS2L_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB4_TLBSYNC, 0x147f0)
DEP_REG32(SMMU_CB4_TLBSTATUS, 0x147f4)
    DEP_FIELD(SMMU_CB4_TLBSTATUS, SACTIVE, 1, 0)
DEP_REG32(SMMU_CB4_PMEVCNTR0, 0x14e00)
DEP_REG32(SMMU_CB4_PMEVCNTR1, 0x14e04)
DEP_REG32(SMMU_CB4_PMEVCNTR2, 0x14e08)
DEP_REG32(SMMU_CB4_PMEVCNTR3, 0x14e0c)
DEP_REG32(SMMU_CB4_PMEVTYPER0, 0x14e80)
    DEP_FIELD(SMMU_CB4_PMEVTYPER0, P, 1, 31)
    DEP_FIELD(SMMU_CB4_PMEVTYPER0, U, 1, 30)
    DEP_FIELD(SMMU_CB4_PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(SMMU_CB4_PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(SMMU_CB4_PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(SMMU_CB4_PMEVTYPER1, 0x14e84)
    DEP_FIELD(SMMU_CB4_PMEVTYPER1, P, 1, 31)
    DEP_FIELD(SMMU_CB4_PMEVTYPER1, U, 1, 30)
    DEP_FIELD(SMMU_CB4_PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(SMMU_CB4_PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(SMMU_CB4_PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(SMMU_CB4_PMEVTYPER2, 0x14e88)
    DEP_FIELD(SMMU_CB4_PMEVTYPER2, P, 1, 31)
    DEP_FIELD(SMMU_CB4_PMEVTYPER2, U, 1, 30)
    DEP_FIELD(SMMU_CB4_PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(SMMU_CB4_PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(SMMU_CB4_PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(SMMU_CB4_PMEVTYPER3, 0x14e8c)
    DEP_FIELD(SMMU_CB4_PMEVTYPER3, P, 1, 31)
    DEP_FIELD(SMMU_CB4_PMEVTYPER3, U, 1, 30)
    DEP_FIELD(SMMU_CB4_PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(SMMU_CB4_PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(SMMU_CB4_PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(SMMU_CB4_PMCFGR, 0x14f00)
    DEP_FIELD(SMMU_CB4_PMCFGR, NCG, 8, 24)
    DEP_FIELD(SMMU_CB4_PMCFGR, UEN, 1, 19)
    DEP_FIELD(SMMU_CB4_PMCFGR, EX, 1, 16)
    DEP_FIELD(SMMU_CB4_PMCFGR, CCD, 1, 15)
    DEP_FIELD(SMMU_CB4_PMCFGR, CC, 1, 14)
    DEP_FIELD(SMMU_CB4_PMCFGR, SIZE, 6, 8)
    DEP_FIELD(SMMU_CB4_PMCFGR, N, 8, 0)
DEP_REG32(SMMU_CB4_PMCR, 0x14f04)
    DEP_FIELD(SMMU_CB4_PMCR, IMP, 8, 24)
    DEP_FIELD(SMMU_CB4_PMCR, X, 1, 4)
    DEP_FIELD(SMMU_CB4_PMCR, P, 1, 1)
    DEP_FIELD(SMMU_CB4_PMCR, E, 1, 0)
DEP_REG32(SMMU_CB4_PMCEID, 0x14f20)
    DEP_FIELD(SMMU_CB4_PMCEID, EVENT0X12, 1, 17)
    DEP_FIELD(SMMU_CB4_PMCEID, EVENT0X11, 1, 16)
    DEP_FIELD(SMMU_CB4_PMCEID, EVENT0X10, 1, 15)
    DEP_FIELD(SMMU_CB4_PMCEID, EVENT0X0A, 1, 9)
    DEP_FIELD(SMMU_CB4_PMCEID, EVENT0X09, 1, 8)
    DEP_FIELD(SMMU_CB4_PMCEID, EVENT0X08, 1, 7)
    DEP_FIELD(SMMU_CB4_PMCEID, EVENT0X01, 1, 1)
    DEP_FIELD(SMMU_CB4_PMCEID, EVENT0X00, 1, 0)
DEP_REG32(SMMU_CB4_PMCNTENSE, 0x14f40)
    DEP_FIELD(SMMU_CB4_PMCNTENSE, P3, 1, 3)
    DEP_FIELD(SMMU_CB4_PMCNTENSE, P2, 1, 2)
    DEP_FIELD(SMMU_CB4_PMCNTENSE, P1, 1, 1)
    DEP_FIELD(SMMU_CB4_PMCNTENSE, P0, 1, 0)
DEP_REG32(SMMU_CB4_PMCNTENCLR, 0x14f44)
    DEP_FIELD(SMMU_CB4_PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB4_PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB4_PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB4_PMCNTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB4_PMCNTENSET, 0x14f48)
    DEP_FIELD(SMMU_CB4_PMCNTENSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB4_PMCNTENSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB4_PMCNTENSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB4_PMCNTENSET, P0, 1, 0)
DEP_REG32(SMMU_CB4_PMINTENCLR, 0x14f4c)
    DEP_FIELD(SMMU_CB4_PMINTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB4_PMINTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB4_PMINTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB4_PMINTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB4_PMOVSCLR, 0x14f50)
    DEP_FIELD(SMMU_CB4_PMOVSCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB4_PMOVSCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB4_PMOVSCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB4_PMOVSCLR, P0, 1, 0)
DEP_REG32(SMMU_CB4_PMOVSSET, 0x14f58)
    DEP_FIELD(SMMU_CB4_PMOVSSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB4_PMOVSSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB4_PMOVSSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB4_PMOVSSET, P0, 1, 0)
DEP_REG32(SMMU_CB4_PMAUTHSTATUS, 0x14fb8)
    DEP_FIELD(SMMU_CB4_PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(SMMU_CB4_PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(SMMU_CB4_PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(SMMU_CB4_PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(SMMU_CB4_PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(SMMU_CB4_PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(SMMU_CB4_PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(SMMU_CB4_PMAUTHSTATUS, NSE, 1, 0)
DEP_REG32(SMMU_CB5_SCTLR, 0x15000)
    DEP_FIELD(SMMU_CB5_SCTLR, NSCFG, 2, 28)
    DEP_FIELD(SMMU_CB5_SCTLR, WACFG, 2, 26)
    DEP_FIELD(SMMU_CB5_SCTLR, RACFG, 2, 24)
    DEP_FIELD(SMMU_CB5_SCTLR, SHCFG, 2, 22)
    DEP_FIELD(SMMU_CB5_SCTLR, FB, 1, 21)
    DEP_FIELD(SMMU_CB5_SCTLR, MTCFG, 1, 20)
    DEP_FIELD(SMMU_CB5_SCTLR, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_CB5_SCTLR, TRANSIENTCFG, 2, 14)
    DEP_FIELD(SMMU_CB5_SCTLR, PTW, 1, 13)
    DEP_FIELD(SMMU_CB5_SCTLR, ASIDPNE, 1, 12)
    DEP_FIELD(SMMU_CB5_SCTLR, UWXN, 1, 10)
    DEP_FIELD(SMMU_CB5_SCTLR, WXN, 1, 9)
    DEP_FIELD(SMMU_CB5_SCTLR, HUPCF, 1, 8)
    DEP_FIELD(SMMU_CB5_SCTLR, CFCFG, 1, 7)
    DEP_FIELD(SMMU_CB5_SCTLR, CFIE, 1, 6)
    DEP_FIELD(SMMU_CB5_SCTLR, CFRE, 1, 5)
    DEP_FIELD(SMMU_CB5_SCTLR, E, 1, 4)
    DEP_FIELD(SMMU_CB5_SCTLR, AFFD, 1, 3)
    DEP_FIELD(SMMU_CB5_SCTLR, AFE, 1, 2)
    DEP_FIELD(SMMU_CB5_SCTLR, TRE, 1, 1)
    DEP_FIELD(SMMU_CB5_SCTLR, M, 1, 0)
DEP_REG32(SMMU_CB5_ACTLR, 0x15004)
    DEP_FIELD(SMMU_CB5_ACTLR, CPRE, 1, 1)
    DEP_FIELD(SMMU_CB5_ACTLR, CMTLB, 1, 0)
DEP_REG32(SMMU_CB5_RESUME, 0x15008)
    DEP_FIELD(SMMU_CB5_RESUME, TNR, 1, 0)
DEP_REG32(SMMU_CB5_TCR2, 0x15010)
    DEP_FIELD(SMMU_CB5_TCR2, NSCFG1, 1, 30)
    DEP_FIELD(SMMU_CB5_TCR2, SEP, 3, 15)
    DEP_FIELD(SMMU_CB5_TCR2, NSCFG0, 1, 14)
    DEP_FIELD(SMMU_CB5_TCR2, TBI1, 1, 6)
    DEP_FIELD(SMMU_CB5_TCR2, TBI0, 1, 5)
    DEP_FIELD(SMMU_CB5_TCR2, AS, 1, 4)
    DEP_FIELD(SMMU_CB5_TCR2, PASIZE, 3, 0)
DEP_REG32(SMMU_CB5_TTBR0_LOW, 0x15020)
    DEP_FIELD(SMMU_CB5_TTBR0_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB5_TTBR0_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB5_TTBR0_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB5_TTBR0_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB5_TTBR0_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB5_TTBR0_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB5_TTBR0_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB5_TTBR0_HIGH, 0x15024)
    DEP_FIELD(SMMU_CB5_TTBR0_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB5_TTBR0_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB5_TTBR1_LOW, 0x15028)
    DEP_FIELD(SMMU_CB5_TTBR1_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB5_TTBR1_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB5_TTBR1_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB5_TTBR1_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB5_TTBR1_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB5_TTBR1_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB5_TTBR1_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB5_TTBR1_HIGH, 0x1502c)
    DEP_FIELD(SMMU_CB5_TTBR1_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB5_TTBR1_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB5_TCR_LPAE, 0x15030)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, EAE, 1, 31)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, NSCFG1_TG1, 1, 30)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, SH1, 2, 28)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, ORGN1, 2, 26)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, IRGN1, 2, 24)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, EPD1, 1, 23)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, A1, 1, 22)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, T1SZ_5_3, 3, 19)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, T1SZ_2_0_PASIZE, 3, 16)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, NSCFG0_TG0, 1, 14)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, SH0, 2, 12)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, ORGN0, 2, 10)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, IRGN0, 2, 8)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, SL0_1_EPD0, 1, 7)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, SL0_0, 1, 6)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, PD1_T0SZ_5, 1, 5)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, S_PD0_T0SZ_4, 1, 4)
    DEP_FIELD(SMMU_CB5_TCR_LPAE, T0SZ_3_0, 4, 0)
DEP_REG32(SMMU_CB5_CONTEXTIDR, 0x15034)
    DEP_FIELD(SMMU_CB5_CONTEXTIDR, PROCID, 24, 8)
    DEP_FIELD(SMMU_CB5_CONTEXTIDR, ASID, 8, 0)
DEP_REG32(SMMU_CB5_PRRR_MAIR0, 0x15038)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, NOS7, 1, 31)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, NOS6, 1, 30)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, NOS5, 1, 29)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, NOS4, 1, 28)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, NOS3, 1, 27)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, NOS2, 1, 26)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, NOS1, 1, 25)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, NOS0, 1, 24)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, NS1, 1, 19)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, NS0, 1, 18)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, DS1, 1, 17)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, DS0, 1, 16)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, TR7, 2, 14)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, TR6, 2, 12)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, TR5, 2, 10)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, TR4, 2, 8)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, TR3, 2, 6)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, TR2, 2, 4)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, TR1, 2, 2)
    DEP_FIELD(SMMU_CB5_PRRR_MAIR0, TR0, 2, 0)
DEP_REG32(SMMU_CB5_NMRR_MAIR1, 0x1503c)
    DEP_FIELD(SMMU_CB5_NMRR_MAIR1, OR7, 2, 30)
    DEP_FIELD(SMMU_CB5_NMRR_MAIR1, OR6, 2, 28)
    DEP_FIELD(SMMU_CB5_NMRR_MAIR1, OR5, 2, 26)
    DEP_FIELD(SMMU_CB5_NMRR_MAIR1, OR4, 2, 24)
    DEP_FIELD(SMMU_CB5_NMRR_MAIR1, OR3, 2, 22)
    DEP_FIELD(SMMU_CB5_NMRR_MAIR1, OR2, 2, 20)
    DEP_FIELD(SMMU_CB5_NMRR_MAIR1, OR1, 2, 18)
    DEP_FIELD(SMMU_CB5_NMRR_MAIR1, OR0, 2, 16)
    DEP_FIELD(SMMU_CB5_NMRR_MAIR1, IR7, 2, 14)
    DEP_FIELD(SMMU_CB5_NMRR_MAIR1, IR6, 2, 12)
    DEP_FIELD(SMMU_CB5_NMRR_MAIR1, IR5, 2, 10)
    DEP_FIELD(SMMU_CB5_NMRR_MAIR1, IR4, 2, 8)
    DEP_FIELD(SMMU_CB5_NMRR_MAIR1, IR3, 2, 6)
    DEP_FIELD(SMMU_CB5_NMRR_MAIR1, IR2, 2, 4)
    DEP_FIELD(SMMU_CB5_NMRR_MAIR1, IR1, 2, 2)
    DEP_FIELD(SMMU_CB5_NMRR_MAIR1, IR0, 2, 0)
DEP_REG32(SMMU_CB5_FSR, 0x15058)
    DEP_FIELD(SMMU_CB5_FSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_CB5_FSR, SS, 1, 30)
    DEP_FIELD(SMMU_CB5_FSR, FORMAT, 2, 9)
    DEP_FIELD(SMMU_CB5_FSR, UUT, 1, 8)
    DEP_FIELD(SMMU_CB5_FSR, ASF, 1, 7)
    DEP_FIELD(SMMU_CB5_FSR, TLBLKF, 1, 6)
    DEP_FIELD(SMMU_CB5_FSR, TLBMCF, 1, 5)
    DEP_FIELD(SMMU_CB5_FSR, EF, 1, 4)
    DEP_FIELD(SMMU_CB5_FSR, PF, 1, 3)
    DEP_FIELD(SMMU_CB5_FSR, AFF, 1, 2)
    DEP_FIELD(SMMU_CB5_FSR, TF, 1, 1)
DEP_REG32(SMMU_CB5_FSRRESTORE, 0x1505c)
DEP_REG32(SMMU_CB5_FAR_LOW, 0x15060)
DEP_REG32(SMMU_CB5_FAR_HIGH, 0x15064)
    DEP_FIELD(SMMU_CB5_FAR_HIGH, BITS, 17, 0)
DEP_REG32(SMMU_CB5_FSYNR0, 0x15068)
    DEP_FIELD(SMMU_CB5_FSYNR0, S1CBNDX, 4, 16)
    DEP_FIELD(SMMU_CB5_FSYNR0, AFR, 1, 11)
    DEP_FIELD(SMMU_CB5_FSYNR0, PTWF, 1, 10)
    DEP_FIELD(SMMU_CB5_FSYNR0, ATOF, 1, 9)
    DEP_FIELD(SMMU_CB5_FSYNR0, NSATTR, 1, 8)
    DEP_FIELD(SMMU_CB5_FSYNR0, IND, 1, 6)
    DEP_FIELD(SMMU_CB5_FSYNR0, PNU, 1, 5)
    DEP_FIELD(SMMU_CB5_FSYNR0, WNR, 1, 4)
    DEP_FIELD(SMMU_CB5_FSYNR0, PLVL, 2, 0)
DEP_REG32(SMMU_CB5_IPAFAR_LOW, 0x15070)
    DEP_FIELD(SMMU_CB5_IPAFAR_LOW, IPAFAR_L, 20, 12)
    DEP_FIELD(SMMU_CB5_IPAFAR_LOW, FAR_RO, 12, 0)
DEP_REG32(SMMU_CB5_IPAFAR_HIGH, 0x15074)
    DEP_FIELD(SMMU_CB5_IPAFAR_HIGH, BITS, 16, 0)
DEP_REG32(SMMU_CB5_TLBIVA_LOW, 0x15600)
DEP_REG32(SMMU_CB5_TLBIVA_HIGH, 0x15604)
    DEP_FIELD(SMMU_CB5_TLBIVA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB5_TLBIVA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB5_TLBIVAA_LOW, 0x15608)
DEP_REG32(SMMU_CB5_TLBIVAA_HIGH, 0x1560c)
    DEP_FIELD(SMMU_CB5_TLBIVAA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB5_TLBIVAA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB5_TLBIASID, 0x15610)
    DEP_FIELD(SMMU_CB5_TLBIASID, ASID, 16, 0)
DEP_REG32(SMMU_CB5_TLBIALL, 0x15618)
DEP_REG32(SMMU_CB5_TLBIVAL_LOW, 0x15620)
DEP_REG32(SMMU_CB5_TLBIVAL_HIGH, 0x15624)
    DEP_FIELD(SMMU_CB5_TLBIVAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB5_TLBIVAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB5_TLBIVAAL_LOW, 0x15628)
DEP_REG32(SMMU_CB5_TLBIVAAL_HIGH, 0x1562c)
    DEP_FIELD(SMMU_CB5_TLBIVAAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB5_TLBIVAAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB5_TLBIIPAS2_LOW, 0x15630)
DEP_REG32(SMMU_CB5_TLBIIPAS2_HIGH, 0x15634)
    DEP_FIELD(SMMU_CB5_TLBIIPAS2_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB5_TLBIIPAS2L_LOW, 0x15638)
DEP_REG32(SMMU_CB5_TLBIIPAS2L_HIGH, 0x1563c)
    DEP_FIELD(SMMU_CB5_TLBIIPAS2L_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB5_TLBSYNC, 0x157f0)
DEP_REG32(SMMU_CB5_TLBSTATUS, 0x157f4)
    DEP_FIELD(SMMU_CB5_TLBSTATUS, SACTIVE, 1, 0)
DEP_REG32(SMMU_CB5_PMEVCNTR0, 0x15e00)
DEP_REG32(SMMU_CB5_PMEVCNTR1, 0x15e04)
DEP_REG32(SMMU_CB5_PMEVCNTR2, 0x15e08)
DEP_REG32(SMMU_CB5_PMEVCNTR3, 0x15e0c)
DEP_REG32(SMMU_CB5_PMEVTYPER0, 0x15e80)
    DEP_FIELD(SMMU_CB5_PMEVTYPER0, P, 1, 31)
    DEP_FIELD(SMMU_CB5_PMEVTYPER0, U, 1, 30)
    DEP_FIELD(SMMU_CB5_PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(SMMU_CB5_PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(SMMU_CB5_PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(SMMU_CB5_PMEVTYPER1, 0x15e84)
    DEP_FIELD(SMMU_CB5_PMEVTYPER1, P, 1, 31)
    DEP_FIELD(SMMU_CB5_PMEVTYPER1, U, 1, 30)
    DEP_FIELD(SMMU_CB5_PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(SMMU_CB5_PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(SMMU_CB5_PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(SMMU_CB5_PMEVTYPER2, 0x15e88)
    DEP_FIELD(SMMU_CB5_PMEVTYPER2, P, 1, 31)
    DEP_FIELD(SMMU_CB5_PMEVTYPER2, U, 1, 30)
    DEP_FIELD(SMMU_CB5_PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(SMMU_CB5_PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(SMMU_CB5_PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(SMMU_CB5_PMEVTYPER3, 0x15e8c)
    DEP_FIELD(SMMU_CB5_PMEVTYPER3, P, 1, 31)
    DEP_FIELD(SMMU_CB5_PMEVTYPER3, U, 1, 30)
    DEP_FIELD(SMMU_CB5_PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(SMMU_CB5_PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(SMMU_CB5_PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(SMMU_CB5_PMCFGR, 0x15f00)
    DEP_FIELD(SMMU_CB5_PMCFGR, NCG, 8, 24)
    DEP_FIELD(SMMU_CB5_PMCFGR, UEN, 1, 19)
    DEP_FIELD(SMMU_CB5_PMCFGR, EX, 1, 16)
    DEP_FIELD(SMMU_CB5_PMCFGR, CCD, 1, 15)
    DEP_FIELD(SMMU_CB5_PMCFGR, CC, 1, 14)
    DEP_FIELD(SMMU_CB5_PMCFGR, SIZE, 6, 8)
    DEP_FIELD(SMMU_CB5_PMCFGR, N, 8, 0)
DEP_REG32(SMMU_CB5_PMCR, 0x15f04)
    DEP_FIELD(SMMU_CB5_PMCR, IMP, 8, 24)
    DEP_FIELD(SMMU_CB5_PMCR, X, 1, 4)
    DEP_FIELD(SMMU_CB5_PMCR, P, 1, 1)
    DEP_FIELD(SMMU_CB5_PMCR, E, 1, 0)
DEP_REG32(SMMU_CB5_PMCEID, 0x15f20)
    DEP_FIELD(SMMU_CB5_PMCEID, EVENT0X12, 1, 17)
    DEP_FIELD(SMMU_CB5_PMCEID, EVENT0X11, 1, 16)
    DEP_FIELD(SMMU_CB5_PMCEID, EVENT0X10, 1, 15)
    DEP_FIELD(SMMU_CB5_PMCEID, EVENT0X0A, 1, 9)
    DEP_FIELD(SMMU_CB5_PMCEID, EVENT0X09, 1, 8)
    DEP_FIELD(SMMU_CB5_PMCEID, EVENT0X08, 1, 7)
    DEP_FIELD(SMMU_CB5_PMCEID, EVENT0X01, 1, 1)
    DEP_FIELD(SMMU_CB5_PMCEID, EVENT0X00, 1, 0)
DEP_REG32(SMMU_CB5_PMCNTENSE, 0x15f40)
    DEP_FIELD(SMMU_CB5_PMCNTENSE, P3, 1, 3)
    DEP_FIELD(SMMU_CB5_PMCNTENSE, P2, 1, 2)
    DEP_FIELD(SMMU_CB5_PMCNTENSE, P1, 1, 1)
    DEP_FIELD(SMMU_CB5_PMCNTENSE, P0, 1, 0)
DEP_REG32(SMMU_CB5_PMCNTENCLR, 0x15f44)
    DEP_FIELD(SMMU_CB5_PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB5_PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB5_PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB5_PMCNTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB5_PMCNTENSET, 0x15f48)
    DEP_FIELD(SMMU_CB5_PMCNTENSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB5_PMCNTENSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB5_PMCNTENSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB5_PMCNTENSET, P0, 1, 0)
DEP_REG32(SMMU_CB5_PMINTENCLR, 0x15f4c)
    DEP_FIELD(SMMU_CB5_PMINTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB5_PMINTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB5_PMINTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB5_PMINTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB5_PMOVSCLR, 0x15f50)
    DEP_FIELD(SMMU_CB5_PMOVSCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB5_PMOVSCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB5_PMOVSCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB5_PMOVSCLR, P0, 1, 0)
DEP_REG32(SMMU_CB5_PMOVSSET, 0x15f58)
    DEP_FIELD(SMMU_CB5_PMOVSSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB5_PMOVSSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB5_PMOVSSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB5_PMOVSSET, P0, 1, 0)
DEP_REG32(SMMU_CB5_PMAUTHSTATUS, 0x15fb8)
    DEP_FIELD(SMMU_CB5_PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(SMMU_CB5_PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(SMMU_CB5_PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(SMMU_CB5_PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(SMMU_CB5_PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(SMMU_CB5_PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(SMMU_CB5_PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(SMMU_CB5_PMAUTHSTATUS, NSE, 1, 0)
DEP_REG32(SMMU_CB6_SCTLR, 0x16000)
    DEP_FIELD(SMMU_CB6_SCTLR, NSCFG, 2, 28)
    DEP_FIELD(SMMU_CB6_SCTLR, WACFG, 2, 26)
    DEP_FIELD(SMMU_CB6_SCTLR, RACFG, 2, 24)
    DEP_FIELD(SMMU_CB6_SCTLR, SHCFG, 2, 22)
    DEP_FIELD(SMMU_CB6_SCTLR, FB, 1, 21)
    DEP_FIELD(SMMU_CB6_SCTLR, MTCFG, 1, 20)
    DEP_FIELD(SMMU_CB6_SCTLR, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_CB6_SCTLR, TRANSIENTCFG, 2, 14)
    DEP_FIELD(SMMU_CB6_SCTLR, PTW, 1, 13)
    DEP_FIELD(SMMU_CB6_SCTLR, ASIDPNE, 1, 12)
    DEP_FIELD(SMMU_CB6_SCTLR, UWXN, 1, 10)
    DEP_FIELD(SMMU_CB6_SCTLR, WXN, 1, 9)
    DEP_FIELD(SMMU_CB6_SCTLR, HUPCF, 1, 8)
    DEP_FIELD(SMMU_CB6_SCTLR, CFCFG, 1, 7)
    DEP_FIELD(SMMU_CB6_SCTLR, CFIE, 1, 6)
    DEP_FIELD(SMMU_CB6_SCTLR, CFRE, 1, 5)
    DEP_FIELD(SMMU_CB6_SCTLR, E, 1, 4)
    DEP_FIELD(SMMU_CB6_SCTLR, AFFD, 1, 3)
    DEP_FIELD(SMMU_CB6_SCTLR, AFE, 1, 2)
    DEP_FIELD(SMMU_CB6_SCTLR, TRE, 1, 1)
    DEP_FIELD(SMMU_CB6_SCTLR, M, 1, 0)
DEP_REG32(SMMU_CB6_ACTLR, 0x16004)
    DEP_FIELD(SMMU_CB6_ACTLR, CPRE, 1, 1)
    DEP_FIELD(SMMU_CB6_ACTLR, CMTLB, 1, 0)
DEP_REG32(SMMU_CB6_RESUME, 0x16008)
    DEP_FIELD(SMMU_CB6_RESUME, TNR, 1, 0)
DEP_REG32(SMMU_CB6_TCR2, 0x16010)
    DEP_FIELD(SMMU_CB6_TCR2, NSCFG1, 1, 30)
    DEP_FIELD(SMMU_CB6_TCR2, SEP, 3, 15)
    DEP_FIELD(SMMU_CB6_TCR2, NSCFG0, 1, 14)
    DEP_FIELD(SMMU_CB6_TCR2, TBI1, 1, 6)
    DEP_FIELD(SMMU_CB6_TCR2, TBI0, 1, 5)
    DEP_FIELD(SMMU_CB6_TCR2, AS, 1, 4)
    DEP_FIELD(SMMU_CB6_TCR2, PASIZE, 3, 0)
DEP_REG32(SMMU_CB6_TTBR0_LOW, 0x16020)
    DEP_FIELD(SMMU_CB6_TTBR0_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB6_TTBR0_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB6_TTBR0_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB6_TTBR0_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB6_TTBR0_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB6_TTBR0_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB6_TTBR0_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB6_TTBR0_HIGH, 0x16024)
    DEP_FIELD(SMMU_CB6_TTBR0_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB6_TTBR0_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB6_TTBR1_LOW, 0x16028)
    DEP_FIELD(SMMU_CB6_TTBR1_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB6_TTBR1_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB6_TTBR1_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB6_TTBR1_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB6_TTBR1_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB6_TTBR1_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB6_TTBR1_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB6_TTBR1_HIGH, 0x1602c)
    DEP_FIELD(SMMU_CB6_TTBR1_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB6_TTBR1_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB6_TCR_LPAE, 0x16030)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, EAE, 1, 31)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, NSCFG1_TG1, 1, 30)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, SH1, 2, 28)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, ORGN1, 2, 26)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, IRGN1, 2, 24)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, EPD1, 1, 23)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, A1, 1, 22)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, T1SZ_5_3, 3, 19)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, T1SZ_2_0_PASIZE, 3, 16)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, NSCFG0_TG0, 1, 14)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, SH0, 2, 12)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, ORGN0, 2, 10)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, IRGN0, 2, 8)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, SL0_1_EPD0, 1, 7)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, SL0_0, 1, 6)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, PD1_T0SZ_5, 1, 5)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, S_PD0_T0SZ_4, 1, 4)
    DEP_FIELD(SMMU_CB6_TCR_LPAE, T0SZ_3_0, 4, 0)
DEP_REG32(SMMU_CB6_CONTEXTIDR, 0x16034)
    DEP_FIELD(SMMU_CB6_CONTEXTIDR, PROCID, 24, 8)
    DEP_FIELD(SMMU_CB6_CONTEXTIDR, ASID, 8, 0)
DEP_REG32(SMMU_CB6_PRRR_MAIR0, 0x16038)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, NOS7, 1, 31)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, NOS6, 1, 30)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, NOS5, 1, 29)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, NOS4, 1, 28)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, NOS3, 1, 27)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, NOS2, 1, 26)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, NOS1, 1, 25)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, NOS0, 1, 24)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, NS1, 1, 19)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, NS0, 1, 18)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, DS1, 1, 17)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, DS0, 1, 16)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, TR7, 2, 14)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, TR6, 2, 12)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, TR5, 2, 10)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, TR4, 2, 8)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, TR3, 2, 6)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, TR2, 2, 4)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, TR1, 2, 2)
    DEP_FIELD(SMMU_CB6_PRRR_MAIR0, TR0, 2, 0)
DEP_REG32(SMMU_CB6_NMRR_MAIR1, 0x1603c)
    DEP_FIELD(SMMU_CB6_NMRR_MAIR1, OR7, 2, 30)
    DEP_FIELD(SMMU_CB6_NMRR_MAIR1, OR6, 2, 28)
    DEP_FIELD(SMMU_CB6_NMRR_MAIR1, OR5, 2, 26)
    DEP_FIELD(SMMU_CB6_NMRR_MAIR1, OR4, 2, 24)
    DEP_FIELD(SMMU_CB6_NMRR_MAIR1, OR3, 2, 22)
    DEP_FIELD(SMMU_CB6_NMRR_MAIR1, OR2, 2, 20)
    DEP_FIELD(SMMU_CB6_NMRR_MAIR1, OR1, 2, 18)
    DEP_FIELD(SMMU_CB6_NMRR_MAIR1, OR0, 2, 16)
    DEP_FIELD(SMMU_CB6_NMRR_MAIR1, IR7, 2, 14)
    DEP_FIELD(SMMU_CB6_NMRR_MAIR1, IR6, 2, 12)
    DEP_FIELD(SMMU_CB6_NMRR_MAIR1, IR5, 2, 10)
    DEP_FIELD(SMMU_CB6_NMRR_MAIR1, IR4, 2, 8)
    DEP_FIELD(SMMU_CB6_NMRR_MAIR1, IR3, 2, 6)
    DEP_FIELD(SMMU_CB6_NMRR_MAIR1, IR2, 2, 4)
    DEP_FIELD(SMMU_CB6_NMRR_MAIR1, IR1, 2, 2)
    DEP_FIELD(SMMU_CB6_NMRR_MAIR1, IR0, 2, 0)
DEP_REG32(SMMU_CB6_FSR, 0x16058)
    DEP_FIELD(SMMU_CB6_FSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_CB6_FSR, SS, 1, 30)
    DEP_FIELD(SMMU_CB6_FSR, FORMAT, 2, 9)
    DEP_FIELD(SMMU_CB6_FSR, UUT, 1, 8)
    DEP_FIELD(SMMU_CB6_FSR, ASF, 1, 7)
    DEP_FIELD(SMMU_CB6_FSR, TLBLKF, 1, 6)
    DEP_FIELD(SMMU_CB6_FSR, TLBMCF, 1, 5)
    DEP_FIELD(SMMU_CB6_FSR, EF, 1, 4)
    DEP_FIELD(SMMU_CB6_FSR, PF, 1, 3)
    DEP_FIELD(SMMU_CB6_FSR, AFF, 1, 2)
    DEP_FIELD(SMMU_CB6_FSR, TF, 1, 1)
DEP_REG32(SMMU_CB6_FSRRESTORE, 0x1605c)
DEP_REG32(SMMU_CB6_FAR_LOW, 0x16060)
DEP_REG32(SMMU_CB6_FAR_HIGH, 0x16064)
    DEP_FIELD(SMMU_CB6_FAR_HIGH, BITS, 17, 0)
DEP_REG32(SMMU_CB6_FSYNR0, 0x16068)
    DEP_FIELD(SMMU_CB6_FSYNR0, S1CBNDX, 4, 16)
    DEP_FIELD(SMMU_CB6_FSYNR0, AFR, 1, 11)
    DEP_FIELD(SMMU_CB6_FSYNR0, PTWF, 1, 10)
    DEP_FIELD(SMMU_CB6_FSYNR0, ATOF, 1, 9)
    DEP_FIELD(SMMU_CB6_FSYNR0, NSATTR, 1, 8)
    DEP_FIELD(SMMU_CB6_FSYNR0, IND, 1, 6)
    DEP_FIELD(SMMU_CB6_FSYNR0, PNU, 1, 5)
    DEP_FIELD(SMMU_CB6_FSYNR0, WNR, 1, 4)
    DEP_FIELD(SMMU_CB6_FSYNR0, PLVL, 2, 0)
DEP_REG32(SMMU_CB6_IPAFAR_LOW, 0x16070)
    DEP_FIELD(SMMU_CB6_IPAFAR_LOW, IPAFAR_L, 20, 12)
    DEP_FIELD(SMMU_CB6_IPAFAR_LOW, FAR_RO, 12, 0)
DEP_REG32(SMMU_CB6_IPAFAR_HIGH, 0x16074)
    DEP_FIELD(SMMU_CB6_IPAFAR_HIGH, BITS, 16, 0)
DEP_REG32(SMMU_CB6_TLBIVA_LOW, 0x16600)
DEP_REG32(SMMU_CB6_TLBIVA_HIGH, 0x16604)
    DEP_FIELD(SMMU_CB6_TLBIVA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB6_TLBIVA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB6_TLBIVAA_LOW, 0x16608)
DEP_REG32(SMMU_CB6_TLBIVAA_HIGH, 0x1660c)
    DEP_FIELD(SMMU_CB6_TLBIVAA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB6_TLBIVAA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB6_TLBIASID, 0x16610)
    DEP_FIELD(SMMU_CB6_TLBIASID, ASID, 16, 0)
DEP_REG32(SMMU_CB6_TLBIALL, 0x16618)
DEP_REG32(SMMU_CB6_TLBIVAL_LOW, 0x16620)
DEP_REG32(SMMU_CB6_TLBIVAL_HIGH, 0x16624)
    DEP_FIELD(SMMU_CB6_TLBIVAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB6_TLBIVAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB6_TLBIVAAL_LOW, 0x16628)
DEP_REG32(SMMU_CB6_TLBIVAAL_HIGH, 0x1662c)
    DEP_FIELD(SMMU_CB6_TLBIVAAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB6_TLBIVAAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB6_TLBIIPAS2_LOW, 0x16630)
DEP_REG32(SMMU_CB6_TLBIIPAS2_HIGH, 0x16634)
    DEP_FIELD(SMMU_CB6_TLBIIPAS2_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB6_TLBIIPAS2L_LOW, 0x16638)
DEP_REG32(SMMU_CB6_TLBIIPAS2L_HIGH, 0x1663c)
    DEP_FIELD(SMMU_CB6_TLBIIPAS2L_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB6_TLBSYNC, 0x167f0)
DEP_REG32(SMMU_CB6_TLBSTATUS, 0x167f4)
    DEP_FIELD(SMMU_CB6_TLBSTATUS, SACTIVE, 1, 0)
DEP_REG32(SMMU_CB6_PMEVCNTR0, 0x16e00)
DEP_REG32(SMMU_CB6_PMEVCNTR1, 0x16e04)
DEP_REG32(SMMU_CB6_PMEVCNTR2, 0x16e08)
DEP_REG32(SMMU_CB6_PMEVCNTR3, 0x16e0c)
DEP_REG32(SMMU_CB6_PMEVTYPER0, 0x16e80)
    DEP_FIELD(SMMU_CB6_PMEVTYPER0, P, 1, 31)
    DEP_FIELD(SMMU_CB6_PMEVTYPER0, U, 1, 30)
    DEP_FIELD(SMMU_CB6_PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(SMMU_CB6_PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(SMMU_CB6_PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(SMMU_CB6_PMEVTYPER1, 0x16e84)
    DEP_FIELD(SMMU_CB6_PMEVTYPER1, P, 1, 31)
    DEP_FIELD(SMMU_CB6_PMEVTYPER1, U, 1, 30)
    DEP_FIELD(SMMU_CB6_PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(SMMU_CB6_PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(SMMU_CB6_PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(SMMU_CB6_PMEVTYPER2, 0x16e88)
    DEP_FIELD(SMMU_CB6_PMEVTYPER2, P, 1, 31)
    DEP_FIELD(SMMU_CB6_PMEVTYPER2, U, 1, 30)
    DEP_FIELD(SMMU_CB6_PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(SMMU_CB6_PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(SMMU_CB6_PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(SMMU_CB6_PMEVTYPER3, 0x16e8c)
    DEP_FIELD(SMMU_CB6_PMEVTYPER3, P, 1, 31)
    DEP_FIELD(SMMU_CB6_PMEVTYPER3, U, 1, 30)
    DEP_FIELD(SMMU_CB6_PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(SMMU_CB6_PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(SMMU_CB6_PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(SMMU_CB6_PMCFGR, 0x16f00)
    DEP_FIELD(SMMU_CB6_PMCFGR, NCG, 8, 24)
    DEP_FIELD(SMMU_CB6_PMCFGR, UEN, 1, 19)
    DEP_FIELD(SMMU_CB6_PMCFGR, EX, 1, 16)
    DEP_FIELD(SMMU_CB6_PMCFGR, CCD, 1, 15)
    DEP_FIELD(SMMU_CB6_PMCFGR, CC, 1, 14)
    DEP_FIELD(SMMU_CB6_PMCFGR, SIZE, 6, 8)
    DEP_FIELD(SMMU_CB6_PMCFGR, N, 8, 0)
DEP_REG32(SMMU_CB6_PMCR, 0x16f04)
    DEP_FIELD(SMMU_CB6_PMCR, IMP, 8, 24)
    DEP_FIELD(SMMU_CB6_PMCR, X, 1, 4)
    DEP_FIELD(SMMU_CB6_PMCR, P, 1, 1)
    DEP_FIELD(SMMU_CB6_PMCR, E, 1, 0)
DEP_REG32(SMMU_CB6_PMCEID, 0x16f20)
    DEP_FIELD(SMMU_CB6_PMCEID, EVENT0X12, 1, 17)
    DEP_FIELD(SMMU_CB6_PMCEID, EVENT0X11, 1, 16)
    DEP_FIELD(SMMU_CB6_PMCEID, EVENT0X10, 1, 15)
    DEP_FIELD(SMMU_CB6_PMCEID, EVENT0X0A, 1, 9)
    DEP_FIELD(SMMU_CB6_PMCEID, EVENT0X09, 1, 8)
    DEP_FIELD(SMMU_CB6_PMCEID, EVENT0X08, 1, 7)
    DEP_FIELD(SMMU_CB6_PMCEID, EVENT0X01, 1, 1)
    DEP_FIELD(SMMU_CB6_PMCEID, EVENT0X00, 1, 0)
DEP_REG32(SMMU_CB6_PMCNTENSE, 0x16f40)
    DEP_FIELD(SMMU_CB6_PMCNTENSE, P3, 1, 3)
    DEP_FIELD(SMMU_CB6_PMCNTENSE, P2, 1, 2)
    DEP_FIELD(SMMU_CB6_PMCNTENSE, P1, 1, 1)
    DEP_FIELD(SMMU_CB6_PMCNTENSE, P0, 1, 0)
DEP_REG32(SMMU_CB6_PMCNTENCLR, 0x16f44)
    DEP_FIELD(SMMU_CB6_PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB6_PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB6_PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB6_PMCNTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB6_PMCNTENSET, 0x16f48)
    DEP_FIELD(SMMU_CB6_PMCNTENSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB6_PMCNTENSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB6_PMCNTENSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB6_PMCNTENSET, P0, 1, 0)
DEP_REG32(SMMU_CB6_PMINTENCLR, 0x16f4c)
    DEP_FIELD(SMMU_CB6_PMINTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB6_PMINTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB6_PMINTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB6_PMINTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB6_PMOVSCLR, 0x16f50)
    DEP_FIELD(SMMU_CB6_PMOVSCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB6_PMOVSCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB6_PMOVSCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB6_PMOVSCLR, P0, 1, 0)
DEP_REG32(SMMU_CB6_PMOVSSET, 0x16f58)
    DEP_FIELD(SMMU_CB6_PMOVSSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB6_PMOVSSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB6_PMOVSSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB6_PMOVSSET, P0, 1, 0)
DEP_REG32(SMMU_CB6_PMAUTHSTATUS, 0x16fb8)
    DEP_FIELD(SMMU_CB6_PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(SMMU_CB6_PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(SMMU_CB6_PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(SMMU_CB6_PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(SMMU_CB6_PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(SMMU_CB6_PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(SMMU_CB6_PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(SMMU_CB6_PMAUTHSTATUS, NSE, 1, 0)
DEP_REG32(SMMU_CB7_SCTLR, 0x17000)
    DEP_FIELD(SMMU_CB7_SCTLR, NSCFG, 2, 28)
    DEP_FIELD(SMMU_CB7_SCTLR, WACFG, 2, 26)
    DEP_FIELD(SMMU_CB7_SCTLR, RACFG, 2, 24)
    DEP_FIELD(SMMU_CB7_SCTLR, SHCFG, 2, 22)
    DEP_FIELD(SMMU_CB7_SCTLR, FB, 1, 21)
    DEP_FIELD(SMMU_CB7_SCTLR, MTCFG, 1, 20)
    DEP_FIELD(SMMU_CB7_SCTLR, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_CB7_SCTLR, TRANSIENTCFG, 2, 14)
    DEP_FIELD(SMMU_CB7_SCTLR, PTW, 1, 13)
    DEP_FIELD(SMMU_CB7_SCTLR, ASIDPNE, 1, 12)
    DEP_FIELD(SMMU_CB7_SCTLR, UWXN, 1, 10)
    DEP_FIELD(SMMU_CB7_SCTLR, WXN, 1, 9)
    DEP_FIELD(SMMU_CB7_SCTLR, HUPCF, 1, 8)
    DEP_FIELD(SMMU_CB7_SCTLR, CFCFG, 1, 7)
    DEP_FIELD(SMMU_CB7_SCTLR, CFIE, 1, 6)
    DEP_FIELD(SMMU_CB7_SCTLR, CFRE, 1, 5)
    DEP_FIELD(SMMU_CB7_SCTLR, E, 1, 4)
    DEP_FIELD(SMMU_CB7_SCTLR, AFFD, 1, 3)
    DEP_FIELD(SMMU_CB7_SCTLR, AFE, 1, 2)
    DEP_FIELD(SMMU_CB7_SCTLR, TRE, 1, 1)
    DEP_FIELD(SMMU_CB7_SCTLR, M, 1, 0)
DEP_REG32(SMMU_CB7_ACTLR, 0x17004)
    DEP_FIELD(SMMU_CB7_ACTLR, CPRE, 1, 1)
    DEP_FIELD(SMMU_CB7_ACTLR, CMTLB, 1, 0)
DEP_REG32(SMMU_CB7_RESUME, 0x17008)
    DEP_FIELD(SMMU_CB7_RESUME, TNR, 1, 0)
DEP_REG32(SMMU_CB7_TCR2, 0x17010)
    DEP_FIELD(SMMU_CB7_TCR2, NSCFG1, 1, 30)
    DEP_FIELD(SMMU_CB7_TCR2, SEP, 3, 15)
    DEP_FIELD(SMMU_CB7_TCR2, NSCFG0, 1, 14)
    DEP_FIELD(SMMU_CB7_TCR2, TBI1, 1, 6)
    DEP_FIELD(SMMU_CB7_TCR2, TBI0, 1, 5)
    DEP_FIELD(SMMU_CB7_TCR2, AS, 1, 4)
    DEP_FIELD(SMMU_CB7_TCR2, PASIZE, 3, 0)
DEP_REG32(SMMU_CB7_TTBR0_LOW, 0x17020)
    DEP_FIELD(SMMU_CB7_TTBR0_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB7_TTBR0_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB7_TTBR0_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB7_TTBR0_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB7_TTBR0_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB7_TTBR0_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB7_TTBR0_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB7_TTBR0_HIGH, 0x17024)
    DEP_FIELD(SMMU_CB7_TTBR0_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB7_TTBR0_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB7_TTBR1_LOW, 0x17028)
    DEP_FIELD(SMMU_CB7_TTBR1_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB7_TTBR1_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB7_TTBR1_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB7_TTBR1_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB7_TTBR1_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB7_TTBR1_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB7_TTBR1_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB7_TTBR1_HIGH, 0x1702c)
    DEP_FIELD(SMMU_CB7_TTBR1_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB7_TTBR1_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB7_TCR_LPAE, 0x17030)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, EAE, 1, 31)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, NSCFG1_TG1, 1, 30)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, SH1, 2, 28)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, ORGN1, 2, 26)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, IRGN1, 2, 24)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, EPD1, 1, 23)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, A1, 1, 22)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, T1SZ_5_3, 3, 19)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, T1SZ_2_0_PASIZE, 3, 16)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, NSCFG0_TG0, 1, 14)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, SH0, 2, 12)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, ORGN0, 2, 10)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, IRGN0, 2, 8)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, SL0_1_EPD0, 1, 7)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, SL0_0, 1, 6)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, PD1_T0SZ_5, 1, 5)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, S_PD0_T0SZ_4, 1, 4)
    DEP_FIELD(SMMU_CB7_TCR_LPAE, T0SZ_3_0, 4, 0)
DEP_REG32(SMMU_CB7_CONTEXTIDR, 0x17034)
    DEP_FIELD(SMMU_CB7_CONTEXTIDR, PROCID, 24, 8)
    DEP_FIELD(SMMU_CB7_CONTEXTIDR, ASID, 8, 0)
DEP_REG32(SMMU_CB7_PRRR_MAIR0, 0x17038)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, NOS7, 1, 31)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, NOS6, 1, 30)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, NOS5, 1, 29)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, NOS4, 1, 28)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, NOS3, 1, 27)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, NOS2, 1, 26)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, NOS1, 1, 25)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, NOS0, 1, 24)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, NS1, 1, 19)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, NS0, 1, 18)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, DS1, 1, 17)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, DS0, 1, 16)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, TR7, 2, 14)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, TR6, 2, 12)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, TR5, 2, 10)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, TR4, 2, 8)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, TR3, 2, 6)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, TR2, 2, 4)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, TR1, 2, 2)
    DEP_FIELD(SMMU_CB7_PRRR_MAIR0, TR0, 2, 0)
DEP_REG32(SMMU_CB7_NMRR_MAIR1, 0x1703c)
    DEP_FIELD(SMMU_CB7_NMRR_MAIR1, OR7, 2, 30)
    DEP_FIELD(SMMU_CB7_NMRR_MAIR1, OR6, 2, 28)
    DEP_FIELD(SMMU_CB7_NMRR_MAIR1, OR5, 2, 26)
    DEP_FIELD(SMMU_CB7_NMRR_MAIR1, OR4, 2, 24)
    DEP_FIELD(SMMU_CB7_NMRR_MAIR1, OR3, 2, 22)
    DEP_FIELD(SMMU_CB7_NMRR_MAIR1, OR2, 2, 20)
    DEP_FIELD(SMMU_CB7_NMRR_MAIR1, OR1, 2, 18)
    DEP_FIELD(SMMU_CB7_NMRR_MAIR1, OR0, 2, 16)
    DEP_FIELD(SMMU_CB7_NMRR_MAIR1, IR7, 2, 14)
    DEP_FIELD(SMMU_CB7_NMRR_MAIR1, IR6, 2, 12)
    DEP_FIELD(SMMU_CB7_NMRR_MAIR1, IR5, 2, 10)
    DEP_FIELD(SMMU_CB7_NMRR_MAIR1, IR4, 2, 8)
    DEP_FIELD(SMMU_CB7_NMRR_MAIR1, IR3, 2, 6)
    DEP_FIELD(SMMU_CB7_NMRR_MAIR1, IR2, 2, 4)
    DEP_FIELD(SMMU_CB7_NMRR_MAIR1, IR1, 2, 2)
    DEP_FIELD(SMMU_CB7_NMRR_MAIR1, IR0, 2, 0)
DEP_REG32(SMMU_CB7_FSR, 0x17058)
    DEP_FIELD(SMMU_CB7_FSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_CB7_FSR, SS, 1, 30)
    DEP_FIELD(SMMU_CB7_FSR, FORMAT, 2, 9)
    DEP_FIELD(SMMU_CB7_FSR, UUT, 1, 8)
    DEP_FIELD(SMMU_CB7_FSR, ASF, 1, 7)
    DEP_FIELD(SMMU_CB7_FSR, TLBLKF, 1, 6)
    DEP_FIELD(SMMU_CB7_FSR, TLBMCF, 1, 5)
    DEP_FIELD(SMMU_CB7_FSR, EF, 1, 4)
    DEP_FIELD(SMMU_CB7_FSR, PF, 1, 3)
    DEP_FIELD(SMMU_CB7_FSR, AFF, 1, 2)
    DEP_FIELD(SMMU_CB7_FSR, TF, 1, 1)
DEP_REG32(SMMU_CB7_FSRRESTORE, 0x1705c)
DEP_REG32(SMMU_CB7_FAR_LOW, 0x17060)
DEP_REG32(SMMU_CB7_FAR_HIGH, 0x17064)
    DEP_FIELD(SMMU_CB7_FAR_HIGH, BITS, 17, 0)
DEP_REG32(SMMU_CB7_FSYNR0, 0x17068)
    DEP_FIELD(SMMU_CB7_FSYNR0, S1CBNDX, 4, 16)
    DEP_FIELD(SMMU_CB7_FSYNR0, AFR, 1, 11)
    DEP_FIELD(SMMU_CB7_FSYNR0, PTWF, 1, 10)
    DEP_FIELD(SMMU_CB7_FSYNR0, ATOF, 1, 9)
    DEP_FIELD(SMMU_CB7_FSYNR0, NSATTR, 1, 8)
    DEP_FIELD(SMMU_CB7_FSYNR0, IND, 1, 6)
    DEP_FIELD(SMMU_CB7_FSYNR0, PNU, 1, 5)
    DEP_FIELD(SMMU_CB7_FSYNR0, WNR, 1, 4)
    DEP_FIELD(SMMU_CB7_FSYNR0, PLVL, 2, 0)
DEP_REG32(SMMU_CB7_IPAFAR_LOW, 0x17070)
    DEP_FIELD(SMMU_CB7_IPAFAR_LOW, IPAFAR_L, 20, 12)
    DEP_FIELD(SMMU_CB7_IPAFAR_LOW, FAR_RO, 12, 0)
DEP_REG32(SMMU_CB7_IPAFAR_HIGH, 0x17074)
    DEP_FIELD(SMMU_CB7_IPAFAR_HIGH, BITS, 16, 0)
DEP_REG32(SMMU_CB7_TLBIVA_LOW, 0x17600)
DEP_REG32(SMMU_CB7_TLBIVA_HIGH, 0x17604)
    DEP_FIELD(SMMU_CB7_TLBIVA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB7_TLBIVA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB7_TLBIVAA_LOW, 0x17608)
DEP_REG32(SMMU_CB7_TLBIVAA_HIGH, 0x1760c)
    DEP_FIELD(SMMU_CB7_TLBIVAA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB7_TLBIVAA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB7_TLBIASID, 0x17610)
    DEP_FIELD(SMMU_CB7_TLBIASID, ASID, 16, 0)
DEP_REG32(SMMU_CB7_TLBIALL, 0x17618)
DEP_REG32(SMMU_CB7_TLBIVAL_LOW, 0x17620)
DEP_REG32(SMMU_CB7_TLBIVAL_HIGH, 0x17624)
    DEP_FIELD(SMMU_CB7_TLBIVAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB7_TLBIVAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB7_TLBIVAAL_LOW, 0x17628)
DEP_REG32(SMMU_CB7_TLBIVAAL_HIGH, 0x1762c)
    DEP_FIELD(SMMU_CB7_TLBIVAAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB7_TLBIVAAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB7_TLBIIPAS2_LOW, 0x17630)
DEP_REG32(SMMU_CB7_TLBIIPAS2_HIGH, 0x17634)
    DEP_FIELD(SMMU_CB7_TLBIIPAS2_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB7_TLBIIPAS2L_LOW, 0x17638)
DEP_REG32(SMMU_CB7_TLBIIPAS2L_HIGH, 0x1763c)
    DEP_FIELD(SMMU_CB7_TLBIIPAS2L_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB7_TLBSYNC, 0x177f0)
DEP_REG32(SMMU_CB7_TLBSTATUS, 0x177f4)
    DEP_FIELD(SMMU_CB7_TLBSTATUS, SACTIVE, 1, 0)
DEP_REG32(SMMU_CB7_PMEVCNTR0, 0x17e00)
DEP_REG32(SMMU_CB7_PMEVCNTR1, 0x17e04)
DEP_REG32(SMMU_CB7_PMEVCNTR2, 0x17e08)
DEP_REG32(SMMU_CB7_PMEVCNTR3, 0x17e0c)
DEP_REG32(SMMU_CB7_PMEVTYPER0, 0x17e80)
    DEP_FIELD(SMMU_CB7_PMEVTYPER0, P, 1, 31)
    DEP_FIELD(SMMU_CB7_PMEVTYPER0, U, 1, 30)
    DEP_FIELD(SMMU_CB7_PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(SMMU_CB7_PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(SMMU_CB7_PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(SMMU_CB7_PMEVTYPER1, 0x17e84)
    DEP_FIELD(SMMU_CB7_PMEVTYPER1, P, 1, 31)
    DEP_FIELD(SMMU_CB7_PMEVTYPER1, U, 1, 30)
    DEP_FIELD(SMMU_CB7_PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(SMMU_CB7_PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(SMMU_CB7_PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(SMMU_CB7_PMEVTYPER2, 0x17e88)
    DEP_FIELD(SMMU_CB7_PMEVTYPER2, P, 1, 31)
    DEP_FIELD(SMMU_CB7_PMEVTYPER2, U, 1, 30)
    DEP_FIELD(SMMU_CB7_PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(SMMU_CB7_PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(SMMU_CB7_PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(SMMU_CB7_PMEVTYPER3, 0x17e8c)
    DEP_FIELD(SMMU_CB7_PMEVTYPER3, P, 1, 31)
    DEP_FIELD(SMMU_CB7_PMEVTYPER3, U, 1, 30)
    DEP_FIELD(SMMU_CB7_PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(SMMU_CB7_PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(SMMU_CB7_PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(SMMU_CB7_PMCFGR, 0x17f00)
    DEP_FIELD(SMMU_CB7_PMCFGR, NCG, 8, 24)
    DEP_FIELD(SMMU_CB7_PMCFGR, UEN, 1, 19)
    DEP_FIELD(SMMU_CB7_PMCFGR, EX, 1, 16)
    DEP_FIELD(SMMU_CB7_PMCFGR, CCD, 1, 15)
    DEP_FIELD(SMMU_CB7_PMCFGR, CC, 1, 14)
    DEP_FIELD(SMMU_CB7_PMCFGR, SIZE, 6, 8)
    DEP_FIELD(SMMU_CB7_PMCFGR, N, 8, 0)
DEP_REG32(SMMU_CB7_PMCR, 0x17f04)
    DEP_FIELD(SMMU_CB7_PMCR, IMP, 8, 24)
    DEP_FIELD(SMMU_CB7_PMCR, X, 1, 4)
    DEP_FIELD(SMMU_CB7_PMCR, P, 1, 1)
    DEP_FIELD(SMMU_CB7_PMCR, E, 1, 0)
DEP_REG32(SMMU_CB7_PMCEID, 0x17f20)
    DEP_FIELD(SMMU_CB7_PMCEID, EVENT0X12, 1, 17)
    DEP_FIELD(SMMU_CB7_PMCEID, EVENT0X11, 1, 16)
    DEP_FIELD(SMMU_CB7_PMCEID, EVENT0X10, 1, 15)
    DEP_FIELD(SMMU_CB7_PMCEID, EVENT0X0A, 1, 9)
    DEP_FIELD(SMMU_CB7_PMCEID, EVENT0X09, 1, 8)
    DEP_FIELD(SMMU_CB7_PMCEID, EVENT0X08, 1, 7)
    DEP_FIELD(SMMU_CB7_PMCEID, EVENT0X01, 1, 1)
    DEP_FIELD(SMMU_CB7_PMCEID, EVENT0X00, 1, 0)
DEP_REG32(SMMU_CB7_PMCNTENSE, 0x17f40)
    DEP_FIELD(SMMU_CB7_PMCNTENSE, P3, 1, 3)
    DEP_FIELD(SMMU_CB7_PMCNTENSE, P2, 1, 2)
    DEP_FIELD(SMMU_CB7_PMCNTENSE, P1, 1, 1)
    DEP_FIELD(SMMU_CB7_PMCNTENSE, P0, 1, 0)
DEP_REG32(SMMU_CB7_PMCNTENCLR, 0x17f44)
    DEP_FIELD(SMMU_CB7_PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB7_PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB7_PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB7_PMCNTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB7_PMCNTENSET, 0x17f48)
    DEP_FIELD(SMMU_CB7_PMCNTENSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB7_PMCNTENSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB7_PMCNTENSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB7_PMCNTENSET, P0, 1, 0)
DEP_REG32(SMMU_CB7_PMINTENCLR, 0x17f4c)
    DEP_FIELD(SMMU_CB7_PMINTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB7_PMINTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB7_PMINTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB7_PMINTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB7_PMOVSCLR, 0x17f50)
    DEP_FIELD(SMMU_CB7_PMOVSCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB7_PMOVSCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB7_PMOVSCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB7_PMOVSCLR, P0, 1, 0)
DEP_REG32(SMMU_CB7_PMOVSSET, 0x17f58)
    DEP_FIELD(SMMU_CB7_PMOVSSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB7_PMOVSSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB7_PMOVSSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB7_PMOVSSET, P0, 1, 0)
DEP_REG32(SMMU_CB7_PMAUTHSTATUS, 0x17fb8)
    DEP_FIELD(SMMU_CB7_PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(SMMU_CB7_PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(SMMU_CB7_PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(SMMU_CB7_PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(SMMU_CB7_PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(SMMU_CB7_PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(SMMU_CB7_PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(SMMU_CB7_PMAUTHSTATUS, NSE, 1, 0)
DEP_REG32(SMMU_CB8_SCTLR, 0x18000)
    DEP_FIELD(SMMU_CB8_SCTLR, NSCFG, 2, 28)
    DEP_FIELD(SMMU_CB8_SCTLR, WACFG, 2, 26)
    DEP_FIELD(SMMU_CB8_SCTLR, RACFG, 2, 24)
    DEP_FIELD(SMMU_CB8_SCTLR, SHCFG, 2, 22)
    DEP_FIELD(SMMU_CB8_SCTLR, FB, 1, 21)
    DEP_FIELD(SMMU_CB8_SCTLR, MTCFG, 1, 20)
    DEP_FIELD(SMMU_CB8_SCTLR, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_CB8_SCTLR, TRANSIENTCFG, 2, 14)
    DEP_FIELD(SMMU_CB8_SCTLR, PTW, 1, 13)
    DEP_FIELD(SMMU_CB8_SCTLR, ASIDPNE, 1, 12)
    DEP_FIELD(SMMU_CB8_SCTLR, UWXN, 1, 10)
    DEP_FIELD(SMMU_CB8_SCTLR, WXN, 1, 9)
    DEP_FIELD(SMMU_CB8_SCTLR, HUPCF, 1, 8)
    DEP_FIELD(SMMU_CB8_SCTLR, CFCFG, 1, 7)
    DEP_FIELD(SMMU_CB8_SCTLR, CFIE, 1, 6)
    DEP_FIELD(SMMU_CB8_SCTLR, CFRE, 1, 5)
    DEP_FIELD(SMMU_CB8_SCTLR, E, 1, 4)
    DEP_FIELD(SMMU_CB8_SCTLR, AFFD, 1, 3)
    DEP_FIELD(SMMU_CB8_SCTLR, AFE, 1, 2)
    DEP_FIELD(SMMU_CB8_SCTLR, TRE, 1, 1)
    DEP_FIELD(SMMU_CB8_SCTLR, M, 1, 0)
DEP_REG32(SMMU_CB8_ACTLR, 0x18004)
    DEP_FIELD(SMMU_CB8_ACTLR, CPRE, 1, 1)
    DEP_FIELD(SMMU_CB8_ACTLR, CMTLB, 1, 0)
DEP_REG32(SMMU_CB8_RESUME, 0x18008)
    DEP_FIELD(SMMU_CB8_RESUME, TNR, 1, 0)
DEP_REG32(SMMU_CB8_TCR2, 0x18010)
    DEP_FIELD(SMMU_CB8_TCR2, NSCFG1, 1, 30)
    DEP_FIELD(SMMU_CB8_TCR2, SEP, 3, 15)
    DEP_FIELD(SMMU_CB8_TCR2, NSCFG0, 1, 14)
    DEP_FIELD(SMMU_CB8_TCR2, TBI1, 1, 6)
    DEP_FIELD(SMMU_CB8_TCR2, TBI0, 1, 5)
    DEP_FIELD(SMMU_CB8_TCR2, AS, 1, 4)
    DEP_FIELD(SMMU_CB8_TCR2, PASIZE, 3, 0)
DEP_REG32(SMMU_CB8_TTBR0_LOW, 0x18020)
    DEP_FIELD(SMMU_CB8_TTBR0_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB8_TTBR0_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB8_TTBR0_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB8_TTBR0_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB8_TTBR0_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB8_TTBR0_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB8_TTBR0_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB8_TTBR0_HIGH, 0x18024)
    DEP_FIELD(SMMU_CB8_TTBR0_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB8_TTBR0_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB8_TTBR1_LOW, 0x18028)
    DEP_FIELD(SMMU_CB8_TTBR1_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB8_TTBR1_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB8_TTBR1_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB8_TTBR1_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB8_TTBR1_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB8_TTBR1_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB8_TTBR1_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB8_TTBR1_HIGH, 0x1802c)
    DEP_FIELD(SMMU_CB8_TTBR1_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB8_TTBR1_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB8_TCR_LPAE, 0x18030)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, EAE, 1, 31)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, NSCFG1_TG1, 1, 30)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, SH1, 2, 28)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, ORGN1, 2, 26)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, IRGN1, 2, 24)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, EPD1, 1, 23)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, A1, 1, 22)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, T1SZ_5_3, 3, 19)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, T1SZ_2_0_PASIZE, 3, 16)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, NSCFG0_TG0, 1, 14)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, SH0, 2, 12)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, ORGN0, 2, 10)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, IRGN0, 2, 8)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, SL0_1_EPD0, 1, 7)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, SL0_0, 1, 6)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, PD1_T0SZ_5, 1, 5)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, S_PD0_T0SZ_4, 1, 4)
    DEP_FIELD(SMMU_CB8_TCR_LPAE, T0SZ_3_0, 4, 0)
DEP_REG32(SMMU_CB8_CONTEXTIDR, 0x18034)
    DEP_FIELD(SMMU_CB8_CONTEXTIDR, PROCID, 24, 8)
    DEP_FIELD(SMMU_CB8_CONTEXTIDR, ASID, 8, 0)
DEP_REG32(SMMU_CB8_PRRR_MAIR0, 0x18038)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, NOS7, 1, 31)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, NOS6, 1, 30)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, NOS5, 1, 29)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, NOS4, 1, 28)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, NOS3, 1, 27)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, NOS2, 1, 26)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, NOS1, 1, 25)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, NOS0, 1, 24)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, NS1, 1, 19)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, NS0, 1, 18)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, DS1, 1, 17)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, DS0, 1, 16)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, TR7, 2, 14)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, TR6, 2, 12)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, TR5, 2, 10)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, TR4, 2, 8)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, TR3, 2, 6)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, TR2, 2, 4)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, TR1, 2, 2)
    DEP_FIELD(SMMU_CB8_PRRR_MAIR0, TR0, 2, 0)
DEP_REG32(SMMU_CB8_NMRR_MAIR1, 0x1803c)
    DEP_FIELD(SMMU_CB8_NMRR_MAIR1, OR7, 2, 30)
    DEP_FIELD(SMMU_CB8_NMRR_MAIR1, OR6, 2, 28)
    DEP_FIELD(SMMU_CB8_NMRR_MAIR1, OR5, 2, 26)
    DEP_FIELD(SMMU_CB8_NMRR_MAIR1, OR4, 2, 24)
    DEP_FIELD(SMMU_CB8_NMRR_MAIR1, OR3, 2, 22)
    DEP_FIELD(SMMU_CB8_NMRR_MAIR1, OR2, 2, 20)
    DEP_FIELD(SMMU_CB8_NMRR_MAIR1, OR1, 2, 18)
    DEP_FIELD(SMMU_CB8_NMRR_MAIR1, OR0, 2, 16)
    DEP_FIELD(SMMU_CB8_NMRR_MAIR1, IR7, 2, 14)
    DEP_FIELD(SMMU_CB8_NMRR_MAIR1, IR6, 2, 12)
    DEP_FIELD(SMMU_CB8_NMRR_MAIR1, IR5, 2, 10)
    DEP_FIELD(SMMU_CB8_NMRR_MAIR1, IR4, 2, 8)
    DEP_FIELD(SMMU_CB8_NMRR_MAIR1, IR3, 2, 6)
    DEP_FIELD(SMMU_CB8_NMRR_MAIR1, IR2, 2, 4)
    DEP_FIELD(SMMU_CB8_NMRR_MAIR1, IR1, 2, 2)
    DEP_FIELD(SMMU_CB8_NMRR_MAIR1, IR0, 2, 0)
DEP_REG32(SMMU_CB8_FSR, 0x18058)
    DEP_FIELD(SMMU_CB8_FSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_CB8_FSR, SS, 1, 30)
    DEP_FIELD(SMMU_CB8_FSR, FORMAT, 2, 9)
    DEP_FIELD(SMMU_CB8_FSR, UUT, 1, 8)
    DEP_FIELD(SMMU_CB8_FSR, ASF, 1, 7)
    DEP_FIELD(SMMU_CB8_FSR, TLBLKF, 1, 6)
    DEP_FIELD(SMMU_CB8_FSR, TLBMCF, 1, 5)
    DEP_FIELD(SMMU_CB8_FSR, EF, 1, 4)
    DEP_FIELD(SMMU_CB8_FSR, PF, 1, 3)
    DEP_FIELD(SMMU_CB8_FSR, AFF, 1, 2)
    DEP_FIELD(SMMU_CB8_FSR, TF, 1, 1)
DEP_REG32(SMMU_CB8_FSRRESTORE, 0x1805c)
DEP_REG32(SMMU_CB8_FAR_LOW, 0x18060)
DEP_REG32(SMMU_CB8_FAR_HIGH, 0x18064)
    DEP_FIELD(SMMU_CB8_FAR_HIGH, BITS, 17, 0)
DEP_REG32(SMMU_CB8_FSYNR0, 0x18068)
    DEP_FIELD(SMMU_CB8_FSYNR0, S1CBNDX, 4, 16)
    DEP_FIELD(SMMU_CB8_FSYNR0, AFR, 1, 11)
    DEP_FIELD(SMMU_CB8_FSYNR0, PTWF, 1, 10)
    DEP_FIELD(SMMU_CB8_FSYNR0, ATOF, 1, 9)
    DEP_FIELD(SMMU_CB8_FSYNR0, NSATTR, 1, 8)
    DEP_FIELD(SMMU_CB8_FSYNR0, IND, 1, 6)
    DEP_FIELD(SMMU_CB8_FSYNR0, PNU, 1, 5)
    DEP_FIELD(SMMU_CB8_FSYNR0, WNR, 1, 4)
    DEP_FIELD(SMMU_CB8_FSYNR0, PLVL, 2, 0)
DEP_REG32(SMMU_CB8_IPAFAR_LOW, 0x18070)
    DEP_FIELD(SMMU_CB8_IPAFAR_LOW, IPAFAR_L, 20, 12)
    DEP_FIELD(SMMU_CB8_IPAFAR_LOW, FAR_RO, 12, 0)
DEP_REG32(SMMU_CB8_IPAFAR_HIGH, 0x18074)
    DEP_FIELD(SMMU_CB8_IPAFAR_HIGH, BITS, 16, 0)
DEP_REG32(SMMU_CB8_TLBIVA_LOW, 0x18600)
DEP_REG32(SMMU_CB8_TLBIVA_HIGH, 0x18604)
    DEP_FIELD(SMMU_CB8_TLBIVA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB8_TLBIVA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB8_TLBIVAA_LOW, 0x18608)
DEP_REG32(SMMU_CB8_TLBIVAA_HIGH, 0x1860c)
    DEP_FIELD(SMMU_CB8_TLBIVAA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB8_TLBIVAA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB8_TLBIASID, 0x18610)
    DEP_FIELD(SMMU_CB8_TLBIASID, ASID, 16, 0)
DEP_REG32(SMMU_CB8_TLBIALL, 0x18618)
DEP_REG32(SMMU_CB8_TLBIVAL_LOW, 0x18620)
DEP_REG32(SMMU_CB8_TLBIVAL_HIGH, 0x18624)
    DEP_FIELD(SMMU_CB8_TLBIVAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB8_TLBIVAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB8_TLBIVAAL_LOW, 0x18628)
DEP_REG32(SMMU_CB8_TLBIVAAL_HIGH, 0x1862c)
    DEP_FIELD(SMMU_CB8_TLBIVAAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB8_TLBIVAAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB8_TLBIIPAS2_LOW, 0x18630)
DEP_REG32(SMMU_CB8_TLBIIPAS2_HIGH, 0x18634)
    DEP_FIELD(SMMU_CB8_TLBIIPAS2_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB8_TLBIIPAS2L_LOW, 0x18638)
DEP_REG32(SMMU_CB8_TLBIIPAS2L_HIGH, 0x1863c)
    DEP_FIELD(SMMU_CB8_TLBIIPAS2L_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB8_TLBSYNC, 0x187f0)
DEP_REG32(SMMU_CB8_TLBSTATUS, 0x187f4)
    DEP_FIELD(SMMU_CB8_TLBSTATUS, SACTIVE, 1, 0)
DEP_REG32(SMMU_CB8_PMEVCNTR0, 0x18e00)
DEP_REG32(SMMU_CB8_PMEVCNTR1, 0x18e04)
DEP_REG32(SMMU_CB8_PMEVCNTR2, 0x18e08)
DEP_REG32(SMMU_CB8_PMEVCNTR3, 0x18e0c)
DEP_REG32(SMMU_CB8_PMEVTYPER0, 0x18e80)
    DEP_FIELD(SMMU_CB8_PMEVTYPER0, P, 1, 31)
    DEP_FIELD(SMMU_CB8_PMEVTYPER0, U, 1, 30)
    DEP_FIELD(SMMU_CB8_PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(SMMU_CB8_PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(SMMU_CB8_PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(SMMU_CB8_PMEVTYPER1, 0x18e84)
    DEP_FIELD(SMMU_CB8_PMEVTYPER1, P, 1, 31)
    DEP_FIELD(SMMU_CB8_PMEVTYPER1, U, 1, 30)
    DEP_FIELD(SMMU_CB8_PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(SMMU_CB8_PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(SMMU_CB8_PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(SMMU_CB8_PMEVTYPER2, 0x18e88)
    DEP_FIELD(SMMU_CB8_PMEVTYPER2, P, 1, 31)
    DEP_FIELD(SMMU_CB8_PMEVTYPER2, U, 1, 30)
    DEP_FIELD(SMMU_CB8_PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(SMMU_CB8_PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(SMMU_CB8_PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(SMMU_CB8_PMEVTYPER3, 0x18e8c)
    DEP_FIELD(SMMU_CB8_PMEVTYPER3, P, 1, 31)
    DEP_FIELD(SMMU_CB8_PMEVTYPER3, U, 1, 30)
    DEP_FIELD(SMMU_CB8_PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(SMMU_CB8_PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(SMMU_CB8_PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(SMMU_CB8_PMCFGR, 0x18f00)
    DEP_FIELD(SMMU_CB8_PMCFGR, NCG, 8, 24)
    DEP_FIELD(SMMU_CB8_PMCFGR, UEN, 1, 19)
    DEP_FIELD(SMMU_CB8_PMCFGR, EX, 1, 16)
    DEP_FIELD(SMMU_CB8_PMCFGR, CCD, 1, 15)
    DEP_FIELD(SMMU_CB8_PMCFGR, CC, 1, 14)
    DEP_FIELD(SMMU_CB8_PMCFGR, SIZE, 6, 8)
    DEP_FIELD(SMMU_CB8_PMCFGR, N, 8, 0)
DEP_REG32(SMMU_CB8_PMCR, 0x18f04)
    DEP_FIELD(SMMU_CB8_PMCR, IMP, 8, 24)
    DEP_FIELD(SMMU_CB8_PMCR, X, 1, 4)
    DEP_FIELD(SMMU_CB8_PMCR, P, 1, 1)
    DEP_FIELD(SMMU_CB8_PMCR, E, 1, 0)
DEP_REG32(SMMU_CB8_PMCEID, 0x18f20)
    DEP_FIELD(SMMU_CB8_PMCEID, EVENT0X12, 1, 17)
    DEP_FIELD(SMMU_CB8_PMCEID, EVENT0X11, 1, 16)
    DEP_FIELD(SMMU_CB8_PMCEID, EVENT0X10, 1, 15)
    DEP_FIELD(SMMU_CB8_PMCEID, EVENT0X0A, 1, 9)
    DEP_FIELD(SMMU_CB8_PMCEID, EVENT0X09, 1, 8)
    DEP_FIELD(SMMU_CB8_PMCEID, EVENT0X08, 1, 7)
    DEP_FIELD(SMMU_CB8_PMCEID, EVENT0X01, 1, 1)
    DEP_FIELD(SMMU_CB8_PMCEID, EVENT0X00, 1, 0)
DEP_REG32(SMMU_CB8_PMCNTENSE, 0x18f40)
    DEP_FIELD(SMMU_CB8_PMCNTENSE, P3, 1, 3)
    DEP_FIELD(SMMU_CB8_PMCNTENSE, P2, 1, 2)
    DEP_FIELD(SMMU_CB8_PMCNTENSE, P1, 1, 1)
    DEP_FIELD(SMMU_CB8_PMCNTENSE, P0, 1, 0)
DEP_REG32(SMMU_CB8_PMCNTENCLR, 0x18f44)
    DEP_FIELD(SMMU_CB8_PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB8_PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB8_PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB8_PMCNTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB8_PMCNTENSET, 0x18f48)
    DEP_FIELD(SMMU_CB8_PMCNTENSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB8_PMCNTENSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB8_PMCNTENSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB8_PMCNTENSET, P0, 1, 0)
DEP_REG32(SMMU_CB8_PMINTENCLR, 0x18f4c)
    DEP_FIELD(SMMU_CB8_PMINTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB8_PMINTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB8_PMINTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB8_PMINTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB8_PMOVSCLR, 0x18f50)
    DEP_FIELD(SMMU_CB8_PMOVSCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB8_PMOVSCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB8_PMOVSCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB8_PMOVSCLR, P0, 1, 0)
DEP_REG32(SMMU_CB8_PMOVSSET, 0x18f58)
    DEP_FIELD(SMMU_CB8_PMOVSSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB8_PMOVSSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB8_PMOVSSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB8_PMOVSSET, P0, 1, 0)
DEP_REG32(SMMU_CB8_PMAUTHSTATUS, 0x18fb8)
    DEP_FIELD(SMMU_CB8_PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(SMMU_CB8_PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(SMMU_CB8_PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(SMMU_CB8_PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(SMMU_CB8_PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(SMMU_CB8_PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(SMMU_CB8_PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(SMMU_CB8_PMAUTHSTATUS, NSE, 1, 0)
DEP_REG32(SMMU_CB9_SCTLR, 0x19000)
    DEP_FIELD(SMMU_CB9_SCTLR, NSCFG, 2, 28)
    DEP_FIELD(SMMU_CB9_SCTLR, WACFG, 2, 26)
    DEP_FIELD(SMMU_CB9_SCTLR, RACFG, 2, 24)
    DEP_FIELD(SMMU_CB9_SCTLR, SHCFG, 2, 22)
    DEP_FIELD(SMMU_CB9_SCTLR, FB, 1, 21)
    DEP_FIELD(SMMU_CB9_SCTLR, MTCFG, 1, 20)
    DEP_FIELD(SMMU_CB9_SCTLR, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_CB9_SCTLR, TRANSIENTCFG, 2, 14)
    DEP_FIELD(SMMU_CB9_SCTLR, PTW, 1, 13)
    DEP_FIELD(SMMU_CB9_SCTLR, ASIDPNE, 1, 12)
    DEP_FIELD(SMMU_CB9_SCTLR, UWXN, 1, 10)
    DEP_FIELD(SMMU_CB9_SCTLR, WXN, 1, 9)
    DEP_FIELD(SMMU_CB9_SCTLR, HUPCF, 1, 8)
    DEP_FIELD(SMMU_CB9_SCTLR, CFCFG, 1, 7)
    DEP_FIELD(SMMU_CB9_SCTLR, CFIE, 1, 6)
    DEP_FIELD(SMMU_CB9_SCTLR, CFRE, 1, 5)
    DEP_FIELD(SMMU_CB9_SCTLR, E, 1, 4)
    DEP_FIELD(SMMU_CB9_SCTLR, AFFD, 1, 3)
    DEP_FIELD(SMMU_CB9_SCTLR, AFE, 1, 2)
    DEP_FIELD(SMMU_CB9_SCTLR, TRE, 1, 1)
    DEP_FIELD(SMMU_CB9_SCTLR, M, 1, 0)
DEP_REG32(SMMU_CB9_ACTLR, 0x19004)
    DEP_FIELD(SMMU_CB9_ACTLR, CPRE, 1, 1)
    DEP_FIELD(SMMU_CB9_ACTLR, CMTLB, 1, 0)
DEP_REG32(SMMU_CB9_RESUME, 0x19008)
    DEP_FIELD(SMMU_CB9_RESUME, TNR, 1, 0)
DEP_REG32(SMMU_CB9_TCR2, 0x19010)
    DEP_FIELD(SMMU_CB9_TCR2, NSCFG1, 1, 30)
    DEP_FIELD(SMMU_CB9_TCR2, SEP, 3, 15)
    DEP_FIELD(SMMU_CB9_TCR2, NSCFG0, 1, 14)
    DEP_FIELD(SMMU_CB9_TCR2, TBI1, 1, 6)
    DEP_FIELD(SMMU_CB9_TCR2, TBI0, 1, 5)
    DEP_FIELD(SMMU_CB9_TCR2, AS, 1, 4)
    DEP_FIELD(SMMU_CB9_TCR2, PASIZE, 3, 0)
DEP_REG32(SMMU_CB9_TTBR0_LOW, 0x19020)
    DEP_FIELD(SMMU_CB9_TTBR0_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB9_TTBR0_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB9_TTBR0_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB9_TTBR0_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB9_TTBR0_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB9_TTBR0_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB9_TTBR0_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB9_TTBR0_HIGH, 0x19024)
    DEP_FIELD(SMMU_CB9_TTBR0_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB9_TTBR0_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB9_TTBR1_LOW, 0x19028)
    DEP_FIELD(SMMU_CB9_TTBR1_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB9_TTBR1_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB9_TTBR1_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB9_TTBR1_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB9_TTBR1_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB9_TTBR1_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB9_TTBR1_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB9_TTBR1_HIGH, 0x1902c)
    DEP_FIELD(SMMU_CB9_TTBR1_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB9_TTBR1_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB9_TCR_LPAE, 0x19030)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, EAE, 1, 31)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, NSCFG1_TG1, 1, 30)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, SH1, 2, 28)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, ORGN1, 2, 26)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, IRGN1, 2, 24)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, EPD1, 1, 23)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, A1, 1, 22)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, T1SZ_5_3, 3, 19)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, T1SZ_2_0_PASIZE, 3, 16)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, NSCFG0_TG0, 1, 14)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, SH0, 2, 12)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, ORGN0, 2, 10)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, IRGN0, 2, 8)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, SL0_1_EPD0, 1, 7)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, SL0_0, 1, 6)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, PD1_T0SZ_5, 1, 5)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, S_PD0_T0SZ_4, 1, 4)
    DEP_FIELD(SMMU_CB9_TCR_LPAE, T0SZ_3_0, 4, 0)
DEP_REG32(SMMU_CB9_CONTEXTIDR, 0x19034)
    DEP_FIELD(SMMU_CB9_CONTEXTIDR, PROCID, 24, 8)
    DEP_FIELD(SMMU_CB9_CONTEXTIDR, ASID, 8, 0)
DEP_REG32(SMMU_CB9_PRRR_MAIR0, 0x19038)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, NOS7, 1, 31)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, NOS6, 1, 30)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, NOS5, 1, 29)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, NOS4, 1, 28)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, NOS3, 1, 27)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, NOS2, 1, 26)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, NOS1, 1, 25)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, NOS0, 1, 24)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, NS1, 1, 19)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, NS0, 1, 18)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, DS1, 1, 17)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, DS0, 1, 16)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, TR7, 2, 14)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, TR6, 2, 12)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, TR5, 2, 10)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, TR4, 2, 8)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, TR3, 2, 6)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, TR2, 2, 4)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, TR1, 2, 2)
    DEP_FIELD(SMMU_CB9_PRRR_MAIR0, TR0, 2, 0)
DEP_REG32(SMMU_CB9_NMRR_MAIR1, 0x1903c)
    DEP_FIELD(SMMU_CB9_NMRR_MAIR1, OR7, 2, 30)
    DEP_FIELD(SMMU_CB9_NMRR_MAIR1, OR6, 2, 28)
    DEP_FIELD(SMMU_CB9_NMRR_MAIR1, OR5, 2, 26)
    DEP_FIELD(SMMU_CB9_NMRR_MAIR1, OR4, 2, 24)
    DEP_FIELD(SMMU_CB9_NMRR_MAIR1, OR3, 2, 22)
    DEP_FIELD(SMMU_CB9_NMRR_MAIR1, OR2, 2, 20)
    DEP_FIELD(SMMU_CB9_NMRR_MAIR1, OR1, 2, 18)
    DEP_FIELD(SMMU_CB9_NMRR_MAIR1, OR0, 2, 16)
    DEP_FIELD(SMMU_CB9_NMRR_MAIR1, IR7, 2, 14)
    DEP_FIELD(SMMU_CB9_NMRR_MAIR1, IR6, 2, 12)
    DEP_FIELD(SMMU_CB9_NMRR_MAIR1, IR5, 2, 10)
    DEP_FIELD(SMMU_CB9_NMRR_MAIR1, IR4, 2, 8)
    DEP_FIELD(SMMU_CB9_NMRR_MAIR1, IR3, 2, 6)
    DEP_FIELD(SMMU_CB9_NMRR_MAIR1, IR2, 2, 4)
    DEP_FIELD(SMMU_CB9_NMRR_MAIR1, IR1, 2, 2)
    DEP_FIELD(SMMU_CB9_NMRR_MAIR1, IR0, 2, 0)
DEP_REG32(SMMU_CB9_FSR, 0x19058)
    DEP_FIELD(SMMU_CB9_FSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_CB9_FSR, SS, 1, 30)
    DEP_FIELD(SMMU_CB9_FSR, FORMAT, 2, 9)
    DEP_FIELD(SMMU_CB9_FSR, UUT, 1, 8)
    DEP_FIELD(SMMU_CB9_FSR, ASF, 1, 7)
    DEP_FIELD(SMMU_CB9_FSR, TLBLKF, 1, 6)
    DEP_FIELD(SMMU_CB9_FSR, TLBMCF, 1, 5)
    DEP_FIELD(SMMU_CB9_FSR, EF, 1, 4)
    DEP_FIELD(SMMU_CB9_FSR, PF, 1, 3)
    DEP_FIELD(SMMU_CB9_FSR, AFF, 1, 2)
    DEP_FIELD(SMMU_CB9_FSR, TF, 1, 1)
DEP_REG32(SMMU_CB9_FSRRESTORE, 0x1905c)
DEP_REG32(SMMU_CB9_FAR_LOW, 0x19060)
DEP_REG32(SMMU_CB9_FAR_HIGH, 0x19064)
    DEP_FIELD(SMMU_CB9_FAR_HIGH, BITS, 17, 0)
DEP_REG32(SMMU_CB9_FSYNR0, 0x19068)
    DEP_FIELD(SMMU_CB9_FSYNR0, S1CBNDX, 4, 16)
    DEP_FIELD(SMMU_CB9_FSYNR0, AFR, 1, 11)
    DEP_FIELD(SMMU_CB9_FSYNR0, PTWF, 1, 10)
    DEP_FIELD(SMMU_CB9_FSYNR0, ATOF, 1, 9)
    DEP_FIELD(SMMU_CB9_FSYNR0, NSATTR, 1, 8)
    DEP_FIELD(SMMU_CB9_FSYNR0, IND, 1, 6)
    DEP_FIELD(SMMU_CB9_FSYNR0, PNU, 1, 5)
    DEP_FIELD(SMMU_CB9_FSYNR0, WNR, 1, 4)
    DEP_FIELD(SMMU_CB9_FSYNR0, PLVL, 2, 0)
DEP_REG32(SMMU_CB9_IPAFAR_LOW, 0x19070)
    DEP_FIELD(SMMU_CB9_IPAFAR_LOW, IPAFAR_L, 20, 12)
    DEP_FIELD(SMMU_CB9_IPAFAR_LOW, FAR_RO, 12, 0)
DEP_REG32(SMMU_CB9_IPAFAR_HIGH, 0x19074)
    DEP_FIELD(SMMU_CB9_IPAFAR_HIGH, BITS, 16, 0)
DEP_REG32(SMMU_CB9_TLBIVA_LOW, 0x19600)
DEP_REG32(SMMU_CB9_TLBIVA_HIGH, 0x19604)
    DEP_FIELD(SMMU_CB9_TLBIVA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB9_TLBIVA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB9_TLBIVAA_LOW, 0x19608)
DEP_REG32(SMMU_CB9_TLBIVAA_HIGH, 0x1960c)
    DEP_FIELD(SMMU_CB9_TLBIVAA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB9_TLBIVAA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB9_TLBIASID, 0x19610)
    DEP_FIELD(SMMU_CB9_TLBIASID, ASID, 16, 0)
DEP_REG32(SMMU_CB9_TLBIALL, 0x19618)
DEP_REG32(SMMU_CB9_TLBIVAL_LOW, 0x19620)
DEP_REG32(SMMU_CB9_TLBIVAL_HIGH, 0x19624)
    DEP_FIELD(SMMU_CB9_TLBIVAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB9_TLBIVAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB9_TLBIVAAL_LOW, 0x19628)
DEP_REG32(SMMU_CB9_TLBIVAAL_HIGH, 0x1962c)
    DEP_FIELD(SMMU_CB9_TLBIVAAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB9_TLBIVAAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB9_TLBIIPAS2_LOW, 0x19630)
DEP_REG32(SMMU_CB9_TLBIIPAS2_HIGH, 0x19634)
    DEP_FIELD(SMMU_CB9_TLBIIPAS2_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB9_TLBIIPAS2L_LOW, 0x19638)
DEP_REG32(SMMU_CB9_TLBIIPAS2L_HIGH, 0x1963c)
    DEP_FIELD(SMMU_CB9_TLBIIPAS2L_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB9_TLBSYNC, 0x197f0)
DEP_REG32(SMMU_CB9_TLBSTATUS, 0x197f4)
    DEP_FIELD(SMMU_CB9_TLBSTATUS, SACTIVE, 1, 0)
DEP_REG32(SMMU_CB9_PMEVCNTR0, 0x19e00)
DEP_REG32(SMMU_CB9_PMEVCNTR1, 0x19e04)
DEP_REG32(SMMU_CB9_PMEVCNTR2, 0x19e08)
DEP_REG32(SMMU_CB9_PMEVCNTR3, 0x19e0c)
DEP_REG32(SMMU_CB9_PMEVTYPER0, 0x19e80)
    DEP_FIELD(SMMU_CB9_PMEVTYPER0, P, 1, 31)
    DEP_FIELD(SMMU_CB9_PMEVTYPER0, U, 1, 30)
    DEP_FIELD(SMMU_CB9_PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(SMMU_CB9_PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(SMMU_CB9_PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(SMMU_CB9_PMEVTYPER1, 0x19e84)
    DEP_FIELD(SMMU_CB9_PMEVTYPER1, P, 1, 31)
    DEP_FIELD(SMMU_CB9_PMEVTYPER1, U, 1, 30)
    DEP_FIELD(SMMU_CB9_PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(SMMU_CB9_PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(SMMU_CB9_PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(SMMU_CB9_PMEVTYPER2, 0x19e88)
    DEP_FIELD(SMMU_CB9_PMEVTYPER2, P, 1, 31)
    DEP_FIELD(SMMU_CB9_PMEVTYPER2, U, 1, 30)
    DEP_FIELD(SMMU_CB9_PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(SMMU_CB9_PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(SMMU_CB9_PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(SMMU_CB9_PMEVTYPER3, 0x19e8c)
    DEP_FIELD(SMMU_CB9_PMEVTYPER3, P, 1, 31)
    DEP_FIELD(SMMU_CB9_PMEVTYPER3, U, 1, 30)
    DEP_FIELD(SMMU_CB9_PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(SMMU_CB9_PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(SMMU_CB9_PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(SMMU_CB9_PMCFGR, 0x19f00)
    DEP_FIELD(SMMU_CB9_PMCFGR, NCG, 8, 24)
    DEP_FIELD(SMMU_CB9_PMCFGR, UEN, 1, 19)
    DEP_FIELD(SMMU_CB9_PMCFGR, EX, 1, 16)
    DEP_FIELD(SMMU_CB9_PMCFGR, CCD, 1, 15)
    DEP_FIELD(SMMU_CB9_PMCFGR, CC, 1, 14)
    DEP_FIELD(SMMU_CB9_PMCFGR, SIZE, 6, 8)
    DEP_FIELD(SMMU_CB9_PMCFGR, N, 8, 0)
DEP_REG32(SMMU_CB9_PMCR, 0x19f04)
    DEP_FIELD(SMMU_CB9_PMCR, IMP, 8, 24)
    DEP_FIELD(SMMU_CB9_PMCR, X, 1, 4)
    DEP_FIELD(SMMU_CB9_PMCR, P, 1, 1)
    DEP_FIELD(SMMU_CB9_PMCR, E, 1, 0)
DEP_REG32(SMMU_CB9_PMCEID, 0x19f20)
    DEP_FIELD(SMMU_CB9_PMCEID, EVENT0X12, 1, 17)
    DEP_FIELD(SMMU_CB9_PMCEID, EVENT0X11, 1, 16)
    DEP_FIELD(SMMU_CB9_PMCEID, EVENT0X10, 1, 15)
    DEP_FIELD(SMMU_CB9_PMCEID, EVENT0X0A, 1, 9)
    DEP_FIELD(SMMU_CB9_PMCEID, EVENT0X09, 1, 8)
    DEP_FIELD(SMMU_CB9_PMCEID, EVENT0X08, 1, 7)
    DEP_FIELD(SMMU_CB9_PMCEID, EVENT0X01, 1, 1)
    DEP_FIELD(SMMU_CB9_PMCEID, EVENT0X00, 1, 0)
DEP_REG32(SMMU_CB9_PMCNTENSE, 0x19f40)
    DEP_FIELD(SMMU_CB9_PMCNTENSE, P3, 1, 3)
    DEP_FIELD(SMMU_CB9_PMCNTENSE, P2, 1, 2)
    DEP_FIELD(SMMU_CB9_PMCNTENSE, P1, 1, 1)
    DEP_FIELD(SMMU_CB9_PMCNTENSE, P0, 1, 0)
DEP_REG32(SMMU_CB9_PMCNTENCLR, 0x19f44)
    DEP_FIELD(SMMU_CB9_PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB9_PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB9_PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB9_PMCNTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB9_PMCNTENSET, 0x19f48)
    DEP_FIELD(SMMU_CB9_PMCNTENSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB9_PMCNTENSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB9_PMCNTENSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB9_PMCNTENSET, P0, 1, 0)
DEP_REG32(SMMU_CB9_PMINTENCLR, 0x19f4c)
    DEP_FIELD(SMMU_CB9_PMINTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB9_PMINTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB9_PMINTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB9_PMINTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB9_PMOVSCLR, 0x19f50)
    DEP_FIELD(SMMU_CB9_PMOVSCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB9_PMOVSCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB9_PMOVSCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB9_PMOVSCLR, P0, 1, 0)
DEP_REG32(SMMU_CB9_PMOVSSET, 0x19f58)
    DEP_FIELD(SMMU_CB9_PMOVSSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB9_PMOVSSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB9_PMOVSSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB9_PMOVSSET, P0, 1, 0)
DEP_REG32(SMMU_CB9_PMAUTHSTATUS, 0x19fb8)
    DEP_FIELD(SMMU_CB9_PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(SMMU_CB9_PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(SMMU_CB9_PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(SMMU_CB9_PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(SMMU_CB9_PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(SMMU_CB9_PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(SMMU_CB9_PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(SMMU_CB9_PMAUTHSTATUS, NSE, 1, 0)
DEP_REG32(SMMU_CB10_SCTLR, 0x1a000)
    DEP_FIELD(SMMU_CB10_SCTLR, NSCFG, 2, 28)
    DEP_FIELD(SMMU_CB10_SCTLR, WACFG, 2, 26)
    DEP_FIELD(SMMU_CB10_SCTLR, RACFG, 2, 24)
    DEP_FIELD(SMMU_CB10_SCTLR, SHCFG, 2, 22)
    DEP_FIELD(SMMU_CB10_SCTLR, FB, 1, 21)
    DEP_FIELD(SMMU_CB10_SCTLR, MTCFG, 1, 20)
    DEP_FIELD(SMMU_CB10_SCTLR, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_CB10_SCTLR, TRANSIENTCFG, 2, 14)
    DEP_FIELD(SMMU_CB10_SCTLR, PTW, 1, 13)
    DEP_FIELD(SMMU_CB10_SCTLR, ASIDPNE, 1, 12)
    DEP_FIELD(SMMU_CB10_SCTLR, UWXN, 1, 10)
    DEP_FIELD(SMMU_CB10_SCTLR, WXN, 1, 9)
    DEP_FIELD(SMMU_CB10_SCTLR, HUPCF, 1, 8)
    DEP_FIELD(SMMU_CB10_SCTLR, CFCFG, 1, 7)
    DEP_FIELD(SMMU_CB10_SCTLR, CFIE, 1, 6)
    DEP_FIELD(SMMU_CB10_SCTLR, CFRE, 1, 5)
    DEP_FIELD(SMMU_CB10_SCTLR, E, 1, 4)
    DEP_FIELD(SMMU_CB10_SCTLR, AFFD, 1, 3)
    DEP_FIELD(SMMU_CB10_SCTLR, AFE, 1, 2)
    DEP_FIELD(SMMU_CB10_SCTLR, TRE, 1, 1)
    DEP_FIELD(SMMU_CB10_SCTLR, M, 1, 0)
DEP_REG32(SMMU_CB10_ACTLR, 0x1a004)
    DEP_FIELD(SMMU_CB10_ACTLR, CPRE, 1, 1)
    DEP_FIELD(SMMU_CB10_ACTLR, CMTLB, 1, 0)
DEP_REG32(SMMU_CB10_RESUME, 0x1a008)
    DEP_FIELD(SMMU_CB10_RESUME, TNR, 1, 0)
DEP_REG32(SMMU_CB10_TCR2, 0x1a010)
    DEP_FIELD(SMMU_CB10_TCR2, NSCFG1, 1, 30)
    DEP_FIELD(SMMU_CB10_TCR2, SEP, 3, 15)
    DEP_FIELD(SMMU_CB10_TCR2, NSCFG0, 1, 14)
    DEP_FIELD(SMMU_CB10_TCR2, TBI1, 1, 6)
    DEP_FIELD(SMMU_CB10_TCR2, TBI0, 1, 5)
    DEP_FIELD(SMMU_CB10_TCR2, AS, 1, 4)
    DEP_FIELD(SMMU_CB10_TCR2, PASIZE, 3, 0)
DEP_REG32(SMMU_CB10_TTBR0_LOW, 0x1a020)
    DEP_FIELD(SMMU_CB10_TTBR0_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB10_TTBR0_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB10_TTBR0_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB10_TTBR0_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB10_TTBR0_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB10_TTBR0_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB10_TTBR0_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB10_TTBR0_HIGH, 0x1a024)
    DEP_FIELD(SMMU_CB10_TTBR0_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB10_TTBR0_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB10_TTBR1_LOW, 0x1a028)
    DEP_FIELD(SMMU_CB10_TTBR1_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB10_TTBR1_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB10_TTBR1_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB10_TTBR1_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB10_TTBR1_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB10_TTBR1_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB10_TTBR1_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB10_TTBR1_HIGH, 0x1a02c)
    DEP_FIELD(SMMU_CB10_TTBR1_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB10_TTBR1_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB10_TCR_LPAE, 0x1a030)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, EAE, 1, 31)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, NSCFG1_TG1, 1, 30)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, SH1, 2, 28)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, ORGN1, 2, 26)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, IRGN1, 2, 24)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, EPD1, 1, 23)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, A1, 1, 22)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, T1SZ_5_3, 3, 19)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, T1SZ_2_0_PASIZE, 3, 16)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, NSCFG0_TG0, 1, 14)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, SH0, 2, 12)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, ORGN0, 2, 10)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, IRGN0, 2, 8)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, SL0_1_EPD0, 1, 7)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, SL0_0, 1, 6)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, PD1_T0SZ_5, 1, 5)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, S_PD0_T0SZ_4, 1, 4)
    DEP_FIELD(SMMU_CB10_TCR_LPAE, T0SZ_3_0, 4, 0)
DEP_REG32(SMMU_CB10_CONTEXTIDR, 0x1a034)
    DEP_FIELD(SMMU_CB10_CONTEXTIDR, PROCID, 24, 8)
    DEP_FIELD(SMMU_CB10_CONTEXTIDR, ASID, 8, 0)
DEP_REG32(SMMU_CB10_PRRR_MAIR0, 0x1a038)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, NOS7, 1, 31)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, NOS6, 1, 30)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, NOS5, 1, 29)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, NOS4, 1, 28)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, NOS3, 1, 27)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, NOS2, 1, 26)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, NOS1, 1, 25)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, NOS0, 1, 24)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, NS1, 1, 19)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, NS0, 1, 18)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, DS1, 1, 17)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, DS0, 1, 16)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, TR7, 2, 14)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, TR6, 2, 12)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, TR5, 2, 10)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, TR4, 2, 8)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, TR3, 2, 6)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, TR2, 2, 4)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, TR1, 2, 2)
    DEP_FIELD(SMMU_CB10_PRRR_MAIR0, TR0, 2, 0)
DEP_REG32(SMMU_CB10_NMRR_MAIR1, 0x1a03c)
    DEP_FIELD(SMMU_CB10_NMRR_MAIR1, OR7, 2, 30)
    DEP_FIELD(SMMU_CB10_NMRR_MAIR1, OR6, 2, 28)
    DEP_FIELD(SMMU_CB10_NMRR_MAIR1, OR5, 2, 26)
    DEP_FIELD(SMMU_CB10_NMRR_MAIR1, OR4, 2, 24)
    DEP_FIELD(SMMU_CB10_NMRR_MAIR1, OR3, 2, 22)
    DEP_FIELD(SMMU_CB10_NMRR_MAIR1, OR2, 2, 20)
    DEP_FIELD(SMMU_CB10_NMRR_MAIR1, OR1, 2, 18)
    DEP_FIELD(SMMU_CB10_NMRR_MAIR1, OR0, 2, 16)
    DEP_FIELD(SMMU_CB10_NMRR_MAIR1, IR7, 2, 14)
    DEP_FIELD(SMMU_CB10_NMRR_MAIR1, IR6, 2, 12)
    DEP_FIELD(SMMU_CB10_NMRR_MAIR1, IR5, 2, 10)
    DEP_FIELD(SMMU_CB10_NMRR_MAIR1, IR4, 2, 8)
    DEP_FIELD(SMMU_CB10_NMRR_MAIR1, IR3, 2, 6)
    DEP_FIELD(SMMU_CB10_NMRR_MAIR1, IR2, 2, 4)
    DEP_FIELD(SMMU_CB10_NMRR_MAIR1, IR1, 2, 2)
    DEP_FIELD(SMMU_CB10_NMRR_MAIR1, IR0, 2, 0)
DEP_REG32(SMMU_CB10_FSR, 0x1a058)
    DEP_FIELD(SMMU_CB10_FSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_CB10_FSR, SS, 1, 30)
    DEP_FIELD(SMMU_CB10_FSR, FORMAT, 2, 9)
    DEP_FIELD(SMMU_CB10_FSR, UUT, 1, 8)
    DEP_FIELD(SMMU_CB10_FSR, ASF, 1, 7)
    DEP_FIELD(SMMU_CB10_FSR, TLBLKF, 1, 6)
    DEP_FIELD(SMMU_CB10_FSR, TLBMCF, 1, 5)
    DEP_FIELD(SMMU_CB10_FSR, EF, 1, 4)
    DEP_FIELD(SMMU_CB10_FSR, PF, 1, 3)
    DEP_FIELD(SMMU_CB10_FSR, AFF, 1, 2)
    DEP_FIELD(SMMU_CB10_FSR, TF, 1, 1)
DEP_REG32(SMMU_CB10_FSRRESTORE, 0x1a05c)
DEP_REG32(SMMU_CB10_FAR_LOW, 0x1a060)
DEP_REG32(SMMU_CB10_FAR_HIGH, 0x1a064)
    DEP_FIELD(SMMU_CB10_FAR_HIGH, BITS, 17, 0)
DEP_REG32(SMMU_CB10_FSYNR0, 0x1a068)
    DEP_FIELD(SMMU_CB10_FSYNR0, S1CBNDX, 4, 16)
    DEP_FIELD(SMMU_CB10_FSYNR0, AFR, 1, 11)
    DEP_FIELD(SMMU_CB10_FSYNR0, PTWF, 1, 10)
    DEP_FIELD(SMMU_CB10_FSYNR0, ATOF, 1, 9)
    DEP_FIELD(SMMU_CB10_FSYNR0, NSATTR, 1, 8)
    DEP_FIELD(SMMU_CB10_FSYNR0, IND, 1, 6)
    DEP_FIELD(SMMU_CB10_FSYNR0, PNU, 1, 5)
    DEP_FIELD(SMMU_CB10_FSYNR0, WNR, 1, 4)
    DEP_FIELD(SMMU_CB10_FSYNR0, PLVL, 2, 0)
DEP_REG32(SMMU_CB10_IPAFAR_LOW, 0x1a070)
    DEP_FIELD(SMMU_CB10_IPAFAR_LOW, IPAFAR_L, 20, 12)
    DEP_FIELD(SMMU_CB10_IPAFAR_LOW, FAR_RO, 12, 0)
DEP_REG32(SMMU_CB10_IPAFAR_HIGH, 0x1a074)
    DEP_FIELD(SMMU_CB10_IPAFAR_HIGH, BITS, 16, 0)
DEP_REG32(SMMU_CB10_TLBIVA_LOW, 0x1a600)
DEP_REG32(SMMU_CB10_TLBIVA_HIGH, 0x1a604)
    DEP_FIELD(SMMU_CB10_TLBIVA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB10_TLBIVA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB10_TLBIVAA_LOW, 0x1a608)
DEP_REG32(SMMU_CB10_TLBIVAA_HIGH, 0x1a60c)
    DEP_FIELD(SMMU_CB10_TLBIVAA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB10_TLBIVAA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB10_TLBIASID, 0x1a610)
    DEP_FIELD(SMMU_CB10_TLBIASID, ASID, 16, 0)
DEP_REG32(SMMU_CB10_TLBIALL, 0x1a618)
DEP_REG32(SMMU_CB10_TLBIVAL_LOW, 0x1a620)
DEP_REG32(SMMU_CB10_TLBIVAL_HIGH, 0x1a624)
    DEP_FIELD(SMMU_CB10_TLBIVAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB10_TLBIVAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB10_TLBIVAAL_LOW, 0x1a628)
DEP_REG32(SMMU_CB10_TLBIVAAL_HIGH, 0x1a62c)
    DEP_FIELD(SMMU_CB10_TLBIVAAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB10_TLBIVAAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB10_TLBIIPAS2_LOW, 0x1a630)
DEP_REG32(SMMU_CB10_TLBIIPAS2_HIGH, 0x1a634)
    DEP_FIELD(SMMU_CB10_TLBIIPAS2_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB10_TLBIIPAS2L_LOW, 0x1a638)
DEP_REG32(SMMU_CB10_TLBIIPAS2L_HIGH, 0x1a63c)
    DEP_FIELD(SMMU_CB10_TLBIIPAS2L_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB10_TLBSYNC, 0x1a7f0)
DEP_REG32(SMMU_CB10_TLBSTATUS, 0x1a7f4)
    DEP_FIELD(SMMU_CB10_TLBSTATUS, SACTIVE, 1, 0)
DEP_REG32(SMMU_CB10_PMEVCNTR0, 0x1ae00)
DEP_REG32(SMMU_CB10_PMEVCNTR1, 0x1ae04)
DEP_REG32(SMMU_CB10_PMEVCNTR2, 0x1ae08)
DEP_REG32(SMMU_CB10_PMEVCNTR3, 0x1ae0c)
DEP_REG32(SMMU_CB10_PMEVTYPER0, 0x1ae80)
    DEP_FIELD(SMMU_CB10_PMEVTYPER0, P, 1, 31)
    DEP_FIELD(SMMU_CB10_PMEVTYPER0, U, 1, 30)
    DEP_FIELD(SMMU_CB10_PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(SMMU_CB10_PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(SMMU_CB10_PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(SMMU_CB10_PMEVTYPER1, 0x1ae84)
    DEP_FIELD(SMMU_CB10_PMEVTYPER1, P, 1, 31)
    DEP_FIELD(SMMU_CB10_PMEVTYPER1, U, 1, 30)
    DEP_FIELD(SMMU_CB10_PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(SMMU_CB10_PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(SMMU_CB10_PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(SMMU_CB10_PMEVTYPER2, 0x1ae88)
    DEP_FIELD(SMMU_CB10_PMEVTYPER2, P, 1, 31)
    DEP_FIELD(SMMU_CB10_PMEVTYPER2, U, 1, 30)
    DEP_FIELD(SMMU_CB10_PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(SMMU_CB10_PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(SMMU_CB10_PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(SMMU_CB10_PMEVTYPER3, 0x1ae8c)
    DEP_FIELD(SMMU_CB10_PMEVTYPER3, P, 1, 31)
    DEP_FIELD(SMMU_CB10_PMEVTYPER3, U, 1, 30)
    DEP_FIELD(SMMU_CB10_PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(SMMU_CB10_PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(SMMU_CB10_PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(SMMU_CB10_PMCFGR, 0x1af00)
    DEP_FIELD(SMMU_CB10_PMCFGR, NCG, 8, 24)
    DEP_FIELD(SMMU_CB10_PMCFGR, UEN, 1, 19)
    DEP_FIELD(SMMU_CB10_PMCFGR, EX, 1, 16)
    DEP_FIELD(SMMU_CB10_PMCFGR, CCD, 1, 15)
    DEP_FIELD(SMMU_CB10_PMCFGR, CC, 1, 14)
    DEP_FIELD(SMMU_CB10_PMCFGR, SIZE, 6, 8)
    DEP_FIELD(SMMU_CB10_PMCFGR, N, 8, 0)
DEP_REG32(SMMU_CB10_PMCR, 0x1af04)
    DEP_FIELD(SMMU_CB10_PMCR, IMP, 8, 24)
    DEP_FIELD(SMMU_CB10_PMCR, X, 1, 4)
    DEP_FIELD(SMMU_CB10_PMCR, P, 1, 1)
    DEP_FIELD(SMMU_CB10_PMCR, E, 1, 0)
DEP_REG32(SMMU_CB10_PMCEID, 0x1af20)
    DEP_FIELD(SMMU_CB10_PMCEID, EVENT0X12, 1, 17)
    DEP_FIELD(SMMU_CB10_PMCEID, EVENT0X11, 1, 16)
    DEP_FIELD(SMMU_CB10_PMCEID, EVENT0X10, 1, 15)
    DEP_FIELD(SMMU_CB10_PMCEID, EVENT0X0A, 1, 9)
    DEP_FIELD(SMMU_CB10_PMCEID, EVENT0X09, 1, 8)
    DEP_FIELD(SMMU_CB10_PMCEID, EVENT0X08, 1, 7)
    DEP_FIELD(SMMU_CB10_PMCEID, EVENT0X01, 1, 1)
    DEP_FIELD(SMMU_CB10_PMCEID, EVENT0X00, 1, 0)
DEP_REG32(SMMU_CB10_PMCNTENSE, 0x1af40)
    DEP_FIELD(SMMU_CB10_PMCNTENSE, P3, 1, 3)
    DEP_FIELD(SMMU_CB10_PMCNTENSE, P2, 1, 2)
    DEP_FIELD(SMMU_CB10_PMCNTENSE, P1, 1, 1)
    DEP_FIELD(SMMU_CB10_PMCNTENSE, P0, 1, 0)
DEP_REG32(SMMU_CB10_PMCNTENCLR, 0x1af44)
    DEP_FIELD(SMMU_CB10_PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB10_PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB10_PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB10_PMCNTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB10_PMCNTENSET, 0x1af48)
    DEP_FIELD(SMMU_CB10_PMCNTENSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB10_PMCNTENSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB10_PMCNTENSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB10_PMCNTENSET, P0, 1, 0)
DEP_REG32(SMMU_CB10_PMINTENCLR, 0x1af4c)
    DEP_FIELD(SMMU_CB10_PMINTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB10_PMINTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB10_PMINTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB10_PMINTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB10_PMOVSCLR, 0x1af50)
    DEP_FIELD(SMMU_CB10_PMOVSCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB10_PMOVSCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB10_PMOVSCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB10_PMOVSCLR, P0, 1, 0)
DEP_REG32(SMMU_CB10_PMOVSSET, 0x1af58)
    DEP_FIELD(SMMU_CB10_PMOVSSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB10_PMOVSSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB10_PMOVSSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB10_PMOVSSET, P0, 1, 0)
DEP_REG32(SMMU_CB10_PMAUTHSTATUS, 0x1afb8)
    DEP_FIELD(SMMU_CB10_PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(SMMU_CB10_PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(SMMU_CB10_PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(SMMU_CB10_PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(SMMU_CB10_PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(SMMU_CB10_PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(SMMU_CB10_PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(SMMU_CB10_PMAUTHSTATUS, NSE, 1, 0)
DEP_REG32(SMMU_CB11_SCTLR, 0x1b000)
    DEP_FIELD(SMMU_CB11_SCTLR, NSCFG, 2, 28)
    DEP_FIELD(SMMU_CB11_SCTLR, WACFG, 2, 26)
    DEP_FIELD(SMMU_CB11_SCTLR, RACFG, 2, 24)
    DEP_FIELD(SMMU_CB11_SCTLR, SHCFG, 2, 22)
    DEP_FIELD(SMMU_CB11_SCTLR, FB, 1, 21)
    DEP_FIELD(SMMU_CB11_SCTLR, MTCFG, 1, 20)
    DEP_FIELD(SMMU_CB11_SCTLR, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_CB11_SCTLR, TRANSIENTCFG, 2, 14)
    DEP_FIELD(SMMU_CB11_SCTLR, PTW, 1, 13)
    DEP_FIELD(SMMU_CB11_SCTLR, ASIDPNE, 1, 12)
    DEP_FIELD(SMMU_CB11_SCTLR, UWXN, 1, 10)
    DEP_FIELD(SMMU_CB11_SCTLR, WXN, 1, 9)
    DEP_FIELD(SMMU_CB11_SCTLR, HUPCF, 1, 8)
    DEP_FIELD(SMMU_CB11_SCTLR, CFCFG, 1, 7)
    DEP_FIELD(SMMU_CB11_SCTLR, CFIE, 1, 6)
    DEP_FIELD(SMMU_CB11_SCTLR, CFRE, 1, 5)
    DEP_FIELD(SMMU_CB11_SCTLR, E, 1, 4)
    DEP_FIELD(SMMU_CB11_SCTLR, AFFD, 1, 3)
    DEP_FIELD(SMMU_CB11_SCTLR, AFE, 1, 2)
    DEP_FIELD(SMMU_CB11_SCTLR, TRE, 1, 1)
    DEP_FIELD(SMMU_CB11_SCTLR, M, 1, 0)
DEP_REG32(SMMU_CB11_ACTLR, 0x1b004)
    DEP_FIELD(SMMU_CB11_ACTLR, CPRE, 1, 1)
    DEP_FIELD(SMMU_CB11_ACTLR, CMTLB, 1, 0)
DEP_REG32(SMMU_CB11_RESUME, 0x1b008)
    DEP_FIELD(SMMU_CB11_RESUME, TNR, 1, 0)
DEP_REG32(SMMU_CB11_TCR2, 0x1b010)
    DEP_FIELD(SMMU_CB11_TCR2, NSCFG1, 1, 30)
    DEP_FIELD(SMMU_CB11_TCR2, SEP, 3, 15)
    DEP_FIELD(SMMU_CB11_TCR2, NSCFG0, 1, 14)
    DEP_FIELD(SMMU_CB11_TCR2, TBI1, 1, 6)
    DEP_FIELD(SMMU_CB11_TCR2, TBI0, 1, 5)
    DEP_FIELD(SMMU_CB11_TCR2, AS, 1, 4)
    DEP_FIELD(SMMU_CB11_TCR2, PASIZE, 3, 0)
DEP_REG32(SMMU_CB11_TTBR0_LOW, 0x1b020)
    DEP_FIELD(SMMU_CB11_TTBR0_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB11_TTBR0_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB11_TTBR0_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB11_TTBR0_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB11_TTBR0_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB11_TTBR0_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB11_TTBR0_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB11_TTBR0_HIGH, 0x1b024)
    DEP_FIELD(SMMU_CB11_TTBR0_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB11_TTBR0_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB11_TTBR1_LOW, 0x1b028)
    DEP_FIELD(SMMU_CB11_TTBR1_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB11_TTBR1_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB11_TTBR1_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB11_TTBR1_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB11_TTBR1_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB11_TTBR1_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB11_TTBR1_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB11_TTBR1_HIGH, 0x1b02c)
    DEP_FIELD(SMMU_CB11_TTBR1_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB11_TTBR1_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB11_TCR_LPAE, 0x1b030)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, EAE, 1, 31)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, NSCFG1_TG1, 1, 30)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, SH1, 2, 28)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, ORGN1, 2, 26)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, IRGN1, 2, 24)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, EPD1, 1, 23)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, A1, 1, 22)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, T1SZ_5_3, 3, 19)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, T1SZ_2_0_PASIZE, 3, 16)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, NSCFG0_TG0, 1, 14)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, SH0, 2, 12)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, ORGN0, 2, 10)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, IRGN0, 2, 8)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, SL0_1_EPD0, 1, 7)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, SL0_0, 1, 6)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, PD1_T0SZ_5, 1, 5)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, S_PD0_T0SZ_4, 1, 4)
    DEP_FIELD(SMMU_CB11_TCR_LPAE, T0SZ_3_0, 4, 0)
DEP_REG32(SMMU_CB11_CONTEXTIDR, 0x1b034)
    DEP_FIELD(SMMU_CB11_CONTEXTIDR, PROCID, 24, 8)
    DEP_FIELD(SMMU_CB11_CONTEXTIDR, ASID, 8, 0)
DEP_REG32(SMMU_CB11_PRRR_MAIR0, 0x1b038)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, NOS7, 1, 31)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, NOS6, 1, 30)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, NOS5, 1, 29)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, NOS4, 1, 28)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, NOS3, 1, 27)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, NOS2, 1, 26)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, NOS1, 1, 25)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, NOS0, 1, 24)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, NS1, 1, 19)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, NS0, 1, 18)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, DS1, 1, 17)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, DS0, 1, 16)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, TR7, 2, 14)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, TR6, 2, 12)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, TR5, 2, 10)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, TR4, 2, 8)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, TR3, 2, 6)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, TR2, 2, 4)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, TR1, 2, 2)
    DEP_FIELD(SMMU_CB11_PRRR_MAIR0, TR0, 2, 0)
DEP_REG32(SMMU_CB11_NMRR_MAIR1, 0x1b03c)
    DEP_FIELD(SMMU_CB11_NMRR_MAIR1, OR7, 2, 30)
    DEP_FIELD(SMMU_CB11_NMRR_MAIR1, OR6, 2, 28)
    DEP_FIELD(SMMU_CB11_NMRR_MAIR1, OR5, 2, 26)
    DEP_FIELD(SMMU_CB11_NMRR_MAIR1, OR4, 2, 24)
    DEP_FIELD(SMMU_CB11_NMRR_MAIR1, OR3, 2, 22)
    DEP_FIELD(SMMU_CB11_NMRR_MAIR1, OR2, 2, 20)
    DEP_FIELD(SMMU_CB11_NMRR_MAIR1, OR1, 2, 18)
    DEP_FIELD(SMMU_CB11_NMRR_MAIR1, OR0, 2, 16)
    DEP_FIELD(SMMU_CB11_NMRR_MAIR1, IR7, 2, 14)
    DEP_FIELD(SMMU_CB11_NMRR_MAIR1, IR6, 2, 12)
    DEP_FIELD(SMMU_CB11_NMRR_MAIR1, IR5, 2, 10)
    DEP_FIELD(SMMU_CB11_NMRR_MAIR1, IR4, 2, 8)
    DEP_FIELD(SMMU_CB11_NMRR_MAIR1, IR3, 2, 6)
    DEP_FIELD(SMMU_CB11_NMRR_MAIR1, IR2, 2, 4)
    DEP_FIELD(SMMU_CB11_NMRR_MAIR1, IR1, 2, 2)
    DEP_FIELD(SMMU_CB11_NMRR_MAIR1, IR0, 2, 0)
DEP_REG32(SMMU_CB11_FSR, 0x1b058)
    DEP_FIELD(SMMU_CB11_FSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_CB11_FSR, SS, 1, 30)
    DEP_FIELD(SMMU_CB11_FSR, FORMAT, 2, 9)
    DEP_FIELD(SMMU_CB11_FSR, UUT, 1, 8)
    DEP_FIELD(SMMU_CB11_FSR, ASF, 1, 7)
    DEP_FIELD(SMMU_CB11_FSR, TLBLKF, 1, 6)
    DEP_FIELD(SMMU_CB11_FSR, TLBMCF, 1, 5)
    DEP_FIELD(SMMU_CB11_FSR, EF, 1, 4)
    DEP_FIELD(SMMU_CB11_FSR, PF, 1, 3)
    DEP_FIELD(SMMU_CB11_FSR, AFF, 1, 2)
    DEP_FIELD(SMMU_CB11_FSR, TF, 1, 1)
DEP_REG32(SMMU_CB11_FSRRESTORE, 0x1b05c)
DEP_REG32(SMMU_CB11_FAR_LOW, 0x1b060)
DEP_REG32(SMMU_CB11_FAR_HIGH, 0x1b064)
    DEP_FIELD(SMMU_CB11_FAR_HIGH, BITS, 17, 0)
DEP_REG32(SMMU_CB11_FSYNR0, 0x1b068)
    DEP_FIELD(SMMU_CB11_FSYNR0, S1CBNDX, 4, 16)
    DEP_FIELD(SMMU_CB11_FSYNR0, AFR, 1, 11)
    DEP_FIELD(SMMU_CB11_FSYNR0, PTWF, 1, 10)
    DEP_FIELD(SMMU_CB11_FSYNR0, ATOF, 1, 9)
    DEP_FIELD(SMMU_CB11_FSYNR0, NSATTR, 1, 8)
    DEP_FIELD(SMMU_CB11_FSYNR0, IND, 1, 6)
    DEP_FIELD(SMMU_CB11_FSYNR0, PNU, 1, 5)
    DEP_FIELD(SMMU_CB11_FSYNR0, WNR, 1, 4)
    DEP_FIELD(SMMU_CB11_FSYNR0, PLVL, 2, 0)
DEP_REG32(SMMU_CB11_IPAFAR_LOW, 0x1b070)
    DEP_FIELD(SMMU_CB11_IPAFAR_LOW, IPAFAR_L, 20, 12)
    DEP_FIELD(SMMU_CB11_IPAFAR_LOW, FAR_RO, 12, 0)
DEP_REG32(SMMU_CB11_IPAFAR_HIGH, 0x1b074)
    DEP_FIELD(SMMU_CB11_IPAFAR_HIGH, BITS, 16, 0)
DEP_REG32(SMMU_CB11_TLBIVA_LOW, 0x1b600)
DEP_REG32(SMMU_CB11_TLBIVA_HIGH, 0x1b604)
    DEP_FIELD(SMMU_CB11_TLBIVA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB11_TLBIVA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB11_TLBIVAA_LOW, 0x1b608)
DEP_REG32(SMMU_CB11_TLBIVAA_HIGH, 0x1b60c)
    DEP_FIELD(SMMU_CB11_TLBIVAA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB11_TLBIVAA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB11_TLBIASID, 0x1b610)
    DEP_FIELD(SMMU_CB11_TLBIASID, ASID, 16, 0)
DEP_REG32(SMMU_CB11_TLBIALL, 0x1b618)
DEP_REG32(SMMU_CB11_TLBIVAL_LOW, 0x1b620)
DEP_REG32(SMMU_CB11_TLBIVAL_HIGH, 0x1b624)
    DEP_FIELD(SMMU_CB11_TLBIVAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB11_TLBIVAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB11_TLBIVAAL_LOW, 0x1b628)
DEP_REG32(SMMU_CB11_TLBIVAAL_HIGH, 0x1b62c)
    DEP_FIELD(SMMU_CB11_TLBIVAAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB11_TLBIVAAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB11_TLBIIPAS2_LOW, 0x1b630)
DEP_REG32(SMMU_CB11_TLBIIPAS2_HIGH, 0x1b634)
    DEP_FIELD(SMMU_CB11_TLBIIPAS2_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB11_TLBIIPAS2L_LOW, 0x1b638)
DEP_REG32(SMMU_CB11_TLBIIPAS2L_HIGH, 0x1b63c)
    DEP_FIELD(SMMU_CB11_TLBIIPAS2L_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB11_TLBSYNC, 0x1b7f0)
DEP_REG32(SMMU_CB11_TLBSTATUS, 0x1b7f4)
    DEP_FIELD(SMMU_CB11_TLBSTATUS, SACTIVE, 1, 0)
DEP_REG32(SMMU_CB11_PMEVCNTR0, 0x1be00)
DEP_REG32(SMMU_CB11_PMEVCNTR1, 0x1be04)
DEP_REG32(SMMU_CB11_PMEVCNTR2, 0x1be08)
DEP_REG32(SMMU_CB11_PMEVCNTR3, 0x1be0c)
DEP_REG32(SMMU_CB11_PMEVTYPER0, 0x1be80)
    DEP_FIELD(SMMU_CB11_PMEVTYPER0, P, 1, 31)
    DEP_FIELD(SMMU_CB11_PMEVTYPER0, U, 1, 30)
    DEP_FIELD(SMMU_CB11_PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(SMMU_CB11_PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(SMMU_CB11_PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(SMMU_CB11_PMEVTYPER1, 0x1be84)
    DEP_FIELD(SMMU_CB11_PMEVTYPER1, P, 1, 31)
    DEP_FIELD(SMMU_CB11_PMEVTYPER1, U, 1, 30)
    DEP_FIELD(SMMU_CB11_PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(SMMU_CB11_PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(SMMU_CB11_PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(SMMU_CB11_PMEVTYPER2, 0x1be88)
    DEP_FIELD(SMMU_CB11_PMEVTYPER2, P, 1, 31)
    DEP_FIELD(SMMU_CB11_PMEVTYPER2, U, 1, 30)
    DEP_FIELD(SMMU_CB11_PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(SMMU_CB11_PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(SMMU_CB11_PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(SMMU_CB11_PMEVTYPER3, 0x1be8c)
    DEP_FIELD(SMMU_CB11_PMEVTYPER3, P, 1, 31)
    DEP_FIELD(SMMU_CB11_PMEVTYPER3, U, 1, 30)
    DEP_FIELD(SMMU_CB11_PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(SMMU_CB11_PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(SMMU_CB11_PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(SMMU_CB11_PMCFGR, 0x1bf00)
    DEP_FIELD(SMMU_CB11_PMCFGR, NCG, 8, 24)
    DEP_FIELD(SMMU_CB11_PMCFGR, UEN, 1, 19)
    DEP_FIELD(SMMU_CB11_PMCFGR, EX, 1, 16)
    DEP_FIELD(SMMU_CB11_PMCFGR, CCD, 1, 15)
    DEP_FIELD(SMMU_CB11_PMCFGR, CC, 1, 14)
    DEP_FIELD(SMMU_CB11_PMCFGR, SIZE, 6, 8)
    DEP_FIELD(SMMU_CB11_PMCFGR, N, 8, 0)
DEP_REG32(SMMU_CB11_PMCR, 0x1bf04)
    DEP_FIELD(SMMU_CB11_PMCR, IMP, 8, 24)
    DEP_FIELD(SMMU_CB11_PMCR, X, 1, 4)
    DEP_FIELD(SMMU_CB11_PMCR, P, 1, 1)
    DEP_FIELD(SMMU_CB11_PMCR, E, 1, 0)
DEP_REG32(SMMU_CB11_PMCEID, 0x1bf20)
    DEP_FIELD(SMMU_CB11_PMCEID, EVENT0X12, 1, 17)
    DEP_FIELD(SMMU_CB11_PMCEID, EVENT0X11, 1, 16)
    DEP_FIELD(SMMU_CB11_PMCEID, EVENT0X10, 1, 15)
    DEP_FIELD(SMMU_CB11_PMCEID, EVENT0X0A, 1, 9)
    DEP_FIELD(SMMU_CB11_PMCEID, EVENT0X09, 1, 8)
    DEP_FIELD(SMMU_CB11_PMCEID, EVENT0X08, 1, 7)
    DEP_FIELD(SMMU_CB11_PMCEID, EVENT0X01, 1, 1)
    DEP_FIELD(SMMU_CB11_PMCEID, EVENT0X00, 1, 0)
DEP_REG32(SMMU_CB11_PMCNTENSE, 0x1bf40)
    DEP_FIELD(SMMU_CB11_PMCNTENSE, P3, 1, 3)
    DEP_FIELD(SMMU_CB11_PMCNTENSE, P2, 1, 2)
    DEP_FIELD(SMMU_CB11_PMCNTENSE, P1, 1, 1)
    DEP_FIELD(SMMU_CB11_PMCNTENSE, P0, 1, 0)
DEP_REG32(SMMU_CB11_PMCNTENCLR, 0x1bf44)
    DEP_FIELD(SMMU_CB11_PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB11_PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB11_PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB11_PMCNTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB11_PMCNTENSET, 0x1bf48)
    DEP_FIELD(SMMU_CB11_PMCNTENSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB11_PMCNTENSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB11_PMCNTENSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB11_PMCNTENSET, P0, 1, 0)
DEP_REG32(SMMU_CB11_PMINTENCLR, 0x1bf4c)
    DEP_FIELD(SMMU_CB11_PMINTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB11_PMINTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB11_PMINTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB11_PMINTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB11_PMOVSCLR, 0x1bf50)
    DEP_FIELD(SMMU_CB11_PMOVSCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB11_PMOVSCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB11_PMOVSCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB11_PMOVSCLR, P0, 1, 0)
DEP_REG32(SMMU_CB11_PMOVSSET, 0x1bf58)
    DEP_FIELD(SMMU_CB11_PMOVSSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB11_PMOVSSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB11_PMOVSSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB11_PMOVSSET, P0, 1, 0)
DEP_REG32(SMMU_CB11_PMAUTHSTATUS, 0x1bfb8)
    DEP_FIELD(SMMU_CB11_PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(SMMU_CB11_PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(SMMU_CB11_PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(SMMU_CB11_PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(SMMU_CB11_PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(SMMU_CB11_PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(SMMU_CB11_PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(SMMU_CB11_PMAUTHSTATUS, NSE, 1, 0)
DEP_REG32(SMMU_CB12_SCTLR, 0x1c000)
    DEP_FIELD(SMMU_CB12_SCTLR, NSCFG, 2, 28)
    DEP_FIELD(SMMU_CB12_SCTLR, WACFG, 2, 26)
    DEP_FIELD(SMMU_CB12_SCTLR, RACFG, 2, 24)
    DEP_FIELD(SMMU_CB12_SCTLR, SHCFG, 2, 22)
    DEP_FIELD(SMMU_CB12_SCTLR, FB, 1, 21)
    DEP_FIELD(SMMU_CB12_SCTLR, MTCFG, 1, 20)
    DEP_FIELD(SMMU_CB12_SCTLR, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_CB12_SCTLR, TRANSIENTCFG, 2, 14)
    DEP_FIELD(SMMU_CB12_SCTLR, PTW, 1, 13)
    DEP_FIELD(SMMU_CB12_SCTLR, ASIDPNE, 1, 12)
    DEP_FIELD(SMMU_CB12_SCTLR, UWXN, 1, 10)
    DEP_FIELD(SMMU_CB12_SCTLR, WXN, 1, 9)
    DEP_FIELD(SMMU_CB12_SCTLR, HUPCF, 1, 8)
    DEP_FIELD(SMMU_CB12_SCTLR, CFCFG, 1, 7)
    DEP_FIELD(SMMU_CB12_SCTLR, CFIE, 1, 6)
    DEP_FIELD(SMMU_CB12_SCTLR, CFRE, 1, 5)
    DEP_FIELD(SMMU_CB12_SCTLR, E, 1, 4)
    DEP_FIELD(SMMU_CB12_SCTLR, AFFD, 1, 3)
    DEP_FIELD(SMMU_CB12_SCTLR, AFE, 1, 2)
    DEP_FIELD(SMMU_CB12_SCTLR, TRE, 1, 1)
    DEP_FIELD(SMMU_CB12_SCTLR, M, 1, 0)
DEP_REG32(SMMU_CB12_ACTLR, 0x1c004)
    DEP_FIELD(SMMU_CB12_ACTLR, CPRE, 1, 1)
    DEP_FIELD(SMMU_CB12_ACTLR, CMTLB, 1, 0)
DEP_REG32(SMMU_CB12_RESUME, 0x1c008)
    DEP_FIELD(SMMU_CB12_RESUME, TNR, 1, 0)
DEP_REG32(SMMU_CB12_TCR2, 0x1c010)
    DEP_FIELD(SMMU_CB12_TCR2, NSCFG1, 1, 30)
    DEP_FIELD(SMMU_CB12_TCR2, SEP, 3, 15)
    DEP_FIELD(SMMU_CB12_TCR2, NSCFG0, 1, 14)
    DEP_FIELD(SMMU_CB12_TCR2, TBI1, 1, 6)
    DEP_FIELD(SMMU_CB12_TCR2, TBI0, 1, 5)
    DEP_FIELD(SMMU_CB12_TCR2, AS, 1, 4)
    DEP_FIELD(SMMU_CB12_TCR2, PASIZE, 3, 0)
DEP_REG32(SMMU_CB12_TTBR0_LOW, 0x1c020)
    DEP_FIELD(SMMU_CB12_TTBR0_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB12_TTBR0_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB12_TTBR0_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB12_TTBR0_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB12_TTBR0_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB12_TTBR0_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB12_TTBR0_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB12_TTBR0_HIGH, 0x1c024)
    DEP_FIELD(SMMU_CB12_TTBR0_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB12_TTBR0_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB12_TTBR1_LOW, 0x1c028)
    DEP_FIELD(SMMU_CB12_TTBR1_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB12_TTBR1_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB12_TTBR1_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB12_TTBR1_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB12_TTBR1_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB12_TTBR1_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB12_TTBR1_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB12_TTBR1_HIGH, 0x1c02c)
    DEP_FIELD(SMMU_CB12_TTBR1_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB12_TTBR1_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB12_TCR_LPAE, 0x1c030)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, EAE, 1, 31)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, NSCFG1_TG1, 1, 30)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, SH1, 2, 28)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, ORGN1, 2, 26)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, IRGN1, 2, 24)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, EPD1, 1, 23)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, A1, 1, 22)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, T1SZ_5_3, 3, 19)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, T1SZ_2_0_PASIZE, 3, 16)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, NSCFG0_TG0, 1, 14)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, SH0, 2, 12)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, ORGN0, 2, 10)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, IRGN0, 2, 8)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, SL0_1_EPD0, 1, 7)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, SL0_0, 1, 6)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, PD1_T0SZ_5, 1, 5)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, S_PD0_T0SZ_4, 1, 4)
    DEP_FIELD(SMMU_CB12_TCR_LPAE, T0SZ_3_0, 4, 0)
DEP_REG32(SMMU_CB12_CONTEXTIDR, 0x1c034)
    DEP_FIELD(SMMU_CB12_CONTEXTIDR, PROCID, 24, 8)
    DEP_FIELD(SMMU_CB12_CONTEXTIDR, ASID, 8, 0)
DEP_REG32(SMMU_CB12_PRRR_MAIR0, 0x1c038)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, NOS7, 1, 31)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, NOS6, 1, 30)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, NOS5, 1, 29)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, NOS4, 1, 28)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, NOS3, 1, 27)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, NOS2, 1, 26)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, NOS1, 1, 25)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, NOS0, 1, 24)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, NS1, 1, 19)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, NS0, 1, 18)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, DS1, 1, 17)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, DS0, 1, 16)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, TR7, 2, 14)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, TR6, 2, 12)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, TR5, 2, 10)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, TR4, 2, 8)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, TR3, 2, 6)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, TR2, 2, 4)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, TR1, 2, 2)
    DEP_FIELD(SMMU_CB12_PRRR_MAIR0, TR0, 2, 0)
DEP_REG32(SMMU_CB12_NMRR_MAIR1, 0x1c03c)
    DEP_FIELD(SMMU_CB12_NMRR_MAIR1, OR7, 2, 30)
    DEP_FIELD(SMMU_CB12_NMRR_MAIR1, OR6, 2, 28)
    DEP_FIELD(SMMU_CB12_NMRR_MAIR1, OR5, 2, 26)
    DEP_FIELD(SMMU_CB12_NMRR_MAIR1, OR4, 2, 24)
    DEP_FIELD(SMMU_CB12_NMRR_MAIR1, OR3, 2, 22)
    DEP_FIELD(SMMU_CB12_NMRR_MAIR1, OR2, 2, 20)
    DEP_FIELD(SMMU_CB12_NMRR_MAIR1, OR1, 2, 18)
    DEP_FIELD(SMMU_CB12_NMRR_MAIR1, OR0, 2, 16)
    DEP_FIELD(SMMU_CB12_NMRR_MAIR1, IR7, 2, 14)
    DEP_FIELD(SMMU_CB12_NMRR_MAIR1, IR6, 2, 12)
    DEP_FIELD(SMMU_CB12_NMRR_MAIR1, IR5, 2, 10)
    DEP_FIELD(SMMU_CB12_NMRR_MAIR1, IR4, 2, 8)
    DEP_FIELD(SMMU_CB12_NMRR_MAIR1, IR3, 2, 6)
    DEP_FIELD(SMMU_CB12_NMRR_MAIR1, IR2, 2, 4)
    DEP_FIELD(SMMU_CB12_NMRR_MAIR1, IR1, 2, 2)
    DEP_FIELD(SMMU_CB12_NMRR_MAIR1, IR0, 2, 0)
DEP_REG32(SMMU_CB12_FSR, 0x1c058)
    DEP_FIELD(SMMU_CB12_FSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_CB12_FSR, SS, 1, 30)
    DEP_FIELD(SMMU_CB12_FSR, FORMAT, 2, 9)
    DEP_FIELD(SMMU_CB12_FSR, UUT, 1, 8)
    DEP_FIELD(SMMU_CB12_FSR, ASF, 1, 7)
    DEP_FIELD(SMMU_CB12_FSR, TLBLKF, 1, 6)
    DEP_FIELD(SMMU_CB12_FSR, TLBMCF, 1, 5)
    DEP_FIELD(SMMU_CB12_FSR, EF, 1, 4)
    DEP_FIELD(SMMU_CB12_FSR, PF, 1, 3)
    DEP_FIELD(SMMU_CB12_FSR, AFF, 1, 2)
    DEP_FIELD(SMMU_CB12_FSR, TF, 1, 1)
DEP_REG32(SMMU_CB12_FSRRESTORE, 0x1c05c)
DEP_REG32(SMMU_CB12_FAR_LOW, 0x1c060)
DEP_REG32(SMMU_CB12_FAR_HIGH, 0x1c064)
    DEP_FIELD(SMMU_CB12_FAR_HIGH, BITS, 17, 0)
DEP_REG32(SMMU_CB12_FSYNR0, 0x1c068)
    DEP_FIELD(SMMU_CB12_FSYNR0, S1CBNDX, 4, 16)
    DEP_FIELD(SMMU_CB12_FSYNR0, AFR, 1, 11)
    DEP_FIELD(SMMU_CB12_FSYNR0, PTWF, 1, 10)
    DEP_FIELD(SMMU_CB12_FSYNR0, ATOF, 1, 9)
    DEP_FIELD(SMMU_CB12_FSYNR0, NSATTR, 1, 8)
    DEP_FIELD(SMMU_CB12_FSYNR0, IND, 1, 6)
    DEP_FIELD(SMMU_CB12_FSYNR0, PNU, 1, 5)
    DEP_FIELD(SMMU_CB12_FSYNR0, WNR, 1, 4)
    DEP_FIELD(SMMU_CB12_FSYNR0, PLVL, 2, 0)
DEP_REG32(SMMU_CB12_IPAFAR_LOW, 0x1c070)
    DEP_FIELD(SMMU_CB12_IPAFAR_LOW, IPAFAR_L, 20, 12)
    DEP_FIELD(SMMU_CB12_IPAFAR_LOW, FAR_RO, 12, 0)
DEP_REG32(SMMU_CB12_IPAFAR_HIGH, 0x1c074)
    DEP_FIELD(SMMU_CB12_IPAFAR_HIGH, BITS, 16, 0)
DEP_REG32(SMMU_CB12_TLBIVA_LOW, 0x1c600)
DEP_REG32(SMMU_CB12_TLBIVA_HIGH, 0x1c604)
    DEP_FIELD(SMMU_CB12_TLBIVA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB12_TLBIVA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB12_TLBIVAA_LOW, 0x1c608)
DEP_REG32(SMMU_CB12_TLBIVAA_HIGH, 0x1c60c)
    DEP_FIELD(SMMU_CB12_TLBIVAA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB12_TLBIVAA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB12_TLBIASID, 0x1c610)
    DEP_FIELD(SMMU_CB12_TLBIASID, ASID, 16, 0)
DEP_REG32(SMMU_CB12_TLBIALL, 0x1c618)
DEP_REG32(SMMU_CB12_TLBIVAL_LOW, 0x1c620)
DEP_REG32(SMMU_CB12_TLBIVAL_HIGH, 0x1c624)
    DEP_FIELD(SMMU_CB12_TLBIVAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB12_TLBIVAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB12_TLBIVAAL_LOW, 0x1c628)
DEP_REG32(SMMU_CB12_TLBIVAAL_HIGH, 0x1c62c)
    DEP_FIELD(SMMU_CB12_TLBIVAAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB12_TLBIVAAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB12_TLBIIPAS2_LOW, 0x1c630)
DEP_REG32(SMMU_CB12_TLBIIPAS2_HIGH, 0x1c634)
    DEP_FIELD(SMMU_CB12_TLBIIPAS2_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB12_TLBIIPAS2L_LOW, 0x1c638)
DEP_REG32(SMMU_CB12_TLBIIPAS2L_HIGH, 0x1c63c)
    DEP_FIELD(SMMU_CB12_TLBIIPAS2L_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB12_TLBSYNC, 0x1c7f0)
DEP_REG32(SMMU_CB12_TLBSTATUS, 0x1c7f4)
    DEP_FIELD(SMMU_CB12_TLBSTATUS, SACTIVE, 1, 0)
DEP_REG32(SMMU_CB12_PMEVCNTR0, 0x1ce00)
DEP_REG32(SMMU_CB12_PMEVCNTR1, 0x1ce04)
DEP_REG32(SMMU_CB12_PMEVCNTR2, 0x1ce08)
DEP_REG32(SMMU_CB12_PMEVCNTR3, 0x1ce0c)
DEP_REG32(SMMU_CB12_PMEVTYPER0, 0x1ce80)
    DEP_FIELD(SMMU_CB12_PMEVTYPER0, P, 1, 31)
    DEP_FIELD(SMMU_CB12_PMEVTYPER0, U, 1, 30)
    DEP_FIELD(SMMU_CB12_PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(SMMU_CB12_PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(SMMU_CB12_PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(SMMU_CB12_PMEVTYPER1, 0x1ce84)
    DEP_FIELD(SMMU_CB12_PMEVTYPER1, P, 1, 31)
    DEP_FIELD(SMMU_CB12_PMEVTYPER1, U, 1, 30)
    DEP_FIELD(SMMU_CB12_PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(SMMU_CB12_PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(SMMU_CB12_PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(SMMU_CB12_PMEVTYPER2, 0x1ce88)
    DEP_FIELD(SMMU_CB12_PMEVTYPER2, P, 1, 31)
    DEP_FIELD(SMMU_CB12_PMEVTYPER2, U, 1, 30)
    DEP_FIELD(SMMU_CB12_PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(SMMU_CB12_PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(SMMU_CB12_PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(SMMU_CB12_PMEVTYPER3, 0x1ce8c)
    DEP_FIELD(SMMU_CB12_PMEVTYPER3, P, 1, 31)
    DEP_FIELD(SMMU_CB12_PMEVTYPER3, U, 1, 30)
    DEP_FIELD(SMMU_CB12_PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(SMMU_CB12_PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(SMMU_CB12_PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(SMMU_CB12_PMCFGR, 0x1cf00)
    DEP_FIELD(SMMU_CB12_PMCFGR, NCG, 8, 24)
    DEP_FIELD(SMMU_CB12_PMCFGR, UEN, 1, 19)
    DEP_FIELD(SMMU_CB12_PMCFGR, EX, 1, 16)
    DEP_FIELD(SMMU_CB12_PMCFGR, CCD, 1, 15)
    DEP_FIELD(SMMU_CB12_PMCFGR, CC, 1, 14)
    DEP_FIELD(SMMU_CB12_PMCFGR, SIZE, 6, 8)
    DEP_FIELD(SMMU_CB12_PMCFGR, N, 8, 0)
DEP_REG32(SMMU_CB12_PMCR, 0x1cf04)
    DEP_FIELD(SMMU_CB12_PMCR, IMP, 8, 24)
    DEP_FIELD(SMMU_CB12_PMCR, X, 1, 4)
    DEP_FIELD(SMMU_CB12_PMCR, P, 1, 1)
    DEP_FIELD(SMMU_CB12_PMCR, E, 1, 0)
DEP_REG32(SMMU_CB12_PMCEID, 0x1cf20)
    DEP_FIELD(SMMU_CB12_PMCEID, EVENT0X12, 1, 17)
    DEP_FIELD(SMMU_CB12_PMCEID, EVENT0X11, 1, 16)
    DEP_FIELD(SMMU_CB12_PMCEID, EVENT0X10, 1, 15)
    DEP_FIELD(SMMU_CB12_PMCEID, EVENT0X0A, 1, 9)
    DEP_FIELD(SMMU_CB12_PMCEID, EVENT0X09, 1, 8)
    DEP_FIELD(SMMU_CB12_PMCEID, EVENT0X08, 1, 7)
    DEP_FIELD(SMMU_CB12_PMCEID, EVENT0X01, 1, 1)
    DEP_FIELD(SMMU_CB12_PMCEID, EVENT0X00, 1, 0)
DEP_REG32(SMMU_CB12_PMCNTENSE, 0x1cf40)
    DEP_FIELD(SMMU_CB12_PMCNTENSE, P3, 1, 3)
    DEP_FIELD(SMMU_CB12_PMCNTENSE, P2, 1, 2)
    DEP_FIELD(SMMU_CB12_PMCNTENSE, P1, 1, 1)
    DEP_FIELD(SMMU_CB12_PMCNTENSE, P0, 1, 0)
DEP_REG32(SMMU_CB12_PMCNTENCLR, 0x1cf44)
    DEP_FIELD(SMMU_CB12_PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB12_PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB12_PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB12_PMCNTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB12_PMCNTENSET, 0x1cf48)
    DEP_FIELD(SMMU_CB12_PMCNTENSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB12_PMCNTENSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB12_PMCNTENSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB12_PMCNTENSET, P0, 1, 0)
DEP_REG32(SMMU_CB12_PMINTENCLR, 0x1cf4c)
    DEP_FIELD(SMMU_CB12_PMINTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB12_PMINTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB12_PMINTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB12_PMINTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB12_PMOVSCLR, 0x1cf50)
    DEP_FIELD(SMMU_CB12_PMOVSCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB12_PMOVSCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB12_PMOVSCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB12_PMOVSCLR, P0, 1, 0)
DEP_REG32(SMMU_CB12_PMOVSSET, 0x1cf58)
    DEP_FIELD(SMMU_CB12_PMOVSSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB12_PMOVSSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB12_PMOVSSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB12_PMOVSSET, P0, 1, 0)
DEP_REG32(SMMU_CB12_PMAUTHSTATUS, 0x1cfb8)
    DEP_FIELD(SMMU_CB12_PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(SMMU_CB12_PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(SMMU_CB12_PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(SMMU_CB12_PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(SMMU_CB12_PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(SMMU_CB12_PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(SMMU_CB12_PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(SMMU_CB12_PMAUTHSTATUS, NSE, 1, 0)
DEP_REG32(SMMU_CB13_SCTLR, 0x1d000)
    DEP_FIELD(SMMU_CB13_SCTLR, NSCFG, 2, 28)
    DEP_FIELD(SMMU_CB13_SCTLR, WACFG, 2, 26)
    DEP_FIELD(SMMU_CB13_SCTLR, RACFG, 2, 24)
    DEP_FIELD(SMMU_CB13_SCTLR, SHCFG, 2, 22)
    DEP_FIELD(SMMU_CB13_SCTLR, FB, 1, 21)
    DEP_FIELD(SMMU_CB13_SCTLR, MTCFG, 1, 20)
    DEP_FIELD(SMMU_CB13_SCTLR, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_CB13_SCTLR, TRANSIENTCFG, 2, 14)
    DEP_FIELD(SMMU_CB13_SCTLR, PTW, 1, 13)
    DEP_FIELD(SMMU_CB13_SCTLR, ASIDPNE, 1, 12)
    DEP_FIELD(SMMU_CB13_SCTLR, UWXN, 1, 10)
    DEP_FIELD(SMMU_CB13_SCTLR, WXN, 1, 9)
    DEP_FIELD(SMMU_CB13_SCTLR, HUPCF, 1, 8)
    DEP_FIELD(SMMU_CB13_SCTLR, CFCFG, 1, 7)
    DEP_FIELD(SMMU_CB13_SCTLR, CFIE, 1, 6)
    DEP_FIELD(SMMU_CB13_SCTLR, CFRE, 1, 5)
    DEP_FIELD(SMMU_CB13_SCTLR, E, 1, 4)
    DEP_FIELD(SMMU_CB13_SCTLR, AFFD, 1, 3)
    DEP_FIELD(SMMU_CB13_SCTLR, AFE, 1, 2)
    DEP_FIELD(SMMU_CB13_SCTLR, TRE, 1, 1)
    DEP_FIELD(SMMU_CB13_SCTLR, M, 1, 0)
DEP_REG32(SMMU_CB13_ACTLR, 0x1d004)
    DEP_FIELD(SMMU_CB13_ACTLR, CPRE, 1, 1)
    DEP_FIELD(SMMU_CB13_ACTLR, CMTLB, 1, 0)
DEP_REG32(SMMU_CB13_RESUME, 0x1d008)
    DEP_FIELD(SMMU_CB13_RESUME, TNR, 1, 0)
DEP_REG32(SMMU_CB13_TCR2, 0x1d010)
    DEP_FIELD(SMMU_CB13_TCR2, NSCFG1, 1, 30)
    DEP_FIELD(SMMU_CB13_TCR2, SEP, 3, 15)
    DEP_FIELD(SMMU_CB13_TCR2, NSCFG0, 1, 14)
    DEP_FIELD(SMMU_CB13_TCR2, TBI1, 1, 6)
    DEP_FIELD(SMMU_CB13_TCR2, TBI0, 1, 5)
    DEP_FIELD(SMMU_CB13_TCR2, AS, 1, 4)
    DEP_FIELD(SMMU_CB13_TCR2, PASIZE, 3, 0)
DEP_REG32(SMMU_CB13_TTBR0_LOW, 0x1d020)
    DEP_FIELD(SMMU_CB13_TTBR0_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB13_TTBR0_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB13_TTBR0_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB13_TTBR0_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB13_TTBR0_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB13_TTBR0_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB13_TTBR0_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB13_TTBR0_HIGH, 0x1d024)
    DEP_FIELD(SMMU_CB13_TTBR0_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB13_TTBR0_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB13_TTBR1_LOW, 0x1d028)
    DEP_FIELD(SMMU_CB13_TTBR1_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB13_TTBR1_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB13_TTBR1_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB13_TTBR1_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB13_TTBR1_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB13_TTBR1_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB13_TTBR1_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB13_TTBR1_HIGH, 0x1d02c)
    DEP_FIELD(SMMU_CB13_TTBR1_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB13_TTBR1_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB13_TCR_LPAE, 0x1d030)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, EAE, 1, 31)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, NSCFG1_TG1, 1, 30)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, SH1, 2, 28)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, ORGN1, 2, 26)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, IRGN1, 2, 24)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, EPD1, 1, 23)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, A1, 1, 22)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, T1SZ_5_3, 3, 19)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, T1SZ_2_0_PASIZE, 3, 16)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, NSCFG0_TG0, 1, 14)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, SH0, 2, 12)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, ORGN0, 2, 10)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, IRGN0, 2, 8)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, SL0_1_EPD0, 1, 7)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, SL0_0, 1, 6)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, PD1_T0SZ_5, 1, 5)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, S_PD0_T0SZ_4, 1, 4)
    DEP_FIELD(SMMU_CB13_TCR_LPAE, T0SZ_3_0, 4, 0)
DEP_REG32(SMMU_CB13_CONTEXTIDR, 0x1d034)
    DEP_FIELD(SMMU_CB13_CONTEXTIDR, PROCID, 24, 8)
    DEP_FIELD(SMMU_CB13_CONTEXTIDR, ASID, 8, 0)
DEP_REG32(SMMU_CB13_PRRR_MAIR0, 0x1d038)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, NOS7, 1, 31)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, NOS6, 1, 30)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, NOS5, 1, 29)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, NOS4, 1, 28)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, NOS3, 1, 27)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, NOS2, 1, 26)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, NOS1, 1, 25)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, NOS0, 1, 24)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, NS1, 1, 19)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, NS0, 1, 18)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, DS1, 1, 17)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, DS0, 1, 16)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, TR7, 2, 14)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, TR6, 2, 12)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, TR5, 2, 10)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, TR4, 2, 8)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, TR3, 2, 6)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, TR2, 2, 4)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, TR1, 2, 2)
    DEP_FIELD(SMMU_CB13_PRRR_MAIR0, TR0, 2, 0)
DEP_REG32(SMMU_CB13_NMRR_MAIR1, 0x1d03c)
    DEP_FIELD(SMMU_CB13_NMRR_MAIR1, OR7, 2, 30)
    DEP_FIELD(SMMU_CB13_NMRR_MAIR1, OR6, 2, 28)
    DEP_FIELD(SMMU_CB13_NMRR_MAIR1, OR5, 2, 26)
    DEP_FIELD(SMMU_CB13_NMRR_MAIR1, OR4, 2, 24)
    DEP_FIELD(SMMU_CB13_NMRR_MAIR1, OR3, 2, 22)
    DEP_FIELD(SMMU_CB13_NMRR_MAIR1, OR2, 2, 20)
    DEP_FIELD(SMMU_CB13_NMRR_MAIR1, OR1, 2, 18)
    DEP_FIELD(SMMU_CB13_NMRR_MAIR1, OR0, 2, 16)
    DEP_FIELD(SMMU_CB13_NMRR_MAIR1, IR7, 2, 14)
    DEP_FIELD(SMMU_CB13_NMRR_MAIR1, IR6, 2, 12)
    DEP_FIELD(SMMU_CB13_NMRR_MAIR1, IR5, 2, 10)
    DEP_FIELD(SMMU_CB13_NMRR_MAIR1, IR4, 2, 8)
    DEP_FIELD(SMMU_CB13_NMRR_MAIR1, IR3, 2, 6)
    DEP_FIELD(SMMU_CB13_NMRR_MAIR1, IR2, 2, 4)
    DEP_FIELD(SMMU_CB13_NMRR_MAIR1, IR1, 2, 2)
    DEP_FIELD(SMMU_CB13_NMRR_MAIR1, IR0, 2, 0)
DEP_REG32(SMMU_CB13_FSR, 0x1d058)
    DEP_FIELD(SMMU_CB13_FSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_CB13_FSR, SS, 1, 30)
    DEP_FIELD(SMMU_CB13_FSR, FORMAT, 2, 9)
    DEP_FIELD(SMMU_CB13_FSR, UUT, 1, 8)
    DEP_FIELD(SMMU_CB13_FSR, ASF, 1, 7)
    DEP_FIELD(SMMU_CB13_FSR, TLBLKF, 1, 6)
    DEP_FIELD(SMMU_CB13_FSR, TLBMCF, 1, 5)
    DEP_FIELD(SMMU_CB13_FSR, EF, 1, 4)
    DEP_FIELD(SMMU_CB13_FSR, PF, 1, 3)
    DEP_FIELD(SMMU_CB13_FSR, AFF, 1, 2)
    DEP_FIELD(SMMU_CB13_FSR, TF, 1, 1)
DEP_REG32(SMMU_CB13_FSRRESTORE, 0x1d05c)
DEP_REG32(SMMU_CB13_FAR_LOW, 0x1d060)
DEP_REG32(SMMU_CB13_FAR_HIGH, 0x1d064)
    DEP_FIELD(SMMU_CB13_FAR_HIGH, BITS, 17, 0)
DEP_REG32(SMMU_CB13_FSYNR0, 0x1d068)
    DEP_FIELD(SMMU_CB13_FSYNR0, S1CBNDX, 4, 16)
    DEP_FIELD(SMMU_CB13_FSYNR0, AFR, 1, 11)
    DEP_FIELD(SMMU_CB13_FSYNR0, PTWF, 1, 10)
    DEP_FIELD(SMMU_CB13_FSYNR0, ATOF, 1, 9)
    DEP_FIELD(SMMU_CB13_FSYNR0, NSATTR, 1, 8)
    DEP_FIELD(SMMU_CB13_FSYNR0, IND, 1, 6)
    DEP_FIELD(SMMU_CB13_FSYNR0, PNU, 1, 5)
    DEP_FIELD(SMMU_CB13_FSYNR0, WNR, 1, 4)
    DEP_FIELD(SMMU_CB13_FSYNR0, PLVL, 2, 0)
DEP_REG32(SMMU_CB13_IPAFAR_LOW, 0x1d070)
    DEP_FIELD(SMMU_CB13_IPAFAR_LOW, IPAFAR_L, 20, 12)
    DEP_FIELD(SMMU_CB13_IPAFAR_LOW, FAR_RO, 12, 0)
DEP_REG32(SMMU_CB13_IPAFAR_HIGH, 0x1d074)
    DEP_FIELD(SMMU_CB13_IPAFAR_HIGH, BITS, 16, 0)
DEP_REG32(SMMU_CB13_TLBIVA_LOW, 0x1d600)
DEP_REG32(SMMU_CB13_TLBIVA_HIGH, 0x1d604)
    DEP_FIELD(SMMU_CB13_TLBIVA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB13_TLBIVA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB13_TLBIVAA_LOW, 0x1d608)
DEP_REG32(SMMU_CB13_TLBIVAA_HIGH, 0x1d60c)
    DEP_FIELD(SMMU_CB13_TLBIVAA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB13_TLBIVAA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB13_TLBIASID, 0x1d610)
    DEP_FIELD(SMMU_CB13_TLBIASID, ASID, 16, 0)
DEP_REG32(SMMU_CB13_TLBIALL, 0x1d618)
DEP_REG32(SMMU_CB13_TLBIVAL_LOW, 0x1d620)
DEP_REG32(SMMU_CB13_TLBIVAL_HIGH, 0x1d624)
    DEP_FIELD(SMMU_CB13_TLBIVAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB13_TLBIVAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB13_TLBIVAAL_LOW, 0x1d628)
DEP_REG32(SMMU_CB13_TLBIVAAL_HIGH, 0x1d62c)
    DEP_FIELD(SMMU_CB13_TLBIVAAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB13_TLBIVAAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB13_TLBIIPAS2_LOW, 0x1d630)
DEP_REG32(SMMU_CB13_TLBIIPAS2_HIGH, 0x1d634)
    DEP_FIELD(SMMU_CB13_TLBIIPAS2_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB13_TLBIIPAS2L_LOW, 0x1d638)
DEP_REG32(SMMU_CB13_TLBIIPAS2L_HIGH, 0x1d63c)
    DEP_FIELD(SMMU_CB13_TLBIIPAS2L_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB13_TLBSYNC, 0x1d7f0)
DEP_REG32(SMMU_CB13_TLBSTATUS, 0x1d7f4)
    DEP_FIELD(SMMU_CB13_TLBSTATUS, SACTIVE, 1, 0)
DEP_REG32(SMMU_CB13_PMEVCNTR0, 0x1de00)
DEP_REG32(SMMU_CB13_PMEVCNTR1, 0x1de04)
DEP_REG32(SMMU_CB13_PMEVCNTR2, 0x1de08)
DEP_REG32(SMMU_CB13_PMEVCNTR3, 0x1de0c)
DEP_REG32(SMMU_CB13_PMEVTYPER0, 0x1de80)
    DEP_FIELD(SMMU_CB13_PMEVTYPER0, P, 1, 31)
    DEP_FIELD(SMMU_CB13_PMEVTYPER0, U, 1, 30)
    DEP_FIELD(SMMU_CB13_PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(SMMU_CB13_PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(SMMU_CB13_PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(SMMU_CB13_PMEVTYPER1, 0x1de84)
    DEP_FIELD(SMMU_CB13_PMEVTYPER1, P, 1, 31)
    DEP_FIELD(SMMU_CB13_PMEVTYPER1, U, 1, 30)
    DEP_FIELD(SMMU_CB13_PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(SMMU_CB13_PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(SMMU_CB13_PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(SMMU_CB13_PMEVTYPER2, 0x1de88)
    DEP_FIELD(SMMU_CB13_PMEVTYPER2, P, 1, 31)
    DEP_FIELD(SMMU_CB13_PMEVTYPER2, U, 1, 30)
    DEP_FIELD(SMMU_CB13_PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(SMMU_CB13_PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(SMMU_CB13_PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(SMMU_CB13_PMEVTYPER3, 0x1de8c)
    DEP_FIELD(SMMU_CB13_PMEVTYPER3, P, 1, 31)
    DEP_FIELD(SMMU_CB13_PMEVTYPER3, U, 1, 30)
    DEP_FIELD(SMMU_CB13_PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(SMMU_CB13_PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(SMMU_CB13_PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(SMMU_CB13_PMCFGR, 0x1df00)
    DEP_FIELD(SMMU_CB13_PMCFGR, NCG, 8, 24)
    DEP_FIELD(SMMU_CB13_PMCFGR, UEN, 1, 19)
    DEP_FIELD(SMMU_CB13_PMCFGR, EX, 1, 16)
    DEP_FIELD(SMMU_CB13_PMCFGR, CCD, 1, 15)
    DEP_FIELD(SMMU_CB13_PMCFGR, CC, 1, 14)
    DEP_FIELD(SMMU_CB13_PMCFGR, SIZE, 6, 8)
    DEP_FIELD(SMMU_CB13_PMCFGR, N, 8, 0)
DEP_REG32(SMMU_CB13_PMCR, 0x1df04)
    DEP_FIELD(SMMU_CB13_PMCR, IMP, 8, 24)
    DEP_FIELD(SMMU_CB13_PMCR, X, 1, 4)
    DEP_FIELD(SMMU_CB13_PMCR, P, 1, 1)
    DEP_FIELD(SMMU_CB13_PMCR, E, 1, 0)
DEP_REG32(SMMU_CB13_PMCEID, 0x1df20)
    DEP_FIELD(SMMU_CB13_PMCEID, EVENT0X12, 1, 17)
    DEP_FIELD(SMMU_CB13_PMCEID, EVENT0X11, 1, 16)
    DEP_FIELD(SMMU_CB13_PMCEID, EVENT0X10, 1, 15)
    DEP_FIELD(SMMU_CB13_PMCEID, EVENT0X0A, 1, 9)
    DEP_FIELD(SMMU_CB13_PMCEID, EVENT0X09, 1, 8)
    DEP_FIELD(SMMU_CB13_PMCEID, EVENT0X08, 1, 7)
    DEP_FIELD(SMMU_CB13_PMCEID, EVENT0X01, 1, 1)
    DEP_FIELD(SMMU_CB13_PMCEID, EVENT0X00, 1, 0)
DEP_REG32(SMMU_CB13_PMCNTENSE, 0x1df40)
    DEP_FIELD(SMMU_CB13_PMCNTENSE, P3, 1, 3)
    DEP_FIELD(SMMU_CB13_PMCNTENSE, P2, 1, 2)
    DEP_FIELD(SMMU_CB13_PMCNTENSE, P1, 1, 1)
    DEP_FIELD(SMMU_CB13_PMCNTENSE, P0, 1, 0)
DEP_REG32(SMMU_CB13_PMCNTENCLR, 0x1df44)
    DEP_FIELD(SMMU_CB13_PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB13_PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB13_PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB13_PMCNTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB13_PMCNTENSET, 0x1df48)
    DEP_FIELD(SMMU_CB13_PMCNTENSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB13_PMCNTENSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB13_PMCNTENSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB13_PMCNTENSET, P0, 1, 0)
DEP_REG32(SMMU_CB13_PMINTENCLR, 0x1df4c)
    DEP_FIELD(SMMU_CB13_PMINTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB13_PMINTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB13_PMINTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB13_PMINTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB13_PMOVSCLR, 0x1df50)
    DEP_FIELD(SMMU_CB13_PMOVSCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB13_PMOVSCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB13_PMOVSCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB13_PMOVSCLR, P0, 1, 0)
DEP_REG32(SMMU_CB13_PMOVSSET, 0x1df58)
    DEP_FIELD(SMMU_CB13_PMOVSSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB13_PMOVSSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB13_PMOVSSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB13_PMOVSSET, P0, 1, 0)
DEP_REG32(SMMU_CB13_PMAUTHSTATUS, 0x1dfb8)
    DEP_FIELD(SMMU_CB13_PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(SMMU_CB13_PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(SMMU_CB13_PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(SMMU_CB13_PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(SMMU_CB13_PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(SMMU_CB13_PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(SMMU_CB13_PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(SMMU_CB13_PMAUTHSTATUS, NSE, 1, 0)
DEP_REG32(SMMU_CB14_SCTLR, 0x1e000)
    DEP_FIELD(SMMU_CB14_SCTLR, NSCFG, 2, 28)
    DEP_FIELD(SMMU_CB14_SCTLR, WACFG, 2, 26)
    DEP_FIELD(SMMU_CB14_SCTLR, RACFG, 2, 24)
    DEP_FIELD(SMMU_CB14_SCTLR, SHCFG, 2, 22)
    DEP_FIELD(SMMU_CB14_SCTLR, FB, 1, 21)
    DEP_FIELD(SMMU_CB14_SCTLR, MTCFG, 1, 20)
    DEP_FIELD(SMMU_CB14_SCTLR, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_CB14_SCTLR, TRANSIENTCFG, 2, 14)
    DEP_FIELD(SMMU_CB14_SCTLR, PTW, 1, 13)
    DEP_FIELD(SMMU_CB14_SCTLR, ASIDPNE, 1, 12)
    DEP_FIELD(SMMU_CB14_SCTLR, UWXN, 1, 10)
    DEP_FIELD(SMMU_CB14_SCTLR, WXN, 1, 9)
    DEP_FIELD(SMMU_CB14_SCTLR, HUPCF, 1, 8)
    DEP_FIELD(SMMU_CB14_SCTLR, CFCFG, 1, 7)
    DEP_FIELD(SMMU_CB14_SCTLR, CFIE, 1, 6)
    DEP_FIELD(SMMU_CB14_SCTLR, CFRE, 1, 5)
    DEP_FIELD(SMMU_CB14_SCTLR, E, 1, 4)
    DEP_FIELD(SMMU_CB14_SCTLR, AFFD, 1, 3)
    DEP_FIELD(SMMU_CB14_SCTLR, AFE, 1, 2)
    DEP_FIELD(SMMU_CB14_SCTLR, TRE, 1, 1)
    DEP_FIELD(SMMU_CB14_SCTLR, M, 1, 0)
DEP_REG32(SMMU_CB14_ACTLR, 0x1e004)
    DEP_FIELD(SMMU_CB14_ACTLR, CPRE, 1, 1)
    DEP_FIELD(SMMU_CB14_ACTLR, CMTLB, 1, 0)
DEP_REG32(SMMU_CB14_RESUME, 0x1e008)
    DEP_FIELD(SMMU_CB14_RESUME, TNR, 1, 0)
DEP_REG32(SMMU_CB14_TCR2, 0x1e010)
    DEP_FIELD(SMMU_CB14_TCR2, NSCFG1, 1, 30)
    DEP_FIELD(SMMU_CB14_TCR2, SEP, 3, 15)
    DEP_FIELD(SMMU_CB14_TCR2, NSCFG0, 1, 14)
    DEP_FIELD(SMMU_CB14_TCR2, TBI1, 1, 6)
    DEP_FIELD(SMMU_CB14_TCR2, TBI0, 1, 5)
    DEP_FIELD(SMMU_CB14_TCR2, AS, 1, 4)
    DEP_FIELD(SMMU_CB14_TCR2, PASIZE, 3, 0)
DEP_REG32(SMMU_CB14_TTBR0_LOW, 0x1e020)
    DEP_FIELD(SMMU_CB14_TTBR0_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB14_TTBR0_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB14_TTBR0_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB14_TTBR0_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB14_TTBR0_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB14_TTBR0_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB14_TTBR0_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB14_TTBR0_HIGH, 0x1e024)
    DEP_FIELD(SMMU_CB14_TTBR0_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB14_TTBR0_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB14_TTBR1_LOW, 0x1e028)
    DEP_FIELD(SMMU_CB14_TTBR1_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB14_TTBR1_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB14_TTBR1_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB14_TTBR1_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB14_TTBR1_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB14_TTBR1_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB14_TTBR1_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB14_TTBR1_HIGH, 0x1e02c)
    DEP_FIELD(SMMU_CB14_TTBR1_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB14_TTBR1_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB14_TCR_LPAE, 0x1e030)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, EAE, 1, 31)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, NSCFG1_TG1, 1, 30)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, SH1, 2, 28)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, ORGN1, 2, 26)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, IRGN1, 2, 24)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, EPD1, 1, 23)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, A1, 1, 22)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, T1SZ_5_3, 3, 19)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, T1SZ_2_0_PASIZE, 3, 16)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, NSCFG0_TG0, 1, 14)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, SH0, 2, 12)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, ORGN0, 2, 10)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, IRGN0, 2, 8)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, SL0_1_EPD0, 1, 7)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, SL0_0, 1, 6)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, PD1_T0SZ_5, 1, 5)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, S_PD0_T0SZ_4, 1, 4)
    DEP_FIELD(SMMU_CB14_TCR_LPAE, T0SZ_3_0, 4, 0)
DEP_REG32(SMMU_CB14_CONTEXTIDR, 0x1e034)
    DEP_FIELD(SMMU_CB14_CONTEXTIDR, PROCID, 24, 8)
    DEP_FIELD(SMMU_CB14_CONTEXTIDR, ASID, 8, 0)
DEP_REG32(SMMU_CB14_PRRR_MAIR0, 0x1e038)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, NOS7, 1, 31)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, NOS6, 1, 30)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, NOS5, 1, 29)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, NOS4, 1, 28)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, NOS3, 1, 27)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, NOS2, 1, 26)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, NOS1, 1, 25)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, NOS0, 1, 24)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, NS1, 1, 19)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, NS0, 1, 18)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, DS1, 1, 17)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, DS0, 1, 16)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, TR7, 2, 14)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, TR6, 2, 12)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, TR5, 2, 10)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, TR4, 2, 8)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, TR3, 2, 6)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, TR2, 2, 4)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, TR1, 2, 2)
    DEP_FIELD(SMMU_CB14_PRRR_MAIR0, TR0, 2, 0)
DEP_REG32(SMMU_CB14_NMRR_MAIR1, 0x1e03c)
    DEP_FIELD(SMMU_CB14_NMRR_MAIR1, OR7, 2, 30)
    DEP_FIELD(SMMU_CB14_NMRR_MAIR1, OR6, 2, 28)
    DEP_FIELD(SMMU_CB14_NMRR_MAIR1, OR5, 2, 26)
    DEP_FIELD(SMMU_CB14_NMRR_MAIR1, OR4, 2, 24)
    DEP_FIELD(SMMU_CB14_NMRR_MAIR1, OR3, 2, 22)
    DEP_FIELD(SMMU_CB14_NMRR_MAIR1, OR2, 2, 20)
    DEP_FIELD(SMMU_CB14_NMRR_MAIR1, OR1, 2, 18)
    DEP_FIELD(SMMU_CB14_NMRR_MAIR1, OR0, 2, 16)
    DEP_FIELD(SMMU_CB14_NMRR_MAIR1, IR7, 2, 14)
    DEP_FIELD(SMMU_CB14_NMRR_MAIR1, IR6, 2, 12)
    DEP_FIELD(SMMU_CB14_NMRR_MAIR1, IR5, 2, 10)
    DEP_FIELD(SMMU_CB14_NMRR_MAIR1, IR4, 2, 8)
    DEP_FIELD(SMMU_CB14_NMRR_MAIR1, IR3, 2, 6)
    DEP_FIELD(SMMU_CB14_NMRR_MAIR1, IR2, 2, 4)
    DEP_FIELD(SMMU_CB14_NMRR_MAIR1, IR1, 2, 2)
    DEP_FIELD(SMMU_CB14_NMRR_MAIR1, IR0, 2, 0)
DEP_REG32(SMMU_CB14_FSR, 0x1e058)
    DEP_FIELD(SMMU_CB14_FSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_CB14_FSR, SS, 1, 30)
    DEP_FIELD(SMMU_CB14_FSR, FORMAT, 2, 9)
    DEP_FIELD(SMMU_CB14_FSR, UUT, 1, 8)
    DEP_FIELD(SMMU_CB14_FSR, ASF, 1, 7)
    DEP_FIELD(SMMU_CB14_FSR, TLBLKF, 1, 6)
    DEP_FIELD(SMMU_CB14_FSR, TLBMCF, 1, 5)
    DEP_FIELD(SMMU_CB14_FSR, EF, 1, 4)
    DEP_FIELD(SMMU_CB14_FSR, PF, 1, 3)
    DEP_FIELD(SMMU_CB14_FSR, AFF, 1, 2)
    DEP_FIELD(SMMU_CB14_FSR, TF, 1, 1)
DEP_REG32(SMMU_CB14_FSRRESTORE, 0x1e05c)
DEP_REG32(SMMU_CB14_FAR_LOW, 0x1e060)
DEP_REG32(SMMU_CB14_FAR_HIGH, 0x1e064)
    DEP_FIELD(SMMU_CB14_FAR_HIGH, BITS, 17, 0)
DEP_REG32(SMMU_CB14_FSYNR0, 0x1e068)
    DEP_FIELD(SMMU_CB14_FSYNR0, S1CBNDX, 4, 16)
    DEP_FIELD(SMMU_CB14_FSYNR0, AFR, 1, 11)
    DEP_FIELD(SMMU_CB14_FSYNR0, PTWF, 1, 10)
    DEP_FIELD(SMMU_CB14_FSYNR0, ATOF, 1, 9)
    DEP_FIELD(SMMU_CB14_FSYNR0, NSATTR, 1, 8)
    DEP_FIELD(SMMU_CB14_FSYNR0, IND, 1, 6)
    DEP_FIELD(SMMU_CB14_FSYNR0, PNU, 1, 5)
    DEP_FIELD(SMMU_CB14_FSYNR0, WNR, 1, 4)
    DEP_FIELD(SMMU_CB14_FSYNR0, PLVL, 2, 0)
DEP_REG32(SMMU_CB14_IPAFAR_LOW, 0x1e070)
    DEP_FIELD(SMMU_CB14_IPAFAR_LOW, IPAFAR_L, 20, 12)
    DEP_FIELD(SMMU_CB14_IPAFAR_LOW, FAR_RO, 12, 0)
DEP_REG32(SMMU_CB14_IPAFAR_HIGH, 0x1e074)
    DEP_FIELD(SMMU_CB14_IPAFAR_HIGH, BITS, 16, 0)
DEP_REG32(SMMU_CB14_TLBIVA_LOW, 0x1e600)
DEP_REG32(SMMU_CB14_TLBIVA_HIGH, 0x1e604)
    DEP_FIELD(SMMU_CB14_TLBIVA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB14_TLBIVA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB14_TLBIVAA_LOW, 0x1e608)
DEP_REG32(SMMU_CB14_TLBIVAA_HIGH, 0x1e60c)
    DEP_FIELD(SMMU_CB14_TLBIVAA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB14_TLBIVAA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB14_TLBIASID, 0x1e610)
    DEP_FIELD(SMMU_CB14_TLBIASID, ASID, 16, 0)
DEP_REG32(SMMU_CB14_TLBIALL, 0x1e618)
DEP_REG32(SMMU_CB14_TLBIVAL_LOW, 0x1e620)
DEP_REG32(SMMU_CB14_TLBIVAL_HIGH, 0x1e624)
    DEP_FIELD(SMMU_CB14_TLBIVAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB14_TLBIVAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB14_TLBIVAAL_LOW, 0x1e628)
DEP_REG32(SMMU_CB14_TLBIVAAL_HIGH, 0x1e62c)
    DEP_FIELD(SMMU_CB14_TLBIVAAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB14_TLBIVAAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB14_TLBIIPAS2_LOW, 0x1e630)
DEP_REG32(SMMU_CB14_TLBIIPAS2_HIGH, 0x1e634)
    DEP_FIELD(SMMU_CB14_TLBIIPAS2_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB14_TLBIIPAS2L_LOW, 0x1e638)
DEP_REG32(SMMU_CB14_TLBIIPAS2L_HIGH, 0x1e63c)
    DEP_FIELD(SMMU_CB14_TLBIIPAS2L_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB14_TLBSYNC, 0x1e7f0)
DEP_REG32(SMMU_CB14_TLBSTATUS, 0x1e7f4)
    DEP_FIELD(SMMU_CB14_TLBSTATUS, SACTIVE, 1, 0)
DEP_REG32(SMMU_CB14_PMEVCNTR0, 0x1ee00)
DEP_REG32(SMMU_CB14_PMEVCNTR1, 0x1ee04)
DEP_REG32(SMMU_CB14_PMEVCNTR2, 0x1ee08)
DEP_REG32(SMMU_CB14_PMEVCNTR3, 0x1ee0c)
DEP_REG32(SMMU_CB14_PMEVTYPER0, 0x1ee80)
    DEP_FIELD(SMMU_CB14_PMEVTYPER0, P, 1, 31)
    DEP_FIELD(SMMU_CB14_PMEVTYPER0, U, 1, 30)
    DEP_FIELD(SMMU_CB14_PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(SMMU_CB14_PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(SMMU_CB14_PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(SMMU_CB14_PMEVTYPER1, 0x1ee84)
    DEP_FIELD(SMMU_CB14_PMEVTYPER1, P, 1, 31)
    DEP_FIELD(SMMU_CB14_PMEVTYPER1, U, 1, 30)
    DEP_FIELD(SMMU_CB14_PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(SMMU_CB14_PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(SMMU_CB14_PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(SMMU_CB14_PMEVTYPER2, 0x1ee88)
    DEP_FIELD(SMMU_CB14_PMEVTYPER2, P, 1, 31)
    DEP_FIELD(SMMU_CB14_PMEVTYPER2, U, 1, 30)
    DEP_FIELD(SMMU_CB14_PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(SMMU_CB14_PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(SMMU_CB14_PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(SMMU_CB14_PMEVTYPER3, 0x1ee8c)
    DEP_FIELD(SMMU_CB14_PMEVTYPER3, P, 1, 31)
    DEP_FIELD(SMMU_CB14_PMEVTYPER3, U, 1, 30)
    DEP_FIELD(SMMU_CB14_PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(SMMU_CB14_PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(SMMU_CB14_PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(SMMU_CB14_PMCFGR, 0x1ef00)
    DEP_FIELD(SMMU_CB14_PMCFGR, NCG, 8, 24)
    DEP_FIELD(SMMU_CB14_PMCFGR, UEN, 1, 19)
    DEP_FIELD(SMMU_CB14_PMCFGR, EX, 1, 16)
    DEP_FIELD(SMMU_CB14_PMCFGR, CCD, 1, 15)
    DEP_FIELD(SMMU_CB14_PMCFGR, CC, 1, 14)
    DEP_FIELD(SMMU_CB14_PMCFGR, SIZE, 6, 8)
    DEP_FIELD(SMMU_CB14_PMCFGR, N, 8, 0)
DEP_REG32(SMMU_CB14_PMCR, 0x1ef04)
    DEP_FIELD(SMMU_CB14_PMCR, IMP, 8, 24)
    DEP_FIELD(SMMU_CB14_PMCR, X, 1, 4)
    DEP_FIELD(SMMU_CB14_PMCR, P, 1, 1)
    DEP_FIELD(SMMU_CB14_PMCR, E, 1, 0)
DEP_REG32(SMMU_CB14_PMCEID, 0x1ef20)
    DEP_FIELD(SMMU_CB14_PMCEID, EVENT0X12, 1, 17)
    DEP_FIELD(SMMU_CB14_PMCEID, EVENT0X11, 1, 16)
    DEP_FIELD(SMMU_CB14_PMCEID, EVENT0X10, 1, 15)
    DEP_FIELD(SMMU_CB14_PMCEID, EVENT0X0A, 1, 9)
    DEP_FIELD(SMMU_CB14_PMCEID, EVENT0X09, 1, 8)
    DEP_FIELD(SMMU_CB14_PMCEID, EVENT0X08, 1, 7)
    DEP_FIELD(SMMU_CB14_PMCEID, EVENT0X01, 1, 1)
    DEP_FIELD(SMMU_CB14_PMCEID, EVENT0X00, 1, 0)
DEP_REG32(SMMU_CB14_PMCNTENSE, 0x1ef40)
    DEP_FIELD(SMMU_CB14_PMCNTENSE, P3, 1, 3)
    DEP_FIELD(SMMU_CB14_PMCNTENSE, P2, 1, 2)
    DEP_FIELD(SMMU_CB14_PMCNTENSE, P1, 1, 1)
    DEP_FIELD(SMMU_CB14_PMCNTENSE, P0, 1, 0)
DEP_REG32(SMMU_CB14_PMCNTENCLR, 0x1ef44)
    DEP_FIELD(SMMU_CB14_PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB14_PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB14_PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB14_PMCNTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB14_PMCNTENSET, 0x1ef48)
    DEP_FIELD(SMMU_CB14_PMCNTENSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB14_PMCNTENSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB14_PMCNTENSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB14_PMCNTENSET, P0, 1, 0)
DEP_REG32(SMMU_CB14_PMINTENCLR, 0x1ef4c)
    DEP_FIELD(SMMU_CB14_PMINTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB14_PMINTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB14_PMINTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB14_PMINTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB14_PMOVSCLR, 0x1ef50)
    DEP_FIELD(SMMU_CB14_PMOVSCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB14_PMOVSCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB14_PMOVSCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB14_PMOVSCLR, P0, 1, 0)
DEP_REG32(SMMU_CB14_PMOVSSET, 0x1ef58)
    DEP_FIELD(SMMU_CB14_PMOVSSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB14_PMOVSSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB14_PMOVSSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB14_PMOVSSET, P0, 1, 0)
DEP_REG32(SMMU_CB14_PMAUTHSTATUS, 0x1efb8)
    DEP_FIELD(SMMU_CB14_PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(SMMU_CB14_PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(SMMU_CB14_PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(SMMU_CB14_PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(SMMU_CB14_PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(SMMU_CB14_PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(SMMU_CB14_PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(SMMU_CB14_PMAUTHSTATUS, NSE, 1, 0)
DEP_REG32(SMMU_CB15_SCTLR, 0x1f000)
    DEP_FIELD(SMMU_CB15_SCTLR, NSCFG, 2, 28)
    DEP_FIELD(SMMU_CB15_SCTLR, WACFG, 2, 26)
    DEP_FIELD(SMMU_CB15_SCTLR, RACFG, 2, 24)
    DEP_FIELD(SMMU_CB15_SCTLR, SHCFG, 2, 22)
    DEP_FIELD(SMMU_CB15_SCTLR, FB, 1, 21)
    DEP_FIELD(SMMU_CB15_SCTLR, MTCFG, 1, 20)
    DEP_FIELD(SMMU_CB15_SCTLR, MEMATTR, 4, 16)
    DEP_FIELD(SMMU_CB15_SCTLR, TRANSIENTCFG, 2, 14)
    DEP_FIELD(SMMU_CB15_SCTLR, PTW, 1, 13)
    DEP_FIELD(SMMU_CB15_SCTLR, ASIDPNE, 1, 12)
    DEP_FIELD(SMMU_CB15_SCTLR, UWXN, 1, 10)
    DEP_FIELD(SMMU_CB15_SCTLR, WXN, 1, 9)
    DEP_FIELD(SMMU_CB15_SCTLR, HUPCF, 1, 8)
    DEP_FIELD(SMMU_CB15_SCTLR, CFCFG, 1, 7)
    DEP_FIELD(SMMU_CB15_SCTLR, CFIE, 1, 6)
    DEP_FIELD(SMMU_CB15_SCTLR, CFRE, 1, 5)
    DEP_FIELD(SMMU_CB15_SCTLR, E, 1, 4)
    DEP_FIELD(SMMU_CB15_SCTLR, AFFD, 1, 3)
    DEP_FIELD(SMMU_CB15_SCTLR, AFE, 1, 2)
    DEP_FIELD(SMMU_CB15_SCTLR, TRE, 1, 1)
    DEP_FIELD(SMMU_CB15_SCTLR, M, 1, 0)
DEP_REG32(SMMU_CB15_ACTLR, 0x1f004)
    DEP_FIELD(SMMU_CB15_ACTLR, CPRE, 1, 1)
    DEP_FIELD(SMMU_CB15_ACTLR, CMTLB, 1, 0)
DEP_REG32(SMMU_CB15_RESUME, 0x1f008)
    DEP_FIELD(SMMU_CB15_RESUME, TNR, 1, 0)
DEP_REG32(SMMU_CB15_TCR2, 0x1f010)
    DEP_FIELD(SMMU_CB15_TCR2, NSCFG1, 1, 30)
    DEP_FIELD(SMMU_CB15_TCR2, SEP, 3, 15)
    DEP_FIELD(SMMU_CB15_TCR2, NSCFG0, 1, 14)
    DEP_FIELD(SMMU_CB15_TCR2, TBI1, 1, 6)
    DEP_FIELD(SMMU_CB15_TCR2, TBI0, 1, 5)
    DEP_FIELD(SMMU_CB15_TCR2, AS, 1, 4)
    DEP_FIELD(SMMU_CB15_TCR2, PASIZE, 3, 0)
DEP_REG32(SMMU_CB15_TTBR0_LOW, 0x1f020)
    DEP_FIELD(SMMU_CB15_TTBR0_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB15_TTBR0_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB15_TTBR0_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB15_TTBR0_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB15_TTBR0_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB15_TTBR0_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB15_TTBR0_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB15_TTBR0_HIGH, 0x1f024)
    DEP_FIELD(SMMU_CB15_TTBR0_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB15_TTBR0_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB15_TTBR1_LOW, 0x1f028)
    DEP_FIELD(SMMU_CB15_TTBR1_LOW, ADDRESS_31_7, 25, 7)
    DEP_FIELD(SMMU_CB15_TTBR1_LOW, ADDRESS_6_IRGN0, 1, 6)
    DEP_FIELD(SMMU_CB15_TTBR1_LOW, ADDRESS_5_NOS, 1, 5)
    DEP_FIELD(SMMU_CB15_TTBR1_LOW, ADDRESS_4_3_RGN, 2, 3)
    DEP_FIELD(SMMU_CB15_TTBR1_LOW, ADDRESS_2, 1, 2)
    DEP_FIELD(SMMU_CB15_TTBR1_LOW, ADDRESS_1_S, 1, 1)
    DEP_FIELD(SMMU_CB15_TTBR1_LOW, ADDRESS_0_IRGN1, 1, 0)
DEP_REG32(SMMU_CB15_TTBR1_HIGH, 0x1f02c)
    DEP_FIELD(SMMU_CB15_TTBR1_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB15_TTBR1_HIGH, ADDRESS, 16, 0)
DEP_REG32(SMMU_CB15_TCR_LPAE, 0x1f030)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, EAE, 1, 31)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, NSCFG1_TG1, 1, 30)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, SH1, 2, 28)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, ORGN1, 2, 26)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, IRGN1, 2, 24)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, EPD1, 1, 23)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, A1, 1, 22)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, T1SZ_5_3, 3, 19)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, T1SZ_2_0_PASIZE, 3, 16)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, NSCFG0_TG0, 1, 14)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, SH0, 2, 12)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, ORGN0, 2, 10)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, IRGN0, 2, 8)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, SL0_1_EPD0, 1, 7)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, SL0_0, 1, 6)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, PD1_T0SZ_5, 1, 5)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, S_PD0_T0SZ_4, 1, 4)
    DEP_FIELD(SMMU_CB15_TCR_LPAE, T0SZ_3_0, 4, 0)
DEP_REG32(SMMU_CB15_CONTEXTIDR, 0x1f034)
    DEP_FIELD(SMMU_CB15_CONTEXTIDR, PROCID, 24, 8)
    DEP_FIELD(SMMU_CB15_CONTEXTIDR, ASID, 8, 0)
DEP_REG32(SMMU_CB15_PRRR_MAIR0, 0x1f038)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, NOS7, 1, 31)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, NOS6, 1, 30)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, NOS5, 1, 29)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, NOS4, 1, 28)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, NOS3, 1, 27)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, NOS2, 1, 26)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, NOS1, 1, 25)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, NOS0, 1, 24)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, NS1, 1, 19)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, NS0, 1, 18)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, DS1, 1, 17)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, DS0, 1, 16)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, TR7, 2, 14)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, TR6, 2, 12)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, TR5, 2, 10)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, TR4, 2, 8)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, TR3, 2, 6)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, TR2, 2, 4)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, TR1, 2, 2)
    DEP_FIELD(SMMU_CB15_PRRR_MAIR0, TR0, 2, 0)
DEP_REG32(SMMU_CB15_NMRR_MAIR1, 0x1f03c)
    DEP_FIELD(SMMU_CB15_NMRR_MAIR1, OR7, 2, 30)
    DEP_FIELD(SMMU_CB15_NMRR_MAIR1, OR6, 2, 28)
    DEP_FIELD(SMMU_CB15_NMRR_MAIR1, OR5, 2, 26)
    DEP_FIELD(SMMU_CB15_NMRR_MAIR1, OR4, 2, 24)
    DEP_FIELD(SMMU_CB15_NMRR_MAIR1, OR3, 2, 22)
    DEP_FIELD(SMMU_CB15_NMRR_MAIR1, OR2, 2, 20)
    DEP_FIELD(SMMU_CB15_NMRR_MAIR1, OR1, 2, 18)
    DEP_FIELD(SMMU_CB15_NMRR_MAIR1, OR0, 2, 16)
    DEP_FIELD(SMMU_CB15_NMRR_MAIR1, IR7, 2, 14)
    DEP_FIELD(SMMU_CB15_NMRR_MAIR1, IR6, 2, 12)
    DEP_FIELD(SMMU_CB15_NMRR_MAIR1, IR5, 2, 10)
    DEP_FIELD(SMMU_CB15_NMRR_MAIR1, IR4, 2, 8)
    DEP_FIELD(SMMU_CB15_NMRR_MAIR1, IR3, 2, 6)
    DEP_FIELD(SMMU_CB15_NMRR_MAIR1, IR2, 2, 4)
    DEP_FIELD(SMMU_CB15_NMRR_MAIR1, IR1, 2, 2)
    DEP_FIELD(SMMU_CB15_NMRR_MAIR1, IR0, 2, 0)
DEP_REG32(SMMU_CB15_FSR, 0x1f058)
    DEP_FIELD(SMMU_CB15_FSR, MULTI, 1, 31)
    DEP_FIELD(SMMU_CB15_FSR, SS, 1, 30)
    DEP_FIELD(SMMU_CB15_FSR, FORMAT, 2, 9)
    DEP_FIELD(SMMU_CB15_FSR, UUT, 1, 8)
    DEP_FIELD(SMMU_CB15_FSR, ASF, 1, 7)
    DEP_FIELD(SMMU_CB15_FSR, TLBLKF, 1, 6)
    DEP_FIELD(SMMU_CB15_FSR, TLBMCF, 1, 5)
    DEP_FIELD(SMMU_CB15_FSR, EF, 1, 4)
    DEP_FIELD(SMMU_CB15_FSR, PF, 1, 3)
    DEP_FIELD(SMMU_CB15_FSR, AFF, 1, 2)
    DEP_FIELD(SMMU_CB15_FSR, TF, 1, 1)
DEP_REG32(SMMU_CB15_FSRRESTORE, 0x1f05c)
DEP_REG32(SMMU_CB15_FAR_LOW, 0x1f060)
DEP_REG32(SMMU_CB15_FAR_HIGH, 0x1f064)
    DEP_FIELD(SMMU_CB15_FAR_HIGH, BITS, 17, 0)
DEP_REG32(SMMU_CB15_FSYNR0, 0x1f068)
    DEP_FIELD(SMMU_CB15_FSYNR0, S1CBNDX, 4, 16)
    DEP_FIELD(SMMU_CB15_FSYNR0, AFR, 1, 11)
    DEP_FIELD(SMMU_CB15_FSYNR0, PTWF, 1, 10)
    DEP_FIELD(SMMU_CB15_FSYNR0, ATOF, 1, 9)
    DEP_FIELD(SMMU_CB15_FSYNR0, NSATTR, 1, 8)
    DEP_FIELD(SMMU_CB15_FSYNR0, IND, 1, 6)
    DEP_FIELD(SMMU_CB15_FSYNR0, PNU, 1, 5)
    DEP_FIELD(SMMU_CB15_FSYNR0, WNR, 1, 4)
    DEP_FIELD(SMMU_CB15_FSYNR0, PLVL, 2, 0)
DEP_REG32(SMMU_CB15_IPAFAR_LOW, 0x1f070)
    DEP_FIELD(SMMU_CB15_IPAFAR_LOW, IPAFAR_L, 20, 12)
    DEP_FIELD(SMMU_CB15_IPAFAR_LOW, FAR_RO, 12, 0)
DEP_REG32(SMMU_CB15_IPAFAR_HIGH, 0x1f074)
    DEP_FIELD(SMMU_CB15_IPAFAR_HIGH, BITS, 16, 0)
DEP_REG32(SMMU_CB15_TLBIVA_LOW, 0x1f600)
DEP_REG32(SMMU_CB15_TLBIVA_HIGH, 0x1f604)
    DEP_FIELD(SMMU_CB15_TLBIVA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB15_TLBIVA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB15_TLBIVAA_LOW, 0x1f608)
DEP_REG32(SMMU_CB15_TLBIVAA_HIGH, 0x1f60c)
    DEP_FIELD(SMMU_CB15_TLBIVAA_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB15_TLBIVAA_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB15_TLBIASID, 0x1f610)
    DEP_FIELD(SMMU_CB15_TLBIASID, ASID, 16, 0)
DEP_REG32(SMMU_CB15_TLBIALL, 0x1f618)
DEP_REG32(SMMU_CB15_TLBIVAL_LOW, 0x1f620)
DEP_REG32(SMMU_CB15_TLBIVAL_HIGH, 0x1f624)
    DEP_FIELD(SMMU_CB15_TLBIVAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB15_TLBIVAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB15_TLBIVAAL_LOW, 0x1f628)
DEP_REG32(SMMU_CB15_TLBIVAAL_HIGH, 0x1f62c)
    DEP_FIELD(SMMU_CB15_TLBIVAAL_HIGH, ASID, 16, 16)
    DEP_FIELD(SMMU_CB15_TLBIVAAL_HIGH, ADDRESS, 5, 0)
DEP_REG32(SMMU_CB15_TLBIIPAS2_LOW, 0x1f630)
DEP_REG32(SMMU_CB15_TLBIIPAS2_HIGH, 0x1f634)
    DEP_FIELD(SMMU_CB15_TLBIIPAS2_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB15_TLBIIPAS2L_LOW, 0x1f638)
DEP_REG32(SMMU_CB15_TLBIIPAS2L_HIGH, 0x1f63c)
    DEP_FIELD(SMMU_CB15_TLBIIPAS2L_HIGH, ADDRESS, 4, 0)
DEP_REG32(SMMU_CB15_TLBSYNC, 0x1f7f0)
DEP_REG32(SMMU_CB15_TLBSTATUS, 0x1f7f4)
    DEP_FIELD(SMMU_CB15_TLBSTATUS, SACTIVE, 1, 0)
DEP_REG32(SMMU_CB15_PMEVCNTR0, 0x1fe00)
DEP_REG32(SMMU_CB15_PMEVCNTR1, 0x1fe04)
DEP_REG32(SMMU_CB15_PMEVCNTR2, 0x1fe08)
DEP_REG32(SMMU_CB15_PMEVCNTR3, 0x1fe0c)
DEP_REG32(SMMU_CB15_PMEVTYPER0, 0x1fe80)
    DEP_FIELD(SMMU_CB15_PMEVTYPER0, P, 1, 31)
    DEP_FIELD(SMMU_CB15_PMEVTYPER0, U, 1, 30)
    DEP_FIELD(SMMU_CB15_PMEVTYPER0, NSP, 1, 29)
    DEP_FIELD(SMMU_CB15_PMEVTYPER0, NSU, 1, 28)
    DEP_FIELD(SMMU_CB15_PMEVTYPER0, EVENT, 5, 0)
DEP_REG32(SMMU_CB15_PMEVTYPER1, 0x1fe84)
    DEP_FIELD(SMMU_CB15_PMEVTYPER1, P, 1, 31)
    DEP_FIELD(SMMU_CB15_PMEVTYPER1, U, 1, 30)
    DEP_FIELD(SMMU_CB15_PMEVTYPER1, NSP, 1, 29)
    DEP_FIELD(SMMU_CB15_PMEVTYPER1, NSU, 1, 28)
    DEP_FIELD(SMMU_CB15_PMEVTYPER1, EVENT, 5, 0)
DEP_REG32(SMMU_CB15_PMEVTYPER2, 0x1fe88)
    DEP_FIELD(SMMU_CB15_PMEVTYPER2, P, 1, 31)
    DEP_FIELD(SMMU_CB15_PMEVTYPER2, U, 1, 30)
    DEP_FIELD(SMMU_CB15_PMEVTYPER2, NSP, 1, 29)
    DEP_FIELD(SMMU_CB15_PMEVTYPER2, NSU, 1, 28)
    DEP_FIELD(SMMU_CB15_PMEVTYPER2, EVENT, 5, 0)
DEP_REG32(SMMU_CB15_PMEVTYPER3, 0x1fe8c)
    DEP_FIELD(SMMU_CB15_PMEVTYPER3, P, 1, 31)
    DEP_FIELD(SMMU_CB15_PMEVTYPER3, U, 1, 30)
    DEP_FIELD(SMMU_CB15_PMEVTYPER3, NSP, 1, 29)
    DEP_FIELD(SMMU_CB15_PMEVTYPER3, NSU, 1, 28)
    DEP_FIELD(SMMU_CB15_PMEVTYPER3, EVENT, 5, 0)
DEP_REG32(SMMU_CB15_PMCFGR, 0x1ff00)
    DEP_FIELD(SMMU_CB15_PMCFGR, NCG, 8, 24)
    DEP_FIELD(SMMU_CB15_PMCFGR, UEN, 1, 19)
    DEP_FIELD(SMMU_CB15_PMCFGR, EX, 1, 16)
    DEP_FIELD(SMMU_CB15_PMCFGR, CCD, 1, 15)
    DEP_FIELD(SMMU_CB15_PMCFGR, CC, 1, 14)
    DEP_FIELD(SMMU_CB15_PMCFGR, SIZE, 6, 8)
    DEP_FIELD(SMMU_CB15_PMCFGR, N, 8, 0)
DEP_REG32(SMMU_CB15_PMCR, 0x1ff04)
    DEP_FIELD(SMMU_CB15_PMCR, IMP, 8, 24)
    DEP_FIELD(SMMU_CB15_PMCR, X, 1, 4)
    DEP_FIELD(SMMU_CB15_PMCR, P, 1, 1)
    DEP_FIELD(SMMU_CB15_PMCR, E, 1, 0)
DEP_REG32(SMMU_CB15_PMCEID, 0x1ff20)
    DEP_FIELD(SMMU_CB15_PMCEID, EVENT0X12, 1, 17)
    DEP_FIELD(SMMU_CB15_PMCEID, EVENT0X11, 1, 16)
    DEP_FIELD(SMMU_CB15_PMCEID, EVENT0X10, 1, 15)
    DEP_FIELD(SMMU_CB15_PMCEID, EVENT0X0A, 1, 9)
    DEP_FIELD(SMMU_CB15_PMCEID, EVENT0X09, 1, 8)
    DEP_FIELD(SMMU_CB15_PMCEID, EVENT0X08, 1, 7)
    DEP_FIELD(SMMU_CB15_PMCEID, EVENT0X01, 1, 1)
    DEP_FIELD(SMMU_CB15_PMCEID, EVENT0X00, 1, 0)
DEP_REG32(SMMU_CB15_PMCNTENSE, 0x1ff40)
    DEP_FIELD(SMMU_CB15_PMCNTENSE, P3, 1, 3)
    DEP_FIELD(SMMU_CB15_PMCNTENSE, P2, 1, 2)
    DEP_FIELD(SMMU_CB15_PMCNTENSE, P1, 1, 1)
    DEP_FIELD(SMMU_CB15_PMCNTENSE, P0, 1, 0)
DEP_REG32(SMMU_CB15_PMCNTENCLR, 0x1ff44)
    DEP_FIELD(SMMU_CB15_PMCNTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB15_PMCNTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB15_PMCNTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB15_PMCNTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB15_PMCNTENSET, 0x1ff48)
    DEP_FIELD(SMMU_CB15_PMCNTENSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB15_PMCNTENSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB15_PMCNTENSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB15_PMCNTENSET, P0, 1, 0)
DEP_REG32(SMMU_CB15_PMINTENCLR, 0x1ff4c)
    DEP_FIELD(SMMU_CB15_PMINTENCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB15_PMINTENCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB15_PMINTENCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB15_PMINTENCLR, P0, 1, 0)
DEP_REG32(SMMU_CB15_PMOVSCLR, 0x1ff50)
    DEP_FIELD(SMMU_CB15_PMOVSCLR, P3, 1, 3)
    DEP_FIELD(SMMU_CB15_PMOVSCLR, P2, 1, 2)
    DEP_FIELD(SMMU_CB15_PMOVSCLR, P1, 1, 1)
    DEP_FIELD(SMMU_CB15_PMOVSCLR, P0, 1, 0)
DEP_REG32(SMMU_CB15_PMOVSSET, 0x1ff58)
    DEP_FIELD(SMMU_CB15_PMOVSSET, P3, 1, 3)
    DEP_FIELD(SMMU_CB15_PMOVSSET, P2, 1, 2)
    DEP_FIELD(SMMU_CB15_PMOVSSET, P1, 1, 1)
    DEP_FIELD(SMMU_CB15_PMOVSSET, P0, 1, 0)
DEP_REG32(SMMU_CB15_PMAUTHSTATUS, 0x1ffb8)
    DEP_FIELD(SMMU_CB15_PMAUTHSTATUS, SNI, 1, 7)
    DEP_FIELD(SMMU_CB15_PMAUTHSTATUS, SNE, 1, 6)
    DEP_FIELD(SMMU_CB15_PMAUTHSTATUS, SI, 1, 5)
    DEP_FIELD(SMMU_CB15_PMAUTHSTATUS, SE, 1, 4)
    DEP_FIELD(SMMU_CB15_PMAUTHSTATUS, NSNI, 1, 3)
    DEP_FIELD(SMMU_CB15_PMAUTHSTATUS, NSNE, 1, 2)
    DEP_FIELD(SMMU_CB15_PMAUTHSTATUS, NSI, 1, 1)
    DEP_FIELD(SMMU_CB15_PMAUTHSTATUS, NSE, 1, 0)

/* Missing from the regspec.  */
DEP_REG32(SMMU_GATS1PR, 0x110)
DEP_REG32(SMMU_GATS1PR_H, 0x114)
DEP_REG32(SMMU_GATS1PW, 0x118)
DEP_REG32(SMMU_GATS1PW_H, 0x11C)
DEP_REG32(SMMU_GATS12PR, 0x130)
DEP_REG32(SMMU_GATS12PR_H, 0x134)
DEP_REG32(SMMU_GATS12PW, 0x138)
DEP_REG32(SMMU_GATS12PW_H, 0x13C)
DEP_REG32(SMMU_GPAR, 0x180)
DEP_REG32(SMMU_GPAR_H, 0x184)
DEP_REG32(SMMU_GATSR, 0x188)

#define R_MAX (R_SMMU_CB15_PMAUTHSTATUS + 1)

/* This should be configurable per instance.  */
#define PAGESIZE (A_SMMU_CB1_TTBR0_LOW - A_SMMU_CB0_TTBR0_LOW)

/* Maximum number of TBUs supported by this model.  */
#define MAX_TBU 16

typedef struct SMMU SMMU;
typedef struct TBU {
    SMMU *smmu;
    IOMMUMemoryRegion iommu;
    AddressSpace *as;
    MemoryRegion *mr;
} TBU;

struct SMMU {
    SysBusDevice parent_obj;
    MemoryRegion iomem;

    MemoryRegion *dma_mr;
    AddressSpace *dma_as;

    TBU tbu[MAX_TBU];

    struct {
        qemu_irq global;
        qemu_irq context[16];
    } irq;

    struct {
        uint32_t pamax;
    } cfg;

    uint32_t regs[R_MAX];
    DepRegisterInfo regs_info[R_MAX];
};

/* Generic page attributes.  */
typedef struct PageAttr {
    uint64_t pa;
    unsigned int block : 1;
    unsigned int rd : 1;
    unsigned int wr : 1;
    unsigned int ns : 1;
} PageAttr;

typedef struct TransReq {
    uint64_t va;
    uint64_t tcr[3];
    uint64_t ttbr[3][2];
    uint32_t access;
    unsigned int stage;
    bool s2_enabled;
    unsigned int s2_cb;

    uint64_t pa;
    uint32_t prot;

    bool err;
} TransReq;

static void smmu_update_ctx_irq(SMMU *s, unsigned int cb)
{
    unsigned int cb_offset = (cb * PAGESIZE) / 4;
    uint32_t sctlr;
    uint32_t fsr;
    bool ie, tf;
    bool pending;

    fsr = s->regs[R_SMMU_CB0_FSR + cb_offset];
    sctlr = s->regs[R_SMMU_CB0_SCTLR + cb_offset];

    tf = DEP_F_EX32(fsr, SMMU_CB0_FSR, TF);
    ie = DEP_F_EX32(sctlr, SMMU_CB0_SCTLR, CFIE);
    pending = tf && ie;
    qemu_set_irq(s->irq.context[cb], pending);
}

static void smmu_fault(SMMU *s, unsigned int cb, TransReq *req, uint64_t syn)
{
    unsigned int cb_offset = (cb * PAGESIZE) / 4;

    s->regs[R_SMMU_CB0_FSR + cb_offset] |= 1 << 1;

    req->err = true;
    s->regs[R_SMMU_CB0_IPAFAR_LOW + cb_offset] = req->va;
    s->regs[R_SMMU_CB0_IPAFAR_HIGH + cb_offset] = req->va >> 32;
    if (req->stage == 2) {
        s->regs[R_SMMU_CB0_FAR_LOW + cb_offset] = req->va;
        s->regs[R_SMMU_CB0_FAR_HIGH + cb_offset] = req->va >> 32;
    }
    smmu_update_ctx_irq(s, cb);
}

static int smmu_stream_id_match(SMMU *s, uint32_t stream_id)
{
    unsigned int nr_smr = DEP_AF_EX32(s->regs, SMMU_SIDR0, NUMSMRG);
    unsigned int i;
    uint32_t s2cr;
    unsigned int cbndx = -1;

    for (i = 0; i < nr_smr; i++) {
        uint32_t v = s->regs[R_SMMU_SMR0 + i];
        bool valid = DEP_F_EX32(v, SMMU_SMR0, VALID);
        uint16_t mask = DEP_F_EX32(v, SMMU_SMR0, MASK);
        uint16_t id = DEP_F_EX32(v, SMMU_SMR0, ID);


        if (valid && (~mask & id) == (~mask & stream_id)) {
            s2cr = s->regs[R_SMMU_S2CR0 + i];
            cbndx = DEP_F_EX32(s2cr, SMMU_S2CR0, CBNDX_VMID);
            break;
        }
    }
    return cbndx;
}

static bool check_s2_startlevel(bool is_aa64, unsigned int pamax, int level,
                                int inputsize, int stride)
{
    /* Negative levels are never allowed.  */
    if (level < 0) {
        return false;
    }

    if (is_aa64) {
        switch (stride) {
        case 13: /* 64KB Pages.  */
            if (level == 0 || (level == 1 && pamax <= 42)) {
                return false;
            }
            break;
        case 11: /* 16KB Pages.  */
            if (level == 0 || (level == 1 && pamax <= 40)) {
                return false;
            }
            break;
        case 9: /* 4KB Pages.  */
            if (level == 0 && pamax <= 42) {
                return false;
            }
            break;
        default:
            g_assert_not_reached();
        }
    } else {
        const int grainsize = stride + 3;
        int startsizecheck;

        /* AArch32 only supports 4KB pages. Assert on that.  */
        assert(stride == 9);

        if (level == 0) {
            return false;
        }

        startsizecheck = inputsize - ((3 - level) * stride + grainsize);
        if (startsizecheck < 1 || startsizecheck > stride + 4) {
            return false;
        }
    }
    return true;
}

static bool check_out_addr(uint64_t addr, unsigned int outputsize)
{
    if (outputsize != 48 && extract64(addr, outputsize, 48 - outputsize)) {
        return false;
    }
    return true;
}

static void smmu_ptw64(SMMU *s, unsigned int cb, TransReq *req)
{
    static const unsigned int outsize_map[] = {
        [0] = 32,
        [1] = 36,
        [2] = 40,
        [3] = 42,
        [4] = 44,
        [5] = 48,
        [6] = 48,
        [7] = 48,
    };
    unsigned int cb_offset = (cb * PAGESIZE) / 4;
    uint32_t sctlr;
    unsigned int tsz;
    unsigned int t0sz;
    unsigned int t1sz;
    unsigned int inputsize;
    unsigned int outputsize;
    unsigned int grainsize = -1;
    unsigned int stride;
    int level = 0;
    unsigned int firstblocklevel = 0;
    unsigned int tg;
    unsigned int ps;
    unsigned int baselowerbound;
    unsigned int stage = req->stage;
    bool blocktranslate = false;
    bool epd = false;
    bool va64;
    bool type64;
    uint32_t tableattrs = 0;
    uint32_t attrs;
    uint32_t s2attrs;
    uint64_t descmask;
    uint64_t ttbr;
    uint64_t desc;

    req->err = false;
    sctlr = s->regs[R_SMMU_CB0_SCTLR + cb_offset];

    if (DEP_F_EX32(sctlr, SMMU_CB0_SCTLR, M) == 0) {
        req->pa = req->va;
        req->prot = IOMMU_RW;
        D("SMMU disabled for context %d sctlr=%x\n", cb, sctlr);
        return;
    }

    ttbr = req->ttbr[stage][0];
    tg = extract32(req->tcr[stage], 14, 2);
    if (stage == 1) {
        ps = extract64(req->tcr[stage], 32, 3);
    } else {
        ps = extract64(req->tcr[stage], 16, 3);
    }
    t0sz = extract32(req->tcr[stage], 0, 6);
    tsz = t0sz;
    req->pa = req->va;

    va64 = s->regs[R_SMMU_CBA2R0 + cb] & 1;
    if (req->stage == 1) {
        type64 = va64 || extract32(req->tcr[1], 31, 1);
        /* We don't support 32bit page-tables yet.  */
        assert(type64);

        if ((req->va & (1ULL << 63)) == 0) {
        } else {
            static const unsigned int tg1map[] = {
                [0] = 3,
                [1] = 2,
                [2] = 0,
                [3] = 1,
            };
            if (!va64 && type64) {
                /* LPAE uses 4K pages.  */
                tg = extract32(req->tcr[stage], 30, 2);
                tg = tg1map[tg];
                t1sz = extract32(req->tcr[stage], 16, 6);
            } else {
                /* Default to 4K.  */
                tg = 0;
                t1sz = extract32(req->tcr[stage], 16, 3);
            }
            ttbr = req->ttbr[stage][1];
            tsz = t1sz;
        }
        epd = extract32(req->tcr[1], 7, 1);
    } else {
        type64 = true;
    }

    if (epd) {
        /* We've already missed the TLB at this stage so fault.  */
        goto do_fault;
    }

    inputsize = 64 - tsz;
    switch (tg) {
    case 1:
        /* 64KB pages.  */
        grainsize = 16;
        level = 3;
        firstblocklevel = 2;
        break;
    case 2:
        /* 16KB pages.  */
        grainsize = 14;
        level = 3;
        firstblocklevel = 2;
        break;
    case 0:
        /* 4KB pages.  */
        grainsize = 12;
        level = 2;
        firstblocklevel = 1;
        break;
    default:
        qemu_log_mask(LOG_GUEST_ERROR, "SMMU: Wrong pagesize\n");
        break;
    }

    stride = grainsize - 3;
    if (req->stage == 1) {
        if (grainsize < 16 && (inputsize > (grainsize + 3 * stride))) {
            level = 0;
        } else if (inputsize > (grainsize + 2 * stride)) {
            level = 1;
        } else if (inputsize > (grainsize + stride)) {
            level = 2;
        }

        if (inputsize < 25 || inputsize > 48
            || extract64(req->va, inputsize, 64 - inputsize)) {
            goto do_fault;
        }
    } else {
        unsigned int startlevel = extract32(req->tcr[stage], 6, 2);
        bool ok;

        level = 3 - startlevel;
        if (grainsize == 12) {
            level = 2 - startlevel;
        }

        ok = check_s2_startlevel(true, 40, level, inputsize, stride);
        if (!ok) {
            goto do_fault;
        }
    }

    outputsize = outsize_map[ps];
    if (outputsize > s->cfg.pamax) {
        outputsize = s->cfg.pamax;
    }

    baselowerbound = 3 + inputsize - ((3 - level) * stride + grainsize);
    ttbr = extract64(ttbr, 0, 48);
    ttbr &= ~((1ULL << baselowerbound) - 1);

    if (!check_out_addr(ttbr, outputsize)) {
        goto do_fault;
    }

    descmask = (1ULL << grainsize) - 1;
    do {
        unsigned int addrselectbottom = (3 - level) * stride + grainsize;
        uint64_t index;
        uint64_t descaddr;
        unsigned int type;

        index = (req->va >> (addrselectbottom - 3)) & descmask;
        index &= ~7ULL;
        descaddr = ttbr | index;

        /* S1 PTW through the S2 system.  */
        if (req->stage == 1 && req->s2_enabled) {
            TransReq s2req = *req;

            s2req.stage = 2;
            s2req.va = descaddr;
            smmu_ptw64(s, s2req.s2_cb, &s2req);
            if (req->err) {
                s->regs[R_SMMU_CB0_IPAFAR_LOW] = descaddr;
                s->regs[R_SMMU_CB0_IPAFAR_HIGH] = descaddr >> 32;
                goto do_fault;
            }
            descaddr = s2req.pa;
        }
        dma_memory_read(s->dma_as, descaddr, &desc, sizeof(desc));
        type = desc & 3;

        D_PTW("smmu: S%d L%d va=0x%"PRIx64" gz=%d descaddr=0x%"PRIx64" "
              "desc=0x%"PRIx64" asb=%d index=0x%"PRIx64" osize=%d\n",
              req->stage, level, req->va, grainsize, descaddr, desc,
              addrselectbottom, index, outputsize);
        ttbr = extract64(desc, 0, 48);
        ttbr &= ~descmask;

        /* special case.  */
        if (!(type & 2) && level == 3) {
            D("smmu: bad level 3 desc\n");
            goto do_fault;
        }

        if (level == 3) {
            break;
        }
        switch (type) {
        case 2:
        case 0:
            /* Invalid.  */
            D("smmu: bad desc\n");
            goto do_fault;
            break;
        case 1:
            blocktranslate = true;
            if (level < firstblocklevel) {
                goto do_fault;
            }
            break;
        case 3:
            tableattrs |= extract64(desc, 59, 5);
            if (!check_out_addr(ttbr, outputsize)) {
                goto do_fault;
            }
            level++;
            break;
        }
    } while (!blocktranslate);

    if (!check_out_addr(ttbr, outputsize)) {
        goto do_fault;
    }

    {
        unsigned long page_size;
        page_size = (1ULL << ((stride * (4 - level)) + 3));
        ttbr |= (req->va & (page_size - 1));
    }

    s2attrs = attrs = extract64(desc, 2, 10)
                      | (extract64(desc, 52, 12) << 10);
    if (req->stage == 1) {
        attrs |= extract32(tableattrs, 0, 2) << 11; /* XN, PXN */
        attrs |= extract32(tableattrs, 3, 1) << 5; /* APTable[1] => AP[2] */
        /* The sense of AP[1] vs APTable[0] is reversed, as APTable[0] == 1
         * means "force PL1 access only", which means forcing AP[1] to 0.
         */
        if (extract32(tableattrs, 2, 1)) {
            attrs &= ~(1 << 4);
        }
    }

    req->prot = IOMMU_RW;
    if ((attrs & (1 << 8)) == 0) {
        /* Access flag */
        D("smmu: access forbidden %x\n", attrs);
        goto do_fault;
    }

    /* AP[1] SBO.  */
    if (!(attrs & (1 << 4))) {
        D("smmu: AP[1] should be one but set to zero!\n");
        goto do_fault;
    }
    if (req->stage == 1) {
        if (attrs & (1 << 5)) {
            /* Write access forbidden */
            if (req->access == IOMMU_WO) {
                D("smmu: Write access forbidden %x\n", attrs);
                goto do_fault;
            }
            req->prot &= ~IOMMU_WO;
        }
    } else {
        switch ((s2attrs >> 4) & 3) {
        /* None.  */
        case 0:
            goto do_fault;
            break;
        /* RO.  */
        case 1:
            if (req->access == IOMMU_WO) {
                goto do_fault;
            }
            req->prot &= ~IOMMU_WO;
            break;
        /* WO.  */
        case 2:
            if (req->access == IOMMU_RO) {
                goto do_fault;
            }
            req->prot &= ~IOMMU_RO;
            break;
        /* RW.  */
        case 3:
            break;
        }
    }

    req->pa = ttbr;
    D("SMMU: 0x%"PRIx64" -> 0x%"PRIx64"\n", req->va, req->pa);
    return;

do_fault:
    D("smmu fault\n");
    smmu_fault(s, cb, req, level);
}

static bool smmu500_at64(SMMU *s, unsigned int cb, hwaddr va,
                         bool wr, bool s2, hwaddr *pa, int *prot)
{
    unsigned int cb_offset = (cb * PAGESIZE) / 4;
    unsigned int cb2_offset = 0;
    TransReq req;
    uint32_t v;
    unsigned int t;

    v = s->regs[R_SMMU_CBAR0 + cb];
    t = DEP_F_EX32(v, SMMU_CBAR0, TYPE);
    switch (t) {
    case 0:
        req.stage = 2;
        req.s2_enabled = true;
        req.s2_cb = cb;
        cb2_offset = cb_offset;
        break;
    case 1:
        req.stage = 1;
        req.s2_enabled = false;
        break;
    case 2:
        req.stage = 1;
        req.s2_enabled = false;
        /* Invalid ??? */
        break;
    case 3:
        req.stage = 1;
        req.s2_enabled = true;
        req.s2_cb = extract32(v, 8, 8);
        cb2_offset = (req.s2_cb * PAGESIZE) / 4;
        break;
    }

    req.va = va;
    req.tcr[1] = s->regs[R_SMMU_CB0_TCR2 + cb_offset];
    req.tcr[1] <<= 32;
    req.tcr[1] |= s->regs[R_SMMU_CB0_TCR_LPAE + cb_offset];

    req.ttbr[1][0] = s->regs[R_SMMU_CB0_TTBR0_HIGH + cb_offset];
    req.ttbr[1][0] <<= 32;
    req.ttbr[1][0] |= s->regs[R_SMMU_CB0_TTBR0_LOW + cb_offset];

    req.ttbr[1][1] = s->regs[R_SMMU_CB0_TTBR1_HIGH + cb_offset];
    req.ttbr[1][1] <<= 32;
    req.ttbr[1][1] |= s->regs[R_SMMU_CB0_TTBR1_LOW + cb_offset];

    if (req.s2_enabled) {
        req.tcr[2] = s->regs[R_SMMU_CB0_TCR_LPAE + cb2_offset];
        req.ttbr[2][0] = s->regs[R_SMMU_CB0_TTBR0_HIGH + cb2_offset];
        req.ttbr[2][0] <<= 32;
        req.ttbr[2][0] |= s->regs[R_SMMU_CB0_TTBR0_LOW + cb2_offset];
    }

    req.access = wr ? IOMMU_WO : IOMMU_RO;

    if (req.stage == 1) {
        smmu_ptw64(s, cb, &req);
        req.stage++;
    } else {
        req.pa = req.va;
    }

    if (s2 && req.s2_enabled) {
        req.va = req.pa;

        smmu_ptw64(s, cb, &req);
    }

    *pa = req.pa;
    *prot = req.prot;
    return req.err;
}

static bool smmu500_at(SMMU *s, unsigned int cb, hwaddr va,
                       bool wr, bool s2, hwaddr *pa, int *prot)
{
    return smmu500_at64(s, cb, va, wr, s2, pa, prot);
}

#define ADDRMASK    ((1ULL << 12) - 1)

static void smmu500_gat(SMMU *s, uint64_t v, bool wr, bool s2)
{
    uint64_t va = v & ~ADDRMASK;
    unsigned int cb = v & ADDRMASK;
    hwaddr pa;
    int prot;
    bool err;

    D("ATS: va=0x%"PRIx64" cb=%d wr=%d s2=%d\n", va, cb, wr, s2);
    err = smmu500_at(s, cb, va, wr, s2, &pa, &prot);

    s->regs[R_SMMU_GPAR] = pa | err;
    s->regs[R_SMMU_GPAR_H] = pa >> 32;
}

static void smmu_gats1pr(DepRegisterInfo *reg, uint64_t val)
{
    SMMU *s = XILINX_SMMU500(reg->opaque);

    val <<= 32;
    val |= s->regs[(reg->access->decode.addr / 4) - 1];
    smmu500_gat(s, val, false, false);
}

static void smmu_gats1pw(DepRegisterInfo *reg, uint64_t val)
{
    SMMU *s = XILINX_SMMU500(reg->opaque);

    val <<= 32;
    val |= s->regs[(reg->access->decode.addr / 4) - 1];
    smmu500_gat(s, val, true, false);
}

static void smmu_gats12pr(DepRegisterInfo *reg, uint64_t val)
{
    SMMU *s = XILINX_SMMU500(reg->opaque);

    val <<= 32;
    val |= s->regs[(reg->access->decode.addr / 4) - 1];
    smmu500_gat(s, val, false, true);
}

static void smmu_gats12pw(DepRegisterInfo *reg, uint64_t val)
{
    SMMU *s = XILINX_SMMU500(reg->opaque);

    val <<= 32;
    val |= s->regs[(reg->access->decode.addr / 4) - 1];
    smmu500_gat(s, val, true, true);
}

static void smmu_nscr0_pw(DepRegisterInfo *reg, uint64_t val)
{
    SMMU *s = XILINX_SMMU500(reg->opaque);

    /* FIXME: Take care of secure vs non-secure accesses.  */
    s->regs[R_SMMU_SCR0] = val;
    s->regs[R_SMMU_NSCR0] = val;
}

static IOMMUTLBEntry smmu_translate(IOMMUMemoryRegion *mr, hwaddr addr,
                                    bool is_write,
                                    MemTxAttrs *attr)
{
    TBU *tbu = container_of(mr, TBU, iommu);
    SMMU *s = tbu->smmu;
    IOMMUTLBEntry ret = {
        .target_as = tbu->as,
        .translated_addr = addr,
        .addr_mask = (1ULL << 12) - 1,
        .perm = IOMMU_RW,
    };
    int cb;
    uint64_t va = addr & ~ADDRMASK;
    hwaddr pa = va;
    int prot;
    bool err = false;
    uint64_t master_id = attr->requester_id;
    bool clientpd = DEP_AF_EX32(s->regs, SMMU_SCR0, CLIENTPD);

    if (clientpd) {
        return ret;
    }

    cb = smmu_stream_id_match(s, master_id);

    if (cb >= 0) {
        err = smmu500_at(s, cb, va, false, true, &pa, &prot);
        ret.translated_addr = pa;
        ret.perm = prot;
        if (err) {
            memset(&ret, 0, sizeof ret);
            ret.perm = IOMMU_NONE;
        }
    }
    return ret;
}

static void smmu_fsr_pw(DepRegisterInfo *reg, uint64_t val)
{
    SMMU *s = XILINX_SMMU500(reg->opaque);
    unsigned int i;

    for (i = 0; i < 16; i++) {
        smmu_update_ctx_irq(s, i);
    }
}

static DepRegisterAccessInfo smmu500_regs_info[] = {
    /* Manually added.  */
    {   .name = "SMMU_GATS1PR",  .decode.addr = A_SMMU_GATS1PR,
    },
    {   .name = "SMMU_GATS1PR_H",  .decode.addr = A_SMMU_GATS1PR_H,
        .post_write = smmu_gats1pr,
    },

    {   .name = "SMMU_GATS1PW",  .decode.addr = A_SMMU_GATS1PW,
    },
    {   .name = "SMMU_GATS1PW_H",  .decode.addr = A_SMMU_GATS1PW_H,
        .post_write = smmu_gats1pw,
    },

    {   .name = "SMMU_GATS12PR",  .decode.addr = A_SMMU_GATS12PR,
    },
    {   .name = "SMMU_GATS12PR_H",  .decode.addr = A_SMMU_GATS12PR_H,
        .post_write = smmu_gats12pr,
    },

    {   .name = "SMMU_GATS12PW",  .decode.addr = A_SMMU_GATS12PW,
    },
    {   .name = "SMMU_GATS12PW_H",  .decode.addr = A_SMMU_GATS12PW_H,
        .post_write = smmu_gats12pw,
    },

    {   .name = "SMMU_GPAR",  .decode.addr = A_SMMU_GPAR, },
    {   .name = "SMMU_GPAR_H",  .decode.addr = A_SMMU_GPAR_H, },
    {   .name = "SMMU_GATSR",  .decode.addr = A_SMMU_GATSR, },


    {   .name = "SMMU_SCR0",  .decode.addr = A_SMMU_SCR0,
        .reset = 0x200001,
        .ro = 0x200330,
    },{ .name = "SMMU_SCR1",  .decode.addr = A_SMMU_SCR1,
        .reset = 0x2013010,
        .ro = 0x10ff0000,
    },{ .name = "SMMU_SACR",  .decode.addr = A_SMMU_SACR,
        .reset = 0x4000004,
    },{ .name = "SMMU_SIDR0",  .decode.addr = A_SMMU_SIDR0,
        .reset = 0xfc013e30,
        .ro = 0xffff7eff,
    },{ .name = "SMMU_SIDR1",  .decode.addr = A_SMMU_SIDR1,
        .reset = 0x30000f10,
        .ro = 0xf0ff9fff,
    },{ .name = "SMMU_SIDR2",  .decode.addr = A_SMMU_SIDR2,
        .reset = 0x5555,
        .ro = 0x7fff,
    },{ .name = "SMMU_SIDR7",  .decode.addr = A_SMMU_SIDR7,
        .reset = 0x21,
        .ro = 0xff,
    },{ .name = "SMMU_SGFAR_LOW",  .decode.addr = A_SMMU_SGFAR_LOW,
    },{ .name = "SMMU_SGFAR_HIGH",  .decode.addr = A_SMMU_SGFAR_HIGH,
    },{ .name = "SMMU_SGFSR",  .decode.addr = A_SMMU_SGFSR,
    },{ .name = "SMMU_SGFSRRESTORE",  .decode.addr = A_SMMU_SGFSRRESTORE,
    },{ .name = "SMMU_SGFSYNR0",  .decode.addr = A_SMMU_SGFSYNR0,
        .ro = 0x40,
    },{ .name = "SMMU_SGFSYNR1",  .decode.addr = A_SMMU_SGFSYNR1,
    },{ .name = "SMMU_STLBIALL",  .decode.addr = A_SMMU_STLBIALL,
    },{ .name = "SMMU_TLBIVMID",  .decode.addr = A_SMMU_TLBIVMID,
    },{ .name = "SMMU_TLBIALLNSNH",  .decode.addr = A_SMMU_TLBIALLNSNH,
    },{ .name = "SMMU_STLBGSYNC",  .decode.addr = A_SMMU_STLBGSYNC,
    },{ .name = "SMMU_STLBGSTATUS",  .decode.addr = A_SMMU_STLBGSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_DBGRPTRTBU",  .decode.addr = A_SMMU_DBGRPTRTBU,
    },{ .name = "SMMU_DBGRDATATBU",  .decode.addr = A_SMMU_DBGRDATATBU,
        .ro = 0xffffffff,
    },{ .name = "SMMU_DBGRPTRTCU",  .decode.addr = A_SMMU_DBGRPTRTCU,
    },{ .name = "SMMU_DBGRDATATCU",  .decode.addr = A_SMMU_DBGRDATATCU,
        .ro = 0xffffffff,
    },{ .name = "SMMU_STLBIVALM_LOW",  .decode.addr = A_SMMU_STLBIVALM_LOW,
    },{ .name = "SMMU_STLBIVALM_HIGH",  .decode.addr = A_SMMU_STLBIVALM_HIGH,
    },{ .name = "SMMU_STLBIVAM_LOW",  .decode.addr = A_SMMU_STLBIVAM_LOW,
    },{ .name = "SMMU_STLBIVAM_HIGH",  .decode.addr = A_SMMU_STLBIVAM_HIGH,
    },{ .name = "SMMU_STLBIALLM",  .decode.addr = A_SMMU_STLBIALLM,
    },{ .name = "SMMU_NSCR0",  .decode.addr = A_SMMU_NSCR0,
        .reset = 0x200001,
        .ro = 0x200330,
        .post_write = smmu_nscr0_pw,
    },{ .name = "SMMU_NSACR",  .decode.addr = A_SMMU_NSACR,
        .reset = 0x400001c,
    },{ .name = "SMMU_NSGFAR_LOW",  .decode.addr = A_SMMU_NSGFAR_LOW,
    },{ .name = "SMMU_NSGFAR_HIGH",  .decode.addr = A_SMMU_NSGFAR_HIGH,
    },{ .name = "SMMU_NSGFSR",  .decode.addr = A_SMMU_NSGFSR,
    },{ .name = "SMMU_NSGFSRRESTORE",  .decode.addr = A_SMMU_NSGFSRRESTORE,
    },{ .name = "SMMU_NSGFSYNR0",  .decode.addr = A_SMMU_NSGFSYNR0,
        .ro = 0x40,
    },{ .name = "SMMU_NSGFSYNDR1",  .decode.addr = A_SMMU_NSGFSYNDR1,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_NSTLBGSYNC",  .decode.addr = A_SMMU_NSTLBGSYNC,
    },{ .name = "SMMU_NSTLBGSTATUS",  .decode.addr = A_SMMU_NSTLBGSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_SMR0",  .decode.addr = A_SMMU_SMR0,
    },{ .name = "SMMU_SMR1",  .decode.addr = A_SMMU_SMR1,
    },{ .name = "SMMU_SMR2",  .decode.addr = A_SMMU_SMR2,
    },{ .name = "SMMU_SMR3",  .decode.addr = A_SMMU_SMR3,
    },{ .name = "SMMU_SMR4",  .decode.addr = A_SMMU_SMR4,
    },{ .name = "SMMU_SMR5",  .decode.addr = A_SMMU_SMR5,
    },{ .name = "SMMU_SMR6",  .decode.addr = A_SMMU_SMR6,
    },{ .name = "SMMU_SMR7",  .decode.addr = A_SMMU_SMR7,
    },{ .name = "SMMU_SMR8",  .decode.addr = A_SMMU_SMR8,
    },{ .name = "SMMU_SMR9",  .decode.addr = A_SMMU_SMR9,
    },{ .name = "SMMU_SMR10",  .decode.addr = A_SMMU_SMR10,
    },{ .name = "SMMU_SMR11",  .decode.addr = A_SMMU_SMR11,
    },{ .name = "SMMU_SMR12",  .decode.addr = A_SMMU_SMR12,
    },{ .name = "SMMU_SMR13",  .decode.addr = A_SMMU_SMR13,
    },{ .name = "SMMU_SMR14",  .decode.addr = A_SMMU_SMR14,
    },{ .name = "SMMU_SMR15",  .decode.addr = A_SMMU_SMR15,
    },{ .name = "SMMU_SMR16",  .decode.addr = A_SMMU_SMR16,
    },{ .name = "SMMU_SMR17",  .decode.addr = A_SMMU_SMR17,
    },{ .name = "SMMU_SMR18",  .decode.addr = A_SMMU_SMR18,
    },{ .name = "SMMU_SMR19",  .decode.addr = A_SMMU_SMR19,
    },{ .name = "SMMU_SMR20",  .decode.addr = A_SMMU_SMR20,
    },{ .name = "SMMU_SMR21",  .decode.addr = A_SMMU_SMR21,
    },{ .name = "SMMU_SMR22",  .decode.addr = A_SMMU_SMR22,
    },{ .name = "SMMU_SMR23",  .decode.addr = A_SMMU_SMR23,
    },{ .name = "SMMU_SMR24",  .decode.addr = A_SMMU_SMR24,
    },{ .name = "SMMU_SMR25",  .decode.addr = A_SMMU_SMR25,
    },{ .name = "SMMU_SMR26",  .decode.addr = A_SMMU_SMR26,
    },{ .name = "SMMU_SMR27",  .decode.addr = A_SMMU_SMR27,
    },{ .name = "SMMU_SMR28",  .decode.addr = A_SMMU_SMR28,
    },{ .name = "SMMU_SMR29",  .decode.addr = A_SMMU_SMR29,
    },{ .name = "SMMU_SMR30",  .decode.addr = A_SMMU_SMR30,
    },{ .name = "SMMU_SMR31",  .decode.addr = A_SMMU_SMR31,
    },{ .name = "SMMU_SMR32",  .decode.addr = A_SMMU_SMR32,
    },{ .name = "SMMU_SMR33",  .decode.addr = A_SMMU_SMR33,
    },{ .name = "SMMU_SMR34",  .decode.addr = A_SMMU_SMR34,
    },{ .name = "SMMU_SMR35",  .decode.addr = A_SMMU_SMR35,
    },{ .name = "SMMU_SMR36",  .decode.addr = A_SMMU_SMR36,
    },{ .name = "SMMU_SMR37",  .decode.addr = A_SMMU_SMR37,
    },{ .name = "SMMU_SMR38",  .decode.addr = A_SMMU_SMR38,
    },{ .name = "SMMU_SMR39",  .decode.addr = A_SMMU_SMR39,
    },{ .name = "SMMU_SMR40",  .decode.addr = A_SMMU_SMR40,
    },{ .name = "SMMU_SMR41",  .decode.addr = A_SMMU_SMR41,
    },{ .name = "SMMU_SMR42",  .decode.addr = A_SMMU_SMR42,
    },{ .name = "SMMU_SMR43",  .decode.addr = A_SMMU_SMR43,
    },{ .name = "SMMU_SMR44",  .decode.addr = A_SMMU_SMR44,
    },{ .name = "SMMU_SMR45",  .decode.addr = A_SMMU_SMR45,
    },{ .name = "SMMU_SMR46",  .decode.addr = A_SMMU_SMR46,
    },{ .name = "SMMU_SMR47",  .decode.addr = A_SMMU_SMR47,
    },{ .name = "SMMU_S2CR0",  .decode.addr = A_SMMU_S2CR0,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR1",  .decode.addr = A_SMMU_S2CR1,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR2",  .decode.addr = A_SMMU_S2CR2,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR3",  .decode.addr = A_SMMU_S2CR3,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR4",  .decode.addr = A_SMMU_S2CR4,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR5",  .decode.addr = A_SMMU_S2CR5,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR6",  .decode.addr = A_SMMU_S2CR6,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR7",  .decode.addr = A_SMMU_S2CR7,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR8",  .decode.addr = A_SMMU_S2CR8,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR9",  .decode.addr = A_SMMU_S2CR9,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR10",  .decode.addr = A_SMMU_S2CR10,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR11",  .decode.addr = A_SMMU_S2CR11,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR12",  .decode.addr = A_SMMU_S2CR12,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR13",  .decode.addr = A_SMMU_S2CR13,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR14",  .decode.addr = A_SMMU_S2CR14,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR15",  .decode.addr = A_SMMU_S2CR15,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR16",  .decode.addr = A_SMMU_S2CR16,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR17",  .decode.addr = A_SMMU_S2CR17,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR18",  .decode.addr = A_SMMU_S2CR18,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR19",  .decode.addr = A_SMMU_S2CR19,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR20",  .decode.addr = A_SMMU_S2CR20,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR21",  .decode.addr = A_SMMU_S2CR21,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR22",  .decode.addr = A_SMMU_S2CR22,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR23",  .decode.addr = A_SMMU_S2CR23,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR24",  .decode.addr = A_SMMU_S2CR24,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR25",  .decode.addr = A_SMMU_S2CR25,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR26",  .decode.addr = A_SMMU_S2CR26,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR27",  .decode.addr = A_SMMU_S2CR27,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR28",  .decode.addr = A_SMMU_S2CR28,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR29",  .decode.addr = A_SMMU_S2CR29,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR30",  .decode.addr = A_SMMU_S2CR30,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR31",  .decode.addr = A_SMMU_S2CR31,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR32",  .decode.addr = A_SMMU_S2CR32,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR33",  .decode.addr = A_SMMU_S2CR33,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR34",  .decode.addr = A_SMMU_S2CR34,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR35",  .decode.addr = A_SMMU_S2CR35,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR36",  .decode.addr = A_SMMU_S2CR36,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR37",  .decode.addr = A_SMMU_S2CR37,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR38",  .decode.addr = A_SMMU_S2CR38,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR39",  .decode.addr = A_SMMU_S2CR39,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR40",  .decode.addr = A_SMMU_S2CR40,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR41",  .decode.addr = A_SMMU_S2CR41,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR42",  .decode.addr = A_SMMU_S2CR42,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR43",  .decode.addr = A_SMMU_S2CR43,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR44",  .decode.addr = A_SMMU_S2CR44,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR45",  .decode.addr = A_SMMU_S2CR45,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR46",  .decode.addr = A_SMMU_S2CR46,
        .reset = 0x20000,
    },{ .name = "SMMU_S2CR47",  .decode.addr = A_SMMU_S2CR47,
        .reset = 0x20000,
    },{ .name = "SMMU_PIDR4",  .decode.addr = A_SMMU_PIDR4,
        .reset = 0x4,
        .ro = 0xff,
    },{ .name = "SMMU_PIDR5",  .decode.addr = A_SMMU_PIDR5,
        .ro = 0xffffffff,
    },{ .name = "SMMU_PIDR6",  .decode.addr = A_SMMU_PIDR6,
        .ro = 0xffffffff,
    },{ .name = "SMMU_PIDR7",  .decode.addr = A_SMMU_PIDR7,
        .ro = 0xffffffff,
    },{ .name = "SMMU_PIDR0",  .decode.addr = A_SMMU_PIDR0,
        .reset = 0x81,
        .ro = 0xff,
    },{ .name = "SMMU_PIDR1",  .decode.addr = A_SMMU_PIDR1,
        .reset = 0xb4,
        .ro = 0xff,
    },{ .name = "SMMU_PIDR2",  .decode.addr = A_SMMU_PIDR2,
        .reset = 0x1b,
        .ro = 0xff,
    },{ .name = "SMMU_PIDR3",  .decode.addr = A_SMMU_PIDR3,
        .ro = 0xff,
    },{ .name = "SMMU_CIDR0",  .decode.addr = A_SMMU_CIDR0,
        .reset = 0xd,
        .ro = 0xff,
    },{ .name = "SMMU_CIDR1",  .decode.addr = A_SMMU_CIDR1,
        .reset = 0xf0,
        .ro = 0xff,
    },{ .name = "SMMU_CIDR2",  .decode.addr = A_SMMU_CIDR2,
        .reset = 0x5,
        .ro = 0xff,
    },{ .name = "SMMU_CIDR3",  .decode.addr = A_SMMU_CIDR3,
        .reset = 0xb1,
        .ro = 0xff,
    },{ .name = "SMMU_CBAR0",  .decode.addr = A_SMMU_CBAR0,
        .reset = 0x20000,
        .ro = 0xff000000,
    },{ .name = "SMMU_CBAR1",  .decode.addr = A_SMMU_CBAR1,
        .reset = 0x20000,
        .ro = 0xff000000,
    },{ .name = "SMMU_CBAR2",  .decode.addr = A_SMMU_CBAR2,
        .reset = 0x20000,
        .ro = 0xff000000,
    },{ .name = "SMMU_CBAR3",  .decode.addr = A_SMMU_CBAR3,
        .reset = 0x20000,
        .ro = 0xff000000,
    },{ .name = "SMMU_CBAR4",  .decode.addr = A_SMMU_CBAR4,
        .reset = 0x20000,
        .ro = 0xff000000,
    },{ .name = "SMMU_CBAR5",  .decode.addr = A_SMMU_CBAR5,
        .reset = 0x20000,
        .ro = 0xff000000,
    },{ .name = "SMMU_CBAR6",  .decode.addr = A_SMMU_CBAR6,
        .reset = 0x20000,
        .ro = 0xff000000,
    },{ .name = "SMMU_CBAR7",  .decode.addr = A_SMMU_CBAR7,
        .reset = 0x20000,
        .ro = 0xff000000,
    },{ .name = "SMMU_CBAR8",  .decode.addr = A_SMMU_CBAR8,
        .reset = 0x20000,
        .ro = 0xff000000,
    },{ .name = "SMMU_CBAR9",  .decode.addr = A_SMMU_CBAR9,
        .reset = 0x20000,
        .ro = 0xff000000,
    },{ .name = "SMMU_CBAR10",  .decode.addr = A_SMMU_CBAR10,
        .reset = 0x20000,
        .ro = 0xff000000,
    },{ .name = "SMMU_CBAR11",  .decode.addr = A_SMMU_CBAR11,
        .reset = 0x20000,
        .ro = 0xff000000,
    },{ .name = "SMMU_CBAR12",  .decode.addr = A_SMMU_CBAR12,
        .reset = 0x20000,
        .ro = 0xff000000,
    },{ .name = "SMMU_CBAR13",  .decode.addr = A_SMMU_CBAR13,
        .reset = 0x20000,
        .ro = 0xff000000,
    },{ .name = "SMMU_CBAR14",  .decode.addr = A_SMMU_CBAR14,
        .reset = 0x20000,
        .ro = 0xff000000,
    },{ .name = "SMMU_CBAR15",  .decode.addr = A_SMMU_CBAR15,
        .reset = 0x20000,
        .ro = 0xff000000,
    },{ .name = "SMMU_CBFRSYNRA0",  .decode.addr = A_SMMU_CBFRSYNRA0,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_CBFRSYNRA1",  .decode.addr = A_SMMU_CBFRSYNRA1,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_CBFRSYNRA2",  .decode.addr = A_SMMU_CBFRSYNRA2,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_CBFRSYNRA3",  .decode.addr = A_SMMU_CBFRSYNRA3,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_CBFRSYNRA4",  .decode.addr = A_SMMU_CBFRSYNRA4,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_CBFRSYNRA5",  .decode.addr = A_SMMU_CBFRSYNRA5,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_CBFRSYNRA6",  .decode.addr = A_SMMU_CBFRSYNRA6,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_CBFRSYNRA7",  .decode.addr = A_SMMU_CBFRSYNRA7,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_CBFRSYNRA8",  .decode.addr = A_SMMU_CBFRSYNRA8,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_CBFRSYNRA9",  .decode.addr = A_SMMU_CBFRSYNRA9,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_CBFRSYNRA10",  .decode.addr = A_SMMU_CBFRSYNRA10,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_CBFRSYNRA11",  .decode.addr = A_SMMU_CBFRSYNRA11,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_CBFRSYNRA12",  .decode.addr = A_SMMU_CBFRSYNRA12,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_CBFRSYNRA13",  .decode.addr = A_SMMU_CBFRSYNRA13,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_CBFRSYNRA14",  .decode.addr = A_SMMU_CBFRSYNRA14,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_CBFRSYNRA15",  .decode.addr = A_SMMU_CBFRSYNRA15,
        .ro = 0x7fff0000,
    },{ .name = "SMMU_CBA2R0",  .decode.addr = A_SMMU_CBA2R0,
    },{ .name = "SMMU_CBA2R1",  .decode.addr = A_SMMU_CBA2R1,
    },{ .name = "SMMU_CBA2R2",  .decode.addr = A_SMMU_CBA2R2,
    },{ .name = "SMMU_CBA2R3",  .decode.addr = A_SMMU_CBA2R3,
    },{ .name = "SMMU_CBA2R4",  .decode.addr = A_SMMU_CBA2R4,
    },{ .name = "SMMU_CBA2R5",  .decode.addr = A_SMMU_CBA2R5,
    },{ .name = "SMMU_CBA2R6",  .decode.addr = A_SMMU_CBA2R6,
    },{ .name = "SMMU_CBA2R7",  .decode.addr = A_SMMU_CBA2R7,
    },{ .name = "SMMU_CBA2R8",  .decode.addr = A_SMMU_CBA2R8,
    },{ .name = "SMMU_CBA2R9",  .decode.addr = A_SMMU_CBA2R9,
    },{ .name = "SMMU_CBA2R10",  .decode.addr = A_SMMU_CBA2R10,
    },{ .name = "SMMU_CBA2R11",  .decode.addr = A_SMMU_CBA2R11,
    },{ .name = "SMMU_CBA2R12",  .decode.addr = A_SMMU_CBA2R12,
    },{ .name = "SMMU_CBA2R13",  .decode.addr = A_SMMU_CBA2R13,
    },{ .name = "SMMU_CBA2R14",  .decode.addr = A_SMMU_CBA2R14,
    },{ .name = "SMMU_CBA2R15",  .decode.addr = A_SMMU_CBA2R15,
    },{ .name = "SMMU_ITCTRL",  .decode.addr = A_SMMU_ITCTRL,
    },{ .name = "SMMU_ITIP",  .decode.addr = A_SMMU_ITIP,
        .ro = 0x1,
    },{ .name = "SMMU_ITOP_GLBL",  .decode.addr = A_SMMU_ITOP_GLBL,
        .ro = 0x202,
    },{ .name = "SMMU_ITOP_PERF_INDEX",  .decode.addr = A_SMMU_ITOP_PERF_INDEX,
    },{ .name = "SMMU_ITOP_CXT0TO31_RAM0",  .decode.addr = A_SMMU_ITOP_CXT0TO31_RAM0,
    },{ .name = "SMMU_TBUQOS0",  .decode.addr = A_SMMU_TBUQOS0,
    },{ .name = "SMMU_PER",  .decode.addr = A_SMMU_PER,
        .ro = 0xffff,
    },{ .name = "SMMU_TBU_PWR_STATUS",  .decode.addr = A_SMMU_TBU_PWR_STATUS,
        .ro = 0xffffffff,
    },{ .name = "PMEVCNTR0",  .decode.addr = A_PMEVCNTR0,
    },{ .name = "PMEVCNTR1",  .decode.addr = A_PMEVCNTR1,
    },{ .name = "PMEVCNTR2",  .decode.addr = A_PMEVCNTR2,
    },{ .name = "PMEVCNTR3",  .decode.addr = A_PMEVCNTR3,
    },{ .name = "PMEVCNTR4",  .decode.addr = A_PMEVCNTR4,
    },{ .name = "PMEVCNTR5",  .decode.addr = A_PMEVCNTR5,
    },{ .name = "PMEVCNTR6",  .decode.addr = A_PMEVCNTR6,
    },{ .name = "PMEVCNTR7",  .decode.addr = A_PMEVCNTR7,
    },{ .name = "PMEVCNTR8",  .decode.addr = A_PMEVCNTR8,
    },{ .name = "PMEVCNTR9",  .decode.addr = A_PMEVCNTR9,
    },{ .name = "PMEVCNTR10",  .decode.addr = A_PMEVCNTR10,
    },{ .name = "PMEVCNTR11",  .decode.addr = A_PMEVCNTR11,
    },{ .name = "PMEVCNTR12",  .decode.addr = A_PMEVCNTR12,
    },{ .name = "PMEVCNTR13",  .decode.addr = A_PMEVCNTR13,
    },{ .name = "PMEVCNTR14",  .decode.addr = A_PMEVCNTR14,
    },{ .name = "PMEVCNTR15",  .decode.addr = A_PMEVCNTR15,
    },{ .name = "PMEVCNTR16",  .decode.addr = A_PMEVCNTR16,
    },{ .name = "PMEVCNTR17",  .decode.addr = A_PMEVCNTR17,
    },{ .name = "PMEVCNTR18",  .decode.addr = A_PMEVCNTR18,
    },{ .name = "PMEVCNTR19",  .decode.addr = A_PMEVCNTR19,
    },{ .name = "PMEVCNTR20",  .decode.addr = A_PMEVCNTR20,
    },{ .name = "PMEVCNTR21",  .decode.addr = A_PMEVCNTR21,
    },{ .name = "PMEVCNTR22",  .decode.addr = A_PMEVCNTR22,
    },{ .name = "PMEVCNTR23",  .decode.addr = A_PMEVCNTR23,
    },{ .name = "PMEVTYPER0",  .decode.addr = A_PMEVTYPER0,
    },{ .name = "PMEVTYPER1",  .decode.addr = A_PMEVTYPER1,
    },{ .name = "PMEVTYPER2",  .decode.addr = A_PMEVTYPER2,
    },{ .name = "PMEVTYPER3",  .decode.addr = A_PMEVTYPER3,
    },{ .name = "PMEVTYPER4",  .decode.addr = A_PMEVTYPER4,
    },{ .name = "PMEVTYPER5",  .decode.addr = A_PMEVTYPER5,
    },{ .name = "PMEVTYPER6",  .decode.addr = A_PMEVTYPER6,
    },{ .name = "PMEVTYPER7",  .decode.addr = A_PMEVTYPER7,
    },{ .name = "PMEVTYPER8",  .decode.addr = A_PMEVTYPER8,
    },{ .name = "PMEVTYPER9",  .decode.addr = A_PMEVTYPER9,
    },{ .name = "PMEVTYPER10",  .decode.addr = A_PMEVTYPER10,
    },{ .name = "PMEVTYPER11",  .decode.addr = A_PMEVTYPER11,
    },{ .name = "PMEVTYPER12",  .decode.addr = A_PMEVTYPER12,
    },{ .name = "PMEVTYPER13",  .decode.addr = A_PMEVTYPER13,
    },{ .name = "PMEVTYPER14",  .decode.addr = A_PMEVTYPER14,
    },{ .name = "PMEVTYPER15",  .decode.addr = A_PMEVTYPER15,
    },{ .name = "PMEVTYPER16",  .decode.addr = A_PMEVTYPER16,
    },{ .name = "PMEVTYPER17",  .decode.addr = A_PMEVTYPER17,
    },{ .name = "PMEVTYPER18",  .decode.addr = A_PMEVTYPER18,
    },{ .name = "PMEVTYPER19",  .decode.addr = A_PMEVTYPER19,
    },{ .name = "PMEVTYPER20",  .decode.addr = A_PMEVTYPER20,
    },{ .name = "PMEVTYPER21",  .decode.addr = A_PMEVTYPER21,
    },{ .name = "PMEVTYPER22",  .decode.addr = A_PMEVTYPER22,
    },{ .name = "PMEVTYPER23",  .decode.addr = A_PMEVTYPER23,
    },{ .name = "PMCGCR0",  .decode.addr = A_PMCGCR0,
        .reset = 0x4000000,
        .ro = 0xf7f0000,
    },{ .name = "PMCGCR1",  .decode.addr = A_PMCGCR1,
        .reset = 0x4010000,
        .ro = 0xf7f0000,
    },{ .name = "PMCGCR2",  .decode.addr = A_PMCGCR2,
        .reset = 0x4020000,
        .ro = 0xf7f0000,
    },{ .name = "PMCGCR3",  .decode.addr = A_PMCGCR3,
        .reset = 0x4030000,
        .ro = 0xf7f0000,
    },{ .name = "PMCGCR4",  .decode.addr = A_PMCGCR4,
        .reset = 0x4040000,
        .ro = 0xf7f0000,
    },{ .name = "PMCGCR5",  .decode.addr = A_PMCGCR5,
        .reset = 0x4050000,
        .ro = 0xf7f0000,
    },{ .name = "PMCGSMR0",  .decode.addr = A_PMCGSMR0,
    },{ .name = "PMCGSMR1",  .decode.addr = A_PMCGSMR1,
    },{ .name = "PMCGSMR2",  .decode.addr = A_PMCGSMR2,
    },{ .name = "PMCGSMR3",  .decode.addr = A_PMCGSMR3,
    },{ .name = "PMCGSMR4",  .decode.addr = A_PMCGSMR4,
    },{ .name = "PMCGSMR5",  .decode.addr = A_PMCGSMR5,
    },{ .name = "PMCNTENSET",  .decode.addr = A_PMCNTENSET,
    },{ .name = "PMCNTENCLR",  .decode.addr = A_PMCNTENCLR,
    },{ .name = "PMINTENSET",  .decode.addr = A_PMINTENSET,
    },{ .name = "PMINTENCLR",  .decode.addr = A_PMINTENCLR,
    },{ .name = "PMOVSCLR",  .decode.addr = A_PMOVSCLR,
    },{ .name = "PMOVSSET",  .decode.addr = A_PMOVSSET,
    },{ .name = "PMCFGR",  .decode.addr = A_PMCFGR,
        .reset = 0x5011f17,
        .ro = 0xff09ffff,
    },{ .name = "PMCR",  .decode.addr = A_PMCR,
        .ro = 0xff000002,
    },{ .name = "PMCEID0",  .decode.addr = A_PMCEID0,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "PMAUTHSTATUS",  .decode.addr = A_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    },{ .name = "PMDEVTYPE",  .decode.addr = A_PMDEVTYPE,
        .reset = 0x56,
        .ro = 0xff,
    },{ .name = "SMMU_CB0_SCTLR",  .decode.addr = A_SMMU_CB0_SCTLR,
        .reset = 0x100,
        .ro = 0x1000,
    },{ .name = "SMMU_CB0_ACTLR",  .decode.addr = A_SMMU_CB0_ACTLR,
        .reset = 0x3,
    },{ .name = "SMMU_CB0_RESUME",  .decode.addr = A_SMMU_CB0_RESUME,
    },{ .name = "SMMU_CB0_TCR2",  .decode.addr = A_SMMU_CB0_TCR2,
        .reset = 0x60,
        .ro = 0x60,
    },{ .name = "SMMU_CB0_TTBR0_LOW",  .decode.addr = A_SMMU_CB0_TTBR0_LOW,
        .ro = 0x4,
    },{ .name = "SMMU_CB0_TTBR0_HIGH",  .decode.addr = A_SMMU_CB0_TTBR0_HIGH,
    },{ .name = "SMMU_CB0_TTBR1_LOW",  .decode.addr = A_SMMU_CB0_TTBR1_LOW,
    },{ .name = "SMMU_CB0_TTBR1_HIGH",  .decode.addr = A_SMMU_CB0_TTBR1_HIGH,
    },{ .name = "SMMU_CB0_TCR_LPAE",  .decode.addr = A_SMMU_CB0_TCR_LPAE,
    },{ .name = "SMMU_CB0_CONTEXTIDR",  .decode.addr = A_SMMU_CB0_CONTEXTIDR,
    },{ .name = "SMMU_CB0_PRRR_MAIR0",  .decode.addr = A_SMMU_CB0_PRRR_MAIR0,
    },{ .name = "SMMU_CB0_NMRR_MAIR1",  .decode.addr = A_SMMU_CB0_NMRR_MAIR1,
    },{ .name = "SMMU_CB0_FSR",  .decode.addr = A_SMMU_CB0_FSR,
        .w1c = 0xffffffff,
        .post_write = smmu_fsr_pw,
    },{ .name = "SMMU_CB0_FSRRESTORE",  .decode.addr = A_SMMU_CB0_FSRRESTORE,
    },{ .name = "SMMU_CB0_FAR_LOW",  .decode.addr = A_SMMU_CB0_FAR_LOW,
    },{ .name = "SMMU_CB0_FAR_HIGH",  .decode.addr = A_SMMU_CB0_FAR_HIGH,
    },{ .name = "SMMU_CB0_FSYNR0",  .decode.addr = A_SMMU_CB0_FSYNR0,
        .ro = 0x200,
    },{ .name = "SMMU_CB0_IPAFAR_LOW",  .decode.addr = A_SMMU_CB0_IPAFAR_LOW,
        .ro = 0xfff,
    },{ .name = "SMMU_CB0_IPAFAR_HIGH",  .decode.addr = A_SMMU_CB0_IPAFAR_HIGH,
    },{ .name = "SMMU_CB0_TLBIVA_LOW",  .decode.addr = A_SMMU_CB0_TLBIVA_LOW,
    },{ .name = "SMMU_CB0_TLBIVA_HIGH",  .decode.addr = A_SMMU_CB0_TLBIVA_HIGH,
    },{ .name = "SMMU_CB0_TLBIVAA_LOW",  .decode.addr = A_SMMU_CB0_TLBIVAA_LOW,
    },{ .name = "SMMU_CB0_TLBIVAA_HIGH",  .decode.addr = A_SMMU_CB0_TLBIVAA_HIGH,
    },{ .name = "SMMU_CB0_TLBIASID",  .decode.addr = A_SMMU_CB0_TLBIASID,
    },{ .name = "SMMU_CB0_TLBIALL",  .decode.addr = A_SMMU_CB0_TLBIALL,
    },{ .name = "SMMU_CB0_TLBIVAL_LOW",  .decode.addr = A_SMMU_CB0_TLBIVAL_LOW,
    },{ .name = "SMMU_CB0_TLBIVAL_HIGH",  .decode.addr = A_SMMU_CB0_TLBIVAL_HIGH,
    },{ .name = "SMMU_CB0_TLBIVAAL_LOW",  .decode.addr = A_SMMU_CB0_TLBIVAAL_LOW,
    },{ .name = "SMMU_CB0_TLBIVAAL_HIGH",  .decode.addr = A_SMMU_CB0_TLBIVAAL_HIGH,
    },{ .name = "SMMU_CB0_TLBIIPAS2_LOW",  .decode.addr = A_SMMU_CB0_TLBIIPAS2_LOW,
    },{ .name = "SMMU_CB0_TLBIIPAS2_HIGH",  .decode.addr = A_SMMU_CB0_TLBIIPAS2_HIGH,
    },{ .name = "SMMU_CB0_TLBIIPAS2L_LOW",  .decode.addr = A_SMMU_CB0_TLBIIPAS2L_LOW,
    },{ .name = "SMMU_CB0_TLBIIPAS2L_HIGH",  .decode.addr = A_SMMU_CB0_TLBIIPAS2L_HIGH,
    },{ .name = "SMMU_CB0_TLBSYNC",  .decode.addr = A_SMMU_CB0_TLBSYNC,
    },{ .name = "SMMU_CB0_TLBSTATUS",  .decode.addr = A_SMMU_CB0_TLBSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_CB0_PMEVCNTR0",  .decode.addr = A_SMMU_CB0_PMEVCNTR0,
    },{ .name = "SMMU_CB0_PMEVCNTR1",  .decode.addr = A_SMMU_CB0_PMEVCNTR1,
    },{ .name = "SMMU_CB0_PMEVCNTR2",  .decode.addr = A_SMMU_CB0_PMEVCNTR2,
    },{ .name = "SMMU_CB0_PMEVCNTR3",  .decode.addr = A_SMMU_CB0_PMEVCNTR3,
    },{ .name = "SMMU_CB0_PMEVTYPER0",  .decode.addr = A_SMMU_CB0_PMEVTYPER0,
    },{ .name = "SMMU_CB0_PMEVTYPER1",  .decode.addr = A_SMMU_CB0_PMEVTYPER1,
    },{ .name = "SMMU_CB0_PMEVTYPER2",  .decode.addr = A_SMMU_CB0_PMEVTYPER2,
    },{ .name = "SMMU_CB0_PMEVTYPER3",  .decode.addr = A_SMMU_CB0_PMEVTYPER3,
    },{ .name = "SMMU_CB0_PMCFGR",  .decode.addr = A_SMMU_CB0_PMCFGR,
        .reset = 0x11f03,
        .ro = 0xff09ffff,
    },{ .name = "SMMU_CB0_PMCR",  .decode.addr = A_SMMU_CB0_PMCR,
        .ro = 0xff000002,
    },{ .name = "SMMU_CB0_PMCEID",  .decode.addr = A_SMMU_CB0_PMCEID,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "SMMU_CB0_PMCNTENSE",  .decode.addr = A_SMMU_CB0_PMCNTENSE,
    },{ .name = "SMMU_CB0_PMCNTENCLR",  .decode.addr = A_SMMU_CB0_PMCNTENCLR,
    },{ .name = "SMMU_CB0_PMCNTENSET",  .decode.addr = A_SMMU_CB0_PMCNTENSET,
    },{ .name = "SMMU_CB0_PMINTENCLR",  .decode.addr = A_SMMU_CB0_PMINTENCLR,
    },{ .name = "SMMU_CB0_PMOVSCLR",  .decode.addr = A_SMMU_CB0_PMOVSCLR,
    },{ .name = "SMMU_CB0_PMOVSSET",  .decode.addr = A_SMMU_CB0_PMOVSSET,
    },{ .name = "SMMU_CB0_PMAUTHSTATUS",  .decode.addr = A_SMMU_CB0_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    },{ .name = "SMMU_CB1_SCTLR",  .decode.addr = A_SMMU_CB1_SCTLR,
        .reset = 0x100,
        .ro = 0x1000,
    },{ .name = "SMMU_CB1_ACTLR",  .decode.addr = A_SMMU_CB1_ACTLR,
        .reset = 0x3,
    },{ .name = "SMMU_CB1_RESUME",  .decode.addr = A_SMMU_CB1_RESUME,
    },{ .name = "SMMU_CB1_TCR2",  .decode.addr = A_SMMU_CB1_TCR2,
        .reset = 0x60,
        .ro = 0x60,
    },{ .name = "SMMU_CB1_TTBR0_LOW",  .decode.addr = A_SMMU_CB1_TTBR0_LOW,
        .ro = 0x4,
    },{ .name = "SMMU_CB1_TTBR0_HIGH",  .decode.addr = A_SMMU_CB1_TTBR0_HIGH,
    },{ .name = "SMMU_CB1_TTBR1_LOW",  .decode.addr = A_SMMU_CB1_TTBR1_LOW,
    },{ .name = "SMMU_CB1_TTBR1_HIGH",  .decode.addr = A_SMMU_CB1_TTBR1_HIGH,
    },{ .name = "SMMU_CB1_TCR_LPAE",  .decode.addr = A_SMMU_CB1_TCR_LPAE,
    },{ .name = "SMMU_CB1_CONTEXTIDR",  .decode.addr = A_SMMU_CB1_CONTEXTIDR,
    },{ .name = "SMMU_CB1_PRRR_MAIR0",  .decode.addr = A_SMMU_CB1_PRRR_MAIR0,
    },{ .name = "SMMU_CB1_NMRR_MAIR1",  .decode.addr = A_SMMU_CB1_NMRR_MAIR1,
    },{ .name = "SMMU_CB1_FSR",  .decode.addr = A_SMMU_CB1_FSR,
        .w1c = 0xffffffff,
        .post_write = smmu_fsr_pw,
    },{ .name = "SMMU_CB1_FSRRESTORE",  .decode.addr = A_SMMU_CB1_FSRRESTORE,
    },{ .name = "SMMU_CB1_FAR_LOW",  .decode.addr = A_SMMU_CB1_FAR_LOW,
    },{ .name = "SMMU_CB1_FAR_HIGH",  .decode.addr = A_SMMU_CB1_FAR_HIGH,
    },{ .name = "SMMU_CB1_FSYNR0",  .decode.addr = A_SMMU_CB1_FSYNR0,
        .ro = 0x200,
    },{ .name = "SMMU_CB1_IPAFAR_LOW",  .decode.addr = A_SMMU_CB1_IPAFAR_LOW,
        .ro = 0xfff,
    },{ .name = "SMMU_CB1_IPAFAR_HIGH",  .decode.addr = A_SMMU_CB1_IPAFAR_HIGH,
    },{ .name = "SMMU_CB1_TLBIVA_LOW",  .decode.addr = A_SMMU_CB1_TLBIVA_LOW,
    },{ .name = "SMMU_CB1_TLBIVA_HIGH",  .decode.addr = A_SMMU_CB1_TLBIVA_HIGH,
    },{ .name = "SMMU_CB1_TLBIVAA_LOW",  .decode.addr = A_SMMU_CB1_TLBIVAA_LOW,
    },{ .name = "SMMU_CB1_TLBIVAA_HIGH",  .decode.addr = A_SMMU_CB1_TLBIVAA_HIGH,
    },{ .name = "SMMU_CB1_TLBIASID",  .decode.addr = A_SMMU_CB1_TLBIASID,
    },{ .name = "SMMU_CB1_TLBIALL",  .decode.addr = A_SMMU_CB1_TLBIALL,
    },{ .name = "SMMU_CB1_TLBIVAL_LOW",  .decode.addr = A_SMMU_CB1_TLBIVAL_LOW,
    },{ .name = "SMMU_CB1_TLBIVAL_HIGH",  .decode.addr = A_SMMU_CB1_TLBIVAL_HIGH,
    },{ .name = "SMMU_CB1_TLBIVAAL_LOW",  .decode.addr = A_SMMU_CB1_TLBIVAAL_LOW,
    },{ .name = "SMMU_CB1_TLBIVAAL_HIGH",  .decode.addr = A_SMMU_CB1_TLBIVAAL_HIGH,
    },{ .name = "SMMU_CB1_TLBIIPAS2_LOW",  .decode.addr = A_SMMU_CB1_TLBIIPAS2_LOW,
    },{ .name = "SMMU_CB1_TLBIIPAS2_HIGH",  .decode.addr = A_SMMU_CB1_TLBIIPAS2_HIGH,
    },{ .name = "SMMU_CB1_TLBIIPAS2L_LOW",  .decode.addr = A_SMMU_CB1_TLBIIPAS2L_LOW,
    },{ .name = "SMMU_CB1_TLBIIPAS2L_HIGH",  .decode.addr = A_SMMU_CB1_TLBIIPAS2L_HIGH,
    },{ .name = "SMMU_CB1_TLBSYNC",  .decode.addr = A_SMMU_CB1_TLBSYNC,
    },{ .name = "SMMU_CB1_TLBSTATUS",  .decode.addr = A_SMMU_CB1_TLBSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_CB1_PMEVCNTR0",  .decode.addr = A_SMMU_CB1_PMEVCNTR0,
    },{ .name = "SMMU_CB1_PMEVCNTR1",  .decode.addr = A_SMMU_CB1_PMEVCNTR1,
    },{ .name = "SMMU_CB1_PMEVCNTR2",  .decode.addr = A_SMMU_CB1_PMEVCNTR2,
    },{ .name = "SMMU_CB1_PMEVCNTR3",  .decode.addr = A_SMMU_CB1_PMEVCNTR3,
    },{ .name = "SMMU_CB1_PMEVTYPER0",  .decode.addr = A_SMMU_CB1_PMEVTYPER0,
    },{ .name = "SMMU_CB1_PMEVTYPER1",  .decode.addr = A_SMMU_CB1_PMEVTYPER1,
    },{ .name = "SMMU_CB1_PMEVTYPER2",  .decode.addr = A_SMMU_CB1_PMEVTYPER2,
    },{ .name = "SMMU_CB1_PMEVTYPER3",  .decode.addr = A_SMMU_CB1_PMEVTYPER3,
    },{ .name = "SMMU_CB1_PMCFGR",  .decode.addr = A_SMMU_CB1_PMCFGR,
        .reset = 0x11f03,
        .ro = 0xff09ffff,
    },{ .name = "SMMU_CB1_PMCR",  .decode.addr = A_SMMU_CB1_PMCR,
        .ro = 0xff000002,
    },{ .name = "SMMU_CB1_PMCEID",  .decode.addr = A_SMMU_CB1_PMCEID,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "SMMU_CB1_PMCNTENSE",  .decode.addr = A_SMMU_CB1_PMCNTENSE,
    },{ .name = "SMMU_CB1_PMCNTENCLR",  .decode.addr = A_SMMU_CB1_PMCNTENCLR,
    },{ .name = "SMMU_CB1_PMCNTENSET",  .decode.addr = A_SMMU_CB1_PMCNTENSET,
    },{ .name = "SMMU_CB1_PMINTENCLR",  .decode.addr = A_SMMU_CB1_PMINTENCLR,
    },{ .name = "SMMU_CB1_PMOVSCLR",  .decode.addr = A_SMMU_CB1_PMOVSCLR,
    },{ .name = "SMMU_CB1_PMOVSSET",  .decode.addr = A_SMMU_CB1_PMOVSSET,
    },{ .name = "SMMU_CB1_PMAUTHSTATUS",  .decode.addr = A_SMMU_CB1_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    },{ .name = "SMMU_CB2_SCTLR",  .decode.addr = A_SMMU_CB2_SCTLR,
        .reset = 0x100,
        .ro = 0x1000,
    },{ .name = "SMMU_CB2_ACTLR",  .decode.addr = A_SMMU_CB2_ACTLR,
        .reset = 0x3,
    },{ .name = "SMMU_CB2_RESUME",  .decode.addr = A_SMMU_CB2_RESUME,
    },{ .name = "SMMU_CB2_TCR2",  .decode.addr = A_SMMU_CB2_TCR2,
        .reset = 0x60,
        .ro = 0x60,
    },{ .name = "SMMU_CB2_TTBR0_LOW",  .decode.addr = A_SMMU_CB2_TTBR0_LOW,
        .ro = 0x4,
    },{ .name = "SMMU_CB2_TTBR0_HIGH",  .decode.addr = A_SMMU_CB2_TTBR0_HIGH,
    },{ .name = "SMMU_CB2_TTBR1_LOW",  .decode.addr = A_SMMU_CB2_TTBR1_LOW,
    },{ .name = "SMMU_CB2_TTBR1_HIGH",  .decode.addr = A_SMMU_CB2_TTBR1_HIGH,
    },{ .name = "SMMU_CB2_TCR_LPAE",  .decode.addr = A_SMMU_CB2_TCR_LPAE,
    },{ .name = "SMMU_CB2_CONTEXTIDR",  .decode.addr = A_SMMU_CB2_CONTEXTIDR,
    },{ .name = "SMMU_CB2_PRRR_MAIR0",  .decode.addr = A_SMMU_CB2_PRRR_MAIR0,
    },{ .name = "SMMU_CB2_NMRR_MAIR1",  .decode.addr = A_SMMU_CB2_NMRR_MAIR1,
    },{ .name = "SMMU_CB2_FSR",  .decode.addr = A_SMMU_CB2_FSR,
        .w1c = 0xffffffff,
        .post_write = smmu_fsr_pw,
    },{ .name = "SMMU_CB2_FSRRESTORE",  .decode.addr = A_SMMU_CB2_FSRRESTORE,
    },{ .name = "SMMU_CB2_FAR_LOW",  .decode.addr = A_SMMU_CB2_FAR_LOW,
    },{ .name = "SMMU_CB2_FAR_HIGH",  .decode.addr = A_SMMU_CB2_FAR_HIGH,
    },{ .name = "SMMU_CB2_FSYNR0",  .decode.addr = A_SMMU_CB2_FSYNR0,
        .ro = 0x200,
    },{ .name = "SMMU_CB2_IPAFAR_LOW",  .decode.addr = A_SMMU_CB2_IPAFAR_LOW,
        .ro = 0xfff,
    },{ .name = "SMMU_CB2_IPAFAR_HIGH",  .decode.addr = A_SMMU_CB2_IPAFAR_HIGH,
    },{ .name = "SMMU_CB2_TLBIVA_LOW",  .decode.addr = A_SMMU_CB2_TLBIVA_LOW,
    },{ .name = "SMMU_CB2_TLBIVA_HIGH",  .decode.addr = A_SMMU_CB2_TLBIVA_HIGH,
    },{ .name = "SMMU_CB2_TLBIVAA_LOW",  .decode.addr = A_SMMU_CB2_TLBIVAA_LOW,
    },{ .name = "SMMU_CB2_TLBIVAA_HIGH",  .decode.addr = A_SMMU_CB2_TLBIVAA_HIGH,
    },{ .name = "SMMU_CB2_TLBIASID",  .decode.addr = A_SMMU_CB2_TLBIASID,
    },{ .name = "SMMU_CB2_TLBIALL",  .decode.addr = A_SMMU_CB2_TLBIALL,
    },{ .name = "SMMU_CB2_TLBIVAL_LOW",  .decode.addr = A_SMMU_CB2_TLBIVAL_LOW,
    },{ .name = "SMMU_CB2_TLBIVAL_HIGH",  .decode.addr = A_SMMU_CB2_TLBIVAL_HIGH,
    },{ .name = "SMMU_CB2_TLBIVAAL_LOW",  .decode.addr = A_SMMU_CB2_TLBIVAAL_LOW,
    },{ .name = "SMMU_CB2_TLBIVAAL_HIGH",  .decode.addr = A_SMMU_CB2_TLBIVAAL_HIGH,
    },{ .name = "SMMU_CB2_TLBIIPAS2_LOW",  .decode.addr = A_SMMU_CB2_TLBIIPAS2_LOW,
    },{ .name = "SMMU_CB2_TLBIIPAS2_HIGH",  .decode.addr = A_SMMU_CB2_TLBIIPAS2_HIGH,
    },{ .name = "SMMU_CB2_TLBIIPAS2L_LOW",  .decode.addr = A_SMMU_CB2_TLBIIPAS2L_LOW,
    },{ .name = "SMMU_CB2_TLBIIPAS2L_HIGH",  .decode.addr = A_SMMU_CB2_TLBIIPAS2L_HIGH,
    },{ .name = "SMMU_CB2_TLBSYNC",  .decode.addr = A_SMMU_CB2_TLBSYNC,
    },{ .name = "SMMU_CB2_TLBSTATUS",  .decode.addr = A_SMMU_CB2_TLBSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_CB2_PMEVCNTR0",  .decode.addr = A_SMMU_CB2_PMEVCNTR0,
    },{ .name = "SMMU_CB2_PMEVCNTR1",  .decode.addr = A_SMMU_CB2_PMEVCNTR1,
    },{ .name = "SMMU_CB2_PMEVCNTR2",  .decode.addr = A_SMMU_CB2_PMEVCNTR2,
    },{ .name = "SMMU_CB2_PMEVCNTR3",  .decode.addr = A_SMMU_CB2_PMEVCNTR3,
    },{ .name = "SMMU_CB2_PMEVTYPER0",  .decode.addr = A_SMMU_CB2_PMEVTYPER0,
    },{ .name = "SMMU_CB2_PMEVTYPER1",  .decode.addr = A_SMMU_CB2_PMEVTYPER1,
    },{ .name = "SMMU_CB2_PMEVTYPER2",  .decode.addr = A_SMMU_CB2_PMEVTYPER2,
    },{ .name = "SMMU_CB2_PMEVTYPER3",  .decode.addr = A_SMMU_CB2_PMEVTYPER3,
    },{ .name = "SMMU_CB2_PMCFGR",  .decode.addr = A_SMMU_CB2_PMCFGR,
        .reset = 0x11f03,
        .ro = 0xff09ffff,
    },{ .name = "SMMU_CB2_PMCR",  .decode.addr = A_SMMU_CB2_PMCR,
        .ro = 0xff000002,
    },{ .name = "SMMU_CB2_PMCEID",  .decode.addr = A_SMMU_CB2_PMCEID,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "SMMU_CB2_PMCNTENSE",  .decode.addr = A_SMMU_CB2_PMCNTENSE,
    },{ .name = "SMMU_CB2_PMCNTENCLR",  .decode.addr = A_SMMU_CB2_PMCNTENCLR,
    },{ .name = "SMMU_CB2_PMCNTENSET",  .decode.addr = A_SMMU_CB2_PMCNTENSET,
    },{ .name = "SMMU_CB2_PMINTENCLR",  .decode.addr = A_SMMU_CB2_PMINTENCLR,
    },{ .name = "SMMU_CB2_PMOVSCLR",  .decode.addr = A_SMMU_CB2_PMOVSCLR,
    },{ .name = "SMMU_CB2_PMOVSSET",  .decode.addr = A_SMMU_CB2_PMOVSSET,
    },{ .name = "SMMU_CB2_PMAUTHSTATUS",  .decode.addr = A_SMMU_CB2_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    },{ .name = "SMMU_CB3_SCTLR",  .decode.addr = A_SMMU_CB3_SCTLR,
        .reset = 0x100,
        .ro = 0x1000,
    },{ .name = "SMMU_CB3_ACTLR",  .decode.addr = A_SMMU_CB3_ACTLR,
        .reset = 0x3,
    },{ .name = "SMMU_CB3_RESUME",  .decode.addr = A_SMMU_CB3_RESUME,
    },{ .name = "SMMU_CB3_TCR2",  .decode.addr = A_SMMU_CB3_TCR2,
        .reset = 0x60,
        .ro = 0x60,
    },{ .name = "SMMU_CB3_TTBR0_LOW",  .decode.addr = A_SMMU_CB3_TTBR0_LOW,
        .ro = 0x4,
    },{ .name = "SMMU_CB3_TTBR0_HIGH",  .decode.addr = A_SMMU_CB3_TTBR0_HIGH,
    },{ .name = "SMMU_CB3_TTBR1_LOW",  .decode.addr = A_SMMU_CB3_TTBR1_LOW,
    },{ .name = "SMMU_CB3_TTBR1_HIGH",  .decode.addr = A_SMMU_CB3_TTBR1_HIGH,
    },{ .name = "SMMU_CB3_TCR_LPAE",  .decode.addr = A_SMMU_CB3_TCR_LPAE,
    },{ .name = "SMMU_CB3_CONTEXTIDR",  .decode.addr = A_SMMU_CB3_CONTEXTIDR,
    },{ .name = "SMMU_CB3_PRRR_MAIR0",  .decode.addr = A_SMMU_CB3_PRRR_MAIR0,
    },{ .name = "SMMU_CB3_NMRR_MAIR1",  .decode.addr = A_SMMU_CB3_NMRR_MAIR1,
    },{ .name = "SMMU_CB3_FSR",  .decode.addr = A_SMMU_CB3_FSR,
        .w1c = 0xffffffff,
        .post_write = smmu_fsr_pw,
    },{ .name = "SMMU_CB3_FSRRESTORE",  .decode.addr = A_SMMU_CB3_FSRRESTORE,
    },{ .name = "SMMU_CB3_FAR_LOW",  .decode.addr = A_SMMU_CB3_FAR_LOW,
    },{ .name = "SMMU_CB3_FAR_HIGH",  .decode.addr = A_SMMU_CB3_FAR_HIGH,
    },{ .name = "SMMU_CB3_FSYNR0",  .decode.addr = A_SMMU_CB3_FSYNR0,
        .ro = 0x200,
    },{ .name = "SMMU_CB3_IPAFAR_LOW",  .decode.addr = A_SMMU_CB3_IPAFAR_LOW,
        .ro = 0xfff,
    },{ .name = "SMMU_CB3_IPAFAR_HIGH",  .decode.addr = A_SMMU_CB3_IPAFAR_HIGH,
    },{ .name = "SMMU_CB3_TLBIVA_LOW",  .decode.addr = A_SMMU_CB3_TLBIVA_LOW,
    },{ .name = "SMMU_CB3_TLBIVA_HIGH",  .decode.addr = A_SMMU_CB3_TLBIVA_HIGH,
    },{ .name = "SMMU_CB3_TLBIVAA_LOW",  .decode.addr = A_SMMU_CB3_TLBIVAA_LOW,
    },{ .name = "SMMU_CB3_TLBIVAA_HIGH",  .decode.addr = A_SMMU_CB3_TLBIVAA_HIGH,
    },{ .name = "SMMU_CB3_TLBIASID",  .decode.addr = A_SMMU_CB3_TLBIASID,
    },{ .name = "SMMU_CB3_TLBIALL",  .decode.addr = A_SMMU_CB3_TLBIALL,
    },{ .name = "SMMU_CB3_TLBIVAL_LOW",  .decode.addr = A_SMMU_CB3_TLBIVAL_LOW,
    },{ .name = "SMMU_CB3_TLBIVAL_HIGH",  .decode.addr = A_SMMU_CB3_TLBIVAL_HIGH,
    },{ .name = "SMMU_CB3_TLBIVAAL_LOW",  .decode.addr = A_SMMU_CB3_TLBIVAAL_LOW,
    },{ .name = "SMMU_CB3_TLBIVAAL_HIGH",  .decode.addr = A_SMMU_CB3_TLBIVAAL_HIGH,
    },{ .name = "SMMU_CB3_TLBIIPAS2_LOW",  .decode.addr = A_SMMU_CB3_TLBIIPAS2_LOW,
    },{ .name = "SMMU_CB3_TLBIIPAS2_HIGH",  .decode.addr = A_SMMU_CB3_TLBIIPAS2_HIGH,
    },{ .name = "SMMU_CB3_TLBIIPAS2L_LOW",  .decode.addr = A_SMMU_CB3_TLBIIPAS2L_LOW,
    },{ .name = "SMMU_CB3_TLBIIPAS2L_HIGH",  .decode.addr = A_SMMU_CB3_TLBIIPAS2L_HIGH,
    },{ .name = "SMMU_CB3_TLBSYNC",  .decode.addr = A_SMMU_CB3_TLBSYNC,
    },{ .name = "SMMU_CB3_TLBSTATUS",  .decode.addr = A_SMMU_CB3_TLBSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_CB3_PMEVCNTR0",  .decode.addr = A_SMMU_CB3_PMEVCNTR0,
    },{ .name = "SMMU_CB3_PMEVCNTR1",  .decode.addr = A_SMMU_CB3_PMEVCNTR1,
    },{ .name = "SMMU_CB3_PMEVCNTR2",  .decode.addr = A_SMMU_CB3_PMEVCNTR2,
    },{ .name = "SMMU_CB3_PMEVCNTR3",  .decode.addr = A_SMMU_CB3_PMEVCNTR3,
    },{ .name = "SMMU_CB3_PMEVTYPER0",  .decode.addr = A_SMMU_CB3_PMEVTYPER0,
    },{ .name = "SMMU_CB3_PMEVTYPER1",  .decode.addr = A_SMMU_CB3_PMEVTYPER1,
    },{ .name = "SMMU_CB3_PMEVTYPER2",  .decode.addr = A_SMMU_CB3_PMEVTYPER2,
    },{ .name = "SMMU_CB3_PMEVTYPER3",  .decode.addr = A_SMMU_CB3_PMEVTYPER3,
    },{ .name = "SMMU_CB3_PMCFGR",  .decode.addr = A_SMMU_CB3_PMCFGR,
        .reset = 0x11f03,
        .ro = 0xff09ffff,
    },{ .name = "SMMU_CB3_PMCR",  .decode.addr = A_SMMU_CB3_PMCR,
        .ro = 0xff000002,
    },{ .name = "SMMU_CB3_PMCEID",  .decode.addr = A_SMMU_CB3_PMCEID,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "SMMU_CB3_PMCNTENSE",  .decode.addr = A_SMMU_CB3_PMCNTENSE,
    },{ .name = "SMMU_CB3_PMCNTENCLR",  .decode.addr = A_SMMU_CB3_PMCNTENCLR,
    },{ .name = "SMMU_CB3_PMCNTENSET",  .decode.addr = A_SMMU_CB3_PMCNTENSET,
    },{ .name = "SMMU_CB3_PMINTENCLR",  .decode.addr = A_SMMU_CB3_PMINTENCLR,
    },{ .name = "SMMU_CB3_PMOVSCLR",  .decode.addr = A_SMMU_CB3_PMOVSCLR,
    },{ .name = "SMMU_CB3_PMOVSSET",  .decode.addr = A_SMMU_CB3_PMOVSSET,
    },{ .name = "SMMU_CB3_PMAUTHSTATUS",  .decode.addr = A_SMMU_CB3_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    },{ .name = "SMMU_CB4_SCTLR",  .decode.addr = A_SMMU_CB4_SCTLR,
        .reset = 0x100,
        .ro = 0x1000,
    },{ .name = "SMMU_CB4_ACTLR",  .decode.addr = A_SMMU_CB4_ACTLR,
        .reset = 0x3,
    },{ .name = "SMMU_CB4_RESUME",  .decode.addr = A_SMMU_CB4_RESUME,
    },{ .name = "SMMU_CB4_TCR2",  .decode.addr = A_SMMU_CB4_TCR2,
        .reset = 0x60,
        .ro = 0x60,
    },{ .name = "SMMU_CB4_TTBR0_LOW",  .decode.addr = A_SMMU_CB4_TTBR0_LOW,
        .ro = 0x4,
    },{ .name = "SMMU_CB4_TTBR0_HIGH",  .decode.addr = A_SMMU_CB4_TTBR0_HIGH,
    },{ .name = "SMMU_CB4_TTBR1_LOW",  .decode.addr = A_SMMU_CB4_TTBR1_LOW,
    },{ .name = "SMMU_CB4_TTBR1_HIGH",  .decode.addr = A_SMMU_CB4_TTBR1_HIGH,
    },{ .name = "SMMU_CB4_TCR_LPAE",  .decode.addr = A_SMMU_CB4_TCR_LPAE,
    },{ .name = "SMMU_CB4_CONTEXTIDR",  .decode.addr = A_SMMU_CB4_CONTEXTIDR,
    },{ .name = "SMMU_CB4_PRRR_MAIR0",  .decode.addr = A_SMMU_CB4_PRRR_MAIR0,
    },{ .name = "SMMU_CB4_NMRR_MAIR1",  .decode.addr = A_SMMU_CB4_NMRR_MAIR1,
    },{ .name = "SMMU_CB4_FSR",  .decode.addr = A_SMMU_CB4_FSR,
        .w1c = 0xffffffff,
        .post_write = smmu_fsr_pw,
    },{ .name = "SMMU_CB4_FSRRESTORE",  .decode.addr = A_SMMU_CB4_FSRRESTORE,
    },{ .name = "SMMU_CB4_FAR_LOW",  .decode.addr = A_SMMU_CB4_FAR_LOW,
    },{ .name = "SMMU_CB4_FAR_HIGH",  .decode.addr = A_SMMU_CB4_FAR_HIGH,
    },{ .name = "SMMU_CB4_FSYNR0",  .decode.addr = A_SMMU_CB4_FSYNR0,
        .ro = 0x200,
    },{ .name = "SMMU_CB4_IPAFAR_LOW",  .decode.addr = A_SMMU_CB4_IPAFAR_LOW,
        .ro = 0xfff,
    },{ .name = "SMMU_CB4_IPAFAR_HIGH",  .decode.addr = A_SMMU_CB4_IPAFAR_HIGH,
    },{ .name = "SMMU_CB4_TLBIVA_LOW",  .decode.addr = A_SMMU_CB4_TLBIVA_LOW,
    },{ .name = "SMMU_CB4_TLBIVA_HIGH",  .decode.addr = A_SMMU_CB4_TLBIVA_HIGH,
    },{ .name = "SMMU_CB4_TLBIVAA_LOW",  .decode.addr = A_SMMU_CB4_TLBIVAA_LOW,
    },{ .name = "SMMU_CB4_TLBIVAA_HIGH",  .decode.addr = A_SMMU_CB4_TLBIVAA_HIGH,
    },{ .name = "SMMU_CB4_TLBIASID",  .decode.addr = A_SMMU_CB4_TLBIASID,
    },{ .name = "SMMU_CB4_TLBIALL",  .decode.addr = A_SMMU_CB4_TLBIALL,
    },{ .name = "SMMU_CB4_TLBIVAL_LOW",  .decode.addr = A_SMMU_CB4_TLBIVAL_LOW,
    },{ .name = "SMMU_CB4_TLBIVAL_HIGH",  .decode.addr = A_SMMU_CB4_TLBIVAL_HIGH,
    },{ .name = "SMMU_CB4_TLBIVAAL_LOW",  .decode.addr = A_SMMU_CB4_TLBIVAAL_LOW,
    },{ .name = "SMMU_CB4_TLBIVAAL_HIGH",  .decode.addr = A_SMMU_CB4_TLBIVAAL_HIGH,
    },{ .name = "SMMU_CB4_TLBIIPAS2_LOW",  .decode.addr = A_SMMU_CB4_TLBIIPAS2_LOW,
    },{ .name = "SMMU_CB4_TLBIIPAS2_HIGH",  .decode.addr = A_SMMU_CB4_TLBIIPAS2_HIGH,
    },{ .name = "SMMU_CB4_TLBIIPAS2L_LOW",  .decode.addr = A_SMMU_CB4_TLBIIPAS2L_LOW,
    },{ .name = "SMMU_CB4_TLBIIPAS2L_HIGH",  .decode.addr = A_SMMU_CB4_TLBIIPAS2L_HIGH,
    },{ .name = "SMMU_CB4_TLBSYNC",  .decode.addr = A_SMMU_CB4_TLBSYNC,
    },{ .name = "SMMU_CB4_TLBSTATUS",  .decode.addr = A_SMMU_CB4_TLBSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_CB4_PMEVCNTR0",  .decode.addr = A_SMMU_CB4_PMEVCNTR0,
    },{ .name = "SMMU_CB4_PMEVCNTR1",  .decode.addr = A_SMMU_CB4_PMEVCNTR1,
    },{ .name = "SMMU_CB4_PMEVCNTR2",  .decode.addr = A_SMMU_CB4_PMEVCNTR2,
    },{ .name = "SMMU_CB4_PMEVCNTR3",  .decode.addr = A_SMMU_CB4_PMEVCNTR3,
    },{ .name = "SMMU_CB4_PMEVTYPER0",  .decode.addr = A_SMMU_CB4_PMEVTYPER0,
    },{ .name = "SMMU_CB4_PMEVTYPER1",  .decode.addr = A_SMMU_CB4_PMEVTYPER1,
    },{ .name = "SMMU_CB4_PMEVTYPER2",  .decode.addr = A_SMMU_CB4_PMEVTYPER2,
    },{ .name = "SMMU_CB4_PMEVTYPER3",  .decode.addr = A_SMMU_CB4_PMEVTYPER3,
    },{ .name = "SMMU_CB4_PMCFGR",  .decode.addr = A_SMMU_CB4_PMCFGR,
        .reset = 0x11f03,
        .ro = 0xff09ffff,
    },{ .name = "SMMU_CB4_PMCR",  .decode.addr = A_SMMU_CB4_PMCR,
        .ro = 0xff000002,
    },{ .name = "SMMU_CB4_PMCEID",  .decode.addr = A_SMMU_CB4_PMCEID,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "SMMU_CB4_PMCNTENSE",  .decode.addr = A_SMMU_CB4_PMCNTENSE,
    },{ .name = "SMMU_CB4_PMCNTENCLR",  .decode.addr = A_SMMU_CB4_PMCNTENCLR,
    },{ .name = "SMMU_CB4_PMCNTENSET",  .decode.addr = A_SMMU_CB4_PMCNTENSET,
    },{ .name = "SMMU_CB4_PMINTENCLR",  .decode.addr = A_SMMU_CB4_PMINTENCLR,
    },{ .name = "SMMU_CB4_PMOVSCLR",  .decode.addr = A_SMMU_CB4_PMOVSCLR,
    },{ .name = "SMMU_CB4_PMOVSSET",  .decode.addr = A_SMMU_CB4_PMOVSSET,
    },{ .name = "SMMU_CB4_PMAUTHSTATUS",  .decode.addr = A_SMMU_CB4_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    },{ .name = "SMMU_CB5_SCTLR",  .decode.addr = A_SMMU_CB5_SCTLR,
        .reset = 0x100,
        .ro = 0x1000,
    },{ .name = "SMMU_CB5_ACTLR",  .decode.addr = A_SMMU_CB5_ACTLR,
        .reset = 0x3,
    },{ .name = "SMMU_CB5_RESUME",  .decode.addr = A_SMMU_CB5_RESUME,
    },{ .name = "SMMU_CB5_TCR2",  .decode.addr = A_SMMU_CB5_TCR2,
        .reset = 0x60,
        .ro = 0x60,
    },{ .name = "SMMU_CB5_TTBR0_LOW",  .decode.addr = A_SMMU_CB5_TTBR0_LOW,
        .ro = 0x4,
    },{ .name = "SMMU_CB5_TTBR0_HIGH",  .decode.addr = A_SMMU_CB5_TTBR0_HIGH,
    },{ .name = "SMMU_CB5_TTBR1_LOW",  .decode.addr = A_SMMU_CB5_TTBR1_LOW,
    },{ .name = "SMMU_CB5_TTBR1_HIGH",  .decode.addr = A_SMMU_CB5_TTBR1_HIGH,
    },{ .name = "SMMU_CB5_TCR_LPAE",  .decode.addr = A_SMMU_CB5_TCR_LPAE,
    },{ .name = "SMMU_CB5_CONTEXTIDR",  .decode.addr = A_SMMU_CB5_CONTEXTIDR,
    },{ .name = "SMMU_CB5_PRRR_MAIR0",  .decode.addr = A_SMMU_CB5_PRRR_MAIR0,
    },{ .name = "SMMU_CB5_NMRR_MAIR1",  .decode.addr = A_SMMU_CB5_NMRR_MAIR1,
    },{ .name = "SMMU_CB5_FSR",  .decode.addr = A_SMMU_CB5_FSR,
        .w1c = 0xffffffff,
        .post_write = smmu_fsr_pw,
    },{ .name = "SMMU_CB5_FSRRESTORE",  .decode.addr = A_SMMU_CB5_FSRRESTORE,
    },{ .name = "SMMU_CB5_FAR_LOW",  .decode.addr = A_SMMU_CB5_FAR_LOW,
    },{ .name = "SMMU_CB5_FAR_HIGH",  .decode.addr = A_SMMU_CB5_FAR_HIGH,
    },{ .name = "SMMU_CB5_FSYNR0",  .decode.addr = A_SMMU_CB5_FSYNR0,
        .ro = 0x200,
    },{ .name = "SMMU_CB5_IPAFAR_LOW",  .decode.addr = A_SMMU_CB5_IPAFAR_LOW,
        .ro = 0xfff,
    },{ .name = "SMMU_CB5_IPAFAR_HIGH",  .decode.addr = A_SMMU_CB5_IPAFAR_HIGH,
    },{ .name = "SMMU_CB5_TLBIVA_LOW",  .decode.addr = A_SMMU_CB5_TLBIVA_LOW,
    },{ .name = "SMMU_CB5_TLBIVA_HIGH",  .decode.addr = A_SMMU_CB5_TLBIVA_HIGH,
    },{ .name = "SMMU_CB5_TLBIVAA_LOW",  .decode.addr = A_SMMU_CB5_TLBIVAA_LOW,
    },{ .name = "SMMU_CB5_TLBIVAA_HIGH",  .decode.addr = A_SMMU_CB5_TLBIVAA_HIGH,
    },{ .name = "SMMU_CB5_TLBIASID",  .decode.addr = A_SMMU_CB5_TLBIASID,
    },{ .name = "SMMU_CB5_TLBIALL",  .decode.addr = A_SMMU_CB5_TLBIALL,
    },{ .name = "SMMU_CB5_TLBIVAL_LOW",  .decode.addr = A_SMMU_CB5_TLBIVAL_LOW,
    },{ .name = "SMMU_CB5_TLBIVAL_HIGH",  .decode.addr = A_SMMU_CB5_TLBIVAL_HIGH,
    },{ .name = "SMMU_CB5_TLBIVAAL_LOW",  .decode.addr = A_SMMU_CB5_TLBIVAAL_LOW,
    },{ .name = "SMMU_CB5_TLBIVAAL_HIGH",  .decode.addr = A_SMMU_CB5_TLBIVAAL_HIGH,
    },{ .name = "SMMU_CB5_TLBIIPAS2_LOW",  .decode.addr = A_SMMU_CB5_TLBIIPAS2_LOW,
    },{ .name = "SMMU_CB5_TLBIIPAS2_HIGH",  .decode.addr = A_SMMU_CB5_TLBIIPAS2_HIGH,
    },{ .name = "SMMU_CB5_TLBIIPAS2L_LOW",  .decode.addr = A_SMMU_CB5_TLBIIPAS2L_LOW,
    },{ .name = "SMMU_CB5_TLBIIPAS2L_HIGH",  .decode.addr = A_SMMU_CB5_TLBIIPAS2L_HIGH,
    },{ .name = "SMMU_CB5_TLBSYNC",  .decode.addr = A_SMMU_CB5_TLBSYNC,
    },{ .name = "SMMU_CB5_TLBSTATUS",  .decode.addr = A_SMMU_CB5_TLBSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_CB5_PMEVCNTR0",  .decode.addr = A_SMMU_CB5_PMEVCNTR0,
    },{ .name = "SMMU_CB5_PMEVCNTR1",  .decode.addr = A_SMMU_CB5_PMEVCNTR1,
    },{ .name = "SMMU_CB5_PMEVCNTR2",  .decode.addr = A_SMMU_CB5_PMEVCNTR2,
    },{ .name = "SMMU_CB5_PMEVCNTR3",  .decode.addr = A_SMMU_CB5_PMEVCNTR3,
    },{ .name = "SMMU_CB5_PMEVTYPER0",  .decode.addr = A_SMMU_CB5_PMEVTYPER0,
    },{ .name = "SMMU_CB5_PMEVTYPER1",  .decode.addr = A_SMMU_CB5_PMEVTYPER1,
    },{ .name = "SMMU_CB5_PMEVTYPER2",  .decode.addr = A_SMMU_CB5_PMEVTYPER2,
    },{ .name = "SMMU_CB5_PMEVTYPER3",  .decode.addr = A_SMMU_CB5_PMEVTYPER3,
    },{ .name = "SMMU_CB5_PMCFGR",  .decode.addr = A_SMMU_CB5_PMCFGR,
        .reset = 0x11f03,
        .ro = 0xff09ffff,
    },{ .name = "SMMU_CB5_PMCR",  .decode.addr = A_SMMU_CB5_PMCR,
        .ro = 0xff000002,
    },{ .name = "SMMU_CB5_PMCEID",  .decode.addr = A_SMMU_CB5_PMCEID,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "SMMU_CB5_PMCNTENSE",  .decode.addr = A_SMMU_CB5_PMCNTENSE,
    },{ .name = "SMMU_CB5_PMCNTENCLR",  .decode.addr = A_SMMU_CB5_PMCNTENCLR,
    },{ .name = "SMMU_CB5_PMCNTENSET",  .decode.addr = A_SMMU_CB5_PMCNTENSET,
    },{ .name = "SMMU_CB5_PMINTENCLR",  .decode.addr = A_SMMU_CB5_PMINTENCLR,
    },{ .name = "SMMU_CB5_PMOVSCLR",  .decode.addr = A_SMMU_CB5_PMOVSCLR,
    },{ .name = "SMMU_CB5_PMOVSSET",  .decode.addr = A_SMMU_CB5_PMOVSSET,
    },{ .name = "SMMU_CB5_PMAUTHSTATUS",  .decode.addr = A_SMMU_CB5_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    },{ .name = "SMMU_CB6_SCTLR",  .decode.addr = A_SMMU_CB6_SCTLR,
        .reset = 0x100,
        .ro = 0x1000,
    },{ .name = "SMMU_CB6_ACTLR",  .decode.addr = A_SMMU_CB6_ACTLR,
        .reset = 0x3,
    },{ .name = "SMMU_CB6_RESUME",  .decode.addr = A_SMMU_CB6_RESUME,
    },{ .name = "SMMU_CB6_TCR2",  .decode.addr = A_SMMU_CB6_TCR2,
        .reset = 0x60,
        .ro = 0x60,
    },{ .name = "SMMU_CB6_TTBR0_LOW",  .decode.addr = A_SMMU_CB6_TTBR0_LOW,
        .ro = 0x4,
    },{ .name = "SMMU_CB6_TTBR0_HIGH",  .decode.addr = A_SMMU_CB6_TTBR0_HIGH,
    },{ .name = "SMMU_CB6_TTBR1_LOW",  .decode.addr = A_SMMU_CB6_TTBR1_LOW,
    },{ .name = "SMMU_CB6_TTBR1_HIGH",  .decode.addr = A_SMMU_CB6_TTBR1_HIGH,
    },{ .name = "SMMU_CB6_TCR_LPAE",  .decode.addr = A_SMMU_CB6_TCR_LPAE,
    },{ .name = "SMMU_CB6_CONTEXTIDR",  .decode.addr = A_SMMU_CB6_CONTEXTIDR,
    },{ .name = "SMMU_CB6_PRRR_MAIR0",  .decode.addr = A_SMMU_CB6_PRRR_MAIR0,
    },{ .name = "SMMU_CB6_NMRR_MAIR1",  .decode.addr = A_SMMU_CB6_NMRR_MAIR1,
    },{ .name = "SMMU_CB6_FSR",  .decode.addr = A_SMMU_CB6_FSR,
        .w1c = 0xffffffff,
        .post_write = smmu_fsr_pw,
    },{ .name = "SMMU_CB6_FSRRESTORE",  .decode.addr = A_SMMU_CB6_FSRRESTORE,
    },{ .name = "SMMU_CB6_FAR_LOW",  .decode.addr = A_SMMU_CB6_FAR_LOW,
    },{ .name = "SMMU_CB6_FAR_HIGH",  .decode.addr = A_SMMU_CB6_FAR_HIGH,
    },{ .name = "SMMU_CB6_FSYNR0",  .decode.addr = A_SMMU_CB6_FSYNR0,
        .ro = 0x200,
    },{ .name = "SMMU_CB6_IPAFAR_LOW",  .decode.addr = A_SMMU_CB6_IPAFAR_LOW,
        .ro = 0xfff,
    },{ .name = "SMMU_CB6_IPAFAR_HIGH",  .decode.addr = A_SMMU_CB6_IPAFAR_HIGH,
    },{ .name = "SMMU_CB6_TLBIVA_LOW",  .decode.addr = A_SMMU_CB6_TLBIVA_LOW,
    },{ .name = "SMMU_CB6_TLBIVA_HIGH",  .decode.addr = A_SMMU_CB6_TLBIVA_HIGH,
    },{ .name = "SMMU_CB6_TLBIVAA_LOW",  .decode.addr = A_SMMU_CB6_TLBIVAA_LOW,
    },{ .name = "SMMU_CB6_TLBIVAA_HIGH",  .decode.addr = A_SMMU_CB6_TLBIVAA_HIGH,
    },{ .name = "SMMU_CB6_TLBIASID",  .decode.addr = A_SMMU_CB6_TLBIASID,
    },{ .name = "SMMU_CB6_TLBIALL",  .decode.addr = A_SMMU_CB6_TLBIALL,
    },{ .name = "SMMU_CB6_TLBIVAL_LOW",  .decode.addr = A_SMMU_CB6_TLBIVAL_LOW,
    },{ .name = "SMMU_CB6_TLBIVAL_HIGH",  .decode.addr = A_SMMU_CB6_TLBIVAL_HIGH,
    },{ .name = "SMMU_CB6_TLBIVAAL_LOW",  .decode.addr = A_SMMU_CB6_TLBIVAAL_LOW,
    },{ .name = "SMMU_CB6_TLBIVAAL_HIGH",  .decode.addr = A_SMMU_CB6_TLBIVAAL_HIGH,
    },{ .name = "SMMU_CB6_TLBIIPAS2_LOW",  .decode.addr = A_SMMU_CB6_TLBIIPAS2_LOW,
    },{ .name = "SMMU_CB6_TLBIIPAS2_HIGH",  .decode.addr = A_SMMU_CB6_TLBIIPAS2_HIGH,
    },{ .name = "SMMU_CB6_TLBIIPAS2L_LOW",  .decode.addr = A_SMMU_CB6_TLBIIPAS2L_LOW,
    },{ .name = "SMMU_CB6_TLBIIPAS2L_HIGH",  .decode.addr = A_SMMU_CB6_TLBIIPAS2L_HIGH,
    },{ .name = "SMMU_CB6_TLBSYNC",  .decode.addr = A_SMMU_CB6_TLBSYNC,
    },{ .name = "SMMU_CB6_TLBSTATUS",  .decode.addr = A_SMMU_CB6_TLBSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_CB6_PMEVCNTR0",  .decode.addr = A_SMMU_CB6_PMEVCNTR0,
    },{ .name = "SMMU_CB6_PMEVCNTR1",  .decode.addr = A_SMMU_CB6_PMEVCNTR1,
    },{ .name = "SMMU_CB6_PMEVCNTR2",  .decode.addr = A_SMMU_CB6_PMEVCNTR2,
    },{ .name = "SMMU_CB6_PMEVCNTR3",  .decode.addr = A_SMMU_CB6_PMEVCNTR3,
    },{ .name = "SMMU_CB6_PMEVTYPER0",  .decode.addr = A_SMMU_CB6_PMEVTYPER0,
    },{ .name = "SMMU_CB6_PMEVTYPER1",  .decode.addr = A_SMMU_CB6_PMEVTYPER1,
    },{ .name = "SMMU_CB6_PMEVTYPER2",  .decode.addr = A_SMMU_CB6_PMEVTYPER2,
    },{ .name = "SMMU_CB6_PMEVTYPER3",  .decode.addr = A_SMMU_CB6_PMEVTYPER3,
    },{ .name = "SMMU_CB6_PMCFGR",  .decode.addr = A_SMMU_CB6_PMCFGR,
        .reset = 0x11f03,
        .ro = 0xff09ffff,
    },{ .name = "SMMU_CB6_PMCR",  .decode.addr = A_SMMU_CB6_PMCR,
        .ro = 0xff000002,
    },{ .name = "SMMU_CB6_PMCEID",  .decode.addr = A_SMMU_CB6_PMCEID,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "SMMU_CB6_PMCNTENSE",  .decode.addr = A_SMMU_CB6_PMCNTENSE,
    },{ .name = "SMMU_CB6_PMCNTENCLR",  .decode.addr = A_SMMU_CB6_PMCNTENCLR,
    },{ .name = "SMMU_CB6_PMCNTENSET",  .decode.addr = A_SMMU_CB6_PMCNTENSET,
    },{ .name = "SMMU_CB6_PMINTENCLR",  .decode.addr = A_SMMU_CB6_PMINTENCLR,
    },{ .name = "SMMU_CB6_PMOVSCLR",  .decode.addr = A_SMMU_CB6_PMOVSCLR,
    },{ .name = "SMMU_CB6_PMOVSSET",  .decode.addr = A_SMMU_CB6_PMOVSSET,
    },{ .name = "SMMU_CB6_PMAUTHSTATUS",  .decode.addr = A_SMMU_CB6_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    },{ .name = "SMMU_CB7_SCTLR",  .decode.addr = A_SMMU_CB7_SCTLR,
        .reset = 0x100,
        .ro = 0x1000,
    },{ .name = "SMMU_CB7_ACTLR",  .decode.addr = A_SMMU_CB7_ACTLR,
        .reset = 0x3,
    },{ .name = "SMMU_CB7_RESUME",  .decode.addr = A_SMMU_CB7_RESUME,
    },{ .name = "SMMU_CB7_TCR2",  .decode.addr = A_SMMU_CB7_TCR2,
        .reset = 0x60,
        .ro = 0x60,
    },{ .name = "SMMU_CB7_TTBR0_LOW",  .decode.addr = A_SMMU_CB7_TTBR0_LOW,
        .ro = 0x4,
    },{ .name = "SMMU_CB7_TTBR0_HIGH",  .decode.addr = A_SMMU_CB7_TTBR0_HIGH,
    },{ .name = "SMMU_CB7_TTBR1_LOW",  .decode.addr = A_SMMU_CB7_TTBR1_LOW,
    },{ .name = "SMMU_CB7_TTBR1_HIGH",  .decode.addr = A_SMMU_CB7_TTBR1_HIGH,
    },{ .name = "SMMU_CB7_TCR_LPAE",  .decode.addr = A_SMMU_CB7_TCR_LPAE,
    },{ .name = "SMMU_CB7_CONTEXTIDR",  .decode.addr = A_SMMU_CB7_CONTEXTIDR,
    },{ .name = "SMMU_CB7_PRRR_MAIR0",  .decode.addr = A_SMMU_CB7_PRRR_MAIR0,
    },{ .name = "SMMU_CB7_NMRR_MAIR1",  .decode.addr = A_SMMU_CB7_NMRR_MAIR1,
    },{ .name = "SMMU_CB7_FSR",  .decode.addr = A_SMMU_CB7_FSR,
        .w1c = 0xffffffff,
        .post_write = smmu_fsr_pw,
    },{ .name = "SMMU_CB7_FSRRESTORE",  .decode.addr = A_SMMU_CB7_FSRRESTORE,
    },{ .name = "SMMU_CB7_FAR_LOW",  .decode.addr = A_SMMU_CB7_FAR_LOW,
    },{ .name = "SMMU_CB7_FAR_HIGH",  .decode.addr = A_SMMU_CB7_FAR_HIGH,
    },{ .name = "SMMU_CB7_FSYNR0",  .decode.addr = A_SMMU_CB7_FSYNR0,
        .ro = 0x200,
    },{ .name = "SMMU_CB7_IPAFAR_LOW",  .decode.addr = A_SMMU_CB7_IPAFAR_LOW,
        .ro = 0xfff,
    },{ .name = "SMMU_CB7_IPAFAR_HIGH",  .decode.addr = A_SMMU_CB7_IPAFAR_HIGH,
    },{ .name = "SMMU_CB7_TLBIVA_LOW",  .decode.addr = A_SMMU_CB7_TLBIVA_LOW,
    },{ .name = "SMMU_CB7_TLBIVA_HIGH",  .decode.addr = A_SMMU_CB7_TLBIVA_HIGH,
    },{ .name = "SMMU_CB7_TLBIVAA_LOW",  .decode.addr = A_SMMU_CB7_TLBIVAA_LOW,
    },{ .name = "SMMU_CB7_TLBIVAA_HIGH",  .decode.addr = A_SMMU_CB7_TLBIVAA_HIGH,
    },{ .name = "SMMU_CB7_TLBIASID",  .decode.addr = A_SMMU_CB7_TLBIASID,
    },{ .name = "SMMU_CB7_TLBIALL",  .decode.addr = A_SMMU_CB7_TLBIALL,
    },{ .name = "SMMU_CB7_TLBIVAL_LOW",  .decode.addr = A_SMMU_CB7_TLBIVAL_LOW,
    },{ .name = "SMMU_CB7_TLBIVAL_HIGH",  .decode.addr = A_SMMU_CB7_TLBIVAL_HIGH,
    },{ .name = "SMMU_CB7_TLBIVAAL_LOW",  .decode.addr = A_SMMU_CB7_TLBIVAAL_LOW,
    },{ .name = "SMMU_CB7_TLBIVAAL_HIGH",  .decode.addr = A_SMMU_CB7_TLBIVAAL_HIGH,
    },{ .name = "SMMU_CB7_TLBIIPAS2_LOW",  .decode.addr = A_SMMU_CB7_TLBIIPAS2_LOW,
    },{ .name = "SMMU_CB7_TLBIIPAS2_HIGH",  .decode.addr = A_SMMU_CB7_TLBIIPAS2_HIGH,
    },{ .name = "SMMU_CB7_TLBIIPAS2L_LOW",  .decode.addr = A_SMMU_CB7_TLBIIPAS2L_LOW,
    },{ .name = "SMMU_CB7_TLBIIPAS2L_HIGH",  .decode.addr = A_SMMU_CB7_TLBIIPAS2L_HIGH,
    },{ .name = "SMMU_CB7_TLBSYNC",  .decode.addr = A_SMMU_CB7_TLBSYNC,
    },{ .name = "SMMU_CB7_TLBSTATUS",  .decode.addr = A_SMMU_CB7_TLBSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_CB7_PMEVCNTR0",  .decode.addr = A_SMMU_CB7_PMEVCNTR0,
    },{ .name = "SMMU_CB7_PMEVCNTR1",  .decode.addr = A_SMMU_CB7_PMEVCNTR1,
    },{ .name = "SMMU_CB7_PMEVCNTR2",  .decode.addr = A_SMMU_CB7_PMEVCNTR2,
    },{ .name = "SMMU_CB7_PMEVCNTR3",  .decode.addr = A_SMMU_CB7_PMEVCNTR3,
    },{ .name = "SMMU_CB7_PMEVTYPER0",  .decode.addr = A_SMMU_CB7_PMEVTYPER0,
    },{ .name = "SMMU_CB7_PMEVTYPER1",  .decode.addr = A_SMMU_CB7_PMEVTYPER1,
    },{ .name = "SMMU_CB7_PMEVTYPER2",  .decode.addr = A_SMMU_CB7_PMEVTYPER2,
    },{ .name = "SMMU_CB7_PMEVTYPER3",  .decode.addr = A_SMMU_CB7_PMEVTYPER3,
    },{ .name = "SMMU_CB7_PMCFGR",  .decode.addr = A_SMMU_CB7_PMCFGR,
        .reset = 0x11f03,
        .ro = 0xff09ffff,
    },{ .name = "SMMU_CB7_PMCR",  .decode.addr = A_SMMU_CB7_PMCR,
        .ro = 0xff000002,
    },{ .name = "SMMU_CB7_PMCEID",  .decode.addr = A_SMMU_CB7_PMCEID,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "SMMU_CB7_PMCNTENSE",  .decode.addr = A_SMMU_CB7_PMCNTENSE,
    },{ .name = "SMMU_CB7_PMCNTENCLR",  .decode.addr = A_SMMU_CB7_PMCNTENCLR,
    },{ .name = "SMMU_CB7_PMCNTENSET",  .decode.addr = A_SMMU_CB7_PMCNTENSET,
    },{ .name = "SMMU_CB7_PMINTENCLR",  .decode.addr = A_SMMU_CB7_PMINTENCLR,
    },{ .name = "SMMU_CB7_PMOVSCLR",  .decode.addr = A_SMMU_CB7_PMOVSCLR,
    },{ .name = "SMMU_CB7_PMOVSSET",  .decode.addr = A_SMMU_CB7_PMOVSSET,
    },{ .name = "SMMU_CB7_PMAUTHSTATUS",  .decode.addr = A_SMMU_CB7_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    },{ .name = "SMMU_CB8_SCTLR",  .decode.addr = A_SMMU_CB8_SCTLR,
        .reset = 0x100,
        .ro = 0x1000,
    },{ .name = "SMMU_CB8_ACTLR",  .decode.addr = A_SMMU_CB8_ACTLR,
        .reset = 0x3,
    },{ .name = "SMMU_CB8_RESUME",  .decode.addr = A_SMMU_CB8_RESUME,
    },{ .name = "SMMU_CB8_TCR2",  .decode.addr = A_SMMU_CB8_TCR2,
        .reset = 0x60,
        .ro = 0x60,
    },{ .name = "SMMU_CB8_TTBR0_LOW",  .decode.addr = A_SMMU_CB8_TTBR0_LOW,
        .ro = 0x4,
    },{ .name = "SMMU_CB8_TTBR0_HIGH",  .decode.addr = A_SMMU_CB8_TTBR0_HIGH,
    },{ .name = "SMMU_CB8_TTBR1_LOW",  .decode.addr = A_SMMU_CB8_TTBR1_LOW,
    },{ .name = "SMMU_CB8_TTBR1_HIGH",  .decode.addr = A_SMMU_CB8_TTBR1_HIGH,
    },{ .name = "SMMU_CB8_TCR_LPAE",  .decode.addr = A_SMMU_CB8_TCR_LPAE,
    },{ .name = "SMMU_CB8_CONTEXTIDR",  .decode.addr = A_SMMU_CB8_CONTEXTIDR,
    },{ .name = "SMMU_CB8_PRRR_MAIR0",  .decode.addr = A_SMMU_CB8_PRRR_MAIR0,
    },{ .name = "SMMU_CB8_NMRR_MAIR1",  .decode.addr = A_SMMU_CB8_NMRR_MAIR1,
    },{ .name = "SMMU_CB8_FSR",  .decode.addr = A_SMMU_CB8_FSR,
        .w1c = 0xffffffff,
        .post_write = smmu_fsr_pw,
    },{ .name = "SMMU_CB8_FSRRESTORE",  .decode.addr = A_SMMU_CB8_FSRRESTORE,
    },{ .name = "SMMU_CB8_FAR_LOW",  .decode.addr = A_SMMU_CB8_FAR_LOW,
    },{ .name = "SMMU_CB8_FAR_HIGH",  .decode.addr = A_SMMU_CB8_FAR_HIGH,
    },{ .name = "SMMU_CB8_FSYNR0",  .decode.addr = A_SMMU_CB8_FSYNR0,
        .ro = 0x200,
    },{ .name = "SMMU_CB8_IPAFAR_LOW",  .decode.addr = A_SMMU_CB8_IPAFAR_LOW,
        .ro = 0xfff,
    },{ .name = "SMMU_CB8_IPAFAR_HIGH",  .decode.addr = A_SMMU_CB8_IPAFAR_HIGH,
    },{ .name = "SMMU_CB8_TLBIVA_LOW",  .decode.addr = A_SMMU_CB8_TLBIVA_LOW,
    },{ .name = "SMMU_CB8_TLBIVA_HIGH",  .decode.addr = A_SMMU_CB8_TLBIVA_HIGH,
    },{ .name = "SMMU_CB8_TLBIVAA_LOW",  .decode.addr = A_SMMU_CB8_TLBIVAA_LOW,
    },{ .name = "SMMU_CB8_TLBIVAA_HIGH",  .decode.addr = A_SMMU_CB8_TLBIVAA_HIGH,
    },{ .name = "SMMU_CB8_TLBIASID",  .decode.addr = A_SMMU_CB8_TLBIASID,
    },{ .name = "SMMU_CB8_TLBIALL",  .decode.addr = A_SMMU_CB8_TLBIALL,
    },{ .name = "SMMU_CB8_TLBIVAL_LOW",  .decode.addr = A_SMMU_CB8_TLBIVAL_LOW,
    },{ .name = "SMMU_CB8_TLBIVAL_HIGH",  .decode.addr = A_SMMU_CB8_TLBIVAL_HIGH,
    },{ .name = "SMMU_CB8_TLBIVAAL_LOW",  .decode.addr = A_SMMU_CB8_TLBIVAAL_LOW,
    },{ .name = "SMMU_CB8_TLBIVAAL_HIGH",  .decode.addr = A_SMMU_CB8_TLBIVAAL_HIGH,
    },{ .name = "SMMU_CB8_TLBIIPAS2_LOW",  .decode.addr = A_SMMU_CB8_TLBIIPAS2_LOW,
    },{ .name = "SMMU_CB8_TLBIIPAS2_HIGH",  .decode.addr = A_SMMU_CB8_TLBIIPAS2_HIGH,
    },{ .name = "SMMU_CB8_TLBIIPAS2L_LOW",  .decode.addr = A_SMMU_CB8_TLBIIPAS2L_LOW,
    },{ .name = "SMMU_CB8_TLBIIPAS2L_HIGH",  .decode.addr = A_SMMU_CB8_TLBIIPAS2L_HIGH,
    },{ .name = "SMMU_CB8_TLBSYNC",  .decode.addr = A_SMMU_CB8_TLBSYNC,
    },{ .name = "SMMU_CB8_TLBSTATUS",  .decode.addr = A_SMMU_CB8_TLBSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_CB8_PMEVCNTR0",  .decode.addr = A_SMMU_CB8_PMEVCNTR0,
    },{ .name = "SMMU_CB8_PMEVCNTR1",  .decode.addr = A_SMMU_CB8_PMEVCNTR1,
    },{ .name = "SMMU_CB8_PMEVCNTR2",  .decode.addr = A_SMMU_CB8_PMEVCNTR2,
    },{ .name = "SMMU_CB8_PMEVCNTR3",  .decode.addr = A_SMMU_CB8_PMEVCNTR3,
    },{ .name = "SMMU_CB8_PMEVTYPER0",  .decode.addr = A_SMMU_CB8_PMEVTYPER0,
    },{ .name = "SMMU_CB8_PMEVTYPER1",  .decode.addr = A_SMMU_CB8_PMEVTYPER1,
    },{ .name = "SMMU_CB8_PMEVTYPER2",  .decode.addr = A_SMMU_CB8_PMEVTYPER2,
    },{ .name = "SMMU_CB8_PMEVTYPER3",  .decode.addr = A_SMMU_CB8_PMEVTYPER3,
    },{ .name = "SMMU_CB8_PMCFGR",  .decode.addr = A_SMMU_CB8_PMCFGR,
        .reset = 0x11f03,
        .ro = 0xff09ffff,
    },{ .name = "SMMU_CB8_PMCR",  .decode.addr = A_SMMU_CB8_PMCR,
        .ro = 0xff000002,
    },{ .name = "SMMU_CB8_PMCEID",  .decode.addr = A_SMMU_CB8_PMCEID,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "SMMU_CB8_PMCNTENSE",  .decode.addr = A_SMMU_CB8_PMCNTENSE,
    },{ .name = "SMMU_CB8_PMCNTENCLR",  .decode.addr = A_SMMU_CB8_PMCNTENCLR,
    },{ .name = "SMMU_CB8_PMCNTENSET",  .decode.addr = A_SMMU_CB8_PMCNTENSET,
    },{ .name = "SMMU_CB8_PMINTENCLR",  .decode.addr = A_SMMU_CB8_PMINTENCLR,
    },{ .name = "SMMU_CB8_PMOVSCLR",  .decode.addr = A_SMMU_CB8_PMOVSCLR,
    },{ .name = "SMMU_CB8_PMOVSSET",  .decode.addr = A_SMMU_CB8_PMOVSSET,
    },{ .name = "SMMU_CB8_PMAUTHSTATUS",  .decode.addr = A_SMMU_CB8_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    },{ .name = "SMMU_CB9_SCTLR",  .decode.addr = A_SMMU_CB9_SCTLR,
        .reset = 0x100,
        .ro = 0x1000,
    },{ .name = "SMMU_CB9_ACTLR",  .decode.addr = A_SMMU_CB9_ACTLR,
        .reset = 0x3,
    },{ .name = "SMMU_CB9_RESUME",  .decode.addr = A_SMMU_CB9_RESUME,
    },{ .name = "SMMU_CB9_TCR2",  .decode.addr = A_SMMU_CB9_TCR2,
        .reset = 0x60,
        .ro = 0x60,
    },{ .name = "SMMU_CB9_TTBR0_LOW",  .decode.addr = A_SMMU_CB9_TTBR0_LOW,
        .ro = 0x4,
    },{ .name = "SMMU_CB9_TTBR0_HIGH",  .decode.addr = A_SMMU_CB9_TTBR0_HIGH,
    },{ .name = "SMMU_CB9_TTBR1_LOW",  .decode.addr = A_SMMU_CB9_TTBR1_LOW,
    },{ .name = "SMMU_CB9_TTBR1_HIGH",  .decode.addr = A_SMMU_CB9_TTBR1_HIGH,
    },{ .name = "SMMU_CB9_TCR_LPAE",  .decode.addr = A_SMMU_CB9_TCR_LPAE,
    },{ .name = "SMMU_CB9_CONTEXTIDR",  .decode.addr = A_SMMU_CB9_CONTEXTIDR,
    },{ .name = "SMMU_CB9_PRRR_MAIR0",  .decode.addr = A_SMMU_CB9_PRRR_MAIR0,
    },{ .name = "SMMU_CB9_NMRR_MAIR1",  .decode.addr = A_SMMU_CB9_NMRR_MAIR1,
    },{ .name = "SMMU_CB9_FSR",  .decode.addr = A_SMMU_CB9_FSR,
        .w1c = 0xffffffff,
        .post_write = smmu_fsr_pw,
    },{ .name = "SMMU_CB9_FSRRESTORE",  .decode.addr = A_SMMU_CB9_FSRRESTORE,
    },{ .name = "SMMU_CB9_FAR_LOW",  .decode.addr = A_SMMU_CB9_FAR_LOW,
    },{ .name = "SMMU_CB9_FAR_HIGH",  .decode.addr = A_SMMU_CB9_FAR_HIGH,
    },{ .name = "SMMU_CB9_FSYNR0",  .decode.addr = A_SMMU_CB9_FSYNR0,
        .ro = 0x200,
    },{ .name = "SMMU_CB9_IPAFAR_LOW",  .decode.addr = A_SMMU_CB9_IPAFAR_LOW,
        .ro = 0xfff,
    },{ .name = "SMMU_CB9_IPAFAR_HIGH",  .decode.addr = A_SMMU_CB9_IPAFAR_HIGH,
    },{ .name = "SMMU_CB9_TLBIVA_LOW",  .decode.addr = A_SMMU_CB9_TLBIVA_LOW,
    },{ .name = "SMMU_CB9_TLBIVA_HIGH",  .decode.addr = A_SMMU_CB9_TLBIVA_HIGH,
    },{ .name = "SMMU_CB9_TLBIVAA_LOW",  .decode.addr = A_SMMU_CB9_TLBIVAA_LOW,
    },{ .name = "SMMU_CB9_TLBIVAA_HIGH",  .decode.addr = A_SMMU_CB9_TLBIVAA_HIGH,
    },{ .name = "SMMU_CB9_TLBIASID",  .decode.addr = A_SMMU_CB9_TLBIASID,
    },{ .name = "SMMU_CB9_TLBIALL",  .decode.addr = A_SMMU_CB9_TLBIALL,
    },{ .name = "SMMU_CB9_TLBIVAL_LOW",  .decode.addr = A_SMMU_CB9_TLBIVAL_LOW,
    },{ .name = "SMMU_CB9_TLBIVAL_HIGH",  .decode.addr = A_SMMU_CB9_TLBIVAL_HIGH,
    },{ .name = "SMMU_CB9_TLBIVAAL_LOW",  .decode.addr = A_SMMU_CB9_TLBIVAAL_LOW,
    },{ .name = "SMMU_CB9_TLBIVAAL_HIGH",  .decode.addr = A_SMMU_CB9_TLBIVAAL_HIGH,
    },{ .name = "SMMU_CB9_TLBIIPAS2_LOW",  .decode.addr = A_SMMU_CB9_TLBIIPAS2_LOW,
    },{ .name = "SMMU_CB9_TLBIIPAS2_HIGH",  .decode.addr = A_SMMU_CB9_TLBIIPAS2_HIGH,
    },{ .name = "SMMU_CB9_TLBIIPAS2L_LOW",  .decode.addr = A_SMMU_CB9_TLBIIPAS2L_LOW,
    },{ .name = "SMMU_CB9_TLBIIPAS2L_HIGH",  .decode.addr = A_SMMU_CB9_TLBIIPAS2L_HIGH,
    },{ .name = "SMMU_CB9_TLBSYNC",  .decode.addr = A_SMMU_CB9_TLBSYNC,
    },{ .name = "SMMU_CB9_TLBSTATUS",  .decode.addr = A_SMMU_CB9_TLBSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_CB9_PMEVCNTR0",  .decode.addr = A_SMMU_CB9_PMEVCNTR0,
    },{ .name = "SMMU_CB9_PMEVCNTR1",  .decode.addr = A_SMMU_CB9_PMEVCNTR1,
    },{ .name = "SMMU_CB9_PMEVCNTR2",  .decode.addr = A_SMMU_CB9_PMEVCNTR2,
    },{ .name = "SMMU_CB9_PMEVCNTR3",  .decode.addr = A_SMMU_CB9_PMEVCNTR3,
    },{ .name = "SMMU_CB9_PMEVTYPER0",  .decode.addr = A_SMMU_CB9_PMEVTYPER0,
    },{ .name = "SMMU_CB9_PMEVTYPER1",  .decode.addr = A_SMMU_CB9_PMEVTYPER1,
    },{ .name = "SMMU_CB9_PMEVTYPER2",  .decode.addr = A_SMMU_CB9_PMEVTYPER2,
    },{ .name = "SMMU_CB9_PMEVTYPER3",  .decode.addr = A_SMMU_CB9_PMEVTYPER3,
    },{ .name = "SMMU_CB9_PMCFGR",  .decode.addr = A_SMMU_CB9_PMCFGR,
        .reset = 0x11f03,
        .ro = 0xff09ffff,
    },{ .name = "SMMU_CB9_PMCR",  .decode.addr = A_SMMU_CB9_PMCR,
        .ro = 0xff000002,
    },{ .name = "SMMU_CB9_PMCEID",  .decode.addr = A_SMMU_CB9_PMCEID,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "SMMU_CB9_PMCNTENSE",  .decode.addr = A_SMMU_CB9_PMCNTENSE,
    },{ .name = "SMMU_CB9_PMCNTENCLR",  .decode.addr = A_SMMU_CB9_PMCNTENCLR,
    },{ .name = "SMMU_CB9_PMCNTENSET",  .decode.addr = A_SMMU_CB9_PMCNTENSET,
    },{ .name = "SMMU_CB9_PMINTENCLR",  .decode.addr = A_SMMU_CB9_PMINTENCLR,
    },{ .name = "SMMU_CB9_PMOVSCLR",  .decode.addr = A_SMMU_CB9_PMOVSCLR,
    },{ .name = "SMMU_CB9_PMOVSSET",  .decode.addr = A_SMMU_CB9_PMOVSSET,
    },{ .name = "SMMU_CB9_PMAUTHSTATUS",  .decode.addr = A_SMMU_CB9_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    },{ .name = "SMMU_CB10_SCTLR",  .decode.addr = A_SMMU_CB10_SCTLR,
        .reset = 0x100,
        .ro = 0x1000,
    },{ .name = "SMMU_CB10_ACTLR",  .decode.addr = A_SMMU_CB10_ACTLR,
        .reset = 0x3,
    },{ .name = "SMMU_CB10_RESUME",  .decode.addr = A_SMMU_CB10_RESUME,
    },{ .name = "SMMU_CB10_TCR2",  .decode.addr = A_SMMU_CB10_TCR2,
        .reset = 0x60,
        .ro = 0x60,
    },{ .name = "SMMU_CB10_TTBR0_LOW",  .decode.addr = A_SMMU_CB10_TTBR0_LOW,
        .ro = 0x4,
    },{ .name = "SMMU_CB10_TTBR0_HIGH",  .decode.addr = A_SMMU_CB10_TTBR0_HIGH,
    },{ .name = "SMMU_CB10_TTBR1_LOW",  .decode.addr = A_SMMU_CB10_TTBR1_LOW,
    },{ .name = "SMMU_CB10_TTBR1_HIGH",  .decode.addr = A_SMMU_CB10_TTBR1_HIGH,
    },{ .name = "SMMU_CB10_TCR_LPAE",  .decode.addr = A_SMMU_CB10_TCR_LPAE,
    },{ .name = "SMMU_CB10_CONTEXTIDR",  .decode.addr = A_SMMU_CB10_CONTEXTIDR,
    },{ .name = "SMMU_CB10_PRRR_MAIR0",  .decode.addr = A_SMMU_CB10_PRRR_MAIR0,
    },{ .name = "SMMU_CB10_NMRR_MAIR1",  .decode.addr = A_SMMU_CB10_NMRR_MAIR1,
    },{ .name = "SMMU_CB10_FSR",  .decode.addr = A_SMMU_CB10_FSR,
        .w1c = 0xffffffff,
        .post_write = smmu_fsr_pw,
    },{ .name = "SMMU_CB10_FSRRESTORE",  .decode.addr = A_SMMU_CB10_FSRRESTORE,
    },{ .name = "SMMU_CB10_FAR_LOW",  .decode.addr = A_SMMU_CB10_FAR_LOW,
    },{ .name = "SMMU_CB10_FAR_HIGH",  .decode.addr = A_SMMU_CB10_FAR_HIGH,
    },{ .name = "SMMU_CB10_FSYNR0",  .decode.addr = A_SMMU_CB10_FSYNR0,
        .ro = 0x200,
    },{ .name = "SMMU_CB10_IPAFAR_LOW",  .decode.addr = A_SMMU_CB10_IPAFAR_LOW,
        .ro = 0xfff,
    },{ .name = "SMMU_CB10_IPAFAR_HIGH",  .decode.addr = A_SMMU_CB10_IPAFAR_HIGH,
    },{ .name = "SMMU_CB10_TLBIVA_LOW",  .decode.addr = A_SMMU_CB10_TLBIVA_LOW,
    },{ .name = "SMMU_CB10_TLBIVA_HIGH",  .decode.addr = A_SMMU_CB10_TLBIVA_HIGH,
    },{ .name = "SMMU_CB10_TLBIVAA_LOW",  .decode.addr = A_SMMU_CB10_TLBIVAA_LOW,
    },{ .name = "SMMU_CB10_TLBIVAA_HIGH",  .decode.addr = A_SMMU_CB10_TLBIVAA_HIGH,
    },{ .name = "SMMU_CB10_TLBIASID",  .decode.addr = A_SMMU_CB10_TLBIASID,
    },{ .name = "SMMU_CB10_TLBIALL",  .decode.addr = A_SMMU_CB10_TLBIALL,
    },{ .name = "SMMU_CB10_TLBIVAL_LOW",  .decode.addr = A_SMMU_CB10_TLBIVAL_LOW,
    },{ .name = "SMMU_CB10_TLBIVAL_HIGH",  .decode.addr = A_SMMU_CB10_TLBIVAL_HIGH,
    },{ .name = "SMMU_CB10_TLBIVAAL_LOW",  .decode.addr = A_SMMU_CB10_TLBIVAAL_LOW,
    },{ .name = "SMMU_CB10_TLBIVAAL_HIGH",  .decode.addr = A_SMMU_CB10_TLBIVAAL_HIGH,
    },{ .name = "SMMU_CB10_TLBIIPAS2_LOW",  .decode.addr = A_SMMU_CB10_TLBIIPAS2_LOW,
    },{ .name = "SMMU_CB10_TLBIIPAS2_HIGH",  .decode.addr = A_SMMU_CB10_TLBIIPAS2_HIGH,
    },{ .name = "SMMU_CB10_TLBIIPAS2L_LOW",  .decode.addr = A_SMMU_CB10_TLBIIPAS2L_LOW,
    },{ .name = "SMMU_CB10_TLBIIPAS2L_HIGH",  .decode.addr = A_SMMU_CB10_TLBIIPAS2L_HIGH,
    },{ .name = "SMMU_CB10_TLBSYNC",  .decode.addr = A_SMMU_CB10_TLBSYNC,
    },{ .name = "SMMU_CB10_TLBSTATUS",  .decode.addr = A_SMMU_CB10_TLBSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_CB10_PMEVCNTR0",  .decode.addr = A_SMMU_CB10_PMEVCNTR0,
    },{ .name = "SMMU_CB10_PMEVCNTR1",  .decode.addr = A_SMMU_CB10_PMEVCNTR1,
    },{ .name = "SMMU_CB10_PMEVCNTR2",  .decode.addr = A_SMMU_CB10_PMEVCNTR2,
    },{ .name = "SMMU_CB10_PMEVCNTR3",  .decode.addr = A_SMMU_CB10_PMEVCNTR3,
    },{ .name = "SMMU_CB10_PMEVTYPER0",  .decode.addr = A_SMMU_CB10_PMEVTYPER0,
    },{ .name = "SMMU_CB10_PMEVTYPER1",  .decode.addr = A_SMMU_CB10_PMEVTYPER1,
    },{ .name = "SMMU_CB10_PMEVTYPER2",  .decode.addr = A_SMMU_CB10_PMEVTYPER2,
    },{ .name = "SMMU_CB10_PMEVTYPER3",  .decode.addr = A_SMMU_CB10_PMEVTYPER3,
    },{ .name = "SMMU_CB10_PMCFGR",  .decode.addr = A_SMMU_CB10_PMCFGR,
        .reset = 0x11f03,
        .ro = 0xff09ffff,
    },{ .name = "SMMU_CB10_PMCR",  .decode.addr = A_SMMU_CB10_PMCR,
        .ro = 0xff000002,
    },{ .name = "SMMU_CB10_PMCEID",  .decode.addr = A_SMMU_CB10_PMCEID,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "SMMU_CB10_PMCNTENSE",  .decode.addr = A_SMMU_CB10_PMCNTENSE,
    },{ .name = "SMMU_CB10_PMCNTENCLR",  .decode.addr = A_SMMU_CB10_PMCNTENCLR,
    },{ .name = "SMMU_CB10_PMCNTENSET",  .decode.addr = A_SMMU_CB10_PMCNTENSET,
    },{ .name = "SMMU_CB10_PMINTENCLR",  .decode.addr = A_SMMU_CB10_PMINTENCLR,
    },{ .name = "SMMU_CB10_PMOVSCLR",  .decode.addr = A_SMMU_CB10_PMOVSCLR,
    },{ .name = "SMMU_CB10_PMOVSSET",  .decode.addr = A_SMMU_CB10_PMOVSSET,
    },{ .name = "SMMU_CB10_PMAUTHSTATUS",  .decode.addr = A_SMMU_CB10_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    },{ .name = "SMMU_CB11_SCTLR",  .decode.addr = A_SMMU_CB11_SCTLR,
        .reset = 0x100,
        .ro = 0x1000,
    },{ .name = "SMMU_CB11_ACTLR",  .decode.addr = A_SMMU_CB11_ACTLR,
        .reset = 0x3,
    },{ .name = "SMMU_CB11_RESUME",  .decode.addr = A_SMMU_CB11_RESUME,
    },{ .name = "SMMU_CB11_TCR2",  .decode.addr = A_SMMU_CB11_TCR2,
        .reset = 0x60,
        .ro = 0x60,
    },{ .name = "SMMU_CB11_TTBR0_LOW",  .decode.addr = A_SMMU_CB11_TTBR0_LOW,
        .ro = 0x4,
    },{ .name = "SMMU_CB11_TTBR0_HIGH",  .decode.addr = A_SMMU_CB11_TTBR0_HIGH,
    },{ .name = "SMMU_CB11_TTBR1_LOW",  .decode.addr = A_SMMU_CB11_TTBR1_LOW,
    },{ .name = "SMMU_CB11_TTBR1_HIGH",  .decode.addr = A_SMMU_CB11_TTBR1_HIGH,
    },{ .name = "SMMU_CB11_TCR_LPAE",  .decode.addr = A_SMMU_CB11_TCR_LPAE,
    },{ .name = "SMMU_CB11_CONTEXTIDR",  .decode.addr = A_SMMU_CB11_CONTEXTIDR,
    },{ .name = "SMMU_CB11_PRRR_MAIR0",  .decode.addr = A_SMMU_CB11_PRRR_MAIR0,
    },{ .name = "SMMU_CB11_NMRR_MAIR1",  .decode.addr = A_SMMU_CB11_NMRR_MAIR1,
    },{ .name = "SMMU_CB11_FSR",  .decode.addr = A_SMMU_CB11_FSR,
        .w1c = 0xffffffff,
        .post_write = smmu_fsr_pw,
    },{ .name = "SMMU_CB11_FSRRESTORE",  .decode.addr = A_SMMU_CB11_FSRRESTORE,
    },{ .name = "SMMU_CB11_FAR_LOW",  .decode.addr = A_SMMU_CB11_FAR_LOW,
    },{ .name = "SMMU_CB11_FAR_HIGH",  .decode.addr = A_SMMU_CB11_FAR_HIGH,
    },{ .name = "SMMU_CB11_FSYNR0",  .decode.addr = A_SMMU_CB11_FSYNR0,
        .ro = 0x200,
    },{ .name = "SMMU_CB11_IPAFAR_LOW",  .decode.addr = A_SMMU_CB11_IPAFAR_LOW,
        .ro = 0xfff,
    },{ .name = "SMMU_CB11_IPAFAR_HIGH",  .decode.addr = A_SMMU_CB11_IPAFAR_HIGH,
    },{ .name = "SMMU_CB11_TLBIVA_LOW",  .decode.addr = A_SMMU_CB11_TLBIVA_LOW,
    },{ .name = "SMMU_CB11_TLBIVA_HIGH",  .decode.addr = A_SMMU_CB11_TLBIVA_HIGH,
    },{ .name = "SMMU_CB11_TLBIVAA_LOW",  .decode.addr = A_SMMU_CB11_TLBIVAA_LOW,
    },{ .name = "SMMU_CB11_TLBIVAA_HIGH",  .decode.addr = A_SMMU_CB11_TLBIVAA_HIGH,
    },{ .name = "SMMU_CB11_TLBIASID",  .decode.addr = A_SMMU_CB11_TLBIASID,
    },{ .name = "SMMU_CB11_TLBIALL",  .decode.addr = A_SMMU_CB11_TLBIALL,
    },{ .name = "SMMU_CB11_TLBIVAL_LOW",  .decode.addr = A_SMMU_CB11_TLBIVAL_LOW,
    },{ .name = "SMMU_CB11_TLBIVAL_HIGH",  .decode.addr = A_SMMU_CB11_TLBIVAL_HIGH,
    },{ .name = "SMMU_CB11_TLBIVAAL_LOW",  .decode.addr = A_SMMU_CB11_TLBIVAAL_LOW,
    },{ .name = "SMMU_CB11_TLBIVAAL_HIGH",  .decode.addr = A_SMMU_CB11_TLBIVAAL_HIGH,
    },{ .name = "SMMU_CB11_TLBIIPAS2_LOW",  .decode.addr = A_SMMU_CB11_TLBIIPAS2_LOW,
    },{ .name = "SMMU_CB11_TLBIIPAS2_HIGH",  .decode.addr = A_SMMU_CB11_TLBIIPAS2_HIGH,
    },{ .name = "SMMU_CB11_TLBIIPAS2L_LOW",  .decode.addr = A_SMMU_CB11_TLBIIPAS2L_LOW,
    },{ .name = "SMMU_CB11_TLBIIPAS2L_HIGH",  .decode.addr = A_SMMU_CB11_TLBIIPAS2L_HIGH,
    },{ .name = "SMMU_CB11_TLBSYNC",  .decode.addr = A_SMMU_CB11_TLBSYNC,
    },{ .name = "SMMU_CB11_TLBSTATUS",  .decode.addr = A_SMMU_CB11_TLBSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_CB11_PMEVCNTR0",  .decode.addr = A_SMMU_CB11_PMEVCNTR0,
    },{ .name = "SMMU_CB11_PMEVCNTR1",  .decode.addr = A_SMMU_CB11_PMEVCNTR1,
    },{ .name = "SMMU_CB11_PMEVCNTR2",  .decode.addr = A_SMMU_CB11_PMEVCNTR2,
    },{ .name = "SMMU_CB11_PMEVCNTR3",  .decode.addr = A_SMMU_CB11_PMEVCNTR3,
    },{ .name = "SMMU_CB11_PMEVTYPER0",  .decode.addr = A_SMMU_CB11_PMEVTYPER0,
    },{ .name = "SMMU_CB11_PMEVTYPER1",  .decode.addr = A_SMMU_CB11_PMEVTYPER1,
    },{ .name = "SMMU_CB11_PMEVTYPER2",  .decode.addr = A_SMMU_CB11_PMEVTYPER2,
    },{ .name = "SMMU_CB11_PMEVTYPER3",  .decode.addr = A_SMMU_CB11_PMEVTYPER3,
    },{ .name = "SMMU_CB11_PMCFGR",  .decode.addr = A_SMMU_CB11_PMCFGR,
        .reset = 0x11f03,
        .ro = 0xff09ffff,
    },{ .name = "SMMU_CB11_PMCR",  .decode.addr = A_SMMU_CB11_PMCR,
        .ro = 0xff000002,
    },{ .name = "SMMU_CB11_PMCEID",  .decode.addr = A_SMMU_CB11_PMCEID,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "SMMU_CB11_PMCNTENSE",  .decode.addr = A_SMMU_CB11_PMCNTENSE,
    },{ .name = "SMMU_CB11_PMCNTENCLR",  .decode.addr = A_SMMU_CB11_PMCNTENCLR,
    },{ .name = "SMMU_CB11_PMCNTENSET",  .decode.addr = A_SMMU_CB11_PMCNTENSET,
    },{ .name = "SMMU_CB11_PMINTENCLR",  .decode.addr = A_SMMU_CB11_PMINTENCLR,
    },{ .name = "SMMU_CB11_PMOVSCLR",  .decode.addr = A_SMMU_CB11_PMOVSCLR,
    },{ .name = "SMMU_CB11_PMOVSSET",  .decode.addr = A_SMMU_CB11_PMOVSSET,
    },{ .name = "SMMU_CB11_PMAUTHSTATUS",  .decode.addr = A_SMMU_CB11_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    },{ .name = "SMMU_CB12_SCTLR",  .decode.addr = A_SMMU_CB12_SCTLR,
        .reset = 0x100,
        .ro = 0x1000,
    },{ .name = "SMMU_CB12_ACTLR",  .decode.addr = A_SMMU_CB12_ACTLR,
        .reset = 0x3,
    },{ .name = "SMMU_CB12_RESUME",  .decode.addr = A_SMMU_CB12_RESUME,
    },{ .name = "SMMU_CB12_TCR2",  .decode.addr = A_SMMU_CB12_TCR2,
        .reset = 0x60,
        .ro = 0x60,
    },{ .name = "SMMU_CB12_TTBR0_LOW",  .decode.addr = A_SMMU_CB12_TTBR0_LOW,
        .ro = 0x4,
    },{ .name = "SMMU_CB12_TTBR0_HIGH",  .decode.addr = A_SMMU_CB12_TTBR0_HIGH,
    },{ .name = "SMMU_CB12_TTBR1_LOW",  .decode.addr = A_SMMU_CB12_TTBR1_LOW,
    },{ .name = "SMMU_CB12_TTBR1_HIGH",  .decode.addr = A_SMMU_CB12_TTBR1_HIGH,
    },{ .name = "SMMU_CB12_TCR_LPAE",  .decode.addr = A_SMMU_CB12_TCR_LPAE,
    },{ .name = "SMMU_CB12_CONTEXTIDR",  .decode.addr = A_SMMU_CB12_CONTEXTIDR,
    },{ .name = "SMMU_CB12_PRRR_MAIR0",  .decode.addr = A_SMMU_CB12_PRRR_MAIR0,
    },{ .name = "SMMU_CB12_NMRR_MAIR1",  .decode.addr = A_SMMU_CB12_NMRR_MAIR1,
    },{ .name = "SMMU_CB12_FSR",  .decode.addr = A_SMMU_CB12_FSR,
        .w1c = 0xffffffff,
        .post_write = smmu_fsr_pw,
    },{ .name = "SMMU_CB12_FSRRESTORE",  .decode.addr = A_SMMU_CB12_FSRRESTORE,
    },{ .name = "SMMU_CB12_FAR_LOW",  .decode.addr = A_SMMU_CB12_FAR_LOW,
    },{ .name = "SMMU_CB12_FAR_HIGH",  .decode.addr = A_SMMU_CB12_FAR_HIGH,
    },{ .name = "SMMU_CB12_FSYNR0",  .decode.addr = A_SMMU_CB12_FSYNR0,
        .ro = 0x200,
    },{ .name = "SMMU_CB12_IPAFAR_LOW",  .decode.addr = A_SMMU_CB12_IPAFAR_LOW,
        .ro = 0xfff,
    },{ .name = "SMMU_CB12_IPAFAR_HIGH",  .decode.addr = A_SMMU_CB12_IPAFAR_HIGH,
    },{ .name = "SMMU_CB12_TLBIVA_LOW",  .decode.addr = A_SMMU_CB12_TLBIVA_LOW,
    },{ .name = "SMMU_CB12_TLBIVA_HIGH",  .decode.addr = A_SMMU_CB12_TLBIVA_HIGH,
    },{ .name = "SMMU_CB12_TLBIVAA_LOW",  .decode.addr = A_SMMU_CB12_TLBIVAA_LOW,
    },{ .name = "SMMU_CB12_TLBIVAA_HIGH",  .decode.addr = A_SMMU_CB12_TLBIVAA_HIGH,
    },{ .name = "SMMU_CB12_TLBIASID",  .decode.addr = A_SMMU_CB12_TLBIASID,
    },{ .name = "SMMU_CB12_TLBIALL",  .decode.addr = A_SMMU_CB12_TLBIALL,
    },{ .name = "SMMU_CB12_TLBIVAL_LOW",  .decode.addr = A_SMMU_CB12_TLBIVAL_LOW,
    },{ .name = "SMMU_CB12_TLBIVAL_HIGH",  .decode.addr = A_SMMU_CB12_TLBIVAL_HIGH,
    },{ .name = "SMMU_CB12_TLBIVAAL_LOW",  .decode.addr = A_SMMU_CB12_TLBIVAAL_LOW,
    },{ .name = "SMMU_CB12_TLBIVAAL_HIGH",  .decode.addr = A_SMMU_CB12_TLBIVAAL_HIGH,
    },{ .name = "SMMU_CB12_TLBIIPAS2_LOW",  .decode.addr = A_SMMU_CB12_TLBIIPAS2_LOW,
    },{ .name = "SMMU_CB12_TLBIIPAS2_HIGH",  .decode.addr = A_SMMU_CB12_TLBIIPAS2_HIGH,
    },{ .name = "SMMU_CB12_TLBIIPAS2L_LOW",  .decode.addr = A_SMMU_CB12_TLBIIPAS2L_LOW,
    },{ .name = "SMMU_CB12_TLBIIPAS2L_HIGH",  .decode.addr = A_SMMU_CB12_TLBIIPAS2L_HIGH,
    },{ .name = "SMMU_CB12_TLBSYNC",  .decode.addr = A_SMMU_CB12_TLBSYNC,
    },{ .name = "SMMU_CB12_TLBSTATUS",  .decode.addr = A_SMMU_CB12_TLBSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_CB12_PMEVCNTR0",  .decode.addr = A_SMMU_CB12_PMEVCNTR0,
    },{ .name = "SMMU_CB12_PMEVCNTR1",  .decode.addr = A_SMMU_CB12_PMEVCNTR1,
    },{ .name = "SMMU_CB12_PMEVCNTR2",  .decode.addr = A_SMMU_CB12_PMEVCNTR2,
    },{ .name = "SMMU_CB12_PMEVCNTR3",  .decode.addr = A_SMMU_CB12_PMEVCNTR3,
    },{ .name = "SMMU_CB12_PMEVTYPER0",  .decode.addr = A_SMMU_CB12_PMEVTYPER0,
    },{ .name = "SMMU_CB12_PMEVTYPER1",  .decode.addr = A_SMMU_CB12_PMEVTYPER1,
    },{ .name = "SMMU_CB12_PMEVTYPER2",  .decode.addr = A_SMMU_CB12_PMEVTYPER2,
    },{ .name = "SMMU_CB12_PMEVTYPER3",  .decode.addr = A_SMMU_CB12_PMEVTYPER3,
    },{ .name = "SMMU_CB12_PMCFGR",  .decode.addr = A_SMMU_CB12_PMCFGR,
        .reset = 0x11f03,
        .ro = 0xff09ffff,
    },{ .name = "SMMU_CB12_PMCR",  .decode.addr = A_SMMU_CB12_PMCR,
        .ro = 0xff000002,
    },{ .name = "SMMU_CB12_PMCEID",  .decode.addr = A_SMMU_CB12_PMCEID,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "SMMU_CB12_PMCNTENSE",  .decode.addr = A_SMMU_CB12_PMCNTENSE,
    },{ .name = "SMMU_CB12_PMCNTENCLR",  .decode.addr = A_SMMU_CB12_PMCNTENCLR,
    },{ .name = "SMMU_CB12_PMCNTENSET",  .decode.addr = A_SMMU_CB12_PMCNTENSET,
    },{ .name = "SMMU_CB12_PMINTENCLR",  .decode.addr = A_SMMU_CB12_PMINTENCLR,
    },{ .name = "SMMU_CB12_PMOVSCLR",  .decode.addr = A_SMMU_CB12_PMOVSCLR,
    },{ .name = "SMMU_CB12_PMOVSSET",  .decode.addr = A_SMMU_CB12_PMOVSSET,
    },{ .name = "SMMU_CB12_PMAUTHSTATUS",  .decode.addr = A_SMMU_CB12_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    },{ .name = "SMMU_CB13_SCTLR",  .decode.addr = A_SMMU_CB13_SCTLR,
        .reset = 0x100,
        .ro = 0x1000,
    },{ .name = "SMMU_CB13_ACTLR",  .decode.addr = A_SMMU_CB13_ACTLR,
        .reset = 0x3,
    },{ .name = "SMMU_CB13_RESUME",  .decode.addr = A_SMMU_CB13_RESUME,
    },{ .name = "SMMU_CB13_TCR2",  .decode.addr = A_SMMU_CB13_TCR2,
        .reset = 0x60,
        .ro = 0x60,
    },{ .name = "SMMU_CB13_TTBR0_LOW",  .decode.addr = A_SMMU_CB13_TTBR0_LOW,
        .ro = 0x4,
    },{ .name = "SMMU_CB13_TTBR0_HIGH",  .decode.addr = A_SMMU_CB13_TTBR0_HIGH,
    },{ .name = "SMMU_CB13_TTBR1_LOW",  .decode.addr = A_SMMU_CB13_TTBR1_LOW,
    },{ .name = "SMMU_CB13_TTBR1_HIGH",  .decode.addr = A_SMMU_CB13_TTBR1_HIGH,
    },{ .name = "SMMU_CB13_TCR_LPAE",  .decode.addr = A_SMMU_CB13_TCR_LPAE,
    },{ .name = "SMMU_CB13_CONTEXTIDR",  .decode.addr = A_SMMU_CB13_CONTEXTIDR,
    },{ .name = "SMMU_CB13_PRRR_MAIR0",  .decode.addr = A_SMMU_CB13_PRRR_MAIR0,
    },{ .name = "SMMU_CB13_NMRR_MAIR1",  .decode.addr = A_SMMU_CB13_NMRR_MAIR1,
    },{ .name = "SMMU_CB13_FSR",  .decode.addr = A_SMMU_CB13_FSR,
        .w1c = 0xffffffff,
        .post_write = smmu_fsr_pw,
    },{ .name = "SMMU_CB13_FSRRESTORE",  .decode.addr = A_SMMU_CB13_FSRRESTORE,
    },{ .name = "SMMU_CB13_FAR_LOW",  .decode.addr = A_SMMU_CB13_FAR_LOW,
    },{ .name = "SMMU_CB13_FAR_HIGH",  .decode.addr = A_SMMU_CB13_FAR_HIGH,
    },{ .name = "SMMU_CB13_FSYNR0",  .decode.addr = A_SMMU_CB13_FSYNR0,
        .ro = 0x200,
    },{ .name = "SMMU_CB13_IPAFAR_LOW",  .decode.addr = A_SMMU_CB13_IPAFAR_LOW,
        .ro = 0xfff,
    },{ .name = "SMMU_CB13_IPAFAR_HIGH",  .decode.addr = A_SMMU_CB13_IPAFAR_HIGH,
    },{ .name = "SMMU_CB13_TLBIVA_LOW",  .decode.addr = A_SMMU_CB13_TLBIVA_LOW,
    },{ .name = "SMMU_CB13_TLBIVA_HIGH",  .decode.addr = A_SMMU_CB13_TLBIVA_HIGH,
    },{ .name = "SMMU_CB13_TLBIVAA_LOW",  .decode.addr = A_SMMU_CB13_TLBIVAA_LOW,
    },{ .name = "SMMU_CB13_TLBIVAA_HIGH",  .decode.addr = A_SMMU_CB13_TLBIVAA_HIGH,
    },{ .name = "SMMU_CB13_TLBIASID",  .decode.addr = A_SMMU_CB13_TLBIASID,
    },{ .name = "SMMU_CB13_TLBIALL",  .decode.addr = A_SMMU_CB13_TLBIALL,
    },{ .name = "SMMU_CB13_TLBIVAL_LOW",  .decode.addr = A_SMMU_CB13_TLBIVAL_LOW,
    },{ .name = "SMMU_CB13_TLBIVAL_HIGH",  .decode.addr = A_SMMU_CB13_TLBIVAL_HIGH,
    },{ .name = "SMMU_CB13_TLBIVAAL_LOW",  .decode.addr = A_SMMU_CB13_TLBIVAAL_LOW,
    },{ .name = "SMMU_CB13_TLBIVAAL_HIGH",  .decode.addr = A_SMMU_CB13_TLBIVAAL_HIGH,
    },{ .name = "SMMU_CB13_TLBIIPAS2_LOW",  .decode.addr = A_SMMU_CB13_TLBIIPAS2_LOW,
    },{ .name = "SMMU_CB13_TLBIIPAS2_HIGH",  .decode.addr = A_SMMU_CB13_TLBIIPAS2_HIGH,
    },{ .name = "SMMU_CB13_TLBIIPAS2L_LOW",  .decode.addr = A_SMMU_CB13_TLBIIPAS2L_LOW,
    },{ .name = "SMMU_CB13_TLBIIPAS2L_HIGH",  .decode.addr = A_SMMU_CB13_TLBIIPAS2L_HIGH,
    },{ .name = "SMMU_CB13_TLBSYNC",  .decode.addr = A_SMMU_CB13_TLBSYNC,
    },{ .name = "SMMU_CB13_TLBSTATUS",  .decode.addr = A_SMMU_CB13_TLBSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_CB13_PMEVCNTR0",  .decode.addr = A_SMMU_CB13_PMEVCNTR0,
    },{ .name = "SMMU_CB13_PMEVCNTR1",  .decode.addr = A_SMMU_CB13_PMEVCNTR1,
    },{ .name = "SMMU_CB13_PMEVCNTR2",  .decode.addr = A_SMMU_CB13_PMEVCNTR2,
    },{ .name = "SMMU_CB13_PMEVCNTR3",  .decode.addr = A_SMMU_CB13_PMEVCNTR3,
    },{ .name = "SMMU_CB13_PMEVTYPER0",  .decode.addr = A_SMMU_CB13_PMEVTYPER0,
    },{ .name = "SMMU_CB13_PMEVTYPER1",  .decode.addr = A_SMMU_CB13_PMEVTYPER1,
    },{ .name = "SMMU_CB13_PMEVTYPER2",  .decode.addr = A_SMMU_CB13_PMEVTYPER2,
    },{ .name = "SMMU_CB13_PMEVTYPER3",  .decode.addr = A_SMMU_CB13_PMEVTYPER3,
    },{ .name = "SMMU_CB13_PMCFGR",  .decode.addr = A_SMMU_CB13_PMCFGR,
        .reset = 0x11f03,
        .ro = 0xff09ffff,
    },{ .name = "SMMU_CB13_PMCR",  .decode.addr = A_SMMU_CB13_PMCR,
        .ro = 0xff000002,
    },{ .name = "SMMU_CB13_PMCEID",  .decode.addr = A_SMMU_CB13_PMCEID,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "SMMU_CB13_PMCNTENSE",  .decode.addr = A_SMMU_CB13_PMCNTENSE,
    },{ .name = "SMMU_CB13_PMCNTENCLR",  .decode.addr = A_SMMU_CB13_PMCNTENCLR,
    },{ .name = "SMMU_CB13_PMCNTENSET",  .decode.addr = A_SMMU_CB13_PMCNTENSET,
    },{ .name = "SMMU_CB13_PMINTENCLR",  .decode.addr = A_SMMU_CB13_PMINTENCLR,
    },{ .name = "SMMU_CB13_PMOVSCLR",  .decode.addr = A_SMMU_CB13_PMOVSCLR,
    },{ .name = "SMMU_CB13_PMOVSSET",  .decode.addr = A_SMMU_CB13_PMOVSSET,
    },{ .name = "SMMU_CB13_PMAUTHSTATUS",  .decode.addr = A_SMMU_CB13_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    },{ .name = "SMMU_CB14_SCTLR",  .decode.addr = A_SMMU_CB14_SCTLR,
        .reset = 0x100,
        .ro = 0x1000,
    },{ .name = "SMMU_CB14_ACTLR",  .decode.addr = A_SMMU_CB14_ACTLR,
        .reset = 0x3,
    },{ .name = "SMMU_CB14_RESUME",  .decode.addr = A_SMMU_CB14_RESUME,
    },{ .name = "SMMU_CB14_TCR2",  .decode.addr = A_SMMU_CB14_TCR2,
        .reset = 0x60,
        .ro = 0x60,
    },{ .name = "SMMU_CB14_TTBR0_LOW",  .decode.addr = A_SMMU_CB14_TTBR0_LOW,
        .ro = 0x4,
    },{ .name = "SMMU_CB14_TTBR0_HIGH",  .decode.addr = A_SMMU_CB14_TTBR0_HIGH,
    },{ .name = "SMMU_CB14_TTBR1_LOW",  .decode.addr = A_SMMU_CB14_TTBR1_LOW,
    },{ .name = "SMMU_CB14_TTBR1_HIGH",  .decode.addr = A_SMMU_CB14_TTBR1_HIGH,
    },{ .name = "SMMU_CB14_TCR_LPAE",  .decode.addr = A_SMMU_CB14_TCR_LPAE,
    },{ .name = "SMMU_CB14_CONTEXTIDR",  .decode.addr = A_SMMU_CB14_CONTEXTIDR,
    },{ .name = "SMMU_CB14_PRRR_MAIR0",  .decode.addr = A_SMMU_CB14_PRRR_MAIR0,
    },{ .name = "SMMU_CB14_NMRR_MAIR1",  .decode.addr = A_SMMU_CB14_NMRR_MAIR1,
    },{ .name = "SMMU_CB14_FSR",  .decode.addr = A_SMMU_CB14_FSR,
        .w1c = 0xffffffff,
        .post_write = smmu_fsr_pw,
    },{ .name = "SMMU_CB14_FSRRESTORE",  .decode.addr = A_SMMU_CB14_FSRRESTORE,
    },{ .name = "SMMU_CB14_FAR_LOW",  .decode.addr = A_SMMU_CB14_FAR_LOW,
    },{ .name = "SMMU_CB14_FAR_HIGH",  .decode.addr = A_SMMU_CB14_FAR_HIGH,
    },{ .name = "SMMU_CB14_FSYNR0",  .decode.addr = A_SMMU_CB14_FSYNR0,
        .ro = 0x200,
    },{ .name = "SMMU_CB14_IPAFAR_LOW",  .decode.addr = A_SMMU_CB14_IPAFAR_LOW,
        .ro = 0xfff,
    },{ .name = "SMMU_CB14_IPAFAR_HIGH",  .decode.addr = A_SMMU_CB14_IPAFAR_HIGH,
    },{ .name = "SMMU_CB14_TLBIVA_LOW",  .decode.addr = A_SMMU_CB14_TLBIVA_LOW,
    },{ .name = "SMMU_CB14_TLBIVA_HIGH",  .decode.addr = A_SMMU_CB14_TLBIVA_HIGH,
    },{ .name = "SMMU_CB14_TLBIVAA_LOW",  .decode.addr = A_SMMU_CB14_TLBIVAA_LOW,
    },{ .name = "SMMU_CB14_TLBIVAA_HIGH",  .decode.addr = A_SMMU_CB14_TLBIVAA_HIGH,
    },{ .name = "SMMU_CB14_TLBIASID",  .decode.addr = A_SMMU_CB14_TLBIASID,
    },{ .name = "SMMU_CB14_TLBIALL",  .decode.addr = A_SMMU_CB14_TLBIALL,
    },{ .name = "SMMU_CB14_TLBIVAL_LOW",  .decode.addr = A_SMMU_CB14_TLBIVAL_LOW,
    },{ .name = "SMMU_CB14_TLBIVAL_HIGH",  .decode.addr = A_SMMU_CB14_TLBIVAL_HIGH,
    },{ .name = "SMMU_CB14_TLBIVAAL_LOW",  .decode.addr = A_SMMU_CB14_TLBIVAAL_LOW,
    },{ .name = "SMMU_CB14_TLBIVAAL_HIGH",  .decode.addr = A_SMMU_CB14_TLBIVAAL_HIGH,
    },{ .name = "SMMU_CB14_TLBIIPAS2_LOW",  .decode.addr = A_SMMU_CB14_TLBIIPAS2_LOW,
    },{ .name = "SMMU_CB14_TLBIIPAS2_HIGH",  .decode.addr = A_SMMU_CB14_TLBIIPAS2_HIGH,
    },{ .name = "SMMU_CB14_TLBIIPAS2L_LOW",  .decode.addr = A_SMMU_CB14_TLBIIPAS2L_LOW,
    },{ .name = "SMMU_CB14_TLBIIPAS2L_HIGH",  .decode.addr = A_SMMU_CB14_TLBIIPAS2L_HIGH,
    },{ .name = "SMMU_CB14_TLBSYNC",  .decode.addr = A_SMMU_CB14_TLBSYNC,
    },{ .name = "SMMU_CB14_TLBSTATUS",  .decode.addr = A_SMMU_CB14_TLBSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_CB14_PMEVCNTR0",  .decode.addr = A_SMMU_CB14_PMEVCNTR0,
    },{ .name = "SMMU_CB14_PMEVCNTR1",  .decode.addr = A_SMMU_CB14_PMEVCNTR1,
    },{ .name = "SMMU_CB14_PMEVCNTR2",  .decode.addr = A_SMMU_CB14_PMEVCNTR2,
    },{ .name = "SMMU_CB14_PMEVCNTR3",  .decode.addr = A_SMMU_CB14_PMEVCNTR3,
    },{ .name = "SMMU_CB14_PMEVTYPER0",  .decode.addr = A_SMMU_CB14_PMEVTYPER0,
    },{ .name = "SMMU_CB14_PMEVTYPER1",  .decode.addr = A_SMMU_CB14_PMEVTYPER1,
    },{ .name = "SMMU_CB14_PMEVTYPER2",  .decode.addr = A_SMMU_CB14_PMEVTYPER2,
    },{ .name = "SMMU_CB14_PMEVTYPER3",  .decode.addr = A_SMMU_CB14_PMEVTYPER3,
    },{ .name = "SMMU_CB14_PMCFGR",  .decode.addr = A_SMMU_CB14_PMCFGR,
        .reset = 0x11f03,
        .ro = 0xff09ffff,
    },{ .name = "SMMU_CB14_PMCR",  .decode.addr = A_SMMU_CB14_PMCR,
        .ro = 0xff000002,
    },{ .name = "SMMU_CB14_PMCEID",  .decode.addr = A_SMMU_CB14_PMCEID,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "SMMU_CB14_PMCNTENSE",  .decode.addr = A_SMMU_CB14_PMCNTENSE,
    },{ .name = "SMMU_CB14_PMCNTENCLR",  .decode.addr = A_SMMU_CB14_PMCNTENCLR,
    },{ .name = "SMMU_CB14_PMCNTENSET",  .decode.addr = A_SMMU_CB14_PMCNTENSET,
    },{ .name = "SMMU_CB14_PMINTENCLR",  .decode.addr = A_SMMU_CB14_PMINTENCLR,
    },{ .name = "SMMU_CB14_PMOVSCLR",  .decode.addr = A_SMMU_CB14_PMOVSCLR,
    },{ .name = "SMMU_CB14_PMOVSSET",  .decode.addr = A_SMMU_CB14_PMOVSSET,
    },{ .name = "SMMU_CB14_PMAUTHSTATUS",  .decode.addr = A_SMMU_CB14_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    },{ .name = "SMMU_CB15_SCTLR",  .decode.addr = A_SMMU_CB15_SCTLR,
        .reset = 0x100,
        .ro = 0x1000,
    },{ .name = "SMMU_CB15_ACTLR",  .decode.addr = A_SMMU_CB15_ACTLR,
        .reset = 0x3,
    },{ .name = "SMMU_CB15_RESUME",  .decode.addr = A_SMMU_CB15_RESUME,
    },{ .name = "SMMU_CB15_TCR2",  .decode.addr = A_SMMU_CB15_TCR2,
        .reset = 0x60,
        .ro = 0x60,
    },{ .name = "SMMU_CB15_TTBR0_LOW",  .decode.addr = A_SMMU_CB15_TTBR0_LOW,
        .ro = 0x4,
    },{ .name = "SMMU_CB15_TTBR0_HIGH",  .decode.addr = A_SMMU_CB15_TTBR0_HIGH,
    },{ .name = "SMMU_CB15_TTBR1_LOW",  .decode.addr = A_SMMU_CB15_TTBR1_LOW,
    },{ .name = "SMMU_CB15_TTBR1_HIGH",  .decode.addr = A_SMMU_CB15_TTBR1_HIGH,
    },{ .name = "SMMU_CB15_TCR_LPAE",  .decode.addr = A_SMMU_CB15_TCR_LPAE,
    },{ .name = "SMMU_CB15_CONTEXTIDR",  .decode.addr = A_SMMU_CB15_CONTEXTIDR,
    },{ .name = "SMMU_CB15_PRRR_MAIR0",  .decode.addr = A_SMMU_CB15_PRRR_MAIR0,
    },{ .name = "SMMU_CB15_NMRR_MAIR1",  .decode.addr = A_SMMU_CB15_NMRR_MAIR1,
    },{ .name = "SMMU_CB15_FSR",  .decode.addr = A_SMMU_CB15_FSR,
        .w1c = 0xffffffff,
        .post_write = smmu_fsr_pw,
    },{ .name = "SMMU_CB15_FSRRESTORE",  .decode.addr = A_SMMU_CB15_FSRRESTORE,
    },{ .name = "SMMU_CB15_FAR_LOW",  .decode.addr = A_SMMU_CB15_FAR_LOW,
    },{ .name = "SMMU_CB15_FAR_HIGH",  .decode.addr = A_SMMU_CB15_FAR_HIGH,
    },{ .name = "SMMU_CB15_FSYNR0",  .decode.addr = A_SMMU_CB15_FSYNR0,
        .ro = 0x200,
    },{ .name = "SMMU_CB15_IPAFAR_LOW",  .decode.addr = A_SMMU_CB15_IPAFAR_LOW,
        .ro = 0xfff,
    },{ .name = "SMMU_CB15_IPAFAR_HIGH",  .decode.addr = A_SMMU_CB15_IPAFAR_HIGH,
    },{ .name = "SMMU_CB15_TLBIVA_LOW",  .decode.addr = A_SMMU_CB15_TLBIVA_LOW,
    },{ .name = "SMMU_CB15_TLBIVA_HIGH",  .decode.addr = A_SMMU_CB15_TLBIVA_HIGH,
    },{ .name = "SMMU_CB15_TLBIVAA_LOW",  .decode.addr = A_SMMU_CB15_TLBIVAA_LOW,
    },{ .name = "SMMU_CB15_TLBIVAA_HIGH",  .decode.addr = A_SMMU_CB15_TLBIVAA_HIGH,
    },{ .name = "SMMU_CB15_TLBIASID",  .decode.addr = A_SMMU_CB15_TLBIASID,
    },{ .name = "SMMU_CB15_TLBIALL",  .decode.addr = A_SMMU_CB15_TLBIALL,
    },{ .name = "SMMU_CB15_TLBIVAL_LOW",  .decode.addr = A_SMMU_CB15_TLBIVAL_LOW,
    },{ .name = "SMMU_CB15_TLBIVAL_HIGH",  .decode.addr = A_SMMU_CB15_TLBIVAL_HIGH,
    },{ .name = "SMMU_CB15_TLBIVAAL_LOW",  .decode.addr = A_SMMU_CB15_TLBIVAAL_LOW,
    },{ .name = "SMMU_CB15_TLBIVAAL_HIGH",  .decode.addr = A_SMMU_CB15_TLBIVAAL_HIGH,
    },{ .name = "SMMU_CB15_TLBIIPAS2_LOW",  .decode.addr = A_SMMU_CB15_TLBIIPAS2_LOW,
    },{ .name = "SMMU_CB15_TLBIIPAS2_HIGH",  .decode.addr = A_SMMU_CB15_TLBIIPAS2_HIGH,
    },{ .name = "SMMU_CB15_TLBIIPAS2L_LOW",  .decode.addr = A_SMMU_CB15_TLBIIPAS2L_LOW,
    },{ .name = "SMMU_CB15_TLBIIPAS2L_HIGH",  .decode.addr = A_SMMU_CB15_TLBIIPAS2L_HIGH,
    },{ .name = "SMMU_CB15_TLBSYNC",  .decode.addr = A_SMMU_CB15_TLBSYNC,
    },{ .name = "SMMU_CB15_TLBSTATUS",  .decode.addr = A_SMMU_CB15_TLBSTATUS,
        .ro = 0x1,
    },{ .name = "SMMU_CB15_PMEVCNTR0",  .decode.addr = A_SMMU_CB15_PMEVCNTR0,
    },{ .name = "SMMU_CB15_PMEVCNTR1",  .decode.addr = A_SMMU_CB15_PMEVCNTR1,
    },{ .name = "SMMU_CB15_PMEVCNTR2",  .decode.addr = A_SMMU_CB15_PMEVCNTR2,
    },{ .name = "SMMU_CB15_PMEVCNTR3",  .decode.addr = A_SMMU_CB15_PMEVCNTR3,
    },{ .name = "SMMU_CB15_PMEVTYPER0",  .decode.addr = A_SMMU_CB15_PMEVTYPER0,
    },{ .name = "SMMU_CB15_PMEVTYPER1",  .decode.addr = A_SMMU_CB15_PMEVTYPER1,
    },{ .name = "SMMU_CB15_PMEVTYPER2",  .decode.addr = A_SMMU_CB15_PMEVTYPER2,
    },{ .name = "SMMU_CB15_PMEVTYPER3",  .decode.addr = A_SMMU_CB15_PMEVTYPER3,
    },{ .name = "SMMU_CB15_PMCFGR",  .decode.addr = A_SMMU_CB15_PMCFGR,
        .reset = 0x11f03,
        .ro = 0xff09ffff,
    },{ .name = "SMMU_CB15_PMCR",  .decode.addr = A_SMMU_CB15_PMCR,
        .ro = 0xff000002,
    },{ .name = "SMMU_CB15_PMCEID",  .decode.addr = A_SMMU_CB15_PMCEID,
        .reset = 0x30303,
        .ro = 0x38383,
    },{ .name = "SMMU_CB15_PMCNTENSE",  .decode.addr = A_SMMU_CB15_PMCNTENSE,
    },{ .name = "SMMU_CB15_PMCNTENCLR",  .decode.addr = A_SMMU_CB15_PMCNTENCLR,
    },{ .name = "SMMU_CB15_PMCNTENSET",  .decode.addr = A_SMMU_CB15_PMCNTENSET,
    },{ .name = "SMMU_CB15_PMINTENCLR",  .decode.addr = A_SMMU_CB15_PMINTENCLR,
    },{ .name = "SMMU_CB15_PMOVSCLR",  .decode.addr = A_SMMU_CB15_PMOVSCLR,
    },{ .name = "SMMU_CB15_PMOVSSET",  .decode.addr = A_SMMU_CB15_PMOVSSET,
    },{ .name = "SMMU_CB15_PMAUTHSTATUS",  .decode.addr = A_SMMU_CB15_PMAUTHSTATUS,
        .reset = 0x80,
        .ro = 0xff,
    }
};

static void smmu500_reset(DeviceState *dev)
{
    SMMU *s = XILINX_SMMU500(dev);
    unsigned int i;

    for (i = 0; i < ARRAY_SIZE(s->regs_info); ++i) {
        dep_register_reset(&s->regs_info[i]);
    }
}

static uint64_t smmu500_read(void *opaque, hwaddr addr, unsigned size)
{
    SMMU *s = XILINX_SMMU500(opaque);
    DepRegisterInfo *r = &s->regs_info[addr / 4];

    if (!r->data) {
        qemu_log("%s: Decode error: read from %" HWADDR_PRIx "\n",
                 object_get_canonical_path(OBJECT(s)),
                 addr);
        return 0;
    }
    return dep_register_read(r);
}

static void smmu500_write(void *opaque, hwaddr addr, uint64_t value,
                      unsigned size)
{
    SMMU *s = XILINX_SMMU500(opaque);
    DepRegisterInfo *r = &s->regs_info[addr / 4];

    if (!r->data) {
        qemu_log("%s: Decode error: write to %" HWADDR_PRIx "=%" PRIx64 "\n",
                 object_get_canonical_path(OBJECT(s)),
                 addr, value);
        return;
    }
    dep_register_write(r, value, ~0);
}


static const MemoryRegionOps smmu500_ops = {
    .read = smmu500_read,
    .write = smmu500_write,
    .endianness = DEVICE_LITTLE_ENDIAN,
    .valid = {
        .min_access_size = 4,
        .max_access_size = 8,
    },
};

static void smmu500_realize(DeviceState *dev, Error **errp)
{
    SMMU *s = XILINX_SMMU500(dev);
    const char *prefix = object_get_canonical_path(OBJECT(dev));
    unsigned int i;

    for (i = 0; i < ARRAY_SIZE(smmu500_regs_info); ++i) {
        DepRegisterInfo *r = &s->regs_info[smmu500_regs_info[i].decode.addr/4];

        *r = (DepRegisterInfo) {
            .data = (uint8_t *)&s->regs[
                    smmu500_regs_info[i].decode.addr/4],
            .data_size = sizeof(uint32_t),
            .access = &smmu500_regs_info[i],
            .debug = XILINX_SMMU500_ERR_DEBUG,
            .prefix = prefix,
            .opaque = s,
        };
    }

    s->dma_as = s->dma_mr ? address_space_init_shareable(s->dma_mr, NULL)
                          : &address_space_memory;

    assert(s->dma_as);

    sysbus_init_irq(SYS_BUS_DEVICE(dev), &s->irq.global);
    for (i = 0; i < 16; i++) {
        sysbus_init_irq(SYS_BUS_DEVICE(dev), &s->irq.context[i]);
    }
}

static void smmu500_init(Object *obj)
{
    SMMU *s = XILINX_SMMU500(obj);
    int i;

    object_property_add_link(obj, "dma", TYPE_MEMORY_REGION,
                             (Object **)&s->dma_mr,
                             qdev_prop_allow_set_link_before_realize,
                             OBJ_PROP_LINK_UNREF_ON_RELEASE,
                             &error_abort);

    for (i = 0; i < MAX_TBU; i++) {
        char *name = g_strdup_printf("mr-%d", i);
        object_property_add_link(obj, name, TYPE_MEMORY_REGION,
                                 (Object **)&s->tbu[i].mr,
                                 qdev_prop_allow_set_link_before_realize,
                                 OBJ_PROP_LINK_UNREF_ON_RELEASE,
                                 &error_abort);
        g_free(name);
        s->tbu[i].smmu = s;
    }
}

static bool smmu_parse_reg(FDTGenericMMap *obj, FDTGenericRegPropInfo reg,
                           Error **errp)
{
    SMMU *s = XILINX_SMMU500(obj);
    SysBusDevice *sbd = SYS_BUS_DEVICE(obj);
    ObjectClass *klass = object_class_by_name(TYPE_XILINX_SMMU500);
    FDTGenericMMapClass *parent_fmc;
    unsigned int i;

    parent_fmc = FDT_GENERIC_MMAP_CLASS(object_class_get_parent(klass));
    memory_region_init_io(&s->iomem, OBJECT(obj), &smmu500_ops, s,
                          TYPE_XILINX_SMMU500, R_MAX * 4);
    sysbus_init_mmio(sbd, &s->iomem);

    for (i = 0; i < (reg.n - 1); i++) {
        char *name = g_strdup_printf("smmu-tbu%d", i);

        assert(s->tbu[i].mr);
        s->tbu[i].as = address_space_init_shareable(s->tbu[i].mr, NULL);
        memory_region_init_iommu(&s->tbu[i].iommu, sizeof(s->tbu[i].iommu),
                                 TYPE_XILINX_SMMU500_IOMMU_MEMORY_REGION,
                                 OBJECT(sbd),
                                 name, UINT64_MAX);
        sysbus_init_mmio(sbd, MEMORY_REGION(&s->tbu[i].iommu));
        g_free(name);
    }

    return parent_fmc ? parent_fmc->parse_reg(obj, reg, errp) : false;
}

static Property smmu_properties[] = {
    DEFINE_PROP_UINT32("pamax", SMMU, cfg.pamax, 48),
    DEFINE_PROP_END_OF_LIST(),
};

static const VMStateDescription vmstate_smmu500 = {
    .name = TYPE_XILINX_SMMU500,
    .version_id = 1,
    .minimum_version_id = 1,
    .minimum_version_id_old = 1,
    .fields = (VMStateField[]) {
        VMSTATE_UINT32_ARRAY(regs, SMMU, R_MAX),
        VMSTATE_END_OF_LIST(),
    }
};

static void smmu500_class_init(ObjectClass *klass, void *data)
{
    DeviceClass *dc = DEVICE_CLASS(klass);
    FDTGenericMMapClass *fmc = FDT_GENERIC_MMAP_CLASS(klass);

    dc->reset = smmu500_reset;
    dc->realize = smmu500_realize;
    dc->vmsd = &vmstate_smmu500;
    dc->props = smmu_properties;
    fmc->parse_reg = smmu_parse_reg;
}

static void smmu500_iommu_memory_region_class_init(ObjectClass *klass,
                                                   void *data)
{
    IOMMUMemoryRegionClass *imrc = IOMMU_MEMORY_REGION_CLASS(klass);

    imrc->translate_attr = smmu_translate;
}

static const TypeInfo smmu500_info = {
    .name          = TYPE_XILINX_SMMU500,
    .parent        = TYPE_SYS_BUS_DEVICE,
    .instance_size = sizeof(SMMU),
    .class_init    = smmu500_class_init,
    .instance_init = smmu500_init,
    .interfaces    = (InterfaceInfo[]) {
        { TYPE_FDT_GENERIC_MMAP },
        { },
    },
};

static const TypeInfo smmu500_iommu_memory_region_info = {
    .name = TYPE_XILINX_SMMU500_IOMMU_MEMORY_REGION,
    .parent = TYPE_IOMMU_MEMORY_REGION,
    .class_init = smmu500_iommu_memory_region_class_init,
};


static void smmu500_register_types(void)
{
    type_register_static(&smmu500_info);
    type_register_static(&smmu500_iommu_memory_region_info);
}

type_init(smmu500_register_types)
